%%% File    : estone.erl
%%% Author  : RHS Linux User <klacke@erix.ericsson.se>
%%% Purpose : Meassure erlang implementations on a certain machine
%%% Created : 29 Oct 1996 by RHS Linux User <klacke@jb.du.etx.ericsson.se>

-module(estone).
-author('klacke@erix.ericsson.se').

-compile(export_all).


%% This is *the* estone benchmark, it tries to meassure
%% the performance of an erlang system in a way which
%% makes it possible to compare different implementations
%% of erlang as well as different HW.
%% estone consist of a number of micro benchmarks that
%% that can be run independently. If all the microbenchmarks
%% are combined, we get a system where erlang 4.3.1 runs with
%% 1000 estones on a 85 Mhz sparc 5.
%% We always meassure wallclock here, which makes it important
%% that the computer which runs the test is not used for anything
%% else but the bench

%% Each micro bench mark is described by an instance of

%%-record(micro,
%%  {function, %% The name of the function implementing the micro
%%   weight,   %% How important is this in typical applications ??
%%   loops = 100,%% initial data
%%   tt1,      %% time to do one round
%%   str}).    %% Header string

run() ->
    run("Gerl","0.6","","Linux","i386","100x2").

run(Compiler,Version,Flags,Os,CPu,MHz) ->
    run(Compiler,Version,Flags,Os,CPu,MHz,nohalt).

run(Compiler,Version,Flags,Os,CPu,MHz,Halt) ->
    io:format("Running estone ~p~n",[mk_fname([Compiler,Version,Flags,Os,CPu,MHz])]),
    L = estone:macro(estone:micros()),
    output(Compiler,Version,Flags,Os,CPu,MHz,L),
    case Halt of
        halt ->
            halt();
        nohalt ->
            ok
    end.

output(Compiler,Version,Flags,Os,CPu,MHz,L) ->
    Fname = mk_fname([Compiler,Version,Flags,Os,CPu,MHz]),
    {ok,F} = file:open(lists:append(Fname,".EST"),[read,write]),
    L2 = [[{compiler,Compiler},{version,Version},{flags,Flags},{os,Os},{cpu,CPu},{mhz,MHz}]|L],
    io:format(F,"~w.~n",[L2]),
    io:format("Output written to ~s~n",[lists:append(Fname,".EST")]),
    file:close(F).

cmdline_run([Compiler,Version,Flags,Os,CPu,MHz]) ->
    spawn(estone,run,[Compiler,Version,Flags,Os,CPu,MHz,halt]).

tty() ->
    tty(all).

tty(all) ->
    pp(macro(micros()));
tty(MicroName) ->
    case find_micro(MicroName,micros()) of
        {ok,M} ->
            pp(macro([M]));
        Other ->
            Other
    end.

find_micro(N,[M|T]) when element(2,M) == N ->
    {ok,M};
find_micro(N,[_|T]) ->
    find_micro(N,T);
find_micro(N,[]) ->
    notfound.

test() ->
    macro(determine_loop_data()).

iterate(Ofile,Times) ->
    {ok,Out} = file:open(Ofile,write),
    iterate2(Out,Times),
    file:close(Out).

iterate2(_,0) ->
    ok;
iterate2(Out,I) ->
    pp(Out,macro(micros())),
    iterate2(Out,I - 1).

micro(K,Loops) ->
    M = micro(K),
    setelement(4,M,Loops).

micro(lists) ->
    {micro,lists,7,6400,undefined,"list manipulation"};
micro(msgp) ->
    {micro,msgp,10,1515,undefined,"small messages"};
micro(msgp_medium) ->
    {micro,msgp_medium,14,1527,undefined,"medium messages"};
micro(msgp_huge) ->
    {micro,msgp_huge,4,52,undefined,"huge messages"};
micro(pattern) ->
    {micro,pattern,5,1046,undefined,"pattern matching"};
micro(trav) ->
    {micro,trav,4,2834,undefined,"traverse"};
micro(port_io) ->
    {micro,port_io,12,4800,undefined,"Port i/o"};
micro(large_dataset_work) ->
    {micro,large_dataset_work,3,1193,undefined,"Work with large dataset"};
micro(large_local_dataset_work) ->
    {micro,large_local_dataset_work,3,1174,undefined,"Work with large local dataset"};
micro(alloc) ->
    {micro,alloc,2,3710,undefined,"Alloc and dealloc"};
micro(bif_dispatch) ->
    {micro,bif_dispatch,5,1623,undefined,"Bif dispatch"};
micro(binary_h) ->
    {micro,binary_h,4,581,undefined,"Binary handling"};
micro(ets) ->
    {micro,ets,6,342,undefined,"ets datadictionary"};
micro(generic) ->
    {micro,generic,9,7977,undefined,"Generic server (with timeout)"};
micro(int_arith) ->
    {micro,int_arith,3,4157,undefined,"Small Integer arithmetics"};
micro(float_arith) ->
    {micro,float_arith,1,5526,undefined,"Float arithmetics"};
micro(fcalls) ->
    {micro,fcalls,5,882,undefined,"Function calls"};
micro(timer) ->
    {micro,timer,2,2312,undefined,"Timers"};
micro(links) ->
    {micro,links,1,30,undefined,"Links"}.

micros() ->
    [micro(lists),micro(msgp),micro(msgp_medium),micro(msgp_huge),micro(pattern),micro(trav),micro(port_io),micro(large_dataset_work),micro(large_local_dataset_work),micro(alloc),micro(bif_dispatch),micro(binary_h),micro(ets),micro(generic),micro(int_arith),micro(float_arith),micro(fcalls),micro(timer),micro(links)].

macro(Ms) ->
    now(),
    statistics(reductions),
    statistics(runtime),
    lists(500),
    run_micros(Ms).

run_micros([]) ->
    io:nl(),
    [];
run_micros([H|T]) ->
    R = run_micro(H),
    [R|run_micros(T)].

run_micro(M) ->
    Pid = spawn(estone,run_micro,[self(),M]),
    Res = receive
              {Pid,Reply} ->
                  Reply
          end,
    Res.

run_micro(Top,M) ->
    Top ! {self(),apply_micro(M)}.

apply_micro(M) ->
    {GC0,Words0,_} = statistics(garbage_collection),
    statistics(reductions),
    Before = now(),
    Compensate = apply_micro(element(2,M),element(4,M)),
    After = now(),
    {GC1,Words1,_} = statistics(garbage_collection),
    {_,Reds} = statistics(reductions),
    Elapsed = subtr(Before,After),
    MicroSecs = Elapsed - Compensate,
    [{title,element(6,M)},{tt1,element(5,M)},{function,element(2,M)},{weight_percentage,element(3,M)},{loops,element(4,M)},{microsecs,MicroSecs},{estones,element(3,M) * element(3,M) * 31000000 div MicroSecs},{gcs,GC1 - GC0},{kilo_word_reclaimed,(Words1 - Words0) div 1000},{kilo_reductions,Reds div 1000},{gc_intensity,gci(Elapsed,GC1 - GC0,Words1 - Words0)}].

pp(Ms) ->
    pp(standard_io,Ms).

pp(Out,Ms) ->
    {Total,Stones} = sum_micros(Ms,0,0),
    io:format(Out,"**** Total time ~w seconds ****~n",[Total / 1000000]),
    io:format(Out,"**** ESTONES = ~w~n~n",[Stones]),
    io:format(Out,"~-31s      ~-12s  ~-10s   %    ~-10s ~n~n",["    Title","Millis","Estone","Loops"]),
    pp2(Out,Ms).

sum_micros([],Tot,Stones) ->
    {Tot,Stones};
sum_micros([H|T],Tot,Sto) ->
    sum_micros(T,ks(microsecs,H) + Tot,ks(estones,H) + Sto).

pp2(_,[]) ->
    ok;
pp2(Out,[R|Tail]) ->
    io:format(Out,"~-35s  ~-12w    ~-10w   ~-2w    ~-10w ~n",[ks(title,R),round(ks(microsecs,R) / 1000),ks(estones,R),ks(weight_percentage,R),ks(loops,R)]),
    pp2(Out,Tail).

ks(K,L) ->
    {value,{_,V}} = lists:keysearch(K,1,L),
    V.

subtr(Before,After) ->
    element(1,After) * 1000000000000 + element(2,After) * 1000000 + element(3,After) - (element(1,Before) * 1000000000000 + element(2,Before) * 1000000 + element(3,Before)).

gci(Micros,Words,Gcs) ->
    256 * Gcs / Micros + Words / Micros.

apply_micro(Name,Loops) ->
    io:format("~w(~w) ",[Name,Loops]),
    apply(estone,Name,[Loops]).

lists(I) ->
    L1 = "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    L2 = "aaaaaaaaaa",
    lists(I,L1,L2).

lists(0,_,_) ->
    0;
lists(I,L1,L2) ->
    revt(10,L1),
    appt(10,L1,L2),
    lists(I - 1,L1,L2).

revt(0,_) ->
    done;
revt(I,L) ->
    reverse(L),
    revt(I - 1,L).

reverse(L) ->
    reverse(L,[]).

reverse([H|T],Ack) ->
    reverse(T,[H|Ack]);
reverse([],Ack) ->
    Ack.

append([H|T],L) ->
    [H|append(T,L)];
append([],L) ->
    L.

appt(0,L1,L2) ->
    ok;
appt(I,L1,L2) ->
    append(L1,L2),
    appt(I - 1,L1,L2).

msgp(I) ->
    msgp(I,small()).

msgp(0,_) ->
    0;
msgp(I,Msg) ->
    P1 = spawn(estone,p1,[self()]),
    P2 = spawn(estone,p1,[P1]),
    P3 = spawn(estone,p1,[P2]),
    P4 = spawn(estone,p1,[P3]),
    msgp_loop(100,P4,Msg),
    msgp(I - 1,Msg).

p1(To) ->
    receive
        {From,{message,X}} ->
            To ! {self(),{message,X}},
            p1(To);
        stop ->
            To ! stop,
            exit(normal)
    end.

msgp_loop(0,P,_) ->
    P ! stop,
    receive
        stop ->
            ok
    end;
msgp_loop(I,P,Msg) ->
    P ! {self(),{message,Msg}},
    receive
        {From,{message,_}} ->
            msgp_loop(I - 1,P,Msg)
    end.

msgp_medium(I) ->
    msgp_medium(I,big()).

msgp_medium(0,_) ->
    0;
msgp_medium(I,Msg) ->
    P1 = spawn(estone,p1,[self()]),
    P2 = spawn(estone,p1,[P1]),
    P3 = spawn(estone,p1,[P2]),
    P4 = spawn(estone,p1,[P3]),
    msgp_loop(100,P4,Msg),
    msgp_medium(I - 1,Msg).

msgp_huge(I) ->
    msgp_huge(I,very_big(15)).

msgp_huge(0,_) ->
    0;
msgp_huge(I,Msg) ->
    P1 = spawn(estone,p1,[self()]),
    P4 = spawn(estone,p1,[P1]),
    msgp_loop(100,P4,Msg),
    msgp_huge(I - 1,Msg).

pattern(0) ->
    0;
pattern(I) ->
    Tail = "aaabbaaababba",
    P1 = [0,1,2,3,4,5|Tail],
    pat_loop1(100,P1),
    pat_loop2(100,P1),
    pat_loop3(100,P1),
    pat_loop4(100,P1),
    pat_loop5(100,P1),
    pattern(I - 1).

pat_loop1(0,_) ->
    ok;
pat_loop1(I,[_,X,Y,0|T]) ->
    ok;
pat_loop1(I,[_,X,Y,1|T]) ->
    ok;
pat_loop1(I,[_,X,Y,2|T]) ->
    ok;
pat_loop1(I,[_,X,Y,3|T]) ->
    pat_loop1(I - 1,[0,X,Y,3|T]).

pat_loop2(0,_) ->
    ok;
pat_loop2(I,[X,Y|Tail]) when Y bsl 1 == 0 ->
    ok;
pat_loop2(I,[X,Y|Tail]) when Y bsl 2 == 0 ->
    ok;
pat_loop2(I,[X,Y|Tail]) when Y bsl 2 == 4 ->
    pat_loop2(I - 1,[X,Y|Tail]).

pat_loop3(0,_) ->
    ok;
pat_loop3(I,[{c,h}|Tail]) ->
    ok;
pat_loop3(I,[1,0|T]) ->
    ok;
pat_loop3(I,[X,Y|Tail]) when binary(X), size(X) == 1 ->
    ok;
pat_loop3(I,[no,Y|Tail]) ->
    ok;
pat_loop3(I,[]) ->
    ok;
pat_loop3(I,[X,Y|T]) when X /= 0 ->
    ok;
pat_loop3(I,[2,3|T]) ->
    ok;
pat_loop3(I,[1,2]) ->
    ok;
pat_loop3(I,[0,1|T]) ->
    pat_loop3(I - 1,[0,1|T]).

pat_loop4(0,_) ->
    ok;
pat_loop4(I,[20|T]) ->
    ok;
pat_loop4(I,[219|T]) ->
    ok;
pat_loop4(I,[18|T]) ->
    ok;
pat_loop4(I,[17|T]) ->
    ok;
pat_loop4(I,[16|T]) ->
    ok;
pat_loop4(I,[15|T]) ->
    ok;
pat_loop4(I,[14|T]) ->
    ok;
pat_loop4(I,[13|T]) ->
    ok;
pat_loop4(I,[12|T]) ->
    ok;
pat_loop4(I,[11|T]) ->
    ok;
pat_loop4(I,[10|T]) ->
    ok;
pat_loop4(I,[9|T]) ->
    ok;
pat_loop4(I,[8|T]) ->
    ok;
pat_loop4(I,[7|T]) ->
    ok;
pat_loop4(I,[6|T]) ->
    ok;
pat_loop4(I,[5|T]) ->
    ok;
pat_loop4(I,[4|T]) ->
    ok;
pat_loop4(I,[3|T]) ->
    ok;
pat_loop4(I,[1|T]) ->
    ok;
pat_loop4(I,[21|T]) ->
    ok;
pat_loop4(I,[22|T]) ->
    ok;
pat_loop4(I,[23|T]) ->
    ok;
pat_loop4(I,[24|T]) ->
    ok;
pat_loop4(I,[25|T]) ->
    ok;
pat_loop4(I,[26|T]) ->
    ok;
pat_loop4(I,[27|T]) ->
    ok;
pat_loop4(I,[0|T]) ->
    pat_loop4(I - 1,[0|T]).

pat_loop5(0,_) ->
    ok;
pat_loop5(I,[0,20|T]) ->
    ok;
pat_loop5(I,[0,19|T]) ->
    ok;
pat_loop5(I,[0,18|T]) ->
    ok;
pat_loop5(I,[0,17|T]) ->
    ok;
pat_loop5(I,[0,16|T]) ->
    ok;
pat_loop5(I,[0,15|T]) ->
    ok;
pat_loop5(I,[0,14|T]) ->
    ok;
pat_loop5(I,[0,13|T]) ->
    ok;
pat_loop5(I,[0,12|T]) ->
    ok;
pat_loop5(I,[0,11|T]) ->
    ok;
pat_loop5(I,[0,10|T]) ->
    ok;
pat_loop5(I,[0,9|T]) ->
    ok;
pat_loop5(I,[0,8|T]) ->
    ok;
pat_loop5(I,[0,7|T]) ->
    ok;
pat_loop5(I,[0,6|T]) ->
    ok;
pat_loop5(I,[0,1|T]) ->
    pat_loop5(I - 1,[0,1|T]).

trav(I) ->
    X = very_big(10),
    trav(I,X).

trav(0,_) ->
    0;
trav(I,T) ->
    do_trav(T),
    trav(I - 1,T).

do_trav(T) when tuple(T) ->
    tup_trav(T,1,1 + size(T));
do_trav([H|T]) ->
    do_trav(H) + do_trav(T);
do_trav(X) when integer(X) ->
    1;
do_trav(X) ->
    0.

tup_trav(T,P,P) ->
    0;
tup_trav(T,P,End) ->
    do_trav(element(P,T)) + tup_trav(T,P + 1,End).

port_io(I) ->
    Before = now(),
    Pps = make_port_pids(5,I),
    send_procs(Pps,go),
    After = now(),
    wait_for_pids(Pps),
    subtr(Before,After).

make_port_pids(0,_) ->
    [];
make_port_pids(NoPorts,J) ->
    [spawn(estone,ppp,[self(),J])|make_port_pids(NoPorts - 1,J)].

ppp(Top,I) ->
    P = open_port({spawn,cat},[]),
    Str = lists:duplicate(200,88),
    Cmd = {self(),{command,Str}},
    receive
        go ->
            ok
    end,
    ppp_loop(P,I,Cmd),
    Cmd2 = {self(),{command,"abcde"}},
    Res = ppp_loop(P,I,Cmd2),
    P ! {self(),close},
    receive
        {P,closed} ->
            closed
    end,
    Top ! {self(),Res}.

ppp_loop(P,0,_) ->
    ok;
ppp_loop(P,I,Cmd) ->
    P ! Cmd,
    receive
        {P,_} ->
            ppp_loop(P,I - 1,Cmd)
    end.

large_dataset_work(I) ->
    {Minus,Ps} = timer:tc(estone,mk_big_procs,[2]),
    trav(I),
    lists(I),
    send_procs(Ps,stop),
    Minus.

mk_big_procs(0) ->
    [];
mk_big_procs(I) ->
    [mk_big_proc()|mk_big_procs(I - 1)].

mk_big_proc() ->
    P = spawn(estone,big_proc,[]),
    P ! {self(),running},
    receive
        {P,yes} ->
            P
    end.

big_proc() ->
    X = very_big(50),
    Y = very_big(50),
    Z = very_big(50),
    receive
        {From,running} ->
            From ! {self(),yes}
    end,
    receive
        stop ->
            {X,Y,Z}
    end.

large_local_dataset_work(I) ->
    {Minus,Data} = timer:tc(estone,very_big,[50]),
    trav(I),
    lists(I),
    Minus.

alloc(0) ->
    0;
alloc(I) ->
    X11 = very_big(),
    X12 = very_big(),
    X13 = very_big(),
    Z = [X14 = very_big(),X15 = very_big(),X16 = very_big()],
    X17 = very_big(),
    X18 = very_big(),
    X19 = very_big(),
    X20 = very_big(),
    X21 = very_big(),
    X22 = very_big(),
    X23 = very_big(),
    X24 = very_big(),
    alloc(I - 1).

bif_dispatch(0) ->
    0;
bif_dispatch(I) ->
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    disp(),
    bif_dispatch(I - 1).

disp() ->
    Tup = {a},
    L = [x],
    self(),
    self(),
    self(),
    self(),
    self(),
    self(),
    self(),
    self(),
    self(),
    make_ref(),
    atom_to_list(''),
    X = list_to_atom([]),
    tuple_to_list({}),
    X2 = list_to_tuple([]),
    element(1,Tup),
    element(1,Tup),
    Elem = element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    element(1,Tup),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    setelement(1,Tup,k),
    Y = setelement(1,Tup,k),
    Date = date(),
    time(),
    put(a,1),
    get(a),
    erase(a),
    hd(L),
    tl(L),
    Len = length(L),
    length(L),
    length(L),
    length(L),
    node(),
    node(),
    node(),
    node(),
    node(),
    node(),
    node(),
    node(),
    S = self(),
    node(S),
    node(S),
    node(S),
    size(Tup),
    W = whereis(code_server),
    whereis(code_server),
    whereis(code_server),
    whereis(code_server),
    whereis(code_server),
    whereis(code_server),
    W2 = whereis(code_server).

generic(I) ->
    register(funky,spawn(estone,gserv,[funky,estone,[],[]])),
    g_loop(I).

g_loop(0) ->
    exit(whereis(funky),kill),
    0;
g_loop(I) ->
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[xyz]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[xyz]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    estone:req(funky,{call,[abc]}),
    g_loop(I - 1).

req(Name,Req) ->
    R = make_ref(),
    Name ! {self(),R,Req},
    receive
        {Name,R,Reply} ->
            Reply
    after
        2000 ->
            exit(timeout)
    end.

gserv(Name,Mod,State,Debug) ->
    receive
        {From,Ref,{call,Req}} when Debug == [] ->
            case catch apply(Mod,handle_call,[From,State,Req]) of
                {reply,Reply,State2} ->
                    From ! {Name,Ref,Reply},
                    gserv(Name,Mod,State2,Debug);
                {noreply,State2} ->
                    gserv(Name,Mod,State2,Debug);
                {'EXIT',Reason} ->
                    exit(Reason)
            end;
        {From,Ref,Req} when Debug /= [] ->
            exit(nodebug)
    end.

handle_call(From,State,[xyz]) ->
    R = atom_to_list(xyz),
    {reply,R,[]};
handle_call(From,State,[abc]) ->
    R = 1 + 3,
    {reply,R,[R|State]}.

binary_h(I) ->
    Before = now(),
    P = spawn(estone,echo,[self()]),
    B = list_to_binary(lists:duplicate(2000,5)),
    After = now(),
    Compensate = subtr(Before,After),
    binary_h_2(I,P,B),
    Compensate.

binary_h_2(0,P,B) ->
    exit(P,kill);
binary_h_2(I,P,B) ->
    echo_loop(P,20,B),
    split_loop(B,{abc,1,2222,self(),"ancnd"},100),
    binary_h_2(I - 1,P,B).

split_loop(B,_,0) ->
    ok;
split_loop(B,Term,I) ->
    {X,Y} = split_binary(B,I),
    size(X),
    binary_to_list(Y,1,2),
    binary_to_term(term_to_binary(Term)),
    split_loop(B,Term,I - 1).

echo_loop(P,0,B) ->
    k;
echo_loop(P,I,B) ->
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    P ! B,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    receive
        _ ->
            ok
    end,
    echo_loop(P,I - 1,B).

ets(0) ->
    0;
ets(I) ->
    T1 = ets:new(a,[set]),
    T2 = ets:new(c,[bag,private]),
    L = [T1,T2],
    run_tabs(L,L,1),
    ets:delete(T1),
    ets:delete(T2),
    ets(I - 1).

run_tabs(_,_,0) ->
    ok;
run_tabs([],L,I) ->
    run_tabs(L,L,I - 1);
run_tabs([Tab|Tail],L,I) ->
    Begin = I * 20,
    End = (I + 1) * 20,
    run_tab(Tab,Begin,End,I),
    run_tabs(Tail,L,I).

run_tab(Tab,X,X,_) ->
    ok;
run_tab(Tab,Beg,End,J) ->
    ets:insert(Tab,{Beg,J}),
    ets:insert(Tab,{J,Beg}),
    ets:insert(Tab,{{foo,Beg},J}),
    ets:insert(Tab,{{foo,J},Beg}),
    ets:delete(Tab,haha),
    ets:match_delete(Tab,{k,j}),
    ets:match(Tab,{Beg,'$1'}),
    ets:match(Tab,{'$1',J}),
    ets:delete(Tab,Beg),
    K = ets:first(Tab),
    K2 = ets:next(Tab,K),
    run_tab(Tab,Beg + 1,End,J).

int_arith(0) ->
    0;
int_arith(I) ->
    do_arith(I) + do_arith(I) + do_arith(I) + do_arith(I) + do_arith(I) + do_arith(I) + do_arith(I) + do_arith(I) + do_arith(I) + 66,
    int_arith(I - 1).

do_arith(I) ->
    do_arith2(I) - do_arith2(I) - do_arith2(I) - do_arith2(I) - do_arith2(I) - do_arith2(I) - do_arith2(I) - 99.

do_arith2(I) ->
    X = 23,
    Y = 789 + I,
    Z = I + 1,
    U = (X bsl 1 bsr I) * X div 2 bsr 4,
    U1 = Z + Z + Z + Z + X bsl 4 * 2 bsl 2,
    Z - U + U1 div 2.

float_arith(0) ->
    0;
float_arith(I) ->
    f_do_arith(I) + f_do_arith(I) + f_do_arith(I) + f_do_arith(I) + f_do_arith(I) + f_do_arith(I) + f_do_arith(I) + f_do_arith(I) + f_do_arith(I) + 66,
    float_arith(I - 1).

f_do_arith(I) ->
    X = 23.4000,
    Y = 789.990 + I,
    Z = I + 1.88000,
    U = X * 1 / I * X / 2 * 4,
    U1 = Z + Z + Z + Z + X * 4 * 2 / 2,
    Z - U + U1 / 2.

fcalls(0) ->
    0;
fcalls(I) ->
    local0(400),
    remote0(400),
    app0(400),
    local1(400),
    remote1(400),
    app1(400),
    fcalls(I - 1).

local0(0) ->
    0;
local0(N) ->
    local0(N - 1).

local1(0) ->
    0;
local1(N) ->
    1 + local1(N - 1).

remote0(0) ->
    0;
remote0(N) ->
    estone:remote0(N - 1).

remote1(0) ->
    0;
remote1(N) ->
    1 + estone:remote1(N - 1).

app0(0) ->
    0;
app0(N) ->
    apply(estone,app0,[N - 1]).

app1(0) ->
    0;
app1(N) ->
    1 + apply(estone,app1,[N - 1]).

timer(I) ->
    L = [50,50,50,100,1000,3000,8000,50000,100000],
    timer(I,L).

timer(0,_) ->
    0;
timer(N,L) ->
    send_self(100),
    recv(100,L,L),
    timer(N - 1).

recv(0,_,_) ->
    ok;
recv(N,[],L) ->
    recv(N,L,L);
recv(N,[Timeout|Tail],L) ->
    receive
        hi_dude ->
            recv(N - 1,Tail,L)
    after
        Timeout ->
            io:format("XXXXX this wasn't supposed to happen???~n",[]),
            ok
    end.

send_self(0) ->
    ok;
send_self(N) ->
    self() ! hi_dude,
    send_self(N - 1).

links(I) ->
    L = mk_link_procs(100),
    send_procs(L,{procs,L,I}),
    wait_for_pids(L),
    0.

mk_link_procs(0) ->
    [];
mk_link_procs(I) ->
    [spawn_link(estone,lproc,[self()])|mk_link_procs(I - 1)].

lproc(Top) ->
    receive
        {procs,Procs,I} ->
            Top ! {self(),lproc(Procs,Procs,link,I)}
    end.

lproc(_,_,_,0) ->
    done;
lproc([],Procs,link,I) ->
    lproc(Procs,Procs,unlink,I - 1);
lproc([],Procs,unlink,I) ->
    lproc(Procs,Procs,link,I - 1);
lproc([Pid|Tail],Procs,unlink,I) ->
    unlink(Pid),
    lproc(Tail,Procs,unlink,I);
lproc([Pid|Tail],Procs,link,I) ->
    link(Pid),
    lproc(Tail,Procs,unlink,I).

echo(Pid) ->
    receive
        X ->
            Pid ! X,
            echo(Pid)
    end.

very_big() ->
    very_big(2).

very_big(0) ->
    [];
very_big(I) ->
    {1,2,3,a,v,f,r,t,y,u,self(),self(),self(),"22222222222222222",{{"234",self()}},[[very_big(I - 1)]]}.

big() ->
    {self(),funky_stuff,baby,{1,[123,true,[]],"abcdef"}}.

small() ->
    {self(),true}.

wait_for_pids([]) ->
    ok;
wait_for_pids([P|Tail]) ->
    receive
        {P,Res} ->
            wait_for_pids(Tail)
    end.

send_procs([P|Tail],Msg) ->
    P ! Msg,
    send_procs(Tail,Msg);
send_procs([],_) ->
    ok.

loopt() ->
    Before = now(),
    loopt(10000),
    After = now(),
    subtr(Before,After) / 10000.

loopt(0) ->
    ok;
loopt(I) ->
    loopt(I - 1).

determine_loop_data() ->
    determine_loop_data(1).

determine_loop_data(I) ->
    Mics = micros(),
    ppld([Mics|determine_loop_data(I,Mics)]).

determine_loop_data(0,Ms) ->
    [];
determine_loop_data(I,Ms) ->
    io:format("Determining loop counts \n",[]),
    check_sum_micros(Ms,0),
    Tmp = tt1(Ms),
    Ms2 = set_loops(Tmp),
    [Ms2|determine_loop_data(I - 1,Ms2)].

ppld(Mss) ->
    ppld(micros(),Mss).

ppld([],_) ->
    ok;
ppld([M|Tail],Mss) ->
    Str = element(6,M),
    Loops = find_loops(element(2,M),Mss),
    io:format("~-20s  Adjusted Mean: ~-5w   ~w~n",[Str,mean(Loops),Loops]),
    ppld(Tail,Mss).

mean(L0) ->
    L1 = lists:sort(L0),
    L = lists:delete(lists:last(L1),tl(L1)),
    round(sum(L) / length(L)).

sum([]) ->
    0;
sum([H|T]) ->
    H + sum(T).

find_loops(Fn,[MicroList|Tail]) ->
    {ok,M} = find_micro(Fn,MicroList),
    [element(4,M)|find_loops(Fn,Tail)];
find_loops(_,[]) ->
    [].

set_loops([M|Tail]) ->
    TT1 = element(5,M),
    MicroTime = element(3,M) * (3000 * 1000 * 100) * 0.01,
    Loops = round(MicroTime / TT1),
    if
        Loops == element(4,M) ->
            io:format("~s is the same~n",[element(6,M)]);
        true ->
            io:format("Adjusting ~s loop val from ~w to ~w~n",[element(6,M),element(4,M),Loops])
    end,
    [setelement(4,M,Loops)|set_loops(Tail)];
set_loops([]) ->
    [].

check_sum_micros([],100) ->
    ok;
check_sum_micros([],I) ->
    io:format("Micros don't sum up ~w~n",[I]);
check_sum_micros([H|T],Ack) ->
    check_sum_micros(T,Ack + element(3,H)).

tt1([M|Tail]) ->
    X = run_micro(setelement(4,M,500)),
    {value,{_,Ms}} = lists:keysearch(microsecs,1,X),
    New = setelement(5,M,Ms / 500),
    [New|tt1(Tail)];
tt1([]) ->
    io:nl(),
    [].

mk_fname([]) ->
    [];
mk_fname([H|T]) when atom(H) ->
    lists:concat([atom_to_list(H),"-",mk_fname(T)]);
mk_fname([H|T]) ->
    lists:concat([H,"-",mk_fname(T)]).

%record_index(R,F) ->
%    exit({badfield,R,F}).

