-module(huff2).
-export([run/0,
	pack/1,
	unpack/1]).


run() ->
   %garbage_collect(),
   %timing:reset(),
   { _, S, _} = now(),
   pack_unpack("uudecode.doc"),
   { _, E, _} = now(),
   (E - S).                                                      
   %{huff, timing:time(), timing:gc_time()}.



pack_unpack(FileName) ->
   {ok, Binary} = file:read_file(FileName),
   Data = binary_to_list(Binary),
   OrgSize = length(Data),
   FT = build_freq_trees(Data),
   Trees = sort_trees(FT, []),
   CodeTree = build_code_tree(Trees),
   Codes = flatten(make_codes(CodeTree, [])),
   Bits = flatten(pack_data(Codes, Data)),
   Bytes = bits_to_bytes(Bits),
   unpack(OrgSize, CodeTree, bytes_to_bits(Bytes)).



%
% Returns a tuple {PackedData, HuffmanTree}
%

pack(ByteList) ->
   FT = build_freq_trees(ByteList),
   Trees = sort_trees(FT, []),
   CodeTree = build_code_tree(Trees),
   Codes = flatten(make_codes(CodeTree, [])),
   Bits = flatten(pack_data(Codes, ByteList)),
   Bytes = bits_to_bytes(Bits),
   {Bytes, CodeTree}.


%
% Unpacks the result from pack/1.
%

unpack({PackedData, CodeTree}) ->
   {OrgLength, _, _} = CodeTree,
   unpack(OrgLength, CodeTree, bytes_to_bits(PackedData)).

unpack(0, CodeTree, Bits) ->
%   io:fwrite("unpack/1.1\n"),
   [];
unpack(Size, CodeTree, Bits) ->
%   io:fwrite("unpack/1.2\n"),
   {Byte, RestBits} = find_byte(CodeTree, Bits),
   [Byte | unpack(Size-1, CodeTree, RestBits)].



find_byte({_, leaf, Byte}, Bits) ->
%   io:fwrite("find_byte/1.1\n"),
   {Byte, Bits};
find_byte({_, L, R}, [1|Bits]) ->
%   io:fwrite("find_byte/1.2\n"),
   find_byte(L, Bits);
find_byte({_, L, R}, [0|Bits]) ->
%   io:fwrite("find_byte/1.3\n"),
   find_byte(R, Bits).



bytes_to_bits([]) ->
%   io:fwrite("bytes_to_bits/1.1\n"),
    [];
bytes_to_bits([Byte|Bytes]) ->
%   io:fwrite("bytes_to_bits/1.2\n"),
    B7 = Byte div 128,
    B6 = (Byte-128*B7) div 64,
    B5 = (Byte-128*B7-64*B6) div 32,
    B4 = (Byte-128*B7-64*B6-32*B5) div 16,
    B3 = (Byte-128*B7-64*B6-32*B5-16*B4) div 8,
    B2 = (Byte-128*B7-64*B6-32*B5-16*B4-8*B3) div 4,
    B1 = (Byte-128*B7-64*B6-32*B5-16*B4-8*B3-4*B2) div 2,
    B0 = (Byte-128*B7-64*B6-32*B5-16*B4-8*B3-4*B2-2*B1),
    [B7,B6,B5,B4,B3,B2,B1,B0 | bytes_to_bits(Bytes)].


bits_to_bytes([B7, B6, B5, B4, B3, B2, B1, B0 | Rest]) ->
%   io:fwrite("bits_to_bytes/1.1\n"),
    [(B7*128+B6*64+B5*32+B4*16+B3*8+B2*4+B1*2+B0) | bits_to_bytes(Rest)];
bits_to_bytes([B7, B6, B5, B4, B3, B2, B1]) ->
%   io:fwrite("bits_to_bytes/1.2\n"),
    [B7*128+B6*64+B5*32+B4*16+B3*8+B2*4+B1*2];
bits_to_bytes([B7, B6, B5, B4, B3, B2]) ->
%   io:fwrite("bits_to_bytes/1.3\n"),
    [B7*128+B6*64+B5*32+B4*16+B3*8+B2*4];
bits_to_bytes([B7, B6, B5, B4, B3]) ->
%   io:fwrite("bits_to_bytes/1.4\n"),
    [B7*128+B6*64+B5*32+B4*16+B3*8];
bits_to_bytes([B7, B6, B5, B4]) ->
%   io:fwrite("bits_to_bytes/1.5\n"),
    [B7*128+B6*64+B5*32+B4*16];
bits_to_bytes([B7, B6, B5]) ->
%   io:fwrite("bits_to_bytes/1.6\n"),
    [B7*128+B6*64+B5*32];
bits_to_bytes([B7, B6]) ->
%   io:fwrite("bits_to_bytes/1.7\n"),
    [B7*128+B6*64];
bits_to_bytes([B7]) ->
%   io:fwrite("bits_to_bytes/1.8\n"),
    [B7*128];
bits_to_bytes([]) ->
%   io:fwrite("bits_to_bytes/1.9\n"),
    [].



pack_data(Codes, []) ->
%   io:fwrite("pack_data/2.1\n"),
   [];
pack_data(Codes, [Byte|Rest]) ->
%   io:fwrite("pack_data/2.2\n"),
   [get_code(Byte, Codes) | pack_data(Codes, Rest)].



get_code(Index, []) ->
%   io:fwrite("pack_data/2.1\n"),
   error;
get_code(Index, [{I, Bits}|_]) when Index == I ->
%   io:fwrite("pack_data/2.2\n"),
   Bits;
get_code(Index, [_|Rest]) ->
%   io:fwrite("pack_data/2.3\n"),
   get_code(Index, Rest).



make_code_index([], Done) ->
   Done;
make_code_index([C|Rest], Done) ->
   make_code_index(Rest, insert_code(C, Done)).



insert_code(C, []) ->
   [C];
insert_code({V1, B1}, [{V2, B2}|Rest]) when V1 < V2 ->
   [{V1, B1}, {V2, B2}|Rest];
insert_code({V1, B1}, [{V2, B2}|Rest]) ->
   [{V2, B2} | insert_code({V1, B1}, Rest)].



make_codes({_, leaf, Byte}, Bits) ->
%   io:fwrite("make_codes/2.1\n"),
    {Byte, reverse(Bits)};
make_codes({_, R, L}, Bits) ->
%   io:fwrite("make_codes/2.2\n"),
    [make_codes(R, [1|Bits]), make_codes(L, [0|Bits])].


%
% Make one huffman tree out of a list of trees.
%

build_code_tree([Tree]) ->
%   io:fwrite("build_code_tree/1.1\n"),
   Tree;
build_code_tree([{Val1, R1, L1}, {Val2, R2, L2} | Rest]) ->
%   io:fwrite("build_code_tree/1.2\n"),
   build_code_tree(insert_tree({Val1+Val2, {Val1, R1, L1}, {Val2, R2, L2}},
			       Rest)).


%
% Sort a list of leaves so that those with least frequence is first.
% Nodes with frequence 0 are removed.
%

sort_trees([], Sorted) ->
%   io:fwrite("sort_trees/2.1\n"),
   Sorted;
sort_trees([T|Trees], Sorted) ->
%   io:fwrite("sort_trees/2.2\n"),
   sort_trees(Trees, insert_tree(T, Sorted)).


%
% Insert a tree in a sorted list (least frequencey first)
%

insert_tree({0, _, _}, []) ->
%   io:fwrite("insert_tree/2.1\n"),
   [];
insert_tree(Tree, []) ->
%   io:fwrite("insert_tree/2.2\n"),
   [Tree];
insert_tree({0, _, _}, Trees)  ->
%   io:fwrite("insert_tree/2.3\n"),
   Trees;
insert_tree({Val1, R1, L1}, [{Val2, R2, L2}|Rest]) when Val1 < Val2 ->
%   io:fwrite("insert_tree/2.4\n"),
   [{Val1, R1, L1}, {Val2, R2, L2}|Rest];
insert_tree(T1, [T2|Rest]) ->
%   io:fwrite("insert_tree/2.5\n"),
   [T2|insert_tree(T1, Rest)].



%
% Makes a list of 256 leaves each containing the frequency of a bytecode.
%

build_freq_trees(Data) ->
%   io:fwrite("build_freq_table/1\n"),
   build_freq_table(Data, 0).

build_freq_table(_, 256) ->
   [];
build_freq_table(Data, X) ->
%   io:fwrite("build_freq_table/2.3\n"),
   [{occurs(X, Data, 0), leaf, X} | build_freq_table(Data, X+1)].


occurs(X, [], Ack) ->
   Ack;
occurs(X, [Y|Rest], Ack) when X == Y ->
%   io:fwrite("occurs/3.2\n"),
   occurs(X, Rest, Ack+1);
occurs(X, [_|Rest],Ack) ->
%   io:fwrite("occurs/3.3\n"),
   occurs(X, Rest, Ack).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%5
%
% Some utilities
%

reverse(X) ->
%   io:fwrite("reverse/1.1\n"),
   reverse(X, []).
reverse([H|T], Y) ->
%   io:fwrite("reverse/1.2\n"),
   reverse(T, [H|Y]);
reverse([], X) ->
%   io:fwrite("reverse/1.3\n"),
   X.



append([H|T], Z) ->
   [H|append(T, Z)];
append([], X) ->
   X.



flatten(List) ->
%   io:fwrite("flatten/1.1\n"),
   flatten(List, []).
flatten([H|T], Cont) when list(H) ->
%   io:fwrite("flatten/1.2\n"),
   flatten(H, [T|Cont]);
flatten([H|T], Cont) ->
%   io:fwrite("flatten/1.3\n"),
   [H|flatten(T, Cont)];
flatten([], [H|Cont]) ->
%   io:fwrite("flatten/1.4\n"),
   flatten(H, Cont);
flatten([], []) ->
%   io:fwrite("flatten/1.5\n"),
   [].

