%    -*- Erlang -*- 
%    File:	camera.erl  (~jb/projects/erl_tracer/camera.erl)
%    Author:	Johan Bevemyr
%    Created:	Tue Sep 12 21:41:06 1995
%    Purpose:   Implement the camera abstract data structure.
 
-module(camera).
-export([camera_init/1,camera_ray/3,
	 camera_size/1,camera_default/0,
	 default_camera/0]).
-compile(export_all).

-import(vector,[vector_sub/2,vector_add/2,vector_cross/2,
		vector_scale/2,vector_norm/1]).



%
%    The uninitialised camera is described by a tupple of
%
%       Eye point  (location of camera)
%       Look point (the point the camera is directed towards)
%       Up vector  (the up direction the the picture)
%       Zoom       (a zoom value)
%       Width      (the width of the picture in world units)
%       Height     (the height of the picture in world units)
%       Res        (the distance between two points in the picture)
%
%    The initialised camera is represented as a tupple of 
%
%       Eye point (the location of the camera)
%       Corner    (the location of the left upper corner of the picture)
%       Res       (the resolution of the camera)
%       Right     (the unit right vector in the picture)
%       Down      (the unit down vector in the picture)
%       Width     (the width of the picture in world units)
%       Height    (the height of the picture in world units)
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Name: camera_init({Eye,Look,Up,Zoom,Width,Height,Res})
%
% Desc: Returns an initialised camera structure.
%

camera_init({Eye,Look,Up,Zoom,Width,Height,Res}) ->
    C = vector_norm(vector_sub(Look,Eye)),
    Right = vector_norm(vector_cross(Up,C)),
    Down = vector_norm(vector_cross(Right,C)),
    Corner = vector_sub(vector_sub(vector_add(Eye,vector_scale(C,Zoom)),
				   vector_scale(Right,Width/2)),
			vector_scale(Down,Height/2)),
    {Eye,Corner,Res,Right,Down,Width,Height}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Name: camera_ray(X,Y,Camera)
%
% Desc: returns the ray going through pixel (X,Y) in the picture.
%

camera_ray(X,Y,{Eye,Corner,Res,Right,Down,Width,Height}) ->
    Origin = Eye,
    P = vector_add(vector_add(Corner, vector_scale(Right,Res*X)),
		   vector_scale(Down,Res*Y)),
    Direction = vector_norm(vector_sub(P,Eye)),
    {Origin,Direction}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Name: camera_size(Camera)
%
% Desc: returns a tupple of the width and the height of the
%       picture in pixels
%

camera_size({Eye,Corner,Res,Right,Down,Width,Height}) ->
    {trunc(Width/Res),trunc(Height/Res)}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% 
% Name: camera_default(), default_camera()
%
% Desc: returns the default camera
%

camera_default() ->  {{0,0,-8},{0,0,0},{0,1,0},10,4,4,0.1}.

default_camera() -> camera_default().
