-module(lexer).

-export([string/1,string/2,token/2,token/3,tokens/2,tokens/3]).

%% Included code.

fix_string([First|Rest]) ->
	fix_end(Rest).

fix_end([Rest]) ->
	[];

fix_end([C|Rest]) ->
	[C|fix_end(Rest)].

%% Interface functions.

%% string(String)
%% string(String, StartLine)
%%  Scan all the tokens in String.

string(String) ->
    string(String, 1, []).

string(String, Line) ->
    string(String, Line, []).

string([], Line, Ts) ->
    yyrev(Ts, []);
string(Ics0, Line0, Ts) ->
    case yystate(yystate(), Ics0, Line0, [], reject) of
	{accept,Ics,Line,Tcs,A} ->
	    string(Ics, Line, yyaction(A, Tcs, Line0, Ts));
	{{AIcs,ATcs,A},CurS,Ics,Line,Tcs} ->
	    string(AIcs, Line, yyaction(A, ATcs, Line0, Ts));
	{reject,CurS,[],Line,[]} ->		%Reached the end
	    yyrev(Ts, []);
	{reject,CurS,[],Line,Tcs} ->
	    yyrev(Ts, [{error,yyrev(Tcs)}]);
	{reject,CurS,[C|Ics],Line,[]} ->	%Bad character
	    string(Ics, Line, [{error,[C]}|Ts]);
	{reject,CurS,[C|Ics],Line,Tcs} ->	%Bad sequence
	    string([C|Ics], Line, [{error,yyrev(Tcs)}|Ts])
    end.

%% token(Continuation, Chars)
%% token(Continuation, Chars, StartLine)
%%  Reentrant function to scan one token. Intended to be used within
%%  io system. If we could not detect the end of a partial token then
%%  request more characters and try again.

token(Cont, Chars) ->
    token(Cont, Chars, 1).

token([], Chars, Line) ->		%First call
    token(yystate(), Chars, Line, [], reject);
token({State,Line,Tcs,Accept}, Chars, _) ->	%Subsequent calls
    token(State, Chars, Line, Tcs, Accept).

token(State, Ics0, Line0, Tcs0, Accept0) ->
    case yystate(State, Ics0, Line0, Tcs0, Accept0) of
	{accept,eof,Line,Tcs,A} ->
	    token_cont(yyaction(A, Tcs, Line0, []), [], Line);
	{accept,Ics,Line,Tcs,A} ->
	    token_cont(yyaction(A, Tcs, Line0, []), Ics, Line);
	{{AIcs,ATcs,A},CurS,[C|Ics],Line,Tcs} ->	%Could detect end
	    token_cont(yyaction(A, ATcs, Line, []), AIcs, Line);
	{{AIcs,ATcs,A},CurS,[],Line,Tcs} ->		%Could not detect end
	    {more,{CurS,Line,Tcs,{AIcs,ATcs,A}}};
	{{AIcs,ATcs,A},CurS,eof,Line,Tcs} ->		%Could detect end
	    token_cont(yyaction(A, ATcs, Line0, []), AIcs, Line);
	{reject,CurS,[C|Ics],Line,[]} ->
	    {done,{error,[C]},Ics};
	{reject,CurS,[C|Ics],Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[C|Ics]};
	{reject,CurS,[],Line,Tcs} ->			%Could not detect end
	    {more,{CurS,Line,Tcs,reject}};
	{reject,CurS,eof,Line,[]} ->
	    {done,eof,[]};
	{reject,CurS,eof,Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[]}
    end.

token_cont([Tok|_], Rest, Line) ->
    {done,{ok,Tok,Line},Rest};
token_cont([], Rest, Line) ->
    token(yystate(), Rest, Line, [], reject).

%% tokens(Continuation, Chars)
%% tokens(Continuation, Chars, StartLine)
%%  Reentrant function to scan tokens until a '$end' token. Intended to
%%  be used within io system.

tokens(Cont, Chars) ->
    tokens(Cont, Chars, 1).

tokens([], Chars, Line) ->				%First call
    tokens(yystate(), Chars, Line, [], reject, []);
tokens({State,Line,Tcs,Accept,Ts}, Chars, _) ->	%Subsequent calls
    tokens(State, Chars, Line, Tcs, Accept, Ts).

tokens(State, Ics0, Line0, Tcs0, Accept0, Ts) ->
    case yystate(State, Ics0, Line0, Tcs0, Accept0) of
	{accept,eof,Line,Tcs,A} ->
	    tokens_cont(yyaction(A, Tcs, Line0, Ts), [], Line);
	{accept,Ics,Line,Tcs,A} ->
	    tokens_cont(yyaction(A, Tcs, Line0, Ts), Ics, Line);
	{{AIcs,ATcs,A},CurS,[C|Ics],Line,Tcs} ->	%Could detect end
	    tokens_cont(yyaction(A, ATcs, Line0, Ts), AIcs, Line);
	{{AIcs,ATcs,A},CurS,[],Line,Tcs} ->		%Could not detect end
	    {more,{CurS,Line,Tcs,{AIcs,ATcs,A},Ts}};
	{{AIcs,ATcs,A},CurS,eof,Line,Tcs} ->		%Could detect end
	    tokens_cont(yyaction(A, ATcs,Line0, Ts), AIcs, Line);
	{reject,CurS,[C|Ics],Line,[]} ->
	    {done,{error,[C]},Ics};
	{reject,CurS,[C|Ics],Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[C|Ics]};
	{reject,CurS,[],Line,Tcs} ->			%Could not detect end
	    {more,{CurS,Line,Tcs,reject,Ts}};
	{reject,CurS,eof,Line,[]} ->
	    {done,eof,[]};
	{reject,CurS,eof,Line,Tcs} ->
	    {done,{error,yyrev(Tcs)},[]}
    end.

tokens_cont(['$end'|Toks], Rest, Line) ->
    {done,{ok,yyrev(Toks, ['$end']),Line},Rest};
tokens_cont([{'$end',L}|Toks], Rest, Line) ->
    {done,{ok,yyrev(Toks, [{'$end',L}]),Line},Rest};
tokens_cont(Toks, Rest, Line) ->
    tokens(yystate(), Rest, Line, [], reject, Toks).

%% Local utility functions.

yyrev(L) ->
    yyrev(L, []).

yyrev([H|T], Stack) ->
    yyrev(T, [H|Stack]);
yyrev([], Stack) -> Stack.

%% Generated DFA tables and actions.
yystate() -> 0.

%% yystate(State, InChars, LineNo, TokenChars, AcceptState)

yystate(0, [$\t|Ics], Line, Tcs, Accept) ->
    yystate(194, Ics, Line, [$\t|Tcs], Accept);
yystate(0, [$\n|Ics], Line, Tcs, Accept) ->
    yystate(193, Ics, Line+1, [$\n|Tcs], Accept);
yystate(0, [$ |Ics], Line, Tcs, Accept) ->
    yystate(193, Ics, Line, [$ |Tcs], Accept);
yystate(0, [$#|Ics], Line, Tcs, Accept) ->
    yystate(192, Ics, Line, [$#|Tcs], Accept);
yystate(0, [$'|Ics], Line, Tcs, Accept) ->
    yystate(190, Ics, Line, [$'|Tcs], Accept);
yystate(0, [$(|Ics], Line, Tcs, Accept) ->
    yystate(189, Ics, Line, [$(|Tcs], Accept);
yystate(0, [$)|Ics], Line, Tcs, Accept) ->
    yystate(188, Ics, Line, [$)|Tcs], Accept);
yystate(0, [$,|Ics], Line, Tcs, Accept) ->
    yystate(187, Ics, Line, [$,|Tcs], Accept);
yystate(0, [$-|Ics], Line, Tcs, Accept) ->
    yystate(183, Ics, Line, [$-|Tcs], Accept);
yystate(0, [$.|Ics], Line, Tcs, Accept) ->
    yystate(181, Ics, Line, [$.|Tcs], Accept);
yystate(0, [$[|Ics], Line, Tcs, Accept) ->
    yystate(179, Ics, Line, [$[|Tcs], Accept);
yystate(0, [$]|Ics], Line, Tcs, Accept) ->
    yystate(178, Ics, Line, [$]|Tcs], Accept);
yystate(0, [$b|Ics], Line, Tcs, Accept) ->
    yystate(163, Ics, Line, [$b|Tcs], Accept);
yystate(0, [$c|Ics], Line, Tcs, Accept) ->
    yystate(148, Ics, Line, [$c|Tcs], Accept);
yystate(0, [$d|Ics], Line, Tcs, Accept) ->
    yystate(136, Ics, Line, [$d|Tcs], Accept);
yystate(0, [$e|Ics], Line, Tcs, Accept) ->
    yystate(122, Ics, Line, [$e|Tcs], Accept);
yystate(0, [$h|Ics], Line, Tcs, Accept) ->
    yystate(116, Ics, Line, [$h|Tcs], Accept);
yystate(0, [$l|Ics], Line, Tcs, Accept) ->
    yystate(103, Ics, Line, [$l|Tcs], Accept);
yystate(0, [$m|Ics], Line, Tcs, Accept) ->
    yystate(100, Ics, Line, [$m|Tcs], Accept);
yystate(0, [$n|Ics], Line, Tcs, Accept) ->
    yystate(94, Ics, Line, [$n|Tcs], Accept);
yystate(0, [$p|Ics], Line, Tcs, Accept) ->
    yystate(82, Ics, Line, [$p|Tcs], Accept);
yystate(0, [$r|Ics], Line, Tcs, Accept) ->
    yystate(45, Ics, Line, [$r|Tcs], Accept);
yystate(0, [$s|Ics], Line, Tcs, Accept) ->
    yystate(34, Ics, Line, [$s|Tcs], Accept);
yystate(0, [$t|Ics], Line, Tcs, Accept) ->
    yystate(26, Ics, Line, [$t|Tcs], Accept);
yystate(0, [$u|Ics], Line, Tcs, Accept) ->
    yystate(22, Ics, Line, [$u|Tcs], Accept);
yystate(0, [$w|Ics], Line, Tcs, Accept) ->
    yystate(17, Ics, Line, [$w|Tcs], Accept);
yystate(0, [$x|Ics], Line, Tcs, Accept) ->
    yystate(11, Ics, Line, [$x|Tcs], Accept);
yystate(0, [$y|Ics], Line, Tcs, Accept) ->
    yystate(5, Ics, Line, [$y|Tcs], Accept);
yystate(0, [$z|Ics], Line, Tcs, Accept) ->
    yystate(1, Ics, Line, [$z|Tcs], Accept);
yystate(0, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(180, Ics, Line, [C|Tcs], Accept);
yystate(0, Ics, Line, Tcs, Accept) ->
    {Accept,0,Ics,Line,Tcs};
yystate(1, [$o|Ics], Line, Tcs, Accept) ->
    yystate(2, Ics, Line, [$o|Tcs], Accept);
yystate(1, Ics, Line, Tcs, Accept) ->
    {Accept,1,Ics,Line,Tcs};
yystate(2, [$o|Ics], Line, Tcs, Accept) ->
    yystate(3, Ics, Line, [$o|Tcs], Accept);
yystate(2, Ics, Line, Tcs, Accept) ->
    {Accept,2,Ics,Line,Tcs};
yystate(3, [$m|Ics], Line, Tcs, Accept) ->
    yystate(4, Ics, Line, [$m|Tcs], Accept);
yystate(3, Ics, Line, Tcs, Accept) ->
    {Accept,3,Ics,Line,Tcs};
yystate(4, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,13};
yystate(5, [$-|Ics], Line, Tcs, Accept) ->
    yystate(6, Ics, Line, [$-|Tcs], Accept);
yystate(5, Ics, Line, Tcs, Accept) ->
    {Accept,5,Ics,Line,Tcs};
yystate(6, [$a|Ics], Line, Tcs, Accept) ->
    yystate(7, Ics, Line, [$a|Tcs], Accept);
yystate(6, Ics, Line, Tcs, Accept) ->
    {Accept,6,Ics,Line,Tcs};
yystate(7, [$x|Ics], Line, Tcs, Accept) ->
    yystate(8, Ics, Line, [$x|Tcs], Accept);
yystate(7, Ics, Line, Tcs, Accept) ->
    {Accept,7,Ics,Line,Tcs};
yystate(8, [$i|Ics], Line, Tcs, Accept) ->
    yystate(9, Ics, Line, [$i|Tcs], Accept);
yystate(8, Ics, Line, Tcs, Accept) ->
    {Accept,8,Ics,Line,Tcs};
yystate(9, [$s|Ics], Line, Tcs, Accept) ->
    yystate(10, Ics, Line, [$s|Tcs], Accept);
yystate(9, Ics, Line, Tcs, Accept) ->
    {Accept,9,Ics,Line,Tcs};
yystate(10, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,21};
yystate(11, [$-|Ics], Line, Tcs, Accept) ->
    yystate(12, Ics, Line, [$-|Tcs], Accept);
yystate(11, Ics, Line, Tcs, Accept) ->
    {Accept,11,Ics,Line,Tcs};
yystate(12, [$a|Ics], Line, Tcs, Accept) ->
    yystate(13, Ics, Line, [$a|Tcs], Accept);
yystate(12, Ics, Line, Tcs, Accept) ->
    {Accept,12,Ics,Line,Tcs};
yystate(13, [$x|Ics], Line, Tcs, Accept) ->
    yystate(14, Ics, Line, [$x|Tcs], Accept);
yystate(13, Ics, Line, Tcs, Accept) ->
    {Accept,13,Ics,Line,Tcs};
yystate(14, [$i|Ics], Line, Tcs, Accept) ->
    yystate(15, Ics, Line, [$i|Tcs], Accept);
yystate(14, Ics, Line, Tcs, Accept) ->
    {Accept,14,Ics,Line,Tcs};
yystate(15, [$s|Ics], Line, Tcs, Accept) ->
    yystate(16, Ics, Line, [$s|Tcs], Accept);
yystate(15, Ics, Line, Tcs, Accept) ->
    {Accept,15,Ics,Line,Tcs};
yystate(16, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,20};
yystate(17, [$i|Ics], Line, Tcs, Accept) ->
    yystate(18, Ics, Line, [$i|Tcs], Accept);
yystate(17, Ics, Line, Tcs, Accept) ->
    {Accept,17,Ics,Line,Tcs};
yystate(18, [$d|Ics], Line, Tcs, Accept) ->
    yystate(19, Ics, Line, [$d|Tcs], Accept);
yystate(18, Ics, Line, Tcs, Accept) ->
    {Accept,18,Ics,Line,Tcs};
yystate(19, [$t|Ics], Line, Tcs, Accept) ->
    yystate(20, Ics, Line, [$t|Tcs], Accept);
yystate(19, Ics, Line, Tcs, Accept) ->
    {Accept,19,Ics,Line,Tcs};
yystate(20, [$h|Ics], Line, Tcs, Accept) ->
    yystate(21, Ics, Line, [$h|Tcs], Accept);
yystate(20, Ics, Line, Tcs, Accept) ->
    {Accept,20,Ics,Line,Tcs};
yystate(21, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,14};
yystate(22, [$p|Ics], Line, Tcs, Accept) ->
    yystate(25, Ics, Line, [$p|Tcs], Accept);
yystate(22, [$s|Ics], Line, Tcs, Accept) ->
    yystate(23, Ics, Line, [$s|Tcs], Accept);
yystate(22, Ics, Line, Tcs, Accept) ->
    {Accept,22,Ics,Line,Tcs};
yystate(23, [$e|Ics], Line, Tcs, Accept) ->
    yystate(24, Ics, Line, [$e|Tcs], Accept);
yystate(23, Ics, Line, Tcs, Accept) ->
    {Accept,23,Ics,Line,Tcs};
yystate(24, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,25};
yystate(25, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,17};
yystate(26, [$r|Ics], Line, Tcs, Accept) ->
    yystate(27, Ics, Line, [$r|Tcs], Accept);
yystate(26, Ics, Line, Tcs, Accept) ->
    {Accept,26,Ics,Line,Tcs};
yystate(27, [$i|Ics], Line, Tcs, Accept) ->
    yystate(28, Ics, Line, [$i|Tcs], Accept);
yystate(27, Ics, Line, Tcs, Accept) ->
    {Accept,27,Ics,Line,Tcs};
yystate(28, [$a|Ics], Line, Tcs, Accept) ->
    yystate(29, Ics, Line, [$a|Tcs], Accept);
yystate(28, Ics, Line, Tcs, Accept) ->
    {Accept,28,Ics,Line,Tcs};
yystate(29, [$n|Ics], Line, Tcs, Accept) ->
    yystate(30, Ics, Line, [$n|Tcs], Accept);
yystate(29, Ics, Line, Tcs, Accept) ->
    {Accept,29,Ics,Line,Tcs};
yystate(30, [$g|Ics], Line, Tcs, Accept) ->
    yystate(31, Ics, Line, [$g|Tcs], Accept);
yystate(30, Ics, Line, Tcs, Accept) ->
    {Accept,30,Ics,Line,Tcs};
yystate(31, [$l|Ics], Line, Tcs, Accept) ->
    yystate(32, Ics, Line, [$l|Tcs], Accept);
yystate(31, Ics, Line, Tcs, Accept) ->
    {Accept,31,Ics,Line,Tcs};
yystate(32, [$e|Ics], Line, Tcs, Accept) ->
    yystate(33, Ics, Line, [$e|Tcs], Accept);
yystate(32, Ics, Line, Tcs, Accept) ->
    {Accept,32,Ics,Line,Tcs};
yystate(33, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,2};
yystate(34, [$h|Ics], Line, Tcs, Accept) ->
    yystate(40, Ics, Line, [$h|Tcs], Accept);
yystate(34, [$p|Ics], Line, Tcs, Accept) ->
    yystate(35, Ics, Line, [$p|Tcs], Accept);
yystate(34, Ics, Line, Tcs, Accept) ->
    {Accept,34,Ics,Line,Tcs};
yystate(35, [$h|Ics], Line, Tcs, Accept) ->
    yystate(36, Ics, Line, [$h|Tcs], Accept);
yystate(35, Ics, Line, Tcs, Accept) ->
    {Accept,35,Ics,Line,Tcs};
yystate(36, [$e|Ics], Line, Tcs, Accept) ->
    yystate(37, Ics, Line, [$e|Tcs], Accept);
yystate(36, Ics, Line, Tcs, Accept) ->
    {Accept,36,Ics,Line,Tcs};
yystate(37, [$r|Ics], Line, Tcs, Accept) ->
    yystate(38, Ics, Line, [$r|Tcs], Accept);
yystate(37, Ics, Line, Tcs, Accept) ->
    {Accept,37,Ics,Line,Tcs};
yystate(38, [$e|Ics], Line, Tcs, Accept) ->
    yystate(39, Ics, Line, [$e|Tcs], Accept);
yystate(38, Ics, Line, Tcs, Accept) ->
    {Accept,38,Ics,Line,Tcs};
yystate(39, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,1};
yystate(40, [$a|Ics], Line, Tcs, Accept) ->
    yystate(41, Ics, Line, [$a|Tcs], Accept);
yystate(40, Ics, Line, Tcs, Accept) ->
    {Accept,40,Ics,Line,Tcs};
yystate(41, [$d|Ics], Line, Tcs, Accept) ->
    yystate(42, Ics, Line, [$d|Tcs], Accept);
yystate(41, Ics, Line, Tcs, Accept) ->
    {Accept,41,Ics,Line,Tcs};
yystate(42, [$o|Ics], Line, Tcs, Accept) ->
    yystate(43, Ics, Line, [$o|Tcs], Accept);
yystate(42, Ics, Line, Tcs, Accept) ->
    {Accept,42,Ics,Line,Tcs};
yystate(43, [$w|Ics], Line, Tcs, Accept) ->
    yystate(44, Ics, Line, [$w|Tcs], Accept);
yystate(43, Ics, Line, Tcs, Accept) ->
    {Accept,43,Ics,Line,Tcs};
yystate(44, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,27};
yystate(45, [$a|Ics], Line, Tcs, Accept) ->
    yystate(77, Ics, Line, [$a|Tcs], Accept);
yystate(45, [$e|Ics], Line, Tcs, Accept) ->
    yystate(46, Ics, Line, [$e|Tcs], Accept);
yystate(45, Ics, Line, Tcs, Accept) ->
    {Accept,45,Ics,Line,Tcs};
yystate(46, [$c|Ics], Line, Tcs, Accept) ->
    yystate(70, Ics, Line, [$c|Tcs], Accept);
yystate(46, [$f|Ics], Line, Tcs, Accept) ->
    yystate(55, Ics, Line, [$f|Tcs], Accept);
yystate(46, [$s|Ics], Line, Tcs, Accept) ->
    yystate(47, Ics, Line, [$s|Tcs], Accept);
yystate(46, Ics, Line, Tcs, Accept) ->
    {Accept,46,Ics,Line,Tcs};
yystate(47, [$o|Ics], Line, Tcs, Accept) ->
    yystate(48, Ics, Line, [$o|Tcs], Accept);
yystate(47, Ics, Line, Tcs, Accept) ->
    {Accept,47,Ics,Line,Tcs};
yystate(48, [$l|Ics], Line, Tcs, Accept) ->
    yystate(49, Ics, Line, [$l|Tcs], Accept);
yystate(48, Ics, Line, Tcs, Accept) ->
    {Accept,48,Ics,Line,Tcs};
yystate(49, [$u|Ics], Line, Tcs, Accept) ->
    yystate(50, Ics, Line, [$u|Tcs], Accept);
yystate(49, Ics, Line, Tcs, Accept) ->
    {Accept,49,Ics,Line,Tcs};
yystate(50, [$t|Ics], Line, Tcs, Accept) ->
    yystate(51, Ics, Line, [$t|Tcs], Accept);
yystate(50, Ics, Line, Tcs, Accept) ->
    {Accept,50,Ics,Line,Tcs};
yystate(51, [$i|Ics], Line, Tcs, Accept) ->
    yystate(52, Ics, Line, [$i|Tcs], Accept);
yystate(51, Ics, Line, Tcs, Accept) ->
    {Accept,51,Ics,Line,Tcs};
yystate(52, [$o|Ics], Line, Tcs, Accept) ->
    yystate(53, Ics, Line, [$o|Tcs], Accept);
yystate(52, Ics, Line, Tcs, Accept) ->
    {Accept,52,Ics,Line,Tcs};
yystate(53, [$n|Ics], Line, Tcs, Accept) ->
    yystate(54, Ics, Line, [$n|Tcs], Accept);
yystate(53, Ics, Line, Tcs, Accept) ->
    {Accept,53,Ics,Line,Tcs};
yystate(54, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,16};
yystate(55, [$l|Ics], Line, Tcs, Accept) ->
    yystate(63, Ics, Line, [$l|Tcs], Accept);
yystate(55, [$r|Ics], Line, Tcs, Accept) ->
    yystate(56, Ics, Line, [$r|Tcs], Accept);
yystate(55, Ics, Line, Tcs, Accept) ->
    {Accept,55,Ics,Line,Tcs};
yystate(56, [$a|Ics], Line, Tcs, Accept) ->
    yystate(57, Ics, Line, [$a|Tcs], Accept);
yystate(56, Ics, Line, Tcs, Accept) ->
    {Accept,56,Ics,Line,Tcs};
yystate(57, [$c|Ics], Line, Tcs, Accept) ->
    yystate(58, Ics, Line, [$c|Tcs], Accept);
yystate(57, Ics, Line, Tcs, Accept) ->
    {Accept,57,Ics,Line,Tcs};
yystate(58, [$t|Ics], Line, Tcs, Accept) ->
    yystate(59, Ics, Line, [$t|Tcs], Accept);
yystate(58, Ics, Line, Tcs, Accept) ->
    {Accept,58,Ics,Line,Tcs};
yystate(59, [$i|Ics], Line, Tcs, Accept) ->
    yystate(60, Ics, Line, [$i|Tcs], Accept);
yystate(59, Ics, Line, Tcs, Accept) ->
    {Accept,59,Ics,Line,Tcs};
yystate(60, [$o|Ics], Line, Tcs, Accept) ->
    yystate(61, Ics, Line, [$o|Tcs], Accept);
yystate(60, Ics, Line, Tcs, Accept) ->
    {Accept,60,Ics,Line,Tcs};
yystate(61, [$n|Ics], Line, Tcs, Accept) ->
    yystate(62, Ics, Line, [$n|Tcs], Accept);
yystate(61, Ics, Line, Tcs, Accept) ->
    {Accept,61,Ics,Line,Tcs};
yystate(62, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,8};
yystate(63, [$e|Ics], Line, Tcs, Accept) ->
    yystate(64, Ics, Line, [$e|Tcs], Accept);
yystate(63, Ics, Line, Tcs, Accept) ->
    {Accept,63,Ics,Line,Tcs};
yystate(64, [$c|Ics], Line, Tcs, Accept) ->
    yystate(65, Ics, Line, [$c|Tcs], Accept);
yystate(64, Ics, Line, Tcs, Accept) ->
    {Accept,64,Ics,Line,Tcs};
yystate(65, [$t|Ics], Line, Tcs, Accept) ->
    yystate(66, Ics, Line, [$t|Tcs], Accept);
yystate(65, Ics, Line, Tcs, Accept) ->
    {Accept,65,Ics,Line,Tcs};
yystate(66, [$i|Ics], Line, Tcs, Accept) ->
    yystate(67, Ics, Line, [$i|Tcs], Accept);
yystate(66, Ics, Line, Tcs, Accept) ->
    {Accept,66,Ics,Line,Tcs};
yystate(67, [$o|Ics], Line, Tcs, Accept) ->
    yystate(68, Ics, Line, [$o|Tcs], Accept);
yystate(67, Ics, Line, Tcs, Accept) ->
    {Accept,67,Ics,Line,Tcs};
yystate(68, [$n|Ics], Line, Tcs, Accept) ->
    yystate(69, Ics, Line, [$n|Tcs], Accept);
yystate(68, Ics, Line, Tcs, Accept) ->
    {Accept,68,Ics,Line,Tcs};
yystate(69, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,7};
yystate(70, [$t|Ics], Line, Tcs, Accept) ->
    yystate(71, Ics, Line, [$t|Tcs], Accept);
yystate(70, Ics, Line, Tcs, Accept) ->
    {Accept,70,Ics,Line,Tcs};
yystate(71, [$a|Ics], Line, Tcs, Accept) ->
    yystate(72, Ics, Line, [$a|Tcs], Accept);
yystate(71, Ics, Line, Tcs, Accept) ->
    {Accept,71,Ics,Line,Tcs};
yystate(72, [$n|Ics], Line, Tcs, Accept) ->
    yystate(73, Ics, Line, [$n|Tcs], Accept);
yystate(72, Ics, Line, Tcs, Accept) ->
    {Accept,72,Ics,Line,Tcs};
yystate(73, [$g|Ics], Line, Tcs, Accept) ->
    yystate(74, Ics, Line, [$g|Tcs], Accept);
yystate(73, Ics, Line, Tcs, Accept) ->
    {Accept,73,Ics,Line,Tcs};
yystate(74, [$l|Ics], Line, Tcs, Accept) ->
    yystate(75, Ics, Line, [$l|Tcs], Accept);
yystate(74, Ics, Line, Tcs, Accept) ->
    {Accept,74,Ics,Line,Tcs};
yystate(75, [$e|Ics], Line, Tcs, Accept) ->
    yystate(76, Ics, Line, [$e|Tcs], Accept);
yystate(75, Ics, Line, Tcs, Accept) ->
    {Accept,75,Ics,Line,Tcs};
yystate(76, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,3};
yystate(77, [$d|Ics], Line, Tcs, Accept) ->
    yystate(78, Ics, Line, [$d|Tcs], Accept);
yystate(77, Ics, Line, Tcs, Accept) ->
    {Accept,77,Ics,Line,Tcs};
yystate(78, [$i|Ics], Line, Tcs, Accept) ->
    yystate(79, Ics, Line, [$i|Tcs], Accept);
yystate(78, Ics, Line, Tcs, Accept) ->
    {Accept,78,Ics,Line,Tcs};
yystate(79, [$u|Ics], Line, Tcs, Accept) ->
    yystate(80, Ics, Line, [$u|Tcs], Accept);
yystate(79, Ics, Line, Tcs, Accept) ->
    {Accept,79,Ics,Line,Tcs};
yystate(80, [$s|Ics], Line, Tcs, Accept) ->
    yystate(81, Ics, Line, [$s|Tcs], Accept);
yystate(80, Ics, Line, Tcs, Accept) ->
    {Accept,80,Ics,Line,Tcs};
yystate(81, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,18};
yystate(82, [$l|Ics], Line, Tcs, Accept) ->
    yystate(90, Ics, Line, [$l|Tcs], Accept);
yystate(82, [$o|Ics], Line, Tcs, Accept) ->
    yystate(83, Ics, Line, [$o|Tcs], Accept);
yystate(82, Ics, Line, Tcs, Accept) ->
    {Accept,82,Ics,Line,Tcs};
yystate(83, [$l|Ics], Line, Tcs, Accept) ->
    yystate(84, Ics, Line, [$l|Tcs], Accept);
yystate(83, Ics, Line, Tcs, Accept) ->
    {Accept,83,Ics,Line,Tcs};
yystate(84, [$e|Ics], Line, Tcs, Accept) ->
    yystate(89, Ics, Line, [$e|Tcs], Accept);
yystate(84, [$y|Ics], Line, Tcs, Accept) ->
    yystate(85, Ics, Line, [$y|Tcs], Accept);
yystate(84, Ics, Line, Tcs, Accept) ->
    {Accept,84,Ics,Line,Tcs};
yystate(85, [$g|Ics], Line, Tcs, Accept) ->
    yystate(86, Ics, Line, [$g|Tcs], Accept);
yystate(85, Ics, Line, Tcs, Accept) ->
    {Accept,85,Ics,Line,Tcs};
yystate(86, [$o|Ics], Line, Tcs, Accept) ->
    yystate(87, Ics, Line, [$o|Tcs], Accept);
yystate(86, Ics, Line, Tcs, Accept) ->
    {Accept,86,Ics,Line,Tcs};
yystate(87, [$n|Ics], Line, Tcs, Accept) ->
    yystate(88, Ics, Line, [$n|Tcs], Accept);
yystate(87, Ics, Line, Tcs, Accept) ->
    {Accept,87,Ics,Line,Tcs};
yystate(88, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,4};
yystate(89, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,29};
yystate(90, [$a|Ics], Line, Tcs, Accept) ->
    yystate(91, Ics, Line, [$a|Tcs], Accept);
yystate(90, Ics, Line, Tcs, Accept) ->
    {Accept,90,Ics,Line,Tcs};
yystate(91, [$n|Ics], Line, Tcs, Accept) ->
    yystate(92, Ics, Line, [$n|Tcs], Accept);
yystate(91, Ics, Line, Tcs, Accept) ->
    {Accept,91,Ics,Line,Tcs};
yystate(92, [$e|Ics], Line, Tcs, Accept) ->
    yystate(93, Ics, Line, [$e|Tcs], Accept);
yystate(92, Ics, Line, Tcs, Accept) ->
    {Accept,92,Ics,Line,Tcs};
yystate(93, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,0};
yystate(94, [$o|Ics], Line, Tcs, Accept) ->
    yystate(95, Ics, Line, [$o|Tcs], Accept);
yystate(94, Ics, Line, Tcs, Accept) ->
    {Accept,94,Ics,Line,Tcs};
yystate(95, [$r|Ics], Line, Tcs, Accept) ->
    yystate(96, Ics, Line, [$r|Tcs], Accept);
yystate(95, Ics, Line, Tcs, Accept) ->
    {Accept,95,Ics,Line,Tcs};
yystate(96, [$m|Ics], Line, Tcs, Accept) ->
    yystate(97, Ics, Line, [$m|Tcs], Accept);
yystate(96, Ics, Line, Tcs, Accept) ->
    {Accept,96,Ics,Line,Tcs};
yystate(97, [$a|Ics], Line, Tcs, Accept) ->
    yystate(98, Ics, Line, [$a|Tcs], Accept);
yystate(97, Ics, Line, Tcs, Accept) ->
    {Accept,97,Ics,Line,Tcs};
yystate(98, [$l|Ics], Line, Tcs, Accept) ->
    yystate(99, Ics, Line, [$l|Tcs], Accept);
yystate(98, Ics, Line, Tcs, Accept) ->
    {Accept,98,Ics,Line,Tcs};
yystate(99, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,22};
yystate(100, [$a|Ics], Line, Tcs, Accept) ->
    yystate(101, Ics, Line, [$a|Tcs], Accept);
yystate(100, Ics, Line, Tcs, Accept) ->
    {Accept,100,Ics,Line,Tcs};
yystate(101, [$x|Ics], Line, Tcs, Accept) ->
    yystate(102, Ics, Line, [$x|Tcs], Accept);
yystate(101, Ics, Line, Tcs, Accept) ->
    {Accept,101,Ics,Line,Tcs};
yystate(102, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,26};
yystate(103, [$i|Ics], Line, Tcs, Accept) ->
    yystate(112, Ics, Line, [$i|Tcs], Accept);
yystate(103, [$o|Ics], Line, Tcs, Accept) ->
    yystate(104, Ics, Line, [$o|Tcs], Accept);
yystate(103, Ics, Line, Tcs, Accept) ->
    {Accept,103,Ics,Line,Tcs};
yystate(104, [$o|Ics], Line, Tcs, Accept) ->
    yystate(105, Ics, Line, [$o|Tcs], Accept);
yystate(104, Ics, Line, Tcs, Accept) ->
    {Accept,104,Ics,Line,Tcs};
yystate(105, [$k|Ics], Line, Tcs, Accept) ->
    yystate(106, Ics, Line, [$k|Tcs], Accept);
yystate(105, Ics, Line, Tcs, Accept) ->
    {Accept,105,Ics,Line,Tcs};
yystate(106, [$p|Ics], Line, Tcs, Accept) ->
    yystate(107, Ics, Line, [$p|Tcs], Accept);
yystate(106, Ics, Line, Tcs, Accept) ->
    {Accept,106,Ics,Line,Tcs};
yystate(107, [$o|Ics], Line, Tcs, Accept) ->
    yystate(108, Ics, Line, [$o|Tcs], Accept);
yystate(107, Ics, Line, Tcs, Accept) ->
    {Accept,107,Ics,Line,Tcs};
yystate(108, [$i|Ics], Line, Tcs, Accept) ->
    yystate(109, Ics, Line, [$i|Tcs], Accept);
yystate(108, Ics, Line, Tcs, Accept) ->
    {Accept,108,Ics,Line,Tcs};
yystate(109, [$n|Ics], Line, Tcs, Accept) ->
    yystate(110, Ics, Line, [$n|Tcs], Accept);
yystate(109, Ics, Line, Tcs, Accept) ->
    {Accept,109,Ics,Line,Tcs};
yystate(110, [$t|Ics], Line, Tcs, Accept) ->
    yystate(111, Ics, Line, [$t|Tcs], Accept);
yystate(110, Ics, Line, Tcs, Accept) ->
    {Accept,110,Ics,Line,Tcs};
yystate(111, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,12};
yystate(112, [$g|Ics], Line, Tcs, Accept) ->
    yystate(113, Ics, Line, [$g|Tcs], Accept);
yystate(112, Ics, Line, Tcs, Accept) ->
    {Accept,112,Ics,Line,Tcs};
yystate(113, [$h|Ics], Line, Tcs, Accept) ->
    yystate(114, Ics, Line, [$h|Tcs], Accept);
yystate(113, Ics, Line, Tcs, Accept) ->
    {Accept,113,Ics,Line,Tcs};
yystate(114, [$t|Ics], Line, Tcs, Accept) ->
    yystate(115, Ics, Line, [$t|Tcs], Accept);
yystate(114, Ics, Line, Tcs, Accept) ->
    {Accept,114,Ics,Line,Tcs};
yystate(115, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,5};
yystate(116, [$e|Ics], Line, Tcs, Accept) ->
    yystate(117, Ics, Line, [$e|Tcs], Accept);
yystate(116, Ics, Line, Tcs, Accept) ->
    {Accept,116,Ics,Line,Tcs};
yystate(117, [$i|Ics], Line, Tcs, Accept) ->
    yystate(118, Ics, Line, [$i|Tcs], Accept);
yystate(117, Ics, Line, Tcs, Accept) ->
    {Accept,117,Ics,Line,Tcs};
yystate(118, [$g|Ics], Line, Tcs, Accept) ->
    yystate(119, Ics, Line, [$g|Tcs], Accept);
yystate(118, Ics, Line, Tcs, Accept) ->
    {Accept,118,Ics,Line,Tcs};
yystate(119, [$h|Ics], Line, Tcs, Accept) ->
    yystate(120, Ics, Line, [$h|Tcs], Accept);
yystate(119, Ics, Line, Tcs, Accept) ->
    {Accept,119,Ics,Line,Tcs};
yystate(120, [$t|Ics], Line, Tcs, Accept) ->
    yystate(121, Ics, Line, [$t|Tcs], Accept);
yystate(120, Ics, Line, Tcs, Accept) ->
    {Accept,120,Ics,Line,Tcs};
yystate(121, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,15};
yystate(122, [$q|Ics], Line, Tcs, Accept) ->
    yystate(130, Ics, Line, [$q|Tcs], Accept);
yystate(122, [$y|Ics], Line, Tcs, Accept) ->
    yystate(123, Ics, Line, [$y|Tcs], Accept);
yystate(122, Ics, Line, Tcs, Accept) ->
    {Accept,122,Ics,Line,Tcs};
yystate(123, [$e|Ics], Line, Tcs, Accept) ->
    yystate(124, Ics, Line, [$e|Tcs], Accept);
yystate(123, Ics, Line, Tcs, Accept) ->
    {Accept,123,Ics,Line,Tcs};
yystate(124, [$p|Ics], Line, Tcs, Accept) ->
    yystate(125, Ics, Line, [$p|Tcs], Accept);
yystate(124, Ics, Line, Tcs, Accept) ->
    {Accept,124,Ics,Line,Tcs};
yystate(125, [$o|Ics], Line, Tcs, Accept) ->
    yystate(126, Ics, Line, [$o|Tcs], Accept);
yystate(125, Ics, Line, Tcs, Accept) ->
    {Accept,125,Ics,Line,Tcs};
yystate(126, [$i|Ics], Line, Tcs, Accept) ->
    yystate(127, Ics, Line, [$i|Tcs], Accept);
yystate(126, Ics, Line, Tcs, Accept) ->
    {Accept,126,Ics,Line,Tcs};
yystate(127, [$n|Ics], Line, Tcs, Accept) ->
    yystate(128, Ics, Line, [$n|Tcs], Accept);
yystate(127, Ics, Line, Tcs, Accept) ->
    {Accept,127,Ics,Line,Tcs};
yystate(128, [$t|Ics], Line, Tcs, Accept) ->
    yystate(129, Ics, Line, [$t|Tcs], Accept);
yystate(128, Ics, Line, Tcs, Accept) ->
    {Accept,128,Ics,Line,Tcs};
yystate(129, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,11};
yystate(130, [$u|Ics], Line, Tcs, Accept) ->
    yystate(131, Ics, Line, [$u|Tcs], Accept);
yystate(130, Ics, Line, Tcs, Accept) ->
    {Accept,130,Ics,Line,Tcs};
yystate(131, [$a|Ics], Line, Tcs, Accept) ->
    yystate(132, Ics, Line, [$a|Tcs], Accept);
yystate(131, Ics, Line, Tcs, Accept) ->
    {Accept,131,Ics,Line,Tcs};
yystate(132, [$t|Ics], Line, Tcs, Accept) ->
    yystate(133, Ics, Line, [$t|Tcs], Accept);
yystate(132, Ics, Line, Tcs, Accept) ->
    {Accept,132,Ics,Line,Tcs};
yystate(133, [$o|Ics], Line, Tcs, Accept) ->
    yystate(134, Ics, Line, [$o|Tcs], Accept);
yystate(133, Ics, Line, Tcs, Accept) ->
    {Accept,133,Ics,Line,Tcs};
yystate(134, [$r|Ics], Line, Tcs, Accept) ->
    yystate(135, Ics, Line, [$r|Tcs], Accept);
yystate(134, Ics, Line, Tcs, Accept) ->
    {Accept,134,Ics,Line,Tcs};
yystate(135, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,30};
yystate(136, [$e|Ics], Line, Tcs, Accept) ->
    yystate(144, Ics, Line, [$e|Tcs], Accept);
yystate(136, [$i|Ics], Line, Tcs, Accept) ->
    yystate(137, Ics, Line, [$i|Tcs], Accept);
yystate(136, Ics, Line, Tcs, Accept) ->
    {Accept,136,Ics,Line,Tcs};
yystate(137, [$s|Ics], Line, Tcs, Accept) ->
    yystate(138, Ics, Line, [$s|Tcs], Accept);
yystate(137, Ics, Line, Tcs, Accept) ->
    {Accept,137,Ics,Line,Tcs};
yystate(138, [$t|Ics], Line, Tcs, Accept) ->
    yystate(139, Ics, Line, [$t|Tcs], Accept);
yystate(138, Ics, Line, Tcs, Accept) ->
    {Accept,138,Ics,Line,Tcs};
yystate(139, [$a|Ics], Line, Tcs, Accept) ->
    yystate(140, Ics, Line, [$a|Tcs], Accept);
yystate(139, Ics, Line, Tcs, Accept) ->
    {Accept,139,Ics,Line,Tcs};
yystate(140, [$n|Ics], Line, Tcs, Accept) ->
    yystate(141, Ics, Line, [$n|Tcs], Accept);
yystate(140, Ics, Line, Tcs, Accept) ->
    {Accept,140,Ics,Line,Tcs};
yystate(141, [$c|Ics], Line, Tcs, Accept) ->
    yystate(142, Ics, Line, [$c|Tcs], Accept);
yystate(141, Ics, Line, Tcs, Accept) ->
    {Accept,141,Ics,Line,Tcs};
yystate(142, [$e|Ics], Line, Tcs, Accept) ->
    yystate(143, Ics, Line, [$e|Tcs], Accept);
yystate(142, Ics, Line, Tcs, Accept) ->
    {Accept,142,Ics,Line,Tcs};
yystate(143, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,24};
yystate(144, [$p|Ics], Line, Tcs, Accept) ->
    yystate(145, Ics, Line, [$p|Tcs], Accept);
yystate(144, Ics, Line, Tcs, Accept) ->
    {Accept,144,Ics,Line,Tcs};
yystate(145, [$t|Ics], Line, Tcs, Accept) ->
    yystate(146, Ics, Line, [$t|Tcs], Accept);
yystate(145, Ics, Line, Tcs, Accept) ->
    {Accept,145,Ics,Line,Tcs};
yystate(146, [$h|Ics], Line, Tcs, Accept) ->
    yystate(147, Ics, Line, [$h|Tcs], Accept);
yystate(146, Ics, Line, Tcs, Accept) ->
    {Accept,146,Ics,Line,Tcs};
yystate(147, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,28};
yystate(148, [$a|Ics], Line, Tcs, Accept) ->
    yystate(158, Ics, Line, [$a|Tcs], Accept);
yystate(148, [$e|Ics], Line, Tcs, Accept) ->
    yystate(153, Ics, Line, [$e|Tcs], Accept);
yystate(148, [$o|Ics], Line, Tcs, Accept) ->
    yystate(149, Ics, Line, [$o|Tcs], Accept);
yystate(148, Ics, Line, Tcs, Accept) ->
    {Accept,148,Ics,Line,Tcs};
yystate(149, [$l|Ics], Line, Tcs, Accept) ->
    yystate(150, Ics, Line, [$l|Tcs], Accept);
yystate(149, Ics, Line, Tcs, Accept) ->
    {Accept,149,Ics,Line,Tcs};
yystate(150, [$o|Ics], Line, Tcs, Accept) ->
    yystate(151, Ics, Line, [$o|Tcs], Accept);
yystate(150, Ics, Line, Tcs, Accept) ->
    {Accept,150,Ics,Line,Tcs};
yystate(151, [$r|Ics], Line, Tcs, Accept) ->
    yystate(152, Ics, Line, [$r|Tcs], Accept);
yystate(151, Ics, Line, Tcs, Accept) ->
    {Accept,151,Ics,Line,Tcs};
yystate(152, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,6};
yystate(153, [$n|Ics], Line, Tcs, Accept) ->
    yystate(154, Ics, Line, [$n|Tcs], Accept);
yystate(153, Ics, Line, Tcs, Accept) ->
    {Accept,153,Ics,Line,Tcs};
yystate(154, [$t|Ics], Line, Tcs, Accept) ->
    yystate(155, Ics, Line, [$t|Tcs], Accept);
yystate(154, Ics, Line, Tcs, Accept) ->
    {Accept,154,Ics,Line,Tcs};
yystate(155, [$e|Ics], Line, Tcs, Accept) ->
    yystate(156, Ics, Line, [$e|Tcs], Accept);
yystate(155, Ics, Line, Tcs, Accept) ->
    {Accept,155,Ics,Line,Tcs};
yystate(156, [$r|Ics], Line, Tcs, Accept) ->
    yystate(157, Ics, Line, [$r|Tcs], Accept);
yystate(156, Ics, Line, Tcs, Accept) ->
    {Accept,156,Ics,Line,Tcs};
yystate(157, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,19};
yystate(158, [$m|Ics], Line, Tcs, Accept) ->
    yystate(159, Ics, Line, [$m|Tcs], Accept);
yystate(158, Ics, Line, Tcs, Accept) ->
    {Accept,158,Ics,Line,Tcs};
yystate(159, [$e|Ics], Line, Tcs, Accept) ->
    yystate(160, Ics, Line, [$e|Tcs], Accept);
yystate(159, Ics, Line, Tcs, Accept) ->
    {Accept,159,Ics,Line,Tcs};
yystate(160, [$r|Ics], Line, Tcs, Accept) ->
    yystate(161, Ics, Line, [$r|Tcs], Accept);
yystate(160, Ics, Line, Tcs, Accept) ->
    {Accept,160,Ics,Line,Tcs};
yystate(161, [$a|Ics], Line, Tcs, Accept) ->
    yystate(162, Ics, Line, [$a|Tcs], Accept);
yystate(161, Ics, Line, Tcs, Accept) ->
    {Accept,161,Ics,Line,Tcs};
yystate(162, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,10};
yystate(163, [$a|Ics], Line, Tcs, Accept) ->
    yystate(169, Ics, Line, [$a|Tcs], Accept);
yystate(163, [$i|Ics], Line, Tcs, Accept) ->
    yystate(164, Ics, Line, [$i|Tcs], Accept);
yystate(163, Ics, Line, Tcs, Accept) ->
    {Accept,163,Ics,Line,Tcs};
yystate(164, [$t|Ics], Line, Tcs, Accept) ->
    yystate(165, Ics, Line, [$t|Tcs], Accept);
yystate(164, Ics, Line, Tcs, Accept) ->
    {Accept,164,Ics,Line,Tcs};
yystate(165, [$m|Ics], Line, Tcs, Accept) ->
    yystate(166, Ics, Line, [$m|Tcs], Accept);
yystate(165, Ics, Line, Tcs, Accept) ->
    {Accept,165,Ics,Line,Tcs};
yystate(166, [$a|Ics], Line, Tcs, Accept) ->
    yystate(167, Ics, Line, [$a|Tcs], Accept);
yystate(166, Ics, Line, Tcs, Accept) ->
    {Accept,166,Ics,Line,Tcs};
yystate(167, [$p|Ics], Line, Tcs, Accept) ->
    yystate(168, Ics, Line, [$p|Tcs], Accept);
yystate(167, Ics, Line, Tcs, Accept) ->
    {Accept,167,Ics,Line,Tcs};
yystate(168, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,9};
yystate(169, [$c|Ics], Line, Tcs, Accept) ->
    yystate(170, Ics, Line, [$c|Tcs], Accept);
yystate(169, Ics, Line, Tcs, Accept) ->
    {Accept,169,Ics,Line,Tcs};
yystate(170, [$k|Ics], Line, Tcs, Accept) ->
    yystate(171, Ics, Line, [$k|Tcs], Accept);
yystate(170, Ics, Line, Tcs, Accept) ->
    {Accept,170,Ics,Line,Tcs};
yystate(171, [$g|Ics], Line, Tcs, Accept) ->
    yystate(172, Ics, Line, [$g|Tcs], Accept);
yystate(171, Ics, Line, Tcs, Accept) ->
    {Accept,171,Ics,Line,Tcs};
yystate(172, [$r|Ics], Line, Tcs, Accept) ->
    yystate(173, Ics, Line, [$r|Tcs], Accept);
yystate(172, Ics, Line, Tcs, Accept) ->
    {Accept,172,Ics,Line,Tcs};
yystate(173, [$o|Ics], Line, Tcs, Accept) ->
    yystate(174, Ics, Line, [$o|Tcs], Accept);
yystate(173, Ics, Line, Tcs, Accept) ->
    {Accept,173,Ics,Line,Tcs};
yystate(174, [$u|Ics], Line, Tcs, Accept) ->
    yystate(175, Ics, Line, [$u|Tcs], Accept);
yystate(174, Ics, Line, Tcs, Accept) ->
    {Accept,174,Ics,Line,Tcs};
yystate(175, [$n|Ics], Line, Tcs, Accept) ->
    yystate(176, Ics, Line, [$n|Tcs], Accept);
yystate(175, Ics, Line, Tcs, Accept) ->
    {Accept,175,Ics,Line,Tcs};
yystate(176, [$d|Ics], Line, Tcs, Accept) ->
    yystate(177, Ics, Line, [$d|Tcs], Accept);
yystate(176, Ics, Line, Tcs, Accept) ->
    {Accept,176,Ics,Line,Tcs};
yystate(177, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,23};
yystate(178, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,36};
yystate(179, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,35};
yystate(180, [$.|Ics], Line, Tcs, Accept) ->
    yystate(181, Ics, Line, [$.|Tcs], {[$.|Ics],Tcs,32});
yystate(180, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(180, Ics, Line, [C|Tcs], {[C|Ics],Tcs,32});
yystate(180, Ics, Line, Tcs, Accept) ->
    {{Ics,Tcs,32},180,Ics,Line,Tcs};
yystate(181, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(182, Ics, Line, [C|Tcs], Accept);
yystate(181, Ics, Line, Tcs, Accept) ->
    {Accept,181,Ics,Line,Tcs};
yystate(182, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(182, Ics, Line, [C|Tcs], {[C|Ics],Tcs,31});
yystate(182, Ics, Line, Tcs, Accept) ->
    {{Ics,Tcs,31},182,Ics,Line,Tcs};
yystate(183, [$.|Ics], Line, Tcs, Accept) ->
    yystate(185, Ics, Line, [$.|Tcs], Accept);
yystate(183, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(184, Ics, Line, [C|Tcs], Accept);
yystate(183, Ics, Line, Tcs, Accept) ->
    {Accept,183,Ics,Line,Tcs};
yystate(184, [$.|Ics], Line, Tcs, Accept) ->
    yystate(185, Ics, Line, [$.|Tcs], {[$.|Ics],Tcs,32});
yystate(184, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(184, Ics, Line, [C|Tcs], {[C|Ics],Tcs,32});
yystate(184, Ics, Line, Tcs, Accept) ->
    {{Ics,Tcs,32},184,Ics,Line,Tcs};
yystate(185, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(186, Ics, Line, [C|Tcs], Accept);
yystate(185, Ics, Line, Tcs, Accept) ->
    {Accept,185,Ics,Line,Tcs};
yystate(186, [C|Ics], Line, Tcs, Accept) when C >= $0, C =< $9 ->
    yystate(186, Ics, Line, [C|Tcs], {[C|Ics],Tcs,31});
yystate(186, Ics, Line, Tcs, Accept) ->
    {{Ics,Tcs,31},186,Ics,Line,Tcs};
yystate(187, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,37};
yystate(188, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,34};
yystate(189, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,33};
yystate(190, [$\n|Ics], Line, Tcs, Accept) ->
    yystate(190, Ics, Line+1, [$\n|Tcs], Accept);
yystate(190, [$'|Ics], Line, Tcs, Accept) ->
    yystate(191, Ics, Line, [$'|Tcs], Accept);
yystate(190, [C|Ics], Line, Tcs, Accept) when C >= $\000, C =< $\t ->
    yystate(190, Ics, Line, [C|Tcs], Accept);
yystate(190, [C|Ics], Line, Tcs, Accept) when C >= $\v, C =< $& ->
    yystate(190, Ics, Line, [C|Tcs], Accept);
yystate(190, [C|Ics], Line, Tcs, Accept) when C >= $(, C =< $\377 ->
    yystate(190, Ics, Line, [C|Tcs], Accept);
yystate(190, Ics, Line, Tcs, Accept) ->
    {Accept,190,Ics,Line,Tcs};
yystate(191, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,38};
yystate(192, [$\v|Ics], Line, Tcs, Accept) ->
    yystate(192, Ics, Line, [$\v|Tcs], {[$\v|Ics],Tcs,39});
yystate(192, [$\f|Ics], Line, Tcs, Accept) ->
    yystate(192, Ics, Line, [$\f|Tcs], {[$\f|Ics],Tcs,39});
yystate(192, [C|Ics], Line, Tcs, Accept) when C >= $\000, C =< $\t ->
    yystate(192, Ics, Line, [C|Tcs], {[C|Ics],Tcs,39});
yystate(192, [C|Ics], Line, Tcs, Accept) when C >= $\016, C =< $\377 ->
    yystate(192, Ics, Line, [C|Tcs], {[C|Ics],Tcs,39});
yystate(192, Ics, Line, Tcs, Accept) ->
    {{Ics,Tcs,39},192,Ics,Line,Tcs};
yystate(193, Ics, Line, Tcs, Accept) ->
    {accept,Ics,Line,Tcs,40};
yystate(194, [$\t|Ics], Line, Tcs, Accept) ->
    yystate(194, Ics, Line, [$\t|Tcs], {[$\t|Ics],Tcs,40});
yystate(194, Ics, Line, Tcs, Accept) ->
    {{Ics,Tcs,40},194,Ics,Line,Tcs};
yystate(_, _, _, _, _) -> error.

yyaction(0, YYrevtext, YYline, YYts) ->
    [begin  {'Plane',YYline} end|YYts];
yyaction(1, YYrevtext, YYline, YYts) ->
    [begin  {'Sphere',YYline} end|YYts];
yyaction(2, YYrevtext, YYline, YYts) ->
    [begin  {'Triangle',YYline} end|YYts];
yyaction(3, YYrevtext, YYline, YYts) ->
    [begin  {'Rectangle',YYline} end|YYts];
yyaction(4, YYrevtext, YYline, YYts) ->
    [begin  {'Polygon',YYline} end|YYts];
yyaction(5, YYrevtext, YYline, YYts) ->
    [begin  {'Light',YYline} end|YYts];
yyaction(6, YYrevtext, YYline, YYts) ->
    [begin  {'Color',YYline} end|YYts];
yyaction(7, YYrevtext, YYline, YYts) ->
    [begin  {'Reflection',YYline} end|YYts];
yyaction(8, YYrevtext, YYline, YYts) ->
    [begin  {'Refraction',YYline} end|YYts];
yyaction(9, YYrevtext, YYline, YYts) ->
    [begin  {'Bitmap',YYline} end|YYts];
yyaction(10, YYrevtext, YYline, YYts) ->
    [begin  {'Camera',YYline} end|YYts];
yyaction(11, YYrevtext, YYline, YYts) ->
    [begin  {'Eyepoint',YYline} end|YYts];
yyaction(12, YYrevtext, YYline, YYts) ->
    [begin  {'Lookpoint',YYline} end|YYts];
yyaction(13, YYrevtext, YYline, YYts) ->
    [begin  {'Zoom',YYline} end|YYts];
yyaction(14, YYrevtext, YYline, YYts) ->
    [begin  {'Width',YYline} end|YYts];
yyaction(15, YYrevtext, YYline, YYts) ->
    [begin  {'Height',YYline} end|YYts];
yyaction(16, YYrevtext, YYline, YYts) ->
    [begin  {'Resolution',YYline} end|YYts];
yyaction(17, YYrevtext, YYline, YYts) ->
    [begin  {'Up',YYline} end|YYts];
yyaction(18, YYrevtext, YYline, YYts) ->
    [begin  {'Radius',YYline} end|YYts];
yyaction(19, YYrevtext, YYline, YYts) ->
    [begin  {'Center',YYline} end|YYts];
yyaction(20, YYrevtext, YYline, YYts) ->
    [begin  {'X-axis',YYline} end|YYts];
yyaction(21, YYrevtext, YYline, YYts) ->
    [begin  {'Y-axis',YYline} end|YYts];
yyaction(22, YYrevtext, YYline, YYts) ->
    [begin  {'Normal',YYline} end|YYts];
yyaction(23, YYrevtext, YYline, YYts) ->
    [begin  {'Background',YYline} end|YYts];
yyaction(24, YYrevtext, YYline, YYts) ->
    [begin  {'Distance',YYline} end|YYts];
yyaction(25, YYrevtext, YYline, YYts) ->
    [begin  {'Use',YYline} end|YYts];
yyaction(26, YYrevtext, YYline, YYts) ->
    [begin  {'Max',YYline} end|YYts];
yyaction(27, YYrevtext, YYline, YYts) ->
    [begin  {'Shadow',YYline} end|YYts];
yyaction(28, YYrevtext, YYline, YYts) ->
    [begin  {'Depth',YYline} end|YYts];
yyaction(29, YYrevtext, YYline, YYts) ->
    [begin  {'Pole',YYline} end|YYts];
yyaction(30, YYrevtext, YYline, YYts) ->
    [begin  {'Equator',YYline} end|YYts];
yyaction(31, YYrevtext, YYline, YYts) ->
    YYtext = yyrev(YYrevtext),
    [begin  {float,YYline,list_to_float(YYtext)} end|YYts];
yyaction(32, YYrevtext, YYline, YYts) ->
    YYtext = yyrev(YYrevtext),
    [begin  {integer,YYline,list_to_integer(YYtext)} end|YYts];
yyaction(33, YYrevtext, YYline, YYts) ->
    [begin  {lpar,YYline} end|YYts];
yyaction(34, YYrevtext, YYline, YYts) ->
    [begin  {rpar,YYline} end|YYts];
yyaction(35, YYrevtext, YYline, YYts) ->
    [begin  {lhpar,YYline} end|YYts];
yyaction(36, YYrevtext, YYline, YYts) ->
    [begin  {rhpar,YYline} end|YYts];
yyaction(37, YYrevtext, YYline, YYts) ->
    [begin  {comma,YYline} end|YYts];
yyaction(38, YYrevtext, YYline, YYts) ->
    YYtext = yyrev(YYrevtext),
    [begin  {string,YYline,fix_string(YYtext)} end|YYts];
yyaction(39, YYrevtext, YYline, YYts) -> YYts;
yyaction(40, YYrevtext, YYline, YYts) -> YYts;
yyaction(_, _, _, _) -> error.
