-module(parser).
%% Copyright (C) 1995 Ellemtel Telecommunications Systems Laboratories
%% Author: Carl Wilhelm Welin 08/7273557

-export([parse/1, parse_and_scan/1, report_error/3, format_error/1]).

% General parser code for yecc parser generator

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error, {Errorline, yecc, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, yecc, "end_of_file"}};
        {error, Descriptor, Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, yecc, Reason}};
        {ok, Tokens, Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, yecc,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    {error, {0, yecc, "premature end of input string!"}}.

yeccerror(Line, Token) ->
    {error, {Line, yecc, ["syntax error before: ", display_token(Token)]}}.

report_error(Source_file, Line, Message) ->
    io:format('~s', [lists:concat([Source_file, ":", Line, ": ",
				   format_error(Message), "\n"])]).

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

display_token({atom, _, A}) -> io_lib:write(A);
display_token({integer,_,N}) -> io_lib:write(N);
display_token({float,_,F}) -> io_lib:write(F);
display_token({char,_,C}) -> io_lib:write_char(C);
display_token({var,_,V}) -> io_lib:format('~s', [V]);
display_token({string,_,S}) -> io_lib:write_string(S);
display_token({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
display_token({'$end', _}) ->
    [];
display_token({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
display_token(Other) ->
    io_lib:write(Other).

% Grammar specific code follows.


value_of(Token) ->
	element(3,Token).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%

merge_camera_opts(Opts) -> 
    Default = camera:default_camera(),
    init_struct(Opts,Default,[eyepoint,lookpoint,up,
			      zoom,width,height,resolution]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

merge_sphere_opts(Opts) -> 
    Default = default_sphere(),
    init_struct(Opts,Default,[center,radius,color,reflection,
			      pole,equator,cross,bitmap]).

default_sphere() ->
    {{0,0,100},3,{255,0,0},0,{0,1,0},{1,0,0},no_cross,no_bitmap}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

merge_plane_opts(Opts) -> 
    Default = default_plane(),
    init_struct(Opts,Default,[normal,distance,reflection,color]).

default_plane() ->
    {{-1,0,2},150,0,{70,190,70}}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

merge_triangle_opts(Opts) -> 
    Default = default_triangle(),
    init_struct(Opts,Default,[center,xaxis,yaxis,color,reflection,
			      bitmap]).

default_triangle() ->
    {{0,0,90},{30,0,0},{30,30,5},{40,40,120},0,no_bitmap}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

merge_light_opts([]) ->
    default_light();

merge_light_opts([{center,C}|_]) -> C.

default_light() -> {0,100,-100}.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% init_struct(Table,Default,Opts)
%
    
init_struct([],Acc,_Template) ->
    Acc;

init_struct([Opt|Opts],Acc,Template) ->
    {Option,Value} = Opt,
    Pos = get_position(Template,Option,1),
    NewAcc = setelement(Pos,Acc,Value),
    init_struct(Opts,NewAcc,Template).

%

get_position([],Key,_Pos) ->
    io:format("Error: option ~w not found~n",[Key]);

get_position([Option|Rest],Key,Pos) ->
    case Option of
	Key ->
	    Pos;
	Other ->
	    get_position(Rest,Key,Pos+1)
    end.
	


yeccpars2(0, 'Background', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 1, [0 | Ss], [T | Stack]);
yeccpars2(0, 'Use', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 8, [0 | Ss], [T | Stack]);
yeccpars2(0, 'Max', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 4, [0 | Ss], [T | Stack]);
yeccpars2(0, 'Sphere', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 6, [0 | Ss], [T | Stack]);
yeccpars2(0, 'Triangle', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 7, [0 | Ss], [T | Stack]);
yeccpars2(0, 'Plane', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 5, [0 | Ss], [T | Stack]);
yeccpars2(0, 'Light', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 3, [0 | Ss], [T | Stack]);
yeccpars2(0, 'Camera', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 2, [0 | Ss], [T | Stack]);
yeccpars2(0, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(13, Cat, [0 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(1, lpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 39, [1 | Ss], [T | Stack]);
yeccpars2(1, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(2, 'Resolution', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 86, [2 | Ss], [T | Stack]);
yeccpars2(2, 'Height', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 84, [2 | Ss], [T | Stack]);
yeccpars2(2, 'Width', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 88, [2 | Ss], [T | Stack]);
yeccpars2(2, 'Up', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 87, [2 | Ss], [T | Stack]);
yeccpars2(2, 'Zoom', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 89, [2 | Ss], [T | Stack]);
yeccpars2(2, 'Lookpoint', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 85, [2 | Ss], [T | Stack]);
yeccpars2(2, 'Eyepoint', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 83, [2 | Ss], [T | Stack]);
yeccpars2(2, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(90, Cat, [2 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(3, 'Bitmap', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 17, [3 | Ss], [T | Stack]);
yeccpars2(3, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 20, [3 | Ss], [T | Stack]);
yeccpars2(3, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 19, [3 | Ss], [T | Stack]);
yeccpars2(3, 'Y-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 22, [3 | Ss], [T | Stack]);
yeccpars2(3, 'X-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 21, [3 | Ss], [T | Stack]);
yeccpars2(3, 'Center', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 18, [3 | Ss], [T | Stack]);
yeccpars2(3, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(80, Cat, [3 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(4, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 77, [4 | Ss], [T | Stack]);
yeccpars2(4, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(5, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 66, [5 | Ss], [T | Stack]);
yeccpars2(5, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 69, [5 | Ss], [T | Stack]);
yeccpars2(5, 'Distance', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 67, [5 | Ss], [T | Stack]);
yeccpars2(5, 'Normal', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 68, [5 | Ss], [T | Stack]);
yeccpars2(5, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(70, Cat, [5 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(6, 'Equator', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 52, [6 | Ss], [T | Stack]);
yeccpars2(6, 'Pole', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 53, [6 | Ss], [T | Stack]);
yeccpars2(6, 'Bitmap', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 49, [6 | Ss], [T | Stack]);
yeccpars2(6, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 55, [6 | Ss], [T | Stack]);
yeccpars2(6, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 51, [6 | Ss], [T | Stack]);
yeccpars2(6, 'Center', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 50, [6 | Ss], [T | Stack]);
yeccpars2(6, 'Radius', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 54, [6 | Ss], [T | Stack]);
yeccpars2(6, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(57, Cat, [6 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(7, 'Bitmap', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 17, [7 | Ss], [T | Stack]);
yeccpars2(7, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 20, [7 | Ss], [T | Stack]);
yeccpars2(7, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 19, [7 | Ss], [T | Stack]);
yeccpars2(7, 'Y-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 22, [7 | Ss], [T | Stack]);
yeccpars2(7, 'X-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 21, [7 | Ss], [T | Stack]);
yeccpars2(7, 'Center', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 18, [7 | Ss], [T | Stack]);
yeccpars2(7, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(24, Cat, [7 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(8, 'Shadow', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 16, [8 | Ss], [T | Stack]);
yeccpars2(8, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 15, [8 | Ss], [T | Stack]);
yeccpars2(8, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(9, Cat, Ss, Stack, T, Ts, Tzr) ->
 yeccpars2(goto(spec, hd(Ss)), Cat, Ss, Stack, T, Ts, Tzr);
yeccpars2(10, Cat, Ss, Stack, T, Ts, Tzr) ->
 yeccpars2(goto(spec, hd(Ss)), Cat, Ss, Stack, T, Ts, Tzr);
yeccpars2(11, Cat, Ss, Stack, T, Ts, Tzr) ->
 yeccpars2(goto(spec, hd(Ss)), Cat, Ss, Stack, T, Ts, Tzr);
yeccpars2(12, 'Max', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 4, [12 | Ss], [T | Stack]);
yeccpars2(12, 'Use', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 8, [12 | Ss], [T | Stack]);
yeccpars2(12, 'Background', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 1, [12 | Ss], [T | Stack]);
yeccpars2(12, 'Light', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 3, [12 | Ss], [T | Stack]);
yeccpars2(12, 'Plane', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 5, [12 | Ss], [T | Stack]);
yeccpars2(12, 'Triangle', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 7, [12 | Ss], [T | Stack]);
yeccpars2(12, 'Sphere', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 6, [12 | Ss], [T | Stack]);
yeccpars2(12, 'Camera', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 2, [12 | Ss], [T | Stack]);
yeccpars2(12, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(14, Cat, [12 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(13, '$end', _, Stack, _, _, _) ->
 {ok, hd(Stack)};
yeccpars2(13, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(14, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  [_1|_2],
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(specs, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(15, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {use_reflection},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(config, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(16, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {use_shadow},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(config, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(17, string, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 48, [17 | Ss], [T | Stack]);
yeccpars2(17, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(18, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [18 | Ss], [T | Stack]);
yeccpars2(18, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(19, lpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 39, [19 | Ss], [T | Stack]);
yeccpars2(19, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(20, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [20 | Ss], [T | Stack]);
yeccpars2(20, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [20 | Ss], [T | Stack]);
yeccpars2(20, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(21, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [21 | Ss], [T | Stack]);
yeccpars2(21, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(22, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [22 | Ss], [T | Stack]);
yeccpars2(22, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(23, 'Bitmap', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 17, [23 | Ss], [T | Stack]);
yeccpars2(23, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 20, [23 | Ss], [T | Stack]);
yeccpars2(23, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 19, [23 | Ss], [T | Stack]);
yeccpars2(23, 'Y-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 22, [23 | Ss], [T | Stack]);
yeccpars2(23, 'X-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 21, [23 | Ss], [T | Stack]);
yeccpars2(23, 'Center', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 18, [23 | Ss], [T | Stack]);
yeccpars2(23, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(25, Cat, [23 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(24, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {triangle,merge_triangle_opts(_2)},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(object, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(25, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  [_1|_2],
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(triangle_opts, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(26, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [26 | Ss], [T | Stack]);
yeccpars2(26, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [26 | Ss], [T | Stack]);
yeccpars2(26, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(27, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {yaxis,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(topt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(28, Cat, Ss,  [_1|Stack], T, Ts, Tzr) ->
 Val =  value_of(_1),
 yeccpars2(goto(number, hd(Ss)), Cat, Ss, [Val | Stack], T, Ts, Tzr);
yeccpars2(29, Cat, Ss,  [_1|Stack], T, Ts, Tzr) ->
 Val =  value_of(_1),
 yeccpars2(goto(number, hd(Ss)), Cat, Ss, [Val | Stack], T, Ts, Tzr);
yeccpars2(30, comma, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 31, [30 | Ss], [T | Stack]);
yeccpars2(30, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(31, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [31 | Ss], [T | Stack]);
yeccpars2(31, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [31 | Ss], [T | Stack]);
yeccpars2(31, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(32, comma, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 33, [32 | Ss], [T | Stack]);
yeccpars2(32, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(33, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [33 | Ss], [T | Stack]);
yeccpars2(33, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [33 | Ss], [T | Stack]);
yeccpars2(33, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(34, rhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 35, [34 | Ss], [T | Stack]);
yeccpars2(34, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(35, Cat, Ss,  [_7,_6,_5,_4,_3,_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {_2,_4,_6},
 Nss = lists:nthtail(6, Ss),
 yeccpars2(goto(vector, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(36, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {xaxis,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(topt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(37, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {reflection,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(topt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(38, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {color,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(topt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(39, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [39 | Ss], [T | Stack]);
yeccpars2(39, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [39 | Ss], [T | Stack]);
yeccpars2(39, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(40, comma, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 41, [40 | Ss], [T | Stack]);
yeccpars2(40, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(41, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [41 | Ss], [T | Stack]);
yeccpars2(41, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [41 | Ss], [T | Stack]);
yeccpars2(41, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(42, comma, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 43, [42 | Ss], [T | Stack]);
yeccpars2(42, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(43, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [43 | Ss], [T | Stack]);
yeccpars2(43, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [43 | Ss], [T | Stack]);
yeccpars2(43, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(44, rpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 45, [44 | Ss], [T | Stack]);
yeccpars2(44, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(45, Cat, Ss,  [_7,_6,_5,_4,_3,_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {_2,_4,_6},
 Nss = lists:nthtail(6, Ss),
 yeccpars2(goto(color, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(46, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {center,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(topt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(47, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {bitmap,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(topt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(48, Cat, Ss,  [_1|Stack], T, Ts, Tzr) ->
 Val =  value_of(_1),
 yeccpars2(goto(name, hd(Ss)), Cat, Ss, [Val | Stack], T, Ts, Tzr);
yeccpars2(49, string, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 48, [49 | Ss], [T | Stack]);
yeccpars2(49, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(50, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [50 | Ss], [T | Stack]);
yeccpars2(50, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(51, lpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 39, [51 | Ss], [T | Stack]);
yeccpars2(51, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(52, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [52 | Ss], [T | Stack]);
yeccpars2(52, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(53, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [53 | Ss], [T | Stack]);
yeccpars2(53, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(54, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [54 | Ss], [T | Stack]);
yeccpars2(54, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [54 | Ss], [T | Stack]);
yeccpars2(54, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(55, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [55 | Ss], [T | Stack]);
yeccpars2(55, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [55 | Ss], [T | Stack]);
yeccpars2(55, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(56, 'Equator', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 52, [56 | Ss], [T | Stack]);
yeccpars2(56, 'Pole', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 53, [56 | Ss], [T | Stack]);
yeccpars2(56, 'Bitmap', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 49, [56 | Ss], [T | Stack]);
yeccpars2(56, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 55, [56 | Ss], [T | Stack]);
yeccpars2(56, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 51, [56 | Ss], [T | Stack]);
yeccpars2(56, 'Center', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 50, [56 | Ss], [T | Stack]);
yeccpars2(56, 'Radius', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 54, [56 | Ss], [T | Stack]);
yeccpars2(56, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(58, Cat, [56 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(57, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {sphere,merge_sphere_opts(_2)},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(object, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(58, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  [_1|_2],
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sphere_opts, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(59, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {reflection,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sopt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(60, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {radius,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sopt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(61, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {pole,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sopt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(62, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {equator,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sopt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(63, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {color,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sopt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(64, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {center,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sopt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(65, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {bitmap,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(sopt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(66, lpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 39, [66 | Ss], [T | Stack]);
yeccpars2(66, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(67, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [67 | Ss], [T | Stack]);
yeccpars2(67, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [67 | Ss], [T | Stack]);
yeccpars2(67, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(68, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [68 | Ss], [T | Stack]);
yeccpars2(68, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(69, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [69 | Ss], [T | Stack]);
yeccpars2(69, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [69 | Ss], [T | Stack]);
yeccpars2(69, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(70, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {plane,merge_plane_opts(_2)},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(object, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(71, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 66, [71 | Ss], [T | Stack]);
yeccpars2(71, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 69, [71 | Ss], [T | Stack]);
yeccpars2(71, 'Distance', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 67, [71 | Ss], [T | Stack]);
yeccpars2(71, 'Normal', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 68, [71 | Ss], [T | Stack]);
yeccpars2(71, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(72, Cat, [71 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(72, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  [_1|_2],
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(plane_opts, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(73, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {reflection,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(popt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(74, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {normal,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(popt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(75, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {distance,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(popt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(76, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {color,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(popt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(77, 'Depth', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 78, [77 | Ss], [T | Stack]);
yeccpars2(77, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(78, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [78 | Ss], [T | Stack]);
yeccpars2(78, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [78 | Ss], [T | Stack]);
yeccpars2(78, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(79, Cat, Ss,  [_4,_3,_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {max_reflection_depth,_4},
 Nss = lists:nthtail(3, Ss),
 yeccpars2(goto(config, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(80, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {light,merge_light_opts(_2)},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(object, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(81, 'Bitmap', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 17, [81 | Ss], [T | Stack]);
yeccpars2(81, 'Reflection', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 20, [81 | Ss], [T | Stack]);
yeccpars2(81, 'Color', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 19, [81 | Ss], [T | Stack]);
yeccpars2(81, 'Y-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 22, [81 | Ss], [T | Stack]);
yeccpars2(81, 'X-axis', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 21, [81 | Ss], [T | Stack]);
yeccpars2(81, 'Center', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 18, [81 | Ss], [T | Stack]);
yeccpars2(81, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(82, Cat, [81 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(82, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  [_1|_2],
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(light_opts, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(83, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [83 | Ss], [T | Stack]);
yeccpars2(83, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(84, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [84 | Ss], [T | Stack]);
yeccpars2(84, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [84 | Ss], [T | Stack]);
yeccpars2(84, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(85, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [85 | Ss], [T | Stack]);
yeccpars2(85, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(86, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [86 | Ss], [T | Stack]);
yeccpars2(86, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [86 | Ss], [T | Stack]);
yeccpars2(86, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(87, lhpar, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 26, [87 | Ss], [T | Stack]);
yeccpars2(87, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(88, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [88 | Ss], [T | Stack]);
yeccpars2(88, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [88 | Ss], [T | Stack]);
yeccpars2(88, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(89, float, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 28, [89 | Ss], [T | Stack]);
yeccpars2(89, integer, Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 29, [89 | Ss], [T | Stack]);
yeccpars2(89, _, _, _, T, Ts, _) ->
 yeccerror(element(2, T), T);
yeccpars2(90, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {camera,merge_camera_opts(_2)},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(camera, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(91, 'Resolution', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 86, [91 | Ss], [T | Stack]);
yeccpars2(91, 'Height', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 84, [91 | Ss], [T | Stack]);
yeccpars2(91, 'Width', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 88, [91 | Ss], [T | Stack]);
yeccpars2(91, 'Up', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 87, [91 | Ss], [T | Stack]);
yeccpars2(91, 'Zoom', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 89, [91 | Ss], [T | Stack]);
yeccpars2(91, 'Lookpoint', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 85, [91 | Ss], [T | Stack]);
yeccpars2(91, 'Eyepoint', Ss, Stack, T, Ts, Tzr) ->
 yeccpars1(Ts, Tzr, 83, [91 | Ss], [T | Stack]);
yeccpars2(91, Cat, Ss,  Stack, T, Ts, Tzr) ->
 Val =  [],
 yeccpars2(92, Cat, [91 | Ss], [Val | Stack], T, Ts, Tzr);
yeccpars2(92, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  [_1|_2],
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(camera_opts, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(93, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {zoom,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(copt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(94, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {width,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(copt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(95, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {up,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(copt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(96, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {resolution,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(copt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(97, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {lookpoint,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(copt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(98, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {height,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(copt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(99, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {eyepoint,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(copt, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(100, Cat, Ss,  [_2,_1|Stack], T, Ts, Tzr) ->
 Val =  {background,_2},
 Nss = lists:nthtail(1, Ss),
 yeccpars2(goto(config, hd(Nss)), Cat, Nss, [Val | Stack], T, Ts, Tzr);
yeccpars2(Other, _, _, _, _, _, _) ->
 exit({parser, Other, missing_state_in_action_table}).

goto(specs, 0) ->
 13;
goto(specs, 12) ->
 14;
goto(spec, 0) ->
 12;
goto(spec, 12) ->
 12;
goto(camera, 0) ->
 9;
goto(camera, 12) ->
 9;
goto(camera_opts, 2) ->
 90;
goto(camera_opts, 91) ->
 92;
goto(copt, 2) ->
 91;
goto(copt, 91) ->
 91;
goto(object, 0) ->
 11;
goto(object, 12) ->
 11;
goto(sphere_opts, 6) ->
 57;
goto(sphere_opts, 56) ->
 58;
goto(sopt, 6) ->
 56;
goto(sopt, 56) ->
 56;
goto(plane_opts, 5) ->
 70;
goto(plane_opts, 71) ->
 72;
goto(popt, 5) ->
 71;
goto(popt, 71) ->
 71;
goto(triangle_opts, 7) ->
 24;
goto(triangle_opts, 23) ->
 25;
goto(topt, 7) ->
 23;
goto(topt, 23) ->
 23;
goto(topt, 3) ->
 81;
goto(topt, 81) ->
 81;
goto(light_opts, 3) ->
 80;
goto(light_opts, 81) ->
 82;
goto(config, 0) ->
 10;
goto(config, 12) ->
 10;
goto(color, 19) ->
 38;
goto(color, 51) ->
 63;
goto(color, 66) ->
 76;
goto(color, 1) ->
 100;
goto(vector, 22) ->
 27;
goto(vector, 21) ->
 36;
goto(vector, 18) ->
 46;
goto(vector, 53) ->
 61;
goto(vector, 52) ->
 62;
goto(vector, 50) ->
 64;
goto(vector, 68) ->
 74;
goto(vector, 87) ->
 95;
goto(vector, 85) ->
 97;
goto(vector, 83) ->
 99;
goto(name, 17) ->
 47;
goto(name, 49) ->
 65;
goto(number, 26) ->
 30;
goto(number, 31) ->
 32;
goto(number, 33) ->
 34;
goto(number, 20) ->
 37;
goto(number, 39) ->
 40;
goto(number, 41) ->
 42;
goto(number, 43) ->
 44;
goto(number, 55) ->
 59;
goto(number, 54) ->
 60;
goto(number, 69) ->
 73;
goto(number, 67) ->
 75;
goto(number, 78) ->
 79;
goto(number, 89) ->
 93;
goto(number, 88) ->
 94;
goto(number, 86) ->
 96;
goto(number, 84) ->
 98;
goto(Symbol, State) ->
 exit({Symbol, State, missing_in_goto_table}).


