%    -*- Erlang -*- 
%    File:	ppm.erl  (~jb/projects/erl_tracer/ppm.erl)
%    Author:	Johan Bevemyr
%    Created:	Tue Sep 12 14:51:27 1995
%    Purpose:   Read and write bitmaps on PPM-format. For a description of
%               the PPM-format see http://www.csd.uu.se/~jb/ppm.ps
%
 
-module(ppm).
-export([ppm_write/3,
	 ppm_read/1]).
-compile(export_all).

-import(file,[open/2,close/1]).
-import(io,[get_chars/3,put_chars/2,get_line/2,format/2,format/3,nl/1]).
-import(rt_util,[reverse/1]).



% A bitmap is represented as a tuple {Width, Height, ListOfColors}

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Name: ppm_read(FileName)
%
% Desc: Returns a bitmap read from the file FileName on PPM-format.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ppm_read(FileName) ->
    case open(FileName, read) of
	{error, Reason} ->
	    format("Can't open file for reason:~s~n",[Reason]),
	    error ; 
	{ok, FileHandle} ->
	    Bitmap = ppm_read_file(FileHandle),
	    close(FileHandle),
	    Bitmap
    end.

%

ppm_read_file(Stream) ->
    {Format,Width,Height} = read_head(Stream),
    ppm_read_bitmap(Stream,Format,Width,Height).

%

read_head(Stream) -> 
    Format = read_digit(Stream),
    Width = read_digit(Stream),
    Height = read_digit(Stream),
    MaxCol = read_digit(Stream),
    {Format,Width,Height}.

%
    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Reading binary data
%

ppm_read_bitmap(Stream,Format,Width,Height) -> 
    case Format of
	3 ->
	    Bitmap = ppm_read_ascii(Stream,Width,Height,[]),
	    {Width,Height,reverse(Bitmap)};
	6 ->
	    Bitmap = ppm_read_binary(Stream,Width,Height,[]),
	    {Width,Height,reverse(Bitmap)}
    end.

%

ppm_read_binary(Stream,Width,Height,Acc) ->
    case Height of
	0 -> Acc ;
	Any ->
	    Row = read_binary_row(Stream,Width,[]),
	    ppm_read_binary(Stream,Width,Height-1,[reverse(Row)|Acc])
    end.

%

read_binary_row(Stream,Width,Acc) ->
    case Width of
	0 -> Acc;
	Any ->
	    Color = read_binary_color(Stream),
	    read_binary_row(Stream,Width-1,[Color|Acc])
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Reading ascii data
%

ppm_read_ascii(Stream,Width,Height,Acc) ->
    case Height of
	0 -> Acc ;
	Any ->
	    Row = read_ascii_row(Stream,Width,[]),
	    ppm_read_ascii(Stream,Width,Height-1,[reverse(Row)|Acc])
    end.

%

read_ascii_row(Stream,Width,Acc) ->
    case Width of
	0 -> Acc;
	Any ->
	    Color = read_ascii_color(Stream),
	    read_ascii_row(Stream,Width-1,[Color|Acc])
    end.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Name: ppm_write(File,Bitmap,Mode)
%
% Desc: write bitmap to file on PPM-format in either ascii or binary
%       mode.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ppm_write(FileName,Bitmap,Mode) ->
    case open(FileName, write) of
	{error, Reason} ->
	    format("Can't open file for reason:~s~n",[Reason]),
	    error ; 
	{ok, FileHandle} ->
	    ppm_write_file(FileHandle,Bitmap,Mode),
	    close(FileHandle),
            ok
    end.

%

ppm_write_file(Stream,{Width,Height,Pixels},Mode) ->
    write_header(Stream,Width,Height,Mode),
    ppm_write_picture(Pixels,Stream,Mode).

%

write_header(Stream,Width,Height,Mode) ->
    case Mode of
	ascii ->
	    format(Stream,"P3~n# Erlang tracer~n~w ~w~n255~n",[Width,Height]);
	binary ->
	    format(Stream,"P6~n# Erlang tracer~n~w ~w~n255~n",[Width,Height])
    end.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Write pixels
%

ppm_write_picture([],_,_) -> done;

ppm_write_picture([Row|Rows],Stream,Mode) ->
    ppm_write_row(Row,Stream,Mode,0),
    ppm_write_picture(Rows,Stream,Mode).

%

ppm_write_row([],_,_,_) -> done ;

ppm_write_row([Pixel|Row],Stream,Mode,Count) ->
    case Mode of
	ascii ->
	    case Count rem 5 of
		0 ->
		    nl(Stream);
		Other ->
		    put_chars(Stream," ")
	    end;
	binary ->
	    true
    end,
    ppm_write_pixel(Pixel,Stream,Mode),
    ppm_write_row(Row,Stream,Mode,Count+1).

%

ppm_write_pixel(Color,Stream,Mode) ->
    case Mode of
	ascii ->
	    put_chars(Stream," "),
	    write_ascii_color(Stream,Color) ;
	binary ->
	    write_binary_color(Stream,Color)
    end.
	    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% support
%

read_digit(Stream) ->
    FirstDigit = skip_nondigits(Stream),
    list_to_integer([FirstDigit|read_digit_rest(Stream)]).

read_digit_rest(Stream) ->
    [Char] = get_chars(Stream, not_used, 1),
    case is_digit(Char) of
	digit ->
	    [Char|read_digit_rest(Stream)];
	comment ->
	    get_line(Stream, not_used),
	    [];
	nodigit ->
	    []
    end.

skip_nondigits(Stream) ->
    [Char] = get_chars(Stream, not_used, 1),
    case is_digit(Char) of
	digit ->
	    Char;
        comment ->
	    get_line(Stream, not_used),
	    skip_nondigits(Stream);
	nodigit ->
	    skip_nondigits(Stream)
    end.
	

is_digit(Char) ->
    if
	Char >= 48, Char =< 57 ->
	    digit ;
	Char == 35 ->
	    comment ;
	true -> nodigit
    end.
    
read_ascii_color(Stream) ->
    R = read_digit(Stream),
    G = read_digit(Stream),
    B = read_digit(Stream),
    {R,G,B}.

read_binary_color(Stream) ->
    [R,G,B] = get_chars(Stream,not_used,3),
    {R,G,B}.

write_ascii_color(Stream,{R,G,B}) ->
    format(Stream,"~3w ~3w ~3w",[R,G,B]).

write_binary_color(Stream,{R,G,B}) ->
    put_chars(Stream,[R,G,B]).



