%    -*- Erlang -*- 
%    File:	ray_tracer.erl  (~jb/projects/erl_tracer/tracer.erl)
%    Author:	Johan Bevemyr
%    Created:	Tue Sep 12 22:10:28 1995
%    Purpose:   Implement a ray tracer. For a description of the
%               algorithms see http://www.csd.uu.se/~jb/kursinfo.ps
%
 

-module(ray_tracer).
-export([trace_pic/3,intersect/2,shadowray/5]).
-compile(export_all).

-import(vector,[vector_sub/2,vector_add/2,
		vector_scale/2,vector_norm/1,vector_mult/2,vector_length/1]).

-import(camera,[camera_ray/3,camera_size/1]).
-import(rt_util,[map/3,keysort/2,sum/1]).
-import(math,[sqrt/1,acos/1,sin/1,pi/0]).
-import(bitmap,[bitmap_get_color/3]).
-import(color,[color_scale/2]).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Name: trace/3
%
% Returns: a bitmap
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

trace_pic(Camera,World,Light) ->
    {Width,Height} = camera_size(Camera),
    Pixels = trace_picture(Height,Width,Camera,World,Light,[]),
    {Width,Height,Pixels}.

%

trace_picture(Height,Width,Camera,World,Light,Acc) ->
    case Height of
	0 -> Acc ;
	Other ->
	    Row = trace_row(Width,Height,Camera,World,Light,[]),
	    Rows = trace_picture(Height-1,Width,Camera,World,Light,[Row|Acc])
    end.

%

trace_row(Width,Height,Camera,World,Light,Acc) ->
    case Width of
	0 -> Acc;
	Other ->
	    Pixel = trace_ray(camera_ray(Width,Height,Camera),
			      World,Light),
	    trace_row(Width-1,Height,Camera,World,Light,[Pixel|Acc])
    end.

%

trace_ray(Ray,World,Light) ->
    Distances = map({ray_tracer,intersect},[Ray],World),
    [Closest|Rest] = keysort(1,Distances),
    {Distance, HitP, Norm, Surface} = Closest,
    if 
	Distance < 9999999 ->
	    {Refl,Color} = Surface,
	    Shading = map({ray_tracer,shadowray},
			  [HitP,Norm,World,Light],Light),
	    Stotal = sum(Shading),
	    color_scale(Color,Stotal);
	true ->
	    background()
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Shading
%

shadowray(LPoint,HitP,Norm,World,Light) ->
    LDist = vector_length(vector_sub(LPoint,HitP)),
    Ray = make_shadow_ray(HitP,Norm,LPoint),
    {So,Sd} = Ray,
    Distances = map({ray_tracer,intersect},[Ray],World),
    [Closest|Rest] = keysort(1,Distances),
    {Distance, _HitPoint, _HitNorm, _HitSurface} = Closest,
    if
	Distance > LDist ->
	    abs(vector_mult(Norm,Sd)) ;
	true ->
	    0
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Intersection
%    

intersect({sphere,Params},Ray) ->
    Hit = sphere_hit(Params,Ray),
    sphere_check_bitmap(Hit);

intersect({plane,Params},Ray) ->
    plane_hit(Params,Ray);

intersect({triangle,Params},Ray) ->
    triangle_hit(Params,Ray);

intersect(Other,Ray) ->
    io:format("Unknown object ~w~n",[Other]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% sphere intersect
%

sphere_hit(Params,Ray) ->
    {Center,Radius,Surface} = Params,
    {Origin,Direction} = Ray,
    OC = vector_sub(Center,Origin),
    L2oc = vector_mult(OC,OC),
    R2 = Radius * Radius,
    if
	L2oc < R2 ->                               % inside sphere
	    Tca = vector_mult(OC,Direction),
	    T2hc = (Radius*Radius) - L2oc + (Tca*Tca),
	    T = Tca+sqrt(T2hc),
	    HitP = vector_add(Origin,vector_scale(Direction,T)),
	    Norm = vector_scale(vector_sub(HitP,Center),1/Radius),
	    {T,HitP,Norm,Surface};
	true ->                                         % outside sphere
	    Tca = vector_mult(OC,Direction),
	    if
		Tca > 0 ->
		    T2hc = (Radius*Radius) - L2oc + (Tca*Tca),
		    if
			T2hc > 0 ->
			    T = Tca-sqrt(T2hc),
			    HitP = vector_add(Origin,
					      vector_scale(Direction,T)),
			    Norm = vector_scale(vector_sub(HitP,Center),
						1/Radius),
			    {T,HitP,Norm,Surface};
			true ->
			    no_hit()
		    end;
		true ->
		    no_hit()
	    end
    end.

sphere_check_bitmap(Hit) ->
    {Dist,HitP,Norm,Surface} = Hit,
    case Surface of
	{bitmap,Ref,Pole,Equator,Cross,Bitmap} ->
	    Phi = acos(-1*vector_mult(Norm,Pole)),
	    V = Phi/pi(),
	    case V of
		0 -> U = 0;
		1 -> U = 0;
		Other ->
		    Theta = (acos(vector_mult(Equator,Norm)/sin(Phi)))/
			(2*pi()),
		    CRn = vector_mult(Cross,Norm),
		    if
			CRn > 0 ->
			    U = Theta;
			true ->
			    U = 1-Theta
		    end
	    end,
	    Color = bitmap_get_color(1-U,1-V,Bitmap),
	    {Dist,HitP,Norm,{Ref,Color}};
	{color,Ref,Color} ->
	    {Dist,HitP,Norm,{Ref,Color}};
	nohit ->
	    Hit;
	Otherwise ->
	    io:format("Unknown surface of shpere ~w~n",[Otherwise])
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% triangle intersect

triangle_hit(Params,Ray) -> 
    {Center,A,B,Norm,Dist,Surface} = Params,
    {T,Ri,Rn,_Surf} = plane_hit({Norm,Dist,no_surf},Ray),
    case Ri of
	nohit ->
	    no_hit();
	Other ->
	    Q = vector_sub(Ri,Center),
	    AB = vector_mult(A,B),
	    QB = vector_mult(Q,B),
	    BB = vector_mult(B,B),
	    U = (BB*vector_mult(Q,A)-
		 AB*QB)/
		(vector_mult(A,A)*BB- AB*AB),
	    if 
		U < 0 ->
		    no_hit();
		U > 1 -> 
		    no_hit();
		true ->
		    V = (QB-U*AB)/BB,
		    UV = U+V,
		    if
			V < 0 ->
			    no_hit();
			V > 1 ->
			    no_hit();
			UV > 1 ->
			    no_hit();
			true ->
			    {T,Ri,Rn,triangle_surf(U,V,Surface)}
		    end
	    end
    end.
			    
triangle_surf(U,V,{color,Ref,Color}) ->
    {Ref,Color};

triangle_surf(U,V,{bitmap,Ref,Bitmap}) ->
    {Ref,bitmap_get_color(U,V,Bitmap)}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% plane intersect

plane_hit(Params,Ray) -> 
    {Norm,Dist,Surf} = Params,
    {Orig,Dir} = Ray,
    Vd = vector_mult(Norm,Dir),
    if
	Vd == 0.0 ->
	    no_hit();
	true ->
	    V0 = Dist - vector_mult(Norm,Orig),
	    T = V0/Vd,
	    if
		T < 0 ->
		    no_hit();
		true ->
		    Ri = vector_add(Orig,vector_scale(Dir,T)),
		    if
			Vd < 0 ->
			    Rn = Norm;
			true ->
			    Rn = vector_scale(Norm,-1)
		    end,
		    {T,Ri,Rn,Surf}
	    end
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
%

no_hit() -> {9999999,nohit,nohit,nohit}.

background() -> {0,0,0}.

make_shadow_ray(HitP,Norm,LPoint) ->
    Origin = vector_add(HitP,vector_scale(Norm,0.0001)),
    Direction = vector_norm(vector_sub(LPoint,Origin)),
    {Origin,Direction}.

    












