%    -*- Erlang -*- 
%    File:	tracer.erl  (~jb/projects/erl_tracer/lex.erl)
%    Author:	Johan Bevemyr
%    Created:	Thu Sep 14 16:53:59 1995
%    Purpose:   Interface to ray_tracer.

%
%  Example of use:
%
%    Eshell V4.3  (abort with ^G)
%    1> c(tracer).
%    {ok,tracer}
%    2> tracer:ray_trace('test.desc','test.ppm').
%    Picture trace in 2090 ms
%    ok
%    3>
%

-module(tracer).

-export([read_desc/1, sort_desc/4, ray_trace/2, run/0]).
-compile(export_all).

-import(lexer,[string/1]).
-import(file,[read_file/1]).
-import(parser,[parse/1]).
-import(rt_util,[append/2]).
-import(ray_tracer,[trace_pic/3]).
-import(ppm,[ppm_write/3,ppm_read/1]).
-import(vector,[vector_norm/1,vector_cross/2,vector_mult/2]).
-import(camera,[camera_init/1]).
-import(io,[format/2]).




run() ->
    tracer:ray_trace('test.desc', 'test.ppm').



ray_trace(InFile,OutFile) ->
    case read_desc(InFile) of
	{ok, Desc} ->
	    {Cam,Obj,Lig} = sort_desc(Desc,[],[],[]),
	    timing:reset_time(),
  	    Bitmap = trace_pic(Cam,Obj,Lig),
	    R = timing:time(),
	    ppm_write(OutFile,Bitmap,binary),
	    {tracer, R};
	{error,Res} ->
	    {error, Res}
    end.

new_ray_trace(InFile,OutFile) ->
    case read_desc(InFile) of
	{ok, Desc} ->
	    {Cam,Obj,Lig} = sort_desc(Desc,[],[],[]),
	
  	    Bitmap = trace_pic(Cam,Obj,Lig),
	
	    ppm_write(OutFile,Bitmap,binary),
	    {tracer, ok};
	{error,Res} ->
	    {error, Res}
    end.


read_desc(File) ->
    case read_file(File) of
	{ok,Binary} ->
	    String = binary_to_list(Binary),
	    parse(append(string(String),[{'$end',999}]));
	{error,Reason} ->
	    {error,Reason}
    end.

%

sort_desc([],AccC, AccO, AccL)  ->
    {AccC, AccO, AccL};
sort_desc([Obj|Rest],AccC,AccO,AccL) ->
    case Obj of
	{camera,Opts} ->
	    sort_desc(Rest,camera_init(Opts),AccO,AccL);
	{sphere,Param} ->
	    sort_desc(Rest,AccC,[sphere_init(Obj)|AccO],AccL);
	{plane,Param} ->
	    sort_desc(Rest,AccC,[plane_init(Obj)|AccO],AccL);
	{triangle,Param} ->
	    sort_desc(Rest,AccC,[triangle_init(Obj)|AccO],AccL);
	{light,Param} ->
	    sort_desc(Rest,AccC,AccO,[Param|AccL])
    end.
	

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialization
%
	
sphere_init({sphere,Params}) ->
    {Center,Radius,Color,Refl,Pole,Equator,Cross,Bitm} = Params,
    case Bitm of
	no_bitmap ->
	    {sphere,{Center,Radius,{color,Refl,Color}}};
	FileName ->
	    NP = vector_norm(Pole),
	    NE = vector_norm(Equator),
	    {sphere,{Center,Radius,{bitmap,Refl,NP,NE,vector_cross(NP,NE),
				    ppm_read(FileName)}}}
    end.

plane_init({plane,Params}) -> 
    {Norm,Dist,Refl,Color} = Params,
    {plane,{vector_norm(Norm),Dist,{Refl,Color}}}.

triangle_init({triangle,Params}) -> 
    {Center,Xaxis,Yaxis,Color,Reflection,Bitmap} = Params,
    Norm = vector_norm(vector_cross(Xaxis,Yaxis)),
    Dist = vector_mult(Center,Norm),
    case Bitmap of
	no_bitmap ->
	    Surface = {color,Reflection,Color};
	Filename ->
	    Surface = {bitmap,Reflection,ppm_read(Filename)}
    end,
    {triangle,{Center,Xaxis,Yaxis,Norm,Dist,Surface}}.
