%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%% Author: Claes Wikstrom  klacke@erix.ericsson.se
%% Purpose: Implement the distributed erlang authentication protocoll
%% and provide an interface to the unix passwd facility
%% This module is internal to the net_kernel and is not intended
%% to be used by mere mortals.
%% The reason for providing it in a separate module is to
%% make it easier to replace it on systems not supporting
%% unix passwords, or to replace or augment it with a local 
%% authentication schema, since this module does not have to be preloaded.


-module(auth).
-export([start/0, loop0/0, auth/1 , is_auth/1, stop/0,
	 open/1, exists/1, authorize/2, print/3]).
-export([cookie/0, cookie/1, node_cookie/1, 
	 node_cookie/2,doit/2]).

start() -> generic:start(auth, auth, loop0, []).
stop()  -> generic:stop(auth).
open(Name) -> generic:request(auth, {open, Name}).

loop0() ->
    case init:get_flag('-nocookie') of
	{'-nocookie', true} -> 
	    ignore;
	notfound ->
	    case init:get_flag('-setcookie') of
		{'-setcookie', C} ->
		    erlang:set_cookie(node(),C);
		notfound ->
		    %% Here is the default 
		    cookie()
	    end;
	Other ->
	    ignore
    end,
    generic:sync_start(auth, started),  %% synch now
    generic:loop(auth, [], {auth, doit}).

%% Opened is a list of servers we have opened up, 
%% The net kernel will let all message to the auth server 
%% through as is

%% message to handle the passwd auth protocoll

doit({From, {Node, authorize_me}}, O) ->
    spawn(auth, authorize, [From, Node]), 
    {noreply, O};

%% A series of bad messages that may come.
doit({From,badcookie,net_kernel,{From,spawn,M,F,A,Gleader}}, O) ->
    auth:print(node(From) ,"~n** Unauthorized spawn attempt to ~w **~n",
	      [node()]),
    erlang:disconnect_node(node(From)), 
    {noreply, O};

doit({From,badcookie,net_kernel,{From,spawn_link,M,F,A,Gleader}}, O) ->
    auth:print(node(From),
	      "~n** Unauthorized spawn_link attempt to ~w **~n",
	      [node()]),
    erlang:disconnect_node(node(From)), 
    {noreply, O};

%% allways allow print outs
doit({print,What,Args}, O) ->
    error_logger:error_msg(What,Args), 
    {noreply, O};

%% Ignore bad messages to the ddd server, they will be resent
%% If the authentication is succesful
doit({From, badcookie, ddd_server, Mess}, O) ->
    {noreply, O};

doit({From, badcookie, rex, Msg}, O) ->
    auth:print(getnode(From), 
	      "~n** Unauthorized rpc attempt to ~w **~n",[node()]),
    disconnect_node(node(From)), 
    {noreply, O};

%% These two messages has to do with the auth:is_auth() call
	
doit({From, badcookie,net_kernel,{From, {is_auth, Node}}}, O) -> %% ho ho
    From ! {auth_reply,node(), no},
    {noreply, O};

doit({From, badcookie, To, {auth_reply, R}}, O) ->  %% Let auth replys through
    catch To ! {auth_reply, R}, 
    {noreply, O};

doit({From, badcookie, To, {auth_reply, N,R}}, O) ->%% Let auth replys through
    catch To ! {auth_reply, N,R}, 
    {noreply, O};

%% This may be registered send as well as pid send.
doit({From,badcookie,Name,Mess}, Opened) ->
    case lists:member(Name, Opened) of
	true ->
	    catch Name ! Mess;
	false ->
	    case catch lists:member(element(1, Mess), Opened) of
		true ->   
		    catch Name ! Mess;  %% Might be a pid as well
		_ ->
		    auth:print(getnode(From), 
			      "~n** Unauthorized send attempt ~w to ~w **~n",
			      [Mess,node()]),
		    erlang:disconnect_node(getnode(From))
	    end
    end, 
    {noreply, Opened};

doit({From, {open, Name}}, Opened)  ->
    if
	node(From) == node() ->
	    {reply, From, ok, [Name | Opened]};
	true ->
	    {reply, From, error, Opened}
    end;

doit({'EXIT', _, _} , O)->   %% Ignore EXIT signals 
    {noreply, O};

doit({From, stop}, O) ->
    if
	node(From) == node() ->
	    {stopped, From};
	true ->
	    {reply, From, error, O}
    end;

doit({From, echo}, O) -> 
    From ! {auth_reply, hello},
    {noreply, O}.


getnode(P) when pid(P) -> node(P);
getnode(P) -> P.


print(Node,Format,Args) ->
    {auth,Node} ! {print,Format,Args}.


%% This code runs on the clinet side in the client process.
auth(Node) ->
    case {exists(Node), is_auth(Node)} of
	{no, _} -> no_such_node;
	{yes, yes} -> authorized;
	{yes, no} -> 
	    erlang:monitor_node(Node, true),
	    {auth, Node} ! {self(), {node(), authorize_me}},
	    receive
		{auth_reply, {Authpid, authorize, Host, User}} ->
		    S=lists:flatten(io_lib:format( "Password for ~w on ~w:",
						  [User,Host])),
		    Pwd=lib:nonl(io:get_line(list_to_atom(S))),
		    C = erlang:get_cookie(),
		    Authpid ! {auth_reply, node(),{passwd,Pwd,C}},
		    receive
			{nodedown, N} -> 
			    failure;
			{auth_reply, {authorized, Cookie}} ->
			    erlang:set_cookie(Node,Cookie),
			    net_kernel  ! {nodeup,Node},
			    erlang:monitor_node(Node, false),
			    authorized
		    end;
		{nodedown, Node} ->
		    failure
	    end
    end.

is_auth(Node) ->   
    case rpc:server_call(Node, net_kernel, auth_reply, {is_auth, node()}) of
	{error, _} -> no;
	X -> X
    end.

exists(Node) ->
    erlang:monitor_node(Node, true),
    {auth, Node} ! {self(), echo},
    R= receive
	{auth_reply, hello} ->
	    yes;

	{nodedown, Node} -> 
	    no
    end,
    erlang:monitor_node(Node, false),
    R.

%% Run on the server side, also in a separate process

authorize(From, Node) ->
    Host = list_to_atom(lib:nonl(unix:cmd("hostname"))),
    User = list_to_atom(lib:nonl(unix:cmd("echo $USER"))),
    authorize(From, Node,Host,User).

authorize(From, Node,Host,User) ->
    From ! {auth_reply,{self(),authorize,Host,User}},
    receive
	{auth_reply, Node, {passwd, P, Cookie}} ->
	    case verify(User,P) of 
		ok ->
		    erlang:set_cookie(Node,Cookie),
		    net_kernel ! {nodeup,Node},
		    From ! {auth_reply,{authorized,erlang:get_cookie()}};
		nok ->
		    auth:print(node(From), "Auth attempt failed \n", []),
		    erlang:disconnect_node(Node)
	    end
    after 30000 ->
	    auth:print(node(From), "Auth attempt timed out \n", []),
	    erlang:disconnect_node(Node)
    end.

%% Don't have the password on the commandline to the shell
%%
verify(User,Pwd) ->
    Cmd  = lists:flatten(lists:concat(["authchk ",User])),
    Port = open_port({spawn,Cmd},[{packet,2}]),
    Port ! {self(),{command,Pwd}},
    receive
	{Port,{data,"Ok"}} ->
	    ok;
	{Port,_}  ->
	    nok;
	{'EXIT',Port,_} ->
	    exit
    end.

%% Some functions to deal with cookies

%% Read cookie from $HOME/.erlang.cookie and set it

cookie() ->
    P = open_port({spawn,"authchk -cookie"},[use_stdio,{packet,2},eof]),
    R = receive
	{P,{data,[$y|Tail]}} ->
	    {ok,Tail};
	{P,{data,[$n|Tail]}} ->
	    {error,Tail};
	{P,eof} ->
	    {error,"Unexpected exit by authchk~n"};
	{P,_} ->
	    {error,"Bad reply from authchk~n"}
    end,
    P ! {self(),close},
    wait_close(P),
    case R of
	{ok,Cookie} ->
	    erlang:set_cookie(node(),list_to_atom(Cookie));
	{error,Msg} ->
	    error_logger:error_msg(Msg,[]),
	    false
    end.

wait_close(P) ->
    receive
	{P,closed} ->
	    ok;
	{P,_} ->
	    wait_close(P)
    end.

cookie([Cookie]) ->
    erlang:set_cookie(node(),Cookie).

%% Exchange cookies with Node, which uses Cookie
%% (Our own cookie must already be set)
node_cookie(Node, Cookie) ->
    %% First set our idea of his cookie
    erlang:set_cookie(Node, Cookie),
    %% Now we can send messages to him
    %% - but he can't send to us until we tell him our cookie
    %% Actually, we just set his idea of our cookie from here
    rpc:call(Node, erlang, set_cookie, [node(), erlang:get_cookie()]).
%% Version that can be called from the command line
node_cookie([Node, Cookie]) ->
    node_cookie(Node, Cookie).

