%% Copyright (C) 1991, Ellemtel Telecommunications Systems Laboratories
%% File	   : c.erl
%% Author  : Joe Armstrong, Robert Virding, Claes Wikstrom
%% Purpose : Utilities to use from shell.

-module(c).

-export([help/0,lc/1,c/1,c/2,nc/1,nc/2, nl/1,l/1,i/0,ni/0,
	 i/3,pid/3,m/0,m/1,
	 erlangrc/1,bi/1, flush/0, regs/0, 
	 nregs/0,pwd/0,ls/0,ls/1,cd/1]).

-import(io, [format/1, format/2]).

help() ->
    format('c(File)    -- compile and load code in <File>~n'),
%%    format('cd(Dir)    -- cd Dir .. example cd("..")~n'),
    format('flush()    -- flush any messages sent to the shell~n'),
    format('help()     -- help info~n'),
    format('i()        -- information about the system~n'),
    format('ni()       -- information about the networked system~n'),
    format('i(X,Y,X)   -- information about pid <X,Y,Z> ~n'),
    format('l(File)    -- Load module in <File>~n'),
    format('lc([File]) -- Compile a list of Erlang modules~n'),
    format('ls()       -- list files in the current directory~n'),
    format('ls(Dir)    -- list files in directory <Dir>~n'),
    format('m()        -- which modules are loaded~n'),
    format('m(Mod)     -- information about module <Mod>~n'),
    format('nc(File)   -- compile and load code in <File> on all nodes~n'),
    format('nl(File)   -- Load module in <File> on all nodes~n'),
    format('pid(X,Y,Z) -- convert X,Y,Z to a Pid~n'),
    format('pwd()      -- print working directory~n'),
    format('regs()     -- information about registered processes~n'),
    format('nregs()    -- information about all registered processes~n').

%% c(FileName)
%%  Compile a file/module.

c(File) ->
    c(File, []).


c(File, Opt) when atom(Opt) -> c(File, [Opt]);
c(File, Opts0) ->
    Opts = [report_errors, error_summary | Opts0],
    case code:interpreted(File) of
	false ->
	    case compile:file(File, Opts) of
		{ok,Mod} ->
		    code:purge(Mod),
		    check_load(code:load_file(Mod), Mod);
		Other ->                                %Errors go here
		    Other
	    end;
	_ ->
	    format("*** Warning: Module ~p is interpreted.~n",[File]),
	    format("             No object file will be created.~n"),
	    int:i(File,Opts)
    end.

check_load({error, R}, _) -> {error, R};
check_load(_, Mod) -> 
    %% Now check where we found it
    case code:is_loaded(Mod) of
	{file, X} ->
	    Str = atom_to_list(X), 
	    case Str of
		[$.,$/|_] ->
		    ok;
		Other ->
		    format('*** Warning Module ~w loaded from:~s ***\n',
			      [Mod, Str]),
		    format('*** THIS WAS NOT THE CURRENT DIRECTORY ***\n')
	    end;
	_ ->
	    uuugh
    end,
    {ok, Mod}.

%% Compile a list of modules 
%% enables the nice unix shell cmd
%% erl -s c lc f1 f2 f3 -s erlang halt
%% to compile files f1.erl , f2.erl ....... from a unix shell

lc([]) -> [];
lc([F|Tail]) -> [compile:file(F, [fast, report_errors, error_summary]) |
		 lc(Tail)].

nc(File) -> nc(File, []).
nc(File, Opt) when atom(Opt) ->
    nc(File, [Opt]);
nc(File, CompilerOpt) ->
    Opts = lists:append(CompilerOpt, [report_errors, error_summary]),
    case code:interpreted(File) of
	false ->
	    case compile:file(File, Opts) of
		{ok,Mod} ->
		    Fname = lists:concat([File, code:objfile_extension()]),
		    case file:read_file(Fname) of
			{ok,Bin} ->
			    rpc:eval_everywhere(code,load_binary,[Mod,Fname,Bin]),
			    {ok,Mod};
			Other ->
			    Other
		    end;
		Other ->                                %Errors go here
		    Other
	    end;
	_ ->
	    format("*** Warning: Module ~p is interpreted.~n",[File]),
	    format("             No object file will be created.~n"),
	    int:ni(File,Opts)
    end.

%% l(Mod)
%%  Reload module Mod from file of same name

l(Mod) ->
    code:purge(Mod),
    code:load_file(Mod).

%% Network version of l/1
nl(Mod) ->
    case code:get_object_code(Mod) of
	{Module, Bin, Fname} ->
            rpc:eval_everywhere(code,load_binary,[Mod,Fname,Bin]);
	Other ->
	    Other
    end.

all_procs() ->
    all_procs(is_alive()).
all_procs(false) -> processes();
all_procs(true) ->
    X = [node() | nodes()],
    lists:flatten(collect(X)).

collect([]) -> [];
collect([N|Tail]) ->
    [rpc:call(N,erlang,processes,[]) | collect(Tail)].

pinfo(Pid) ->
    pinfo1(is_alive(),Pid).
pinfo1(false,Pid ) -> process_info(Pid);
pinfo1(true,Pid) ->
    N = node(Pid),
    rpc:call(N,erlang,process_info,[Pid]).

ni() ->
    hformat("Pid", "Initial Call", "Current Function", "Reds", "Msgs"),
    display_info(all_procs(), 0, 0).
i() ->
    hformat("Pid", "Initial Call", "Current Function", "Reds", "Msgs"),
    display_info(processes(), 0, 0).

display_info([Pid|T], R, M) ->
    case pinfo(Pid) of
	undefined ->
	    display_info(T, R, M);
	Info ->
	    Call = fetch(initial_call, Info),
	    Curr = case fetch(current_function, Info) of
		       {Mod,F,Args} when list(Args) ->
			   {Mod,F,length(Args)};
		       Other ->
			   Other
		   end,
	    Reds  = fetch(reductions, Info), 
	    LM = length(fetch(messages, Info)),
	    pformat(io_lib:format("~w", [Pid]),
		    io_lib:format("~w", [Call]),
		    io_lib:format("~w", [Curr]), Reds, LM),
	    display_info(T, R+Reds, M + LM)
    end;
display_info([], R, M) ->
    pformat("Total", "", "", R, M).

hformat(A1, A2, A3, A4, A5) ->
    format("~-12s ~-21s ~-21s ~9s ~4s~n", [A1,A2,A3,A4,A5]).

pformat(A1, A2, A3, A4, A5) ->
    format("~-12s ~-21s ~-21s ~9w ~4w~n", [A1,A2,A3,A4,A5]).

fetch(Key, Info) ->
    case lists:keysearch(Key, 1, Info) of
	{value, {_, Val}} ->
	    Val;
	_ ->
	    0
    end.

pid(X,Y,Z) ->
    list_to_pid(lists:concat(['<',integer_to_list(X),'.',
			      integer_to_list(Y),'.',
			      integer_to_list(Z),'>'])).

i(X,Y,Z) ->
    pinfo(pid(X,Y,Z)).

m() ->
    mformat("Module", "File"),
    m1(lists:sort(code:all_loaded())).

m1([{Mod, File}|T]) ->
    mformat(Mod, File),
    m1(T);
m1([]) ->
    true.

mformat(A1, A2) ->
    format("~-20s     ~s\n", [A1,A2]).

%% erlangrc(Home)
%%  Try to run a ".erlang" file, first in the current directory
%%  else in home directory.

erlangrc([Home]) ->
    file:path_eval([".",Home], ".erlang").

bi(I) ->
    case erlang:info(I) of
	X when binary(X) -> io:put_chars(binary_to_list(X));
	X when list(X) -> io:put_chars(X);
	X -> format("~w", [X])
    end.

%%
%% Short and nice form of module info
%%

m(M) ->
    L = apply({M, module_info},[]),
    {value, {exports, E}} = lists:keysearch(exports, 1, L),
    Srcfile = src_file(L),
    Time = get_time(L),
    format("Module ~w compiled: ",[M]), print_time(Time), 
    format("Sourcefile:  ~s~n",[Srcfile]),
    print_object_file(lists:keysearch(M, 1, code:all_loaded())),
    format("Exports: ~n",[]), print_exports(lists:keysort(1,E)).

print_object_file({value,{Mod,File}}) ->
    format("Objectfile:  ~s~n",[to_list(File)]);
print_object_file(_) -> ignore.


print_exports(X) when length(X) > 16 -> 
    split_print_exports(X);
print_exports([]) -> ok;
print_exports([{F, A} |Tail]) -> 
    format("         ~w/~w~n",[F, A]), 
    print_exports(Tail).

get_time(L) -> 
    case lists:keysearch(compiletime, 1, L) of
        {value, {compiletime, Time}} -> 
	    Time;
	false ->
	    notime
    end.

src_file(L) ->
    case lists:keysearch(info, 1, L) of
	{value, {info, I}} ->
	    case lists:keysearch(source_file, 1, I) of
		false -> "No source file info available";
		{value, {source_file, Src}} -> Src
	    end;
	false ->
	    case lists:keysearch(source_file, 1, L) of
		false -> "No source file info available";
		{value, {source_file, Src}} -> Src
	    end
    end.

split_print_exports(L) ->
    Len = length(L),
    Mid = Len div 2,
    L1 = lists:sublist(L, 1, Mid),
    L2 = lists:sublist(L, Mid +1, Len - Mid + 1),
    split_print_exports(L1, L2).

split_print_exports([], [{F, A}|T]) ->
    Str = " ",
    format("~-30s~w/~w~n",[Str, F, A]), 
    split_print_exports([], T);

split_print_exports([{F1, A1}|T1], [{F2, A2} | T2]) ->
    Str = lists:flatten(io_lib:format("~w/~w",[F1, A1])),
    format("~-30s~w/~w~n",[Str, F2, A2]), 
    split_print_exports(T1, T2);

split_print_exports([], []) -> ok.		       

print_time({Year,Month,Day, Hour,Min, Secs}) ->
    format("Date: ~s ~w ~w, ",[month(Month), Day, Year]),
    format("Time: ~s.~s~n",[maybepad(Hour), maybepad(Min)]);

print_time(notime) ->
    format("No compile time info available~n",[]).

maybepad(X) when X < 10  -> lists:flatten(io_lib:format("0~w",[X]));
maybepad(X) -> lists:flatten(io_lib:format("~w",[X])).

month(1) -> "January";
month(2) -> "February";
month(3) -> "March";
month(4) -> "April";
month(5) -> "May";
month(6) -> "June";
month(7) -> "July";
month(8) -> "August";
month(9) -> "September";
month(10) -> "October";
month(11) -> "November";
month(12) -> "December".

to_list(X) when list(X) -> X;
to_list(X) when atom(X) -> atom_to_list(X).

%% Just because we can't eval receive statements...
flush() ->
    receive
	X ->
	    format("Shell got ~w~n",[X]),
	    flush()
    after 0 ->
	    ok
    end.

%% Print formated info about all registered processes in the system
nregs() ->
    regs(nodes()).
regs() ->
    regs([]).
regs([]) ->
    regs2([{node(),registered()}]);
regs(Nodes) ->
    regs2(collect_regs([node() | Nodes])).


collect_regs([]) -> [];
collect_regs([N|Tail]) ->
    [{N, rpc:call(N,erlang,registered,[])} | collect_regs(Tail)].

regs2([]) -> ok;
regs2([N|Tail]) ->
    print_node_regs(N), regs2(Tail).
print_node_regs({N, List}) when list(List) ->
    format("~n** Registered procs on node ~w **~n",[N]),
    rformat("Name", "Pid", "Initial Call", "Reds", "Msgs"),
    plist(lists:sort(List),N).

plist([], N) -> ok;
plist([Name|Tail], N) when N == node() ->
    P = whereis(Name),
    pline(Name, catch process_info(P), P),
    plist(Tail, N);
plist([Name|Tail], N) ->
    P = rpc:call(N, erlang, whereis, [Name]),
    pline(Name, pinfo(P), P),
    plist(Tail, N).
    
pline(Name, {'EXIT', _}, Pid) ->
    rformat(Name, Pid, "dead", 0, 0);
pline(Name, Info, Pid) ->
    Call = fetch(initial_call, Info),
    Reds  = fetch(reductions, Info), 
    LM = length(fetch(messages, Info)),
    rformat(io_lib:format("~w",[Name]), 
	    io_lib:format("~w",[Pid]), 
	    io_lib:format("~w",[Call]), 
	    integer_to_list(Reds), integer_to_list(LM)).

rformat(Name, Pid, Call, Reds, LM) ->
    format("~-20s ~-12s ~-25s ~9s ~4s~n",
	      [Name,
	       Pid, Call,
	       Reds, LM]).


pwd() ->
    {ok, Str} = file:get_cwd(),
    format('~s\n',[to_lower(Str)]).

to_lower([H|T]) when $A =< H, H =< $Z ->
    [H - $A + $a|to_lower(T)];
to_lower([H|T]) ->
    [H|to_lower(T)];
to_lower([]) ->
    [].

cd(Dir) ->
    file:set_cwd(Dir),
    pwd().

%% ls
%%  the strategy is to print in fixed width fileds

ls() ->
	ls(".").

ls(Dir) ->
    case file:list_dir(Dir) of
	{ok, Entries} ->
	    ls_print(Entries);
	_ ->
	    format("Invalid directory\n")
    end.

ls_print(L) ->
    Width = lists:min([lists:max(lengths(L, [])), 40]) + 5,
    ls_print(L, Width, 0).

ls_print(X, Width, Len) when Width + Len > 80 ->
    io:nl(),
    ls_print(X, Width, 0);
ls_print([H|T], Width, Len) ->
    Str = pad(H, Width),
    io:format("~s",[Str]),
    ls_print(T, Width, Len+Width);
ls_print([], _, _) ->
    io:nl().


lengths([H|T], L) -> lengths(T, [length(H)|L]);
lengths([], L)    -> L.

pad(Str, Width) ->
    case Width - length(Str) of
	N when N >= 0 ->
	    lists:append(Str, lists:duplicate(N, $ ));
	_ ->
	    Str
    end.
