%% Copyright (C) 1991, Ellemtel Telecommunications Systems Laboratories
%% File    : code.erl
%% Author  : Joe Armstrong, Robert Virding, Claes Wikstrom, Magnus Froberg
%% Purpose : A simple code server.
%%           Which maybe isn't that simple any longer :-)

-module(code).
-export([objfile_extension/0, 
	 set_path/1, 
	 get_path/0, 
	 load_file/1,
	 ensure_loaded/1,
	 load_abs/1,
	 load_binary/3,
	 delete/1,
	 purge/1,
	 is_loaded/1,
	 all_loaded/0,
	 stop/0,
	 root_dir/0,
	 lib_dir/0,
	 compiler_dir/0,
	 uc_dir/0,
	 stick_dir/1,
	 unstick_dir/1,
	 get_object_code/1,
	 add_path/1,
	 add_pathsz/1,
	 add_paths/1,
	 add_pathsa/1,
	 add_patha/1,
	 add_pathz/1,
	 add_uc/1,
	 start/0, start/1,
	 server0/1,
	 which/1,
	 clash/0]).

%% Following functions are exported for the interpreter
-export([interpret/1,
	 interpret_binary/3,
	 delete_interpret/1,
	 interpreted/0,
	 interpreted/1]).
	 
-import(lists,[concat/1,keydelete/3,keysearch/3,append/2]).
-import(error_logger, [error_msg/2]).

%% User interface.
%%
%% objfile_extension()		-> ".jam" | ".vee" | ...
%% set_path(Dir*)		-> true
%% get_path()			-> Dir*
%% add_path(Dir)		-> true | {error, What}
%% add_patha(Dir)		-> true | {error, What}
%% add_pathz(Dir)		-> true | {error, What}
%% add_paths(DirList)           -> true | {error, What}
%% add_pathsa(DirList)          -> true | {error, What}
%% add_pathsz(DirList)          -> true | {error, What}
%% load_file(File)		-> {error,What} | {module, Mod}
%% load_abs(File)		-> {error,What} | {module, Mod}
%% load_binary(Mod,File,Bin)    -> {error,What} | {module,Mod}
%% ensure_loaded(Module)	-> {error,What} | {module, Mod} |
%%                                 {interpret, Mod}
%% delete(Module)
%% purge(Module)
%% is_loaded(Module)		-> {file, File} | false
%% all_loaded()			-> {Module, File}*
%% add_uc(Name)			-> true
%% get_object_code(Mod)         -> error | {Mod, Filename, Bin}
%% stop()			-> true
%% root_dir()                   
%% compiler_dir()
%% uc_dir()
%% lib_dir()
%% stick_dir(Dir)               -> ok | {error ,Reason}
%% which(Module)                -> Filename
%% clash() ->                   -> print out


%% User interface

objfile_extension() ->
    case erlang:info(machine) of
	"JAM" -> ".jam";
	"VEE" -> ".vee";
	"BEAM" -> ".beam.o"
    end.

set_path(PathList) ->  request(code_server,{set_path,PathList}).
get_path()         ->  request(code_server,get_path).
load_file(Mod)     ->  request(code_server,{load_file,Mod}).
ensure_loaded(Mod) ->  request(code_server,{ensure_loaded,Mod}).
load_abs(File)     ->  request(code_server,{load_abs,File}).
load_binary(Mod,File,Bin) -> request(code_server,{load_binary,Mod,File,Bin}).
delete(Mod)        ->  request(code_server,{delete,Mod}).
purge(Mod)         ->  request(code_server,{purge,Mod}).
is_loaded(Mod)     ->  request(code_server,{is_loaded,Mod}).
get_object_code(M) ->  request(code_server, {get_object_code, M}).
all_loaded()       ->  request(code_server,all_loaded).
stop()             ->  request(code_server,stop).
root_dir()         ->  request(code_server,{dir,root_dir}).
lib_dir()          ->  request(code_server,{dir,lib_dir}).
compiler_dir()     ->  request(code_server,{dir,compiler_dir}).
uc_dir()           ->  request(code_server,{dir,uc_dir}).
stick_dir(Dir)     ->  request(code_server,{stick_dir,Dir}).
unstick_dir(Dir)   ->  request(code_server,{unstick_dir,Dir}).
add_path(Dir)      ->  add_pathz(Dir).
add_pathsz(Dirs)   ->  add_paths(Dirs).
interpreted()      ->  request(code_server,{interpreted}).
interpret(Module)  ->  request(code_server,{interpret,Module}).
interpreted(Module)->  request(code_server,{interpreted,Module}).
delete_interpret(Module) ->  request(code_server,{delete_int,Module}).
interpret_binary(Mod,File,Bin) ->
    request(code_server,{interpret_binary,Mod,File,Bin}).


add_paths([]) ->  ok;
add_paths([Dir|Tail]) -> 
    add_pathz(Dir),
    add_paths(Tail).

add_pathsa([]) ->  ok;
add_pathsa([Dir|Tail]) -> 
    add_patha(Dir),
    add_pathsa(Tail).


add_patha(Dir)     ->
    New = [Dir|get_path()],
    set_path(New).

add_pathz(Dir) ->
    New = append(get_path(), [Dir]),
    set_path(New).

add_uc(Name) ->
    add_pathz(list_to_atom(concat([uc_dir(),"/",Name]))).


request(To,Req) ->
    To ! {self(),Req},
    receive
        {To,What} -> What
    end.

send_ack(To,What) -> To ! {code_server,What}.

start() ->
    start([stick]).

start(Stick) ->
    case init:get_flag('-root') of
	notfound ->
	    error_msg('Can not start code server~n',[]),
	    exit(crash);
	{'-root',Root} ->
	    CompilerDir = lists:concat([Root,"/lib/compiler/ebin"]),
	    StdDir = lists:concat([Root,"/lib/std/ebin"]),
	    Id = spawn(code, server0, [Root]),
	    register(code_server, Id),
	    load_file(int),   %% Load the initial interpreter module.
	    case Stick of
		[stick] ->
		    stick_dir(CompilerDir),
		    stick_dir(StdDir);
		_ ->
		    ok
	    end,
	    Id;
	Other ->
	    error_msg('Can not start code server ~w ~n',[Other]),
	    exit(crash)
    
    end.

server0(Root0) ->
    Root = atom_to_list(Root0),
    LibDir = append(Root,"/lib"),
    {ok,Dirs0} = file:list_dir(LibDir),
    Dirs = lists:map({erlang,list_to_atom},[],Dirs0),
    {Paths, Libs} = make_path(LibDir,Dirs,[],[]),
    InitialPath = ['.'|Paths],
    process_flag(trap_exit, true),
    case init:get_flag('-loaded') of
	{'-loaded', Loaded} ->
	    server(Root, InitialPath, 
		   lists:append(Loaded,fix(erlang:pre_loaded())),[]);
	_ ->
	    server(Root, InitialPath, fix(erlang:pre_loaded()),[])
    end.

make_path(_,[],Res,Bs) -> {Res,Bs};
make_path(BundleDir,[Bundle|Tail],Res,Bs) ->
    Dir = concat([BundleDir,"/",Bundle]),
    Bin = append(Dir,"/ebin"),
    %% First try with /ebin otherwise just add the dir
    case file:file_info(Bin) of
	{ok,{_,directory,_,_,_,_,_}} ->
	    make_path(BundleDir,Tail,[list_to_atom(Bin)|Res],[Bundle|Bs]);
	_ ->
	    case file:file_info(Dir) of
		{ok,{_,directory,_,_,_,_,_}} ->
		    make_path(BundleDir,Tail,
			      [list_to_atom(Dir)|Res],[Bundle|Bs]);
		_ ->
		    make_path(BundleDir,Tail,Res,Bs)
	    end
    end.

fix([]) -> [];
fix([H|Tail]) -> [{H,preloaded}|fix(Tail)].


%% server(Root,Path, CurrentModules, InterpretedModules)

server(Root, Path, Cur, Int) ->
    receive
	{From,{stick_dir,Dir}} ->
	    case file:list_dir(Dir) of
		{ok,Listing} ->
		    stick_libarary(Listing),
		    send_ack(From,ok);
		Other ->
		    send_ack(From,Other)
	    end,
	    server(Root, Path, Cur, Int);
	{From,{unstick_dir,Dir}} ->
	    case file:list_dir(Dir) of
		{ok,Listing} ->
		    unstick_libarary(Listing),
		    send_ack(From,ok);
		Other ->
		    send_ack(From,Other)
	    end,
	    server(Root, Path, Cur, Int);
	{From,{dir,Dir}} when pid(From) ->
            do_dir(From,Root,Dir),
            server(Root, Path, Cur, Int);
	{From,{load_file,Mod}} ->
	    {Status, Cur1} = load_file(Mod, Path, Cur),
	    send_ack(From, Status),
	    server(Root, Path, Cur1, del_interpret(Status,Int));
	{From,{set_path,PathList}} ->
	    {NewPath, Reply} = check_path(PathList, Path),
	    send_ack(From, Reply),
	    server(Root, NewPath, Cur, Int);
	{From,get_path} ->
	    send_ack(From, Path),
	    server(Root, Path, Cur, Int);
	%% Messages to load, delete and purge modules/files.
	{From,{load_abs,File}} ->
	    {Status, Cur1} = load_abs(File, Cur),
	    send_ack(From, Status),
	    server(Root, Path, Cur1, del_interpret(Status,Int));
	{From,{load_binary,Mod,File,Bin}} ->
	    {Status,Cur1} = do_load_binary(Mod,File,Bin,Cur),
	    send_ack(From, Status),
	    server(Root, Path, Cur1, del_interpret(Status,Int));
	{From,{ensure_loaded,Mod}} ->
	    case erlang:module_loaded(to_atom(Mod)) of
		true ->
		    %% Should we check code_server's internal database here?
		    send_ack(From, {module,Mod}),
		    server(Root, Path, Cur, Int);
		false ->
		    case lists:member(to_atom(Mod),Int) of
			true ->
			    send_ack(From, {interpret,Mod}),
			    server(Root, Path, Cur, Int);
			_ ->
			    {Status, Cur1} = load_file(Mod, Path, Cur),
			    send_ack(From, Status),
			    server(Root, Path, Cur1, Int)
		    end
	    end;
	{From,{delete,Mod}} ->
	    case catch erlang:delete_module(to_atom(Mod)) of
		true ->
		    send_ack(From, true),
		    server(Root, Path, keydelete(Mod, 1, Cur), Int);
		_ ->
		    send_ack(From,false),
		    server(Root,Path,Cur,Int)
	    end;
	{From,{purge,Mod}} ->
	    send_ack(From, do_purge(Mod)),
	    server(Root, Path, Cur, Int);
	{From,{is_loaded,Module}} ->
	    Where = is_loaded(Module, Cur),
	    send_ack(From, Where),
	    server(Root, Path, Cur, Int);
	{From,all_loaded} ->
	    send_ack(From, all_loaded(Cur,Int)),
	    server(Root, Path, Cur, Int);

	%% Messages to handle interpretation of modules
	{From,{interpret,Module}} ->
	    {Status,Int1} = add_interpret(Module,Int),
	    send_ack(From, Status),
	    server(Root, Path, Cur, Int1);
	{From,{interpreted,Module}} ->
	    send_ack(From, lists:member(to_atom(Module), Int)),
	    server(Root, Path, Cur, Int);
	{From,{interpreted}} ->
	    send_ack(From, Int),
	    server(Root, Path, Cur, Int);
	{From,{interpret_binary,Mod,File,Bin}} ->
	    {Status,Int1} = add_interpret(Mod,Int),
	    case Status of
		{error,_} ->
		    send_ack(From, Status);
		_ ->
		    %% The interpreter server acknowledges the request.
		    load_interpret(From,File,Mod,Bin)
	    end,
	    server(Root, Path, Cur, Int1);
	{From,{delete_int,Module}} ->
	    Int1 = delete_interpret(Module,Int),
	    send_ack(From, ok),
	    server(Root, Path, Cur, Int1);

	{From, {get_object_code, Mod}} ->
	    send_ack(From, mod_to_bin(Path, Mod)),
	    server(Root, Path, Cur, Int);
	{From,stop} ->
	    send_ack(From, true);
	Other ->			
	    error_msg(' ** Codeserver*** ignoring ~w~n ',[Other]),
	    server(Root, Path, Cur, Int)
    end.

all_loaded(Cur,Int) ->
    lists:append(Cur,ints(Int)).

ints([Mod|Mods]) -> [{Mod,interpreted}|ints(Mods)];
ints([])         -> [].

check_path(PathList, Orig) ->
    check_path(PathList, PathList, Orig).

check_path([], Path, Orig) -> 
    {Path, true};
check_path([Dir |Tail], Path, Orig) ->
    case file:file_info(Dir) of
	{ok,_} -> 
	    check_path(Tail, Path, Orig);
	{error, R} -> 
	    {Orig, {error, bad_directory}}
    end;
check_path(_, _, Orig) ->
    {Orig, {error, bad_path}}.

do_dir(From,Root,compiler_dir) ->
    From ! {code_server,
	    list_to_atom(append(Root,[$/|atom_to_list(compiler)]))};
do_dir(From,Root,lib_dir) ->
    From ! {code_server,list_to_atom(append(Root,[$/|atom_to_list(lib)]))};
do_dir(From,Root,uc_dir) ->
    From ! {code_server,list_to_atom(append(Root,[$/|atom_to_list(uc)]))};
do_dir(From,Root,root_dir) ->
    From ! {code_server,list_to_atom(Root)};
do_dir(From,Root,compiler_dir) ->
    From ! {code_server,
	    list_to_atom(append(Root,[$/|atom_to_list(compiler)]))};
do_dir(From,Root,_) ->
    From ! {code_server,'bad request to code'}.


mod_to_bin([],Mod) ->
    error;
mod_to_bin([Dir|Tail],Mod) ->
    File = concat([Dir,'/',Mod,objfile_extension()]),
    case file:read_file(File) of
	{error,_} -> 
	    mod_to_bin(Tail,Mod);
	{ok,Bin} ->
	    {Mod,Bin,list_to_atom(File)}
    end.

load_file(Mod,Path,Cur) ->
    case mod_to_bin(Path,Mod) of
	error -> {{error,nofile},Cur};
	{Mod,Binary,File} -> try_load_module(File, Mod, Binary, Cur)
    end.

add_interpret(Module,Int) ->
    M = to_atom(Module),
    case sticky(M) of
	true ->     %% Sticky file reject the load (interpretation)
	    error_msg("Can't interpret module that reside in sticky dir\n",[]),
	    {{error, sticky_directory}, Int};
	false ->
	    case lists:member(M,Int) of
		false ->
		    int:tell_interpreter({interpret,M}),
		    {{module,M},[M|Int]};
		true ->
		    {{module,M},Int}
	    end
    end.

del_interpret({module,Mod},Int) -> delete_interpret(Mod,Int);
del_interpret(_,Int)            -> Int.

delete_interpret(Module,Int) ->
    M = to_atom(Module),
    case lists:member(M,Int) of
	true ->
	    NewModules = lists:delete(M,Int),
	    int:tell_interpreter({delete,M}),
	    NewModules;
	_ ->
	    Int
    end.

load_interpret(From,File,Module,Binary) ->
    M = to_atom(Module),
    delete_modules([M]),
    int:tell_interpreter({load,From,M,File,Binary}).

delete_modules([Mod|Modules]) ->
    do_purge(Mod),
    erlang:delete_module(Mod),
    delete_modules(Modules);
delete_modules([]) ->
    true.

do_load_binary(Module,File,Binary,Cur) ->
    case erlang:module_loaded(to_atom(Module)) of
	true ->
	    do_purge(Module);
	false ->
	    ok
    end,
    try_load_module(File, Module, Binary, Cur).

load_abs(File, Cur) ->
    FileName = lists:concat([File,objfile_extension()]),
    Mod = to_mod(lists:reverse(to_list(File))),
    case file:read_file(FileName) of
	{ok,Bin} ->
	    try_load_module(FileName, Mod, Bin, Cur);
	{error,Reason} ->
	    {{error,Reason}, Cur}
    end.


try_load_module(File, Mod, Bin, Cur) ->
    M = to_atom(Mod),
    case sticky(M) of
	true ->                         %% Sticky file reject the load
	    error_msg("Can't load module that reside in sticky dir\n",[]),
	    {{error, sticky_directory}, Cur};
	false ->
	    case erlang:load_module(M, Bin) of
		{module,M} ->
		    Cur1 = add_module(M, File, Cur),
		    {{module,Mod}, Cur1};
		{error,What} ->
		    {{error,What}, Cur}
	    end
    end.

%% The idea here is that we shall sucseed in loading 
%% a lib module the first time and then never be able reload it !!!!!
sticky(Mod) ->
    case erlang:module_loaded(Mod) of
	true ->
	    case get(Mod) of
		true -> true;
		_ -> false
	    end;
	_ -> false
    end.

	    
add_module(Module, FileName, Cur) when atom(FileName) ->
    [{Module,FileName}|keydelete(Module, 1, Cur)];
add_module(Module, FileName, Cur) ->
    [{Module,list_to_atom(FileName)}|keydelete(Module, 1, Cur)].

is_loaded(Module, Cur) ->
    case keysearch(Module, 1, Cur) of
       {value,{_,FileName}} ->
	   {file,FileName};
       false ->
	  false
   end.

%% do_purge(Module)
%%  Kill all processes running code from *old* Module, and then purge the
%%  module. Return true if any processes killed, else false.

do_purge(Mod) ->
    M = to_atom(Mod),
    do_purge(processes(), M, false).

do_purge([P|Ps], Mod, Purged) ->
    case erlang:check_process_code(P, Mod) of
	true ->
	    exit(P, kill),
	    do_purge(Ps, Mod, true);
	false ->
	    do_purge(Ps, Mod, Purged)
    end;
do_purge([], Mod, Purged) ->
    catch erlang:purge_module(Mod),
    Purged.


%% Put/Erase all libarary modules into local process dict

stick_libarary(LibDirListing) ->
    putem(get_mods(LibDirListing, lists:reverse(objfile_extension()))).
unstick_libarary(LibDirListing) ->
    eraseem(get_mods(LibDirListing, lists:reverse(objfile_extension()))).
    

putem([]) -> done;
putem([M|Tail]) -> put(to_atom(M),true), putem(Tail).

eraseem([]) -> done;
eraseem([M|Tail]) -> erase(M), eraseem(Tail).

get_mods([File|Tail], Endpattern) ->
    case is_mod(lists:reverse(File), Endpattern ) of
	{module,M} -> [M|get_mods(Tail, Endpattern)];
	false      -> get_mods(Tail, Endpattern)
    end;
get_mods([], _) ->
    [].

is_mod(L, []) -> {module, to_mod(L)};
is_mod([H|T0], [H|T1]) -> is_mod(T0, T1);
is_mod(_, _) -> false.

to_mod(BackwardChars) -> 
    to_mod(BackwardChars,[]).
to_mod([],Ack) -> list_to_atom(Ack);
to_mod([$/|_],Ack) -> list_to_atom(Ack);
to_mod([X|Tail],Ack) -> to_mod(Tail,[X|Ack]).


%% Find out which version of a particular module we would
%% load if we tried to load it, unless it's allready loaded.    
%% In that case return the name of the file which contains
%% the loaded object code

which(Module) when atom (Module) ->
    case code:is_loaded(Module) of
        false ->
            which2(Module);
        {file, File} ->
            File
    end.

which2(Module) ->
    Ext = code:objfile_extension(),
    File = lists:concat([Module, Ext]),
    Path = code:get_path(),
    list_to_atom(lists:flatten(which(File,Path))).

which(_,[]) ->
    "non_existing";

which(Module,[Directory|Tail]) ->
    {ok,Files} = file:list_dir(Directory),
    case lists:member(Module,Files) of
	true ->
	    io_lib:format("~s/~s",[Directory,Module]);
	false ->
	    which(Module,Tail)
    end.


%% Search the entire path system looking for name clashes

clash() ->
    Path = code:get_path(),
    Struct = lists:flatten(build(Path)),
    Len = length(search(Struct)),
    io:format("** Found ~w name clashes in code paths ~n",[Len]).

search([]) -> [];
search([{Dir,File} | Tail]) ->
    case lists:keysearch(File,2,Tail) of
	false -> 
	    search(Tail);
	{value,{Dir2,File}} ->
	    io:format("** ~s/~s hides ~s/~s~n",[Dir,File,Dir2,File]),
	    [clash | search(Tail)]
    end.

build([]) -> [];
build([Dir|Tail]) ->
    Files = filter(code:objfile_extension(), Dir, file:list_dir(Dir)),
    [decorate(Files, Dir) | build(Tail)].

decorate([], _) -> [];
decorate([File|Tail], Dir) ->
    [{Dir, File} | decorate(Tail, Dir)].

filter(Ext, Dir, {error,_}) ->     
    io:format("** Bad path can't read ~s~n", [Dir]), [];
filter(Ext, _, {ok,Files}) -> 
    filter2(Ext, length(Ext), Files).

filter2(Ext,Extlen, []) -> [];
filter2(Ext, Extlen,[File|Tail]) ->
    case has_ext(Ext,Extlen, File) of 
	true -> [File | filter2(Ext, Extlen, Tail)];
	false -> filter2(Ext, Extlen, Tail)
    end.

has_ext(Ext, Extlen,File) ->
    L = length(File),
    case catch lists:nthtail(L - Extlen, File) of
	Ext -> true;
	_ -> false
    end.

to_list(X) when list(X) -> X;
to_list(X) when atom(X) -> atom_to_list(X).

to_atom(X) when atom(X) -> X;
to_atom(X) when list(X) -> list_to_atom(X).

