%% Copyright (C) 1994, Ellemtel Telecommunications Systems Laboratories
-module(ddd).
-author('tobbe@erix.ericsson.se').
-export([start/0,stop/0,init/0,in/2,out/1,out/2,delete/1,
	 get_table_ref/0]).
-export([doit/2]).

%%-------------------------------------------------------
%% Created:  7 Oct 1994 by tobbe@erix.ericsson.se
%% Function: A Distributed Data Dictionary.
%%           Provides a fully replicated dictionary on all
%%           nodes connected in any configuration, where
%%           the nodes may come and go. We assumes however
%%           that no message is lost by the underlying
%%           communication media.
%% 
%% The algoritm used for ensuring consistency is based on
%% theories in the following classical paper: 
%% 
%%    "Time, Clocks, and the ordering of events 
%%     in a distributed system", Leslie Lamport, 
%%    Comm.of.the ACM, July 1978, Vol 21, No 7.
%%-------------------------------------------------------

start() ->         generic:start(ddd_server, ddd ,init, []).
stop() ->          generic:stop(ddd_server).
in(Key,Value) ->   generic:request(ddd_server, {ddd_in, Key, Value}).
out(Key) ->        generic:request(ddd_server, {ddd_out, Key}).
out({ddd, R}, Key) when reference(R) -> dbget({ddd,R}, Key).
delete(Key) ->     generic:request(ddd_server, {ddd_delete,Key}).
get_table_ref() -> generic:request(ddd_server, ddd_read).
		
%% THE SERVER
init() ->
    Clock   = lclock_new(),
    process_flag(trap_exit, true),
    T = ets:new(ddd,[set, protected]),
    generic:sync_start(ddd_server, started),
    generic:loop(ddd_server, {Clock,T}, {ddd, doit}).


doit(Request, State) ->
    {Clock, T} = State,
    case Request of
	%% --- CLIENT REQUESTS ---
	{From, {ddd_in, Key, Value}} ->    
	    NewClock = lclock_bump(Clock),
	    ets:insert(T, {Key, Value, NewClock}),
	    distribute(NewClock,Key,Value),
	    {reply, From, true, {NewClock, T}};
	
	{From,{ddd_out,Key}} ->
	    Value = dbget(T,Key),
	    {reply, From, Value, State};
	
	{From,{ddd_delete,Key}} ->
	    NewClock = lclock_bump(Clock),
	    Old = ets:delete(T,Key),
	    del_distribute(NewClock,Key),
	    {reply, From, Old, {NewClock, T}};
	
	%% Someone wants a table refence to read from
	%% that is only possible for user defined tables
	{From, ddd_read} ->
	    {reply, From, T, State};
	
	%% --- REMOTE SERVER REQUESTS ---
	%% Note that we might have to adjust our own clock
	{ddd_insert,Msg} ->
	    RemoteClock = remote_insert(T,Msg),
	    NewClock = adjust_clock(Clock,RemoteClock),
	    {noreply, {NewClock, T}};

	{ddd_remove,Msg} ->
	    RemoteClock = remote_remove(T,Msg),
	    NewClock = adjust_clock(Clock,RemoteClock),
	    {noreply, {NewClock,T}};
	
	%% --- NODE HANDLING ---
	%% A new node has been connected, exchange dictionaries.
	{nodeup,Node} ->
	    send_table(ets:tab2list(T),Node),
	    {noreply, State};
	
	%% Here comes the tables from the other side
	{ddd_exchange_tables,Node,Tab} ->
	    store_table(T,Tab,Node),
	    {noreply, State};

	%% A node went down, what a pity...
	{nodedown,Node} ->
	    {noreply, State};
	
	{From, stop} ->
	    {stopped, From}
    end.


%% Check if we have to adjust our clock
adjust_clock(Clock,RClock) when RClock>Clock -> RClock;
adjust_clock(Clock,_) -> Clock.

%% A write has been made at some other node, now check if it
%% is ok to update the local database. Otherwise resend the
%% latest information in order to ensure consistency.
%% Return the remote clock.
remote_insert(Table,{Origin,Key,{Value,TimeStamp},VisitedNodes}) ->
    case ets:lookup(Table,Key) of
	%% No problems..
	[] ->
	    ets:insert(Table,{Key, Value,TimeStamp}),
	    Msg = {Origin,Key,{Value,TimeStamp},VisitedNodes},
	    pass_on(nodes(),{ddd_insert,Msg});
	%% A "younger" value to be inserted
	[{_,V,Ts}] when Ts<TimeStamp ->
	    ets:insert(Table,{Key, Value,TimeStamp}),
	    Msg = {Origin,Key,{Value,TimeStamp},VisitedNodes},
	    pass_on(nodes(),{ddd_insert,Msg});
	%% In case of equal timestamps, then we
	%% have a lexical ordering of all nodes
	[{_,V,TimeStamp}] when Origin<node() ->
	    ets:insert(Table,{Key,Value,TimeStamp}),
	    Msg = {Origin,Key,{Value,TimeStamp},VisitedNodes},
	    pass_on(nodes(),{ddd_insert,Msg});
	%% We either have a "younger" datum ourselves, or our node
	%% is the "lexical winner". Now we must resend that event 
	%% to ensure consistency.
	[{_,V,Ts}] ->
	    distribute(Ts,Key,V)
    end,
    TimeStamp.

%% Works almost as remote_insert/2 above..
remote_remove(Table,{Origin,Key,TimeStamp,VisitedNodes}) ->
    case ets:lookup(Table,Key) of
	%% No problems..
	[] ->
	    pass_on(nodes(),{ddd_remove,{Origin,Key,TimeStamp,VisitedNodes}});
	%% A "younger" remove event
	[{_,V,Ts}] when Ts<TimeStamp ->
	    ets:delete(Table,Key),
	    pass_on(nodes(),{ddd_remove,{Origin,Key,TimeStamp,VisitedNodes}});
	%% In case of equal timestamps, then we
	%% have a lexical ordering of all nodes
	[{_,V,TimeStamp}] when Origin<node() ->
	    ets:delete(Table,Key),
	    pass_on(nodes(),{ddd_remove,{Origin,Key,TimeStamp,VisitedNodes}});
	%% We either have a "younger" datum ourselves, or our node
	%% is the "lexical winner". Now we must resend that event 
	%% to ensure consistency.
	[{_,V,Ts}] ->
	    distribute(Ts,Key,V)
    end,
    TimeStamp.

%% Distribute event to all known servers
distribute(TimeStamp,Key,Value) ->
    Nodes = nodes(),
    Msg = {ddd_insert,{node(),Key,{Value,TimeStamp},[node()]}},
    send_to_nodes(Nodes,Msg).

%% Distribute a delete event to all known servers
del_distribute(TimeStamp,Key) ->
    Nodes = nodes(),
    Msg = {ddd_remove,{node(),Key,TimeStamp,[node()]}},
    send_to_nodes(Nodes,Msg).

%% Pass on the event to all Nodes that have not seen it already
pass_on(Nodes,{Tag,{Origin,Key,Datum,OldNodes}}) ->
    RemainingNodes = diff(Nodes,OldNodes),
    Msg = {Tag,{Origin,Key,Datum,[node()|OldNodes]}},
    send_to_nodes(RemainingNodes,Msg).	       

%% Send message to all Nodes
send_to_nodes([Node|T],Msg) ->
    {ddd_server,Node} ! Msg,
    send_to_nodes(T,Msg);
send_to_nodes([],_) -> true.

%% Remove all elements from the first argument that
%% exists in the second argument.
diff([H|T],X) when list(X) ->
    case lists:member(H,X) of
	true  -> diff(T,X);
	false -> [H|diff(T,X)]
    end;
diff([],_) -> [].


%% Strip off the Key and the timestamp
dbget(Table,Key) ->
    case ets:lookup(Table,Key) of
	[{Key,Val,Ts}] -> Val;
	[] -> undefined
    end.

%% Pack the tables and send them to the node
send_table([],Node) -> 
    ignore;
send_table(List, Node) ->
    {ddd_server,Node} ! {ddd_exchange_tables,node(),List}.


store_table(Table, Items, Node) when tuple(Items) ->
    store_table(Table, tuple_to_list(Items), Node);
store_table(Table,[],Node) -> done;
store_table(Table, [Entry|Tail], Node) ->
    store_table_1(Table,Entry,Node),
    store_table(Table, Tail, Node).

store_table_1(Table,{K,V,Ts},Node) ->
    case ets:lookup(Table,K) of
	%% New value, store aaaaaaa
	[] ->
	    ets:insert(Table,{K,V,Ts});
	%% Same value, continue
	[{K,V,Ts}] ->
	    ok;
	%% The "least" node keeps the old value
	Val when node()<Node ->
	    ignore;
	%% Overwriting...
	Val -> 
	    ets:insert(Table,{K,V,Ts})
    end;
store_table_1(_,[],_) ->
    true.

%%  A logical clock data structure that can be 
%%  increased (almost) forever, and can be compared 
%%  with other Lclock values.

lclock_new() -> {0,0}.

lclock_bump({Msb,Lsb}) ->
    T = Lsb + 1,
    if
	%% Overflow ?
	T<Lsb -> {Msb+1,0};
	true  -> {Msb,T}
    end.
