%% Copyright (C) 1995, Ellemtel Telecommunications Systems Laboratories
%% File    : epp.erl
%% Author  : Robert Virding
%% Purpose : An Erlang code preprocessor.

-module(epp).

-export([open/2,open/3,close/1,format_error/1]).
-export([scan_erl_form/1,parse_erl_form/1,macro_defs/1]).
-export([server/4]).

%% open(FileName, IncludePath)
%% open(FileName, IncludePath, PreDefMacros)
%% close(Epp)
%% scan_erl_form(Epp)
%% parse_erl_form(Epp)
%% macro_defs(Epp)

open(File, Path) ->
    open(File, Path, []).

open(File, Path, Pdm) ->
    Epp = spawn_link(epp, server, [self(),File,Path,Pdm]),
    wait_epp_reply(Epp).

close(Epp) ->
    epp_request(Epp, close),
    wait_epp_reply(Epp).

scan_erl_form(Epp) ->
    epp_request(Epp, scan_erl_form),
    wait_epp_reply(Epp).

parse_erl_form(Epp) ->
    epp_request(Epp, scan_erl_form),
    case wait_epp_reply(Epp) of
	{ok,Toks} ->
	    erl_parse:parse_form(Toks);
	Other ->
	    Other
    end.

macro_defs(Epp) ->
    epp_request(Epp, macro_defs),
    wait_epp_reply(Epp).

%% format_error(ErrorDescriptor) -> String
%%  Return a string describing the error.

format_error({bad,W}) ->
    io_lib:format("badly formed '~s'", [W]);
format_error({call,What}) ->
    io_lib:format("illegal macro call '~s'",[What]);
format_error({undefined,M}) ->
    io_lib:format("undefined macro '~w'", [M]);
format_error({depth,What}) ->
    io_lib:format("~s too deep",[What]);
format_error({mismatch,M}) ->
    io_lib:format("argument mismatch for macro '~w'", [M]);
format_error({arg_error,M}) ->
    io_lib:format("badly formed argument for macro '~w'", [M]);
format_error({redefine,M}) ->
    io_lib:format("redefining macro '~w'", [M]);
format_error({include,F}) ->
    io_lib:format("can't find include file \"~s\"", [F]);
format_error({illegal,How,What}) ->
    io_lib:format("~s '-~s'", [How,What]).

%% server(StarterPid, FileName, Path, PreDefMacros)

server(Pid, Name, Path, Pdm) ->
    process_flag(trap_exit, true),
    case file:open(Name, read) of
	{ok,File} ->
	    put(sys_path, Path),
	    put(user_path, Path),
	    Ms0 = dict:new(),
	    Ms1 = dict:store({var,'EPPFile'}, {none,[{string,1,Name}]}, Ms0),
	    Ms2 = dict:store({var,'EPPLine'}, {none,[{integer,1,1}]}, Ms1),
	    case predef_macros(Pdm, Ms2) of
		{ok,Ms} ->
		    epp_reply(Pid, {ok,self()}),
		    From = wait_request(Ms),
		    epp_reply(From, enter_file_reply(Name, 1)),
		    wait_req_scan(File, 1, Name, [], [], Ms);
		error ->
		    epp_reply(Pid, {error,predef})
	    end;
	{error,E} ->
	    epp_reply(Pid, {error,E})
    end.

%% predef_macros(PreDefMacros, Macros) ->
%%	{ok,MacroDict} | error
%%  Add the predefined macros to the macros dictionary. A macro without a
%%  value gets the value 'true'.

predef_macros([{M,Val}|Pdm], Ms) when atom(M) ->
    case dict:find({atom,M}, Ms) of
	{ok,Def} ->
	    error;
	error ->
	    Exp = erl_parse:tokens(erl_parse:abstract(Val)),
	    predef_macros(Pdm, dict:store({atom,M}, {none,Exp}, Ms))
    end;
predef_macros([M|Pdm], Ms) when atom(M) ->
    case dict:find({atom,M}, Ms) of
	{ok,Def} ->
	    error;
	error ->
	    predef_macros(Pdm, dict:store({atom,M}, {none,[{atom,1,true}]}, Ms))
    end;
predef_macros([Md|Pdm], Ms) -> error;
predef_macros([], Ms) -> {ok,Ms}.

%% wait_request(Macros) ->
%%	RequestFrom
%% wait_req_scan(CurrFile, CurrLine, CurrName, CurrIstack, FileStack, Macros)
%% wait_req_skip(CurrFile, CurrLine, CurrName, CurrIstack,
%%		 FileStack, Macros, SkipIstack)
%%  Handle requests, processing trivial requests directly. Either return
%%  requestor or scan/skip tokens.

wait_request(Ms) ->
    receive
	{epp_request,From,scan_erl_form} ->
	    From;
	{epp_request,From,macro_defs} ->
	    epp_reply(From, dict:dict_to_list(Ms)),
	    wait_request(Ms);
	{epp_request,From,close} ->
	    epp_reply(From, ok),
	    exit(normal);
	{'EXIT',_,R} ->
	    exit(R);
	Other ->
	    io:fwrite("Epp: unknown '~w'\n", [Other]),
	    wait_request(Ms)
    end.

wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms) ->
    From = wait_request(Ms),
    scan_toks(From, Cf, Cl, Cn, Cis, Fs, Ms).

wait_req_skip(Cf, Cl, Cn, Cis, Fs, Ms, Sis) ->
    From = wait_request(Ms),
    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, Sis).
	
%% enter_file(Path, FileName, IncludeLine, From,
%%	      CurrF, CurrLine, CurrName, CurrIstack, FileStack, Macros)
%% leave_file(From, CurrF, CurrLine, CurrName, CurrIstack, FileStack, Macros)
%%  Handle antering and leaving included files. Notify caller when the
%%  current file is changed. Note it is an error to exit a file if we are
%%  in a conditional. These functions never return.

enter_file(Path, NewName, Li, From, Cf, Cl, Cn, Cis, Fs, Ms)
  when length(Fs) >= 8 ->
    epp_reply(From, {error,{Li,epp,{depth,"include"}}}),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms);
enter_file(Path, NewName, Li, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case file:path_open(Path, NewName, read) of
	{ok,NewF,Pname} ->
	    epp_reply(From, enter_file_reply(Pname, 1)),
	    Ms1 = dict:store({var,'EPPFile'}, {none,[{string,1,Pname}]}, Ms),
	    wait_req_scan(NewF, 1, Pname, [], [{Cf,Cl,Cn,Cis}|Fs], Ms1);
	{error,E} ->
	    epp_reply(From, {error,{Li,epp,{include,NewName}}}),
	    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms)
    end.

enter_file_reply(Name, Line) ->
    {ok, [{'-',Line},{atom,Line,file},{'(',Line},
	  {string,Line,Name},{',',Line},{integer,Line,Line},
	  {')',Line},{dot,Line}]}.

leave_file(From, Cf, Cl, Cn, [I|Cis], Fs, Ms) ->
    epp_reply(From, {error,{Cl,epp,{illegal,"unterminated",I}}}),
    From = wait_request(Ms),
    leave_file(From, Cf, Cl, Cn, [], Fs, Ms);
leave_file(From, Cf, Cl, Cn, [], [], Ms) ->
    epp_reply(From, {eof,Cl}),
    wait_req_scan(Cf, Cl, Cn, [], [], Ms);
leave_file(From, Cf0, Cl0, Cn0, [], [{Cf,Cl,Cn,Cis}|Fs], Ms) ->
    file:close(Cf0),
    epp_reply(From, enter_file_reply(Cn, Cl)),
    Ms1 = dict:store({var,'EPPFile'}, {none,[{string,1,Cn}]}, Ms),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms1).

%% scan_toks(From, File, CurrLine, CurrName, CurrIstack, FileStack, Macros)
%% scan_toks(Tokens, From, File, CurrLine, CurrName, CurrIstack,
%%	     FileStack, Macros)

scan_toks(From, Cf, Cl0, Cn, Cis, Fs, Ms) ->
    case io:scan_erl_form(Cf, '', Cl0) of
	{ok,Toks,Cl} ->
	    scan_toks(Toks, From, Cf, Cl, Cn, Cis, Fs, Ms);
	{error,E,Cl} ->
	    epp_reply(From, {error,E}),
	    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms);
	{eof,Cl} ->
	    leave_file(From, Cf, Cl, Cn, Cis, Fs, Ms);
	{error,E} ->
	    epp_reply(From, {error,E}),
	    true				%This serious, just exit!
    end.

scan_toks([{'-',Lh},{atom,Ld,define}|Toks], From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_define(Toks, Ld, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_toks([{'-',Lh},{atom,Ld,undef}|Toks], From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_undef(Toks, Ld, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_toks([{'-',Lh},{atom,Li,include}|Toks], From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_include(Toks, Li, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_toks([{'-',Lh},{atom,Li,ifdef}|Toks], From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_ifdef(Toks, Li, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_toks([{'-',Lh},{atom,Li,ifndef}|Toks], From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_ifndef(Toks, Li, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_toks([{'-',Lh},{atom,Le,else}|Toks], From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_else(Toks, Le, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_toks([{'-',Lh},{atom,Le,endif}|Toks], From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_endif(Toks, Le, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_toks(Toks0, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case catch expand_macros(Toks0, Ms) of
	Toks when list(Toks) ->
	    epp_reply(From, {ok,Toks});
	{error,ErrL,What} ->
	    epp_reply(From, {error,{ErrL,epp,What}})
    end,
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms).

%% scan_define(Tokens, DefineLine, From, File, CurrLine, CurrName, CurrIstack,
%%	       Filestack, Macros)

scan_define([{'(',Lp},{atom,Lm,M},{',',Lc}|Toks], Ld,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({atom,M}, Ms) of
	{ok,Def} ->
	    epp_reply(From, {error,{Lm,epp,{redefine,M}}}),
	    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms);
	error ->
	    Ms1 = dict:store({atom,M}, {none,macro_expansion(Toks)}, Ms),
	    scan_toks(From, Cf, Cl, Cn, Cis, Fs, Ms1)
    end;
scan_define([{'(',Lp},{atom,Lm,M},{'(',Lc}|Toks], Ld,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({atom,M}, Ms) of
	{ok,Def} ->
	    epp_reply(From, {error,{Lm,epp,{redefine,M}}}),
	    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms);
	error ->
	    case catch macro_pars(Toks, []) of
		{ok, {As, Me}} ->
		    Ms1 = dict:store({atom,M}, {As,Me}, Ms),
		    scan_toks(From, Cf, Cl, Cn, Cis, Fs, Ms1);
		_ ->
		    epp_reply(From, {error,{Ld,epp,{bad,define}}}),
		    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms)
	    end
    end;
scan_define([{'(',Lp},{var,Lm,M},{',',Lc}|Toks], Ld,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({var,M}, Ms) of
	{ok,Def} ->
	    epp_reply(From, {error,{Lm,epp,{redefine,M}}}),
	    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms);
	error ->
	    Ms1 = dict:store({var,M}, {none,macro_expansion(Toks)}, Ms),
	    scan_toks(From, Cf, Cl, Cn, Cis, Fs, Ms1)
    end;
scan_define([{'(',Lp},{var,Lm,M},{'(',Lc}|Toks], Ld,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({var,M}, Ms) of
	{ok,Def} ->
	    epp_reply(From, {error,{Lm,epp,{redefine,M}}}),
	    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms);
	error ->
	    case catch macro_pars(Toks, []) of
		{ok, {As, Me}} ->
		    Ms1 = dict:store({var,M}, {As,Me}, Ms),
		    scan_toks(From, Cf, Cl, Cn, Cis, Fs, Ms1);
		_ ->
		    epp_reply(From, {error,{Ld,epp,{bad,define}}}),
		    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms)
	    end
    end;
scan_define(Toks, Ld, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    epp_reply(From, {error,{Ld,epp,{bad,define}}}),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms).

%% scan_undef(Tokens, UndefLine, From, File, CurrLine, CurrName, CurrIstack,
%%	      Filestack, Macros)

scan_undef([{'(',Llp},{atom,Lm,M},{')',Lrp},{dot,Ld}], Lu,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_toks(From, Cf, Cl, Cn, Cis, Fs, dict:erase({atom,M}, Ms));
scan_undef([{'(',Llp},{var,Lm,M},{')',Lrp},{dot,Ld}], Lu,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    scan_toks(From, Cf, Cl, Cn, Cis, Fs, dict:erase({var,M}, Ms));
scan_undef(Toks, Lu, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    epp_reply(From, {error,{Lu,epp,{bad,undef}}}),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms).

%% scan_include(Tokens, IncludeLine, From, File, CurrLine, CurrName,
%%		CurrIstack, Filestack, Macros)

scan_include([{'(',Llp},{string,Lf,NewName},{')',Lrp},{dot,Ld}], Li,
	     From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    enter_file(get(user_path), NewName, Li, From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_include(Toks, Li, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    epp_reply(From, {error,{Li,epp,{bad,include}}}),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms).

%% scan_ifdef(Tokens, IfdefLine, From, CurrFile, CurrLine, CurrName,
%%	      CurrIstack, FileStack, Macros)
%% scan_ifndef(Tokens, IfdefLine, From, CurrFile, CurrLine, CurrName,
%%	       CurrIstack, FileStack, Macros)
%%  Handle the conditional parsing of a file.
%%  Report a badly formed if[n]def test and then treat as undefined macro.

scan_ifdef([{'(',Llp},{atom,Lm,M},{')',Lrp},{dot,Ld}], Li,
	   From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({atom,M}, Ms) of
	{ok,Def} ->
	    scan_toks(From, Cf, Cl, Cn, [ifdef|Cis], Fs, Ms);
	error ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [ifdef])
    end;
scan_ifdef([{'(',Llp},{var,Lm,M},{')',Lrp},{dot,Ld}], Li,
	   From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({var,M}, Ms) of
	{ok,Def} ->
	    scan_toks(From, Cf, Cl, Cn, [ifdef|Cis], Fs, Ms);
	error ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [ifdef])
    end;
scan_ifdef(Toks, Li, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    epp_reply(From, {error,{Li,epp,{bad,ifdef}}}),
    wait_req_skip(Cf, Cl, Cn, Cis, Fs, Ms, [ifdef]).

scan_ifndef([{'(',Llp},{atom,Lm,M},{')',Lrp},{dot,Ld}], Li,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({atom,M}, Ms) of
	{ok,Def} ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [ifndef]);
	error ->
	    scan_toks(From, Cf, Cl, Cn, [ifndef|Cis], Fs, Ms)
    end;
scan_ifndef([{'(',Llp},{var,Lm,M},{')',Lrp},{dot,Ld}], Li,
	    From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    case dict:find({var,M}, Ms) of
	{ok,Def} ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [ifndef]);
	error ->
	    scan_toks(From, Cf, Cl, Cn, [ifndef|Cis], Fs, Ms)
    end;
scan_ifndef(Toks, Li, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    epp_reply(From, {error,{Li,epp,{bad,ifndef}}}),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms).

%% scan_else(Tokens, ElseLine, From, CurrFile, CurrLine, CurrName, CurrIstack,
%%	     FileStack, Macros)
%%  If we are in an if body then convert to else and skip, if we are in an
%%  else or not in anything report an error.

scan_else([{dot,Ld}], Le, From, Cf, Cl, Cn, [else|Cis], Fs, Ms) ->
    epp_reply(From, {error,{Le,epp,{illegal,"repeated",else}}}),
    wait_req_skip(Cf, Cl, Cn, Cis, Fs, Ms, [else]);
scan_else([{dot,Ld}], Le, From, Cf, Cl, Cn, [I|Cis], Fs, Ms) ->
    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [else]);
scan_else([{dot,Ld}], Le, From, Cf, Cl, Cn, [], Fs, Ms) ->
    epp_reply(From, {error,{Le,epp,{illegal,"unbalanced",else}}}),
    wait_req_scan(Cf, Cl, Cn, [], Fs, Ms);
scan_else(Toks, Le, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    epp_reply(From, {error,{Le,epp,{bad,else}}}),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms).

%% scan_endif(Tokens, EndifLine, From, CurrFile, CurrLine, CurrName,
%%	      CurrIstack, FileStack, Macros)
%%  If we are in an if body then exit it, else report an error.

scan_endif([{dot,Ld}], Le, From, Cf, Cl, Cn, [I|Cis], Fs, Ms) ->
    scan_toks(From, Cf, Cl, Cn, Cis, Fs, Ms);
scan_endif([{dot,Ld}], Le, From, Cf, Cl, Cn, [], Fs, Ms) ->
    epp_reply(From, {error,{Le,epp,{illegal,"unbalanced",endif}}}),
    wait_req_scan(Cf, Cl, Cn, [], Fs, Ms);
scan_endif(Toks, Le, From, Cf, Cl, Cn, Cis, Fs, Ms) ->
    epp_reply(From, {error,{Le,epp,{bad,endif}}}),
    wait_req_scan(Cf, Cl, Cn, Cis, Fs, Ms).

%% skip_toks(From, CurrFile, CurrLine, CurrName, CurrIstack,
%%	     FileStack, Macros, SkipIstack)
%%  Skip over forms until current conditional has been exited. Handle
%%  nested conditionals and repeated 'else's.

skip_toks(From, Cf, Cl0, Cn, Cis, Fs, Ms, [I|Sis]) ->
    case io:scan_erl_form(Cf, '', Cl0) of
	{ok,[{'-',Lh},{atom,Li,ifdef}|Toks],Cl} ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [ifdef,I|Sis]);
	{ok,[{'-',Lh},{atom,Li,ifndef}|Toks],Cl} ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [ifndef,I|Sis]);
	{ok,[{'-',Lh},{atom,Le,else}|Toks],Cl}->
	    skip_else(Le, From, Cf, Cl, Cn, Cis, Fs, Ms, [I|Sis]);
	{ok,[{'-',Lh},{atom,Le,endif}|Toks],Cl} ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, Sis);
	{ok,Toks,Cl} ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [I|Sis]);
	{error,E,Cl} ->
	    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, [I|Sis]);
	{eof,Cl} ->
	    leave_file(From, Cf, Cl, Cn, [I|Sis], Fs, Ms);
	{error,E} ->
	    epp_reply(From, {error,E}),
	    true				%This serious, just exit!
    end;
skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, []) ->
    scan_toks(From, Cf, Cl, Cn, Cis, Fs, Ms).

skip_else(Le, From, Cf, Cl, Cn, Cis, Fs, Ms, [else|Sis]) ->
    epp_reply(From, {error,{Le,epp,{illegal,"repeated",else}}}),
    wait_req_skip(Cf, Cl, Cn, Cis, Fs, Ms, [else|Sis]);
skip_else(Le, From, Cf, Cl, Cn, Cis, Fs, Ms, [I]) ->
    scan_toks(From, Cf, Cl, Cn, [else|Cis], Fs, Ms);
skip_else(Le, From, Cf, Cl, Cn, Cis, Fs, Ms, Sis) ->
    skip_toks(From, Cf, Cl, Cn, Cis, Fs, Ms, Sis).

%% macro_pars(Tokens, ArgStack)
%% macro_expansion(Tokens)
%%  Extract the macro parameters and the expansion from a macro definition.

macro_pars([{')',Lp}, {',',Ld}|Ex], Args) ->
    {ok, {lists:reverse(Args), macro_expansion(Ex)}};
macro_pars([{var,_,Name}, {')',Lp}, {',',Ld}|Ex], Args) ->
    false = lists:member(Name, Args),		%Prolog is nice
    {ok, {lists:reverse([Name|Args]), macro_expansion(Ex)}};
macro_pars([{var,L,Name}, {',',_}|Ts], Args) ->
    false = lists:member(Name, Args),           
    macro_pars(Ts, [Name|Args]).

macro_expansion([{')',Lp},{dot,Ld}]) -> [];
macro_expansion([{dot,Ld}]) -> [];		%Be nice, allow no right paren!
macro_expansion([T|Ts]) ->
    [T|macro_expansion(Ts)].

%% expand_macros(Tokens, Macros)
%% expand_macro(Tokens, MacroLine, RestTokens)
%%  Expand the macros in a list of tokens, making sure that an expansion
%%  gets the same line number as the macro call.

%% Macro expansion
expand_macros([{'?',Lq},{atom,Lm,M}|Toks], Ms) ->
    case dict:find({atom,M}, Ms) of
	{ok,{none,Exp}} ->
	    expand_macros(expand_macro(Exp, Lm, Toks, dict:new()), Ms);
	{ok,{As,Exp}} ->
	    {Bs,Toks1} = bind_args(Toks, Lm, M, As, dict:new()),
	    %%io:format("Bound arguments to macro ~w (~w)~n", [M,Bs]),
	    expand_macros(expand_macro(Exp, Lm, Toks1, Bs), Ms);
	error ->
	    throw({error,Lm,{undefined,M}})
    end;
%% Special macros
expand_macros([{'?',Lq},{var,Lm,'EPPLine'}|Toks], Ms) ->
    [{integer,Lm,Lm}|expand_macros(Toks, Ms)];
expand_macros([{'?',Lq},{var,Lm,M}|Toks], Ms) ->
    case dict:find({var,M}, Ms) of
	{ok,{none,Exp}} ->
	    expand_macros(expand_macro(Exp, Lm, Toks, []), Ms);
	{ok,{As,Exp}} ->
	    {Bs,Toks1} = bind_args(Toks, Lm, M, As, dict:new()),
	    %%io:format("Bound arguments to macro ~w (~w)~n", [M,Bs]),
	    expand_macros(expand_macro(Exp, Lm, Toks1, Bs), Ms);
	error ->
	    throw({error,Lm,{call,[$?|atom_to_list(M)]}})
    end;
%% Illegal macros
expand_macros([{'?',Lq},{Type,Lt}|Toks], Ms) ->
    throw({error,Lt,{call,[$?|atom_to_list(Type)]}});
expand_macros([{'?',Lq},{Type,Lt,What}|Toks], Ms) ->
    throw({error,Lt,{call,[$?|io_lib:write(What)]}});

expand_macros([T|Ts], Ms) ->
    [T|expand_macros(Ts, Ms)];
expand_macros([], Ms) -> [].

%% bind_args(Tokens, MacroLine, MacroName, ArgumentVars, Bindings)
%%  Collect the arguments to a macro call and check for correct number.

bind_args([{'(',Llp},{')',Lrp}|Toks], Lm, M, [], Bs) ->
    {Bs,Toks};
bind_args([{'(',Llp}|Toks0], Lm, M, [A|As], Bs) ->
    {Arg,Toks1} = macro_arg(Toks0, [], []),
    macro_args(Toks1, Lm, M, As, dict:store(A, Arg, Bs));
bind_args(Toks, Lm, M, As, Bs) ->
    throw({error,Lm,{mismatch,M}}).

macro_args([{')',Lrp}|Toks], Lm, M, [], Bs) ->
    {Bs,Toks};
macro_args([{',',Lc}|Toks0], Lm, M, [A|As], Bs) ->
    {Arg,Toks1} = macro_arg(Toks0, [], []),
    macro_args(Toks1, Lm, M, As, dict:store(A, Arg, Bs));
macro_args([], Lm, M, As, Bs) ->
    throw({error,Lm,{arg_error,M}});
macro_args(Toks, Lm, M, As, Bs) ->
    throw({error,Lm,{mismatch,M}}).

macro_arg([{',',Lc}|Toks], [], Arg) ->
    {lists:reverse(Arg),[{',',Lc}|Toks]};
macro_arg([{')',Lrp}|Toks], [], Arg) ->
    {lists:reverse(Arg),[{')',Lrp}|Toks]};
macro_arg([{'(',Llp}|Toks], E, Arg) ->
    macro_arg(Toks, [')'|E], [{'(',Llp}|Arg]);
macro_arg([{'[',Lls}|Toks], E, Arg) ->
    macro_arg(Toks, [']'|E], [{'[',Lls}|Arg]);
macro_arg([{'{',Llc}|Toks], E, Arg) ->
    macro_arg(Toks, ['}'|E], [{'{',Llc}|Arg]);
macro_arg([{'begin',Lb}|Toks], E, Arg) ->
    macro_arg(Toks, ['end'|E], [{'begin',Lb}|Arg]);
macro_arg([{'if',Li}|Toks], E, Arg) ->
    macro_arg(Toks, ['end'|E], [{'if',Li}|Arg]);
macro_arg([{'case',Lc}|Toks], E, Arg) ->
    macro_arg(Toks, ['end'|E], [{'case',Lc}|Arg]);
macro_arg([{'receive',Lr}|Toks], E, Arg) ->
    macro_arg(Toks, ['end'|E], [{'receive',Lr}|Arg]);
macro_arg([{Rb,Lrb}|Toks], [Rb|E], Arg) ->
    macro_arg(Toks, E, [{Rb,Lrb}|Arg]);
macro_arg([T|Toks], E, Arg) ->
    macro_arg(Toks, E, [T|Arg]);
macro_arg([], E, Arg) ->
    {lists:reverse(Arg),[]}.

%% expand_macro(MacroDef, MacroLine, RestTokens, Bindings)
%% expand_arg(Argtokens, MacroTokens, MacroLine, RestTokens, Bindings)
%%  Insert the macro expansion replacing macro parameters with their
%%  argument values, inserting the line number of first the macro call
%%  and then the macro arguments, i.e. simulate textual expansion.

expand_macro([{var,Lv,V}|Ts], L, Rest, Bs) ->
    case dict:find(V, Bs) of
	{ok,Val} ->
	    %% lists:append(Val, expand_macro(Ts, L, Rest, Bs));
	    expand_arg(Val, Ts, L, Rest, Bs);
	error ->
	    [{var,L,V}|expand_macro(Ts, L, Rest, Bs)]
    end;
expand_macro([T|Ts], L, Rest, Bs) ->
    [setelement(2, T, L)|expand_macro(Ts, L, Rest, Bs)];
expand_macro([], L, Rest, Bs) -> Rest.

expand_arg([A|As], Ts, L, Rest, Bs) ->
    [A|expand_arg(As, Ts, element(2, A), Rest, Bs)];
expand_arg([], Ts, L, Rest, Bs) ->
    expand_macro(Ts, L, Rest, Bs).

%% epp_request(Epp, Request)
%% epp_reply(From, Reply)
%% wait_epp_reply(Epp)
%%  Handle communication with the epp.

epp_request(Epp, Req) ->
    Epp ! {epp_request,self(),Req}.

epp_reply(From, Rep) ->
    From ! {epp_reply,self(),Rep}.

wait_epp_reply(Epp) ->
    receive
	{epp_reply,Epp,Rep} -> Rep
    end.
