%% Copyright (C) 1990, Ellemtel Telecommunications Systems Laboratories
%% File    : erl_eval.erl
%% Author  : Robert Virding
%% Purpose : An evaluator for Erlang abstract syntax.

-module(erl_eval).

-export([seq/2,seq/3,exprs/2,expr/2,expr/3,arg_list/2,arg_list/3]).
-export([new_bindings/0,bindings/1,binding/2,add_binding/3,del_binding/2]).

%% seq(ExpressionSeq, Bindings)
%% seq(ExpressionSeq, Bindings, LocalFuncHandler)
%% exprs(ExpressionSeq, Bindings)
%%  Exprs is included for backward compatability.
%%  Returns:
%%	{value,Value,NewBindings}

exprs(Seq, Bs) ->
    seq(Seq, Bs).

seq(Seq, Bs) ->
    seq(Seq, Bs, none).

seq(Seq, Bs, Lf) ->
    seq(Seq, Bs, true, Lf).

seq([E|Es], Bs0, _, Lf) ->
    {value,V,Bs} = expr(E, Bs0, Lf),
    seq(Es, Bs, V, Lf);
seq([], Bs, V, Lf) ->
    {value,V,Bs}.

%% expr(Expression, Bindings)
%% expr(Expression, Bindings, LocalFuncHandler)
%%  Returns:
%%	{value,Value,NewBindings}

expr(E, Bs) ->
    expr(E, Bs, none).

expr({atom,_,A}, Bs, Lf) ->
    {value,A,Bs};
expr({integer,_,I}, Bs, Lf) ->
    {value,I,Bs};
expr({float,_,F}, Bs, Lf) ->
    {value,F,Bs};
expr({char,_,C}, Bs, Lf) ->
    {value,C,Bs};
expr({var,L,V}, Bs, Lf) ->
    case binding(V, Bs) of
	{value,Val} ->
	    {value,Val,Bs};
	unbound ->
	    exit({{unbound,V},{erl_eval,expr,3}})
    end;
expr({string,_,S}, Bs, Lf) ->
    {value,S,Bs};
expr({nil, _}, Bs, Lf) ->
    {value,[],Bs};
expr({cons,_,H0,T0}, Bs0, Lf) ->
    {value,H,Bs1} = expr(H0, Bs0, Lf),
    {value,T,Bs2} = expr(T0, Bs0, Lf),
    {value,[H|T],merge_bindings(Bs1, Bs2)};
expr({tuple,_,Es0}, Bs0, Lf) ->
    {Vs,Bs} = arg_list(Es0, Bs0, Lf),
    {value,list_to_tuple(Vs),Bs};
expr({arith,_,Op,L0,R0}, Bs0, Lf) ->
    {value,L,Bs1} = expr(L0, Bs0, Lf),
    {value,R,Bs2} = expr(R0, Bs0, Lf),
    {value,eval_arith(Op, L, R),merge_bindings(Bs1, Bs2)};
expr({arith,_,Op,A0}, Bs0, Lf) ->
    {value,A,Bs} = expr(A0, Bs0, Lf),
    {value,eval_arith(Op, A),Bs};
expr({bif,_,Name,As0}, Bs0, Lf) ->
    {As,Bs} = arg_list(As0, Bs0, Lf),
    {value,bif(Name, As),Bs};
expr({call,_,[],Name,As0}, Bs0, {M,F}) ->	%Local functions handler
    {As,Bs} = arg_list(As0, Bs0, {M,F}),
    {value,apply(M, F, [Name,As]),Bs};
expr({call,_,[],Name,As}, Bs, {M,F,Eas}) ->
    %% This gives complete control of everything to the handler.
    apply(M, F, [Name,As,Bs|Eas]);
expr({call,_,[],Name,As0}, Bs0, Lf) ->		%No local functions
    {As,Bs} = arg_list(As0, Bs0, Lf),
    exit({undef,{erl_eval,Name,As}});
expr({call,_,Mod,Name,As0}, Bs0, Lf) ->
    {As,Bs} = arg_list(As0, Bs0, Lf),
    {value,apply(Mod, Name, As),Bs};
expr({match,_,Lhs,Rhs0}, Bs0, Lf) ->
    {value,Rhs,Bs1} = expr(Rhs0, Bs0, Lf),
    case match(Lhs, Rhs, Bs1) of
	{match,Bs} ->
	    {value,Rhs,Bs};
	nomatch ->
	    exit({{badmatch,Rhs},{erl_eval,expr,3}})
    end;
expr({block,_,Es}, Bs, Lf) ->
    seq(Es, Bs, Lf);
expr({'if',_,Cs}, Bs, Lf) ->
    if_clauses(Cs, Bs, Lf);
expr({'case',_,E,Cs}, Bs0, Lf) ->
    {value,Val,Bs} = expr(E, Bs0, Lf),
    case_clauses(Val, Cs, Bs, Lf);
expr({'receive',_,Cs}, Bs, Lf) ->
    receive_clauses(Cs, Bs, Lf,  []);
expr({'receive',_, Cs, E, TB}, Bs0, Lf) ->
    {value,T,Bs} = expr(E, Bs0, Lf),
    receive_clauses(T, Cs, {TB,Bs}, Bs0, Lf, []);
expr({'catch',_,Expr}, Bs0, Lf) ->
    case catch expr(Expr, Bs0, Lf) of
	{value,Val,Bs} ->
	    {value,Val,Bs};
	Other ->
	    {value,Other,Bs0}
    end;
expr({send,_,T0,M0}, Bs0, Lf) ->
    {value,T,Bs1} = expr(T0, Bs0, Lf),
    {value,M,Bs2} = expr(M0, Bs0, Lf),
    {value,T ! M,merge_bindings(Bs1, Bs2)};
expr({value,_,Val}, Bs, Lf) ->			%Special case straight values.
    {value,Val,Bs};
expr(E, Bs, Lf) ->				%Not Yet Implemented
    exit({'NYI',E}).

%% arg_list(ExpressionList, Bindings)
%% arg_list(ExpressionList, Bindings, LocalFuncHandler)
%%  Evaluate a list of expressions "in parallel" at the same level.

arg_list(Es, Bs) ->
    arg_list(Es, [], Bs, Bs, none).

arg_list(Es, Bs, Lf) ->
    arg_list(Es, [], Bs, Bs, Lf).

arg_list([E|Es], Vs, BsOrig, Bs0, Lf) ->
    {value,V,Bs1} = expr(E, BsOrig, Lf),
    arg_list(Es, [V|Vs], BsOrig, merge_bindings(Bs1, Bs0), Lf);
arg_list([], Vs, _, Bs, Lf) ->
    {reverse(Vs),Bs}.

eval_arith('+', A1, A2) when number(A1), number(A2) ->
    A1 + A2;
eval_arith('-', A1, A2) when number(A1), number(A2) ->
    A1 - A2;
eval_arith('*', A1, A2) when number(A1), number(A2) ->
    A1 * A2;
eval_arith('/', A1, A2) when number(A1), number(A2) ->
    A1 / A2;
eval_arith('div', A1, A2) when integer(A1), integer(A2) ->
    A1 div A2;
eval_arith('rem', A1, A2) when integer(A1), integer(A2) ->
    A1 rem A2;
eval_arith('band', A1, A2) when integer(A1), integer(A2) ->
    A1 band A2;
eval_arith('bor', A1, A2) when integer(A1), integer(A2) ->
    A1 bor A2;
eval_arith('bxor', A1, A2) when integer(A1), integer(A2) ->
    A1 bxor A2;
eval_arith('bsl', A1, A2) when integer(A1), integer(A2) ->
    A1 bsl A2;
eval_arith('bsr', A1, A2) when integer(A1), integer(A2) ->
    A1 bsr A2;
eval_arith('^', A1, A2) when number(A1), number(A2) ->
    math:pow(A1, A2);
eval_arith(Op, A1, A2) ->
    exit({{badarith,{erlang,Op,[A1,A2]}},{erl_eval,expr,3}}).

eval_arith('+', A) when number(A) ->
    A;
eval_arith('-', A) when number(A) ->
    -A;
eval_arith('bnot', A) when integer(A) ->
    bnot A;
eval_arith(Op, A) ->
    exit({{badarith,{erlang,Op,[A]}},{erl_eval,expr,3}}).

%% bif(Name, Arguments)
%%  Evaluate the Erlang builtin function Name. (Special case apply here)

bif(apply, [M,F,As]) ->
    apply(M, F, As);
bif(Name, As) ->
    apply(erlang, Name, As).

%% if_clauses(Clauses, Bindings, LocalFuncHandler)

if_clauses([{clause,_,G,B}|Cs], Bs, Lf) ->
    case guard(G, Bs, Lf) of
	true ->
	    seq(B, Bs, Lf);
	false ->
	    if_clauses(Cs, Bs, Lf)
    end;
if_clauses([], Bs, Lf) ->
    exit({if_clause,{erl_eval,expr,3}}).

%% case_clauses(Value, Clauses, Bindings, LocalFuncHandler)

case_clauses(Val, Cs, Bs, Lf) ->
    case match_clause(Cs, Val, Bs, Lf) of
	{B, Bs1} ->
	    seq(B, Bs1, Lf);
	nomatch ->
	    exit({{case_clause,Val},{erl_eval,expr,3}})
    end.

%%
%% receive(Clauses, Bindings, LocalFuncHandler, Messages) 
%%
receive_clauses(Cs, Bs, Lf, Ms) ->
    receive
	Val ->
	    case match_clause(Cs, Val, Bs, Lf) of
		{B, Bs1} ->
		    merge_queue(Ms),
		    seq(B, Bs1, Lf);
		nomatch ->
		    receive_clauses(Cs, Bs, Lf, [Val|Ms])
	    end
    end.
%%
%% receive_clauses(TimeOut, Clauses, TimeoutBody, Bindings, LocalFuncHandler)
%%
receive_clauses(T, Cs, TB, Bs, Lf, Ms) ->
    {_,_} = statistics(runtime),
    receive
	Val ->
	    case match_clause(Cs, Val, Bs, Lf) of
		{B, Bs1} ->
		    merge_queue(Ms),
		    seq(B, Bs1, Lf);
		nomatch ->
		    {_,T1} = statistics(runtime),
		    if
			T == infinity ->
			    receive_clauses(T, Cs, TB, Bs, Lf, [Val|Ms]);
			T-T1 =< 0 ->
			    receive_clauses(0, Cs, TB, Bs, Lf, [Val|Ms]);
			true ->
			    receive_clauses(T-T1, Cs, TB, Bs, Lf, [Val|Ms])
		    end
	    end
    after T ->
	    merge_queue(Ms),
	    {B, Bs1} = TB,
	    seq(B, Bs1, Lf)
    end.

merge_queue(Ms) ->
    Old = process_flag(pre_empt, false),
    send_all(recv_all(Ms), self()),
    process_flag(pre_empt, Old).

recv_all(Xs) ->
    receive
	X -> recv_all([X|Xs])
    after 0 ->
	    lists:reverse(Xs)
    end.

send_all([X|Xs], Self) ->
    Self ! X,
    send_all(Xs, Self);
send_all([], _) -> true.
	
    
%% match_clause -> {Body, Bindings} or nomatch

match_clause([{clause,_,[P],G,B}|Cs], Val, Bs, Lf) ->
    case match(P, Val, Bs) of
	{match, Bs1} ->
	    case guard(G, Bs1, Lf) of
		true ->
		    {B, Bs1};
		false ->
		    match_clause(Cs, Val, Bs, Lf)
	    end;
	nomatch ->
	    match_clause(Cs, Val, Bs, Lf)
    end;
match_clause([], _, _, _) ->
    nomatch.


%% guard(GuardTests, Bindings)

guard([{test,_,Name,As0}|Gs], Bs0, Lf) ->
    {As,Bs} = arg_list(As0, Bs0, Lf),
    case guard_test(Name, As) of
	true ->
	    guard(Gs, Bs, Lf);
	false ->
	    false
    end;
guard([{comp,_,Op,Lhs0,Rhs0}|Gs], Bs0, Lf) ->
    {value,Lhs,Bs1} = expr(Lhs0, Bs0, Lf),
    {value,Rhs,Bs} = expr(Rhs0, Bs1, Lf),
    case guard_comp(Op, Lhs, Rhs) of
	true ->
	    guard(Gs, Bs, Lf);
	false ->
	    false
    end;
guard([], Bs, Lf) ->
    true.

guard_test(integer, [A]) when integer(A) -> true;
guard_test(float, [A]) when float(A) -> true;
guard_test(number, [A]) when number(A) -> true;
guard_test(atom, [A]) when atom(A) -> true;
guard_test(constant, [A]) when constant(A) -> true;
guard_test(list, [A]) when list(A) -> true;
guard_test(tuple, [A]) when tuple(A) -> true;
guard_test(pid, [A]) when pid(A) -> true;
guard_test(reference, [A]) when reference(A) -> true;
guard_test(port, [A]) when port(A) -> true;
guard_test(_, _) ->
    false.

%% Arithmetic comparison operators
guard_comp('==', Lhs, Rhs) when Lhs == Rhs -> true;
guard_comp('/=', Lhs, Rhs) when Lhs /= Rhs -> true;
guard_comp('=<', Lhs, Rhs) when Lhs =< Rhs -> true;
guard_comp('<', Lhs, Rhs) when Lhs < Rhs -> true;
guard_comp('>=', Lhs, Rhs) when Lhs >= Rhs -> true;
guard_comp('>', Lhs, Rhs) when Lhs > Rhs -> true;
%% Term comparison operators
guard_comp('=:=', Lhs, Rhs) when Lhs =:= Rhs -> true;
guard_comp('=/=', Lhs, Rhs) when Lhs =/= Rhs -> true;
guard_comp(_, _, _) -> false.

%% match(Pattern, Term, Bindings)
%%  Try to match Pattern against Term with the current bindings.
%%  Returns:
%%	{match,NewBindings}
%%	nomatch

match(Pat, Term, Bs) ->
    catch match1(Pat, Term, Bs).

match1({atom,_,A}, A, Bs) ->
    {match,Bs};
match1({integer,_,I}, I, Bs) ->
    {match,Bs};
match1({float,_,F}, F, Bs) ->
    {match,Bs};
match1({char,_,C}, C, Bs) ->
    {match,Bs};
match1({var,_,'_'}, _, Bs) ->			%Anonymous variable matches
    {match,Bs};					% everything, no new bindings
match1({var,_,Name}, Term, Bs) ->
    case binding(Name, Bs) of
	{value,Term} ->
	    {match,Bs};
	{value,V} ->
	    throw(nomatch);
	unbound ->
	    {match,add_binding(Name, Term, Bs)}
    end;
match1({string,_,S}, S, Bs) ->
    {match,Bs};
match1({nil,_}, [], Bs) ->
    {match,Bs};
match1({cons,_,H,T}, [H1|T1], Bs0) ->
    {match,Bs} = match1(H, H1, Bs0),
    match1(T, T1, Bs);
match1({tuple,_,Elts}, Tuple, Bs) when length(Elts) == size(Tuple) ->
    match_list(Elts, Tuple, 1, Bs);
match1(_, _, _) ->
    throw(nomatch).

match_list([E|Es], Tuple, I, Bs0) ->
    {match,Bs} = match1(E, element(I, Tuple), Bs0),
    match_list(Es, Tuple, I+1, Bs);
match_list([], _, _, Bs) ->
    {match,Bs}.

%% new_bindings()
%% bindings(Bindings)
%% binding(Name, Bindings)
%% add_binding(Name, Value, Bindings)
%% del_binding(Name, Bindings)

new_bindings() ->
    [].

bindings(Bs) ->
    Bs.						%Already as list of pairs

binding(Name, [{Name,Val}|_]) ->
    {value,Val};
binding(Name, [{_,_}|Bs]) ->
    binding(Name, Bs);
binding(_, []) ->
    unbound.

add_binding(Name, Val, Bs) ->
    [{Name,Val}|Bs].

del_binding(Name, [{Name,_}|Bs]) ->
    Bs;
del_binding(Name, [B|Bs]) ->
    [B|del_binding(Name, Bs)];
del_binding(Name, []) ->
    [].

%% merge_bindings(Bindings1, Bindings2)
%%  Merge bindings detecting bad matches.

merge_bindings([{Name,V}|B1s], B2s) ->
    case binding(Name, B2s) of
	{value,V} ->				%Already there, and the SAME
	    merge_bindings(B1s, B2s);
	{value,V1} ->				%	but different
	    exit({{badmatch,V1},{erl_eval,expr,3}});
	unbound ->				%Not there, add it
	    [{Name,V}|merge_bindings(B1s, B2s)]
    end;
merge_bindings([], B2s) ->
    B2s.

reverse(List) ->
    reverse(List, []).

reverse([H|T], S) ->
    reverse(T, [H|S]);
reverse([], S) ->
    S.
