%% Copyright (C) 1991, Ellemtel Telecommunications Systems Laboratories
%% File    : sys_lint.erl
%% Author  : Robert Virding
%% Purpose : Do necessary checking of Erlang code.

-module(erl_lint).

-export([func/1,format_error/1]).

-import(ordsets, [new_set/0,list_to_set/1,
		  is_element/2,add_element/2,del_element/2,
		  union/1,union/2,intersection/1,intersection/2,subtract/2]).
-import(lists,   [append/2]).

%% format_error(Error)
%%  Return a string describing the error.

format_error({unbound_variable,Var}) ->
    io_lib:format("unbound variable ~w", [Var]);
format_error({unsafe_variable,Var}) ->
    io_lib:format("unsafe variable ~w", [Var]).

%% func(Function) ->
%%	{[Error],[Warning]}
%%
%%  Checks for legal usage of variables. They must have been given a value
%%  before they are used and not be unsafe. This is really the only check
%%  that MUST be done to guarantee legal code.
%%
%%  Errors is a list:
%%      {unbound_variable, Variable}
%%      {unsafe_variable, Variable}
%%  Warnings is empty.

func({function,Line,Func,Arity,Cs}) ->
    {UnBound, UnSafe} = clauses(Cs, [], [], []),
    L1 = add_errors(unbound_variable, UnBound, []),
    {add_errors(unsafe_variable, UnSafe, L1),[]}.

add_errors(Type, [{Line,Var}|Ts], L) ->
    add_errors(Type, Ts, [{Line,erl_lint,{Type,Var}}|L]);
add_errors(_, [], L) ->
    L.

%% clauses(...) -> {UnBound, Unsafe}
%% i.e. {[], []} 			     if no errors
%%      {{Line,Varaible}*, {Line,Variable*}} if errors

clauses([C|Cs], Imp, UnBound0, UnSafe0) ->
    {UnBound1, UnSafe1} = check_vars(clause(C, Imp)),
    clauses(Cs, Imp, append(UnBound0, UnBound1), append(UnSafe0, UnSafe1));
clauses([], Imp, UnBound, UnSafe) ->
    {UnBound, UnSafe}.

%% check_vars(VariableList)
%% check_vars(VariableList, Unbound, Unsafe)
%%  Check for unbound or unsafe variables in VariableList generating errors
%%  if any exist.

check_vars(Vs) ->
    check_vars(Vs, [], []).

check_vars([{unbound,Line,V}|Vs], Unb, Uns) ->
    check_vars(Vs, [{Line,V}|Unb], Uns);
check_vars([{unsafe,Line,V}|Vs], Unb, Uns) ->
    check_vars(Vs, Unb, [{Line, V}|Uns]);
check_vars([P|Vs], Unb, Uns) ->
    check_vars(Vs, Unb, Uns);
check_vars([], Unb, Uns) ->
    {Unb,Uns}.

%% clause(Clause, ImportedVariables)
%%  Check a clause, return a VariableList. This list contains the variables
%%  that have been seen, {maybe_unsafe,Var} for each variable that may be
%%  unsafe, {unsafe,Var} and {unbound,Var} for unsafe or unbound variables.
%%  These last two are errors. This structure is used throughout.

clause({clause,Line,H,G,B}, Imp) ->
    Vsh = vars_in(H, true, Imp),
    Vsg = seq(G, Vsh),
    seq(B, Vsg);
clause({clause,Line,G,B}, Imp) ->
    Vsg = seq(G, Imp),
    seq(B, Vsg).

%% seq(Sequence, VisibleVariables)
%%  Check a sequence, return UsedVariables.

seq([E|Es], Vs) ->
    Ns = expr(E, Vs),
    seq(Es, union(Vs, Ns));
seq([], Vs) ->
    Vs.

%% expr(Expression, VisibleVariables)
%%  Check an expression, returns NewVariables.

expr({cons,Line,H,T}, Vs) ->
    union(expr(H, Vs), expr(T, Vs));
expr({tuple,Line,Es}, Vs) ->
    expr_list(Es, Vs);
expr({arith,Line,Op,L,R}, Vs) ->
    union(expr(L, Vs), expr(R, Vs));
expr({arith,Line,Op,A}, Vs) ->
    expr(A, Vs);
expr({bif,Line,N,As}, Vs) ->
    expr_list(As, Vs);
expr({call,Line,[],F,As}, Vs) ->
    expr_list(As, Vs);
expr({call,Line,M,F,As}, Vs) ->
    expr_list(As, Vs);
expr({match,Line,P,E}, Vs) ->
    Ms = expr(E, Vs),
    %% Must do some work here to get just new stuff.
    subtract(vars_in(P, true, union(Vs, Ms)), Vs);
expr({block,Line,Es}, Vs) ->
    %% Unfold block into a sequence.
    seq(Es, Vs);
expr({'if',Line,Cs}, Vs) ->
    Cvs = icr_clauses(Cs, Vs),
    {All,Some} = new_in(Vs, Cvs),
    maybe_unsafe(Some, All);
expr({'case',Line,E,Cs}, Vs) ->
    Ns = expr(E, Vs),
    Before = union(Vs, Ns),
    Cvs = icr_clauses(Cs, Before),
    {All,Some} = new_in(Before, Cvs),
    union(Ns, maybe_unsafe(Some, All));
expr({'catch',Line,E}, Vs) ->
    %% No new variables added.
    Cs = expr(E, Vs),
    maybe_unsafe(Cs, []);
expr({send,Line,T,M}, Vs) ->
    union(expr(T, Vs), expr(M, Vs));
expr({'receive',Line,Cs}, Vs) ->
    Cvs = icr_clauses(Cs, Vs),
    {All,Some} = new_in(Vs, Cvs),
    maybe_unsafe(Some, All);
expr({'receive',Line,Cs,To,ToEs}, Vs) ->
    %% Are variables from the timeout expression visible in the clauses?
    Before = union(Vs, Ns = expr(To, Vs)),
    Nse = seq(ToEs, Before),
    Cvs = icr_clauses(Cs, Before),
    {All,Some} = new_in(Before, [Nse|Cvs]),
    union(Ns, maybe_unsafe(Some, All));
expr({record,Line,Name,Pats}, Vs) ->
    record_list(Pats, Vs);
expr({record,Line,Rec,Name,Pats}, Vs) ->
    union(expr(Rec, Vs), record_list(Pats, Vs));
expr({record_field,Line,Rec,Name,Field}, Vs) ->
    expr(Rec, Vs);
%% The next two cases are for guards.
expr({test,Line,Test,As}, Vs) ->
    expr_list(As, Vs);
expr({comp,Line,Op,L,R}, Vs) ->
    union(expr(L, Vs), expr(R, Vs));
expr({var,Line,V}, Vs) ->
    add_var(V, Line, Vs, true);
expr(Const, Vs) ->
    [].

expr_list([E|Es], Vs) ->
    union(expr(E, Vs), expr_list(Es, Vs));
expr_list([], Vs) ->
    [].

record_list([{Field,Val}|Fs], Vs) ->
    union(expr(Val, Vs), record_list(Fs, Vs));
record_list([], Vs) ->
    [].

%% icr_clauses(Clauses, Imports)

icr_clauses([C|Cs], Imp) ->
    Vsc = clause(C, Imp),
    Vss = icr_clauses(Cs, Imp),
    [Vsc|Vss];
icr_clauses([], Imp) ->
    [].

%% new_in(Before, RegionList)
%%  Return the variables new in all clauses and those new in some clauses.

new_in(Before, Region) ->
    InAll = intersection(Region),
    InSome = union(Region),
    NewInAll = subtract(InAll, Before),
    NewInSome = subtract(subtract(InSome, Before), NewInAll),
    {NewInAll,NewInSome}.

%% add_var(Variable, VariableList, UnboundIsError)
%%  Add a variable to the variable list, checking if it is possibly unsafe
%%  or if unbound is an error.

add_var(V, Line, Vs, Unbound) ->
    case is_element(V, Vs) of
	true ->
	    Vs;
	false ->
	    case is_unsafe(V, Vs) of
		true ->
		    add_element(V, add_element({unsafe,Line,V}, Vs));
		false when Unbound == true ->
		    add_element(V, add_element({unbound,Line,V}, Vs));
		false ->
		    add_element(V, Vs)
	    end
    end.

is_unsafe(V, [{maybe_unsafe,V}|Vs]) ->
    true;
is_unsafe(V, [P|Vs]) ->
    is_unsafe(V, Vs);
is_unsafe(V, []) ->
    false.

%% maybe_unsafe(Variables, VariableList)
%%  Add Variables to VariableList marking them as possibly unsafe.

maybe_unsafe([V|Vs], Vars) when atom(V) ->
    maybe_unsafe(Vs, add_element({maybe_unsafe,V}, Vars));
maybe_unsafe([V|Vs], Vars) ->
    maybe_unsafe(Vs, add_element(V, Vars));
maybe_unsafe([], Vars) ->
    Vars.

%% vars_in(Expression, IgnoreDC, KnownVars)
%%  Add variables in Expression to KnownVars checking for unsafe variables
%%  but allowing unbound variables. This is really useful only for patterns.

vars_in({var,_,'_'}, true, Vars) ->
    Vars;
vars_in({var,Line,V}, I, Vars) ->
    add_var(V, Line,Vars, false);
%% Add case for each sized tagged tuple that can occur.
vars_in({Tag,A1}, I, Vs0) ->			%This is for record fields.
    vars_in(A1, I, Vs0);
vars_in({Tag,_,A1}, I, Vs0) ->
    vars_in(A1, I, Vs0);
vars_in({Tag,_,A1,A2}, I, Vs0) ->
    Vs1 = vars_in(A1, I, Vs0),
    vars_in(A2, I, Vs1);
%% Clauses for handling lists of objects.
vars_in([H|T], I, Vars0) ->
    Vars1 = vars_in(H, I, Vars0),
    vars_in(T, I, Vars1);
vars_in([], I, Vars) ->
    Vars;
%% Default case for constants.
vars_in(T, I, Vars) ->
    Vars.
