%% Copyright (C) 1990, Ellemtel Telecommunications Systems Laboratories
%% File    : parse.erl
%% Author  : Robert Virding
%% Purpose : Parse attribute or function definition forms.

%% Modifications:
%%   930326 -- Joe Armstrong
%%      Changed {number, N} wrappers to
%%      {integer, N} or {float, N}
%%   940125 -- Magnus Froberg
%%      Changed parser to include line number information in
%%      the abstract form. The input Tokens is wrapped with line
%%      information. (e.g. {atom,A} ==> {atom,{Line,A}}
%%      Handles also the old format of token lists.

-module(erl_parse).

-export([parse_form/1,parse_exprs/1,parse_term/1,format_error/1]).
-export([is_pattern/1,is_patternlist/1]).
-export([is_term/1,is_termlist/1]).
-export([normalise/1,normalise_list/1,abstract/1,tokens/1,tokens/2]).
-export([inop_prec/1,preop_prec/1,max_prec/0]).
-export([builtins/0,erlang_bif/2,erlang_guard_bif/2,erlang_guard_test/2]).

-import(lists, [append/2,sublist/3]).
-import(io_lib, [write/1,format/2,write_char/1,write_string/1]).

parse_form(Tokens) ->  
     case catch form(Tokens) of
	{error,What,Rest,Line} ->
	    parse_error(What, Rest, Line);
	{Line,Form,Rest} ->
	    all_parsed_form(Rest, Form, Tokens, Line)
    end.

parse_exprs(Tokens) ->
     case catch exprs(Tokens) of
	{error,What,Rest,Line} ->
	    parse_error(What, Rest, Line);
	{Line,Form,Rest} ->
	    all_parsed_form(Rest, Form, Tokens, Line)
    end.

parse_term(Tokens) ->
    case catch term(Tokens) of
	{error,What,Rest,Line} ->
	    parse_error(What, Rest, Line);
	{Line,Term,Rest} ->
	    all_parsed_term(Rest, Term, Tokens, Line)
    end.

%% format_error(Error)
%%  Return a string describing the error.

format_error({before,Error,Tok}) ->
    io_lib:format("~s before '~s'", [format_error(Error),write_token(Tok)]);
format_error({illegal,What}) ->
    format("illegal ~s", [What]);
format_error({bad,What}) ->
    format("bad ~s", [What]);
format_error({missing,What}) ->
    format("missing ~s", [What]);
format_error({terminator,Termin}) ->
    format("',' or '~s' expected", [Termin]);
format_error({expect,T})      ->
    format("~s expected",[T]);
format_error(head) ->
    "head mismatch";
format_error(comp_op) ->
    "comparison operator expected";
format_error('end') ->
    "premature end";
format_error(guard_call) ->
    "illegal call in guard";
format_error(list) ->
    "',' '|' or ']' expected in list";
format_error({prefix_op,Op}) ->
    format("prefix operator '~s' with wrong precedence", [Op]).

write_token({atom,_,A}) -> write(A);
write_token({integer,_,N}) -> write(N);
write_token({float,_,F}) -> write(F);
write_token({char,_,C}) -> write_char(C);
write_token({var,_,V}) -> format('~s', [V]);
write_token({string,_,S}) -> write_string(S);
write_token({Other,_}) when atom(Other) ->	%Puntuation tokens
    format('~s', [Other]);
write_token(Other) ->
    write(Other).

%% all_parsed_form(Rest, Form, Tokens, Line) ->
%%	{ok,Form}
%%	{error,{Line,erl_parse,Error}}

all_parsed_form([{dot,L}], Form, Tokens, Line) ->
    {ok,Form};
all_parsed_form(Ts, Form, Tokens, Line) ->
    parse_error({missing,punctuation}, Ts, Line).

%% all_parsed_term(Rest, Form, Tokens, Line) ->
%%	{ok,Term}
%%	{error,{Line,erl_parse,Error}}

all_parsed_term([{dot,L}], AbsTerm, Tokens, Line) ->
    {ok,normalise(AbsTerm)};
all_parsed_term(Ts, Term, Tokens, Line) ->
    parse_error({missing,punctuation}, Ts, Line).

%% expect(Tokens, Type, CurrentLine) ->
%%	{TokenLine,RestTokens}

expect([{Tok,Line,_}|Toks], Tok, _) ->
    {Line,Toks};
expect([{Tok,Line}|Toks], Tok, _) ->
    {Line,Toks};
expect(Toks, Tok, Line) ->
    syntax_error({expect,Tok}, Toks, Line).

syntax_error(What, Rest, Line) ->
    throw({error,What,Rest,Line}).

parse_error(Error, [T|Ts], Line) ->
    {error,{token_line(T),erl_parse,{before,Error,T}}};
parse_error(Error, [], Line) ->
    {error,{Line,erl_parse,Error}}.

token_line({Type,Line,Value}) -> Line;
token_line({Type,Line}) -> Line.

%% form(Tokens) ->
%%	{LastLine,Form,RestTokens}

form([{'-',Line}|Ts]) ->
    attribute(Ts, Line);
form([T|Ts]) ->
    function([T|Ts], token_line(T));
form([]) ->
    syntax_error({missing,form}, [], 0).

attribute([{atom,L0,module}|Ts0], Line) ->
    {L1,Ts1} = expect(Ts0, '(', L0),
    {L2,Mod,Ts2} = module_name(Ts1, L1),
    {L3,Ts3} = expect(Ts2, ')', L2),
    {L3,{attribute,Line,module,Mod},Ts3};
attribute([{atom,L0,export}|Ts0], Line) ->
    {L1,Es,Ts1} = export(Ts0, L0),
    {L1,{attribute,Line,export,Es},Ts1};
attribute([{atom,L0,import}|Ts0], Line) ->
    {L1,Is,Ts1} = import(Ts0, L0),
    {L1,{attribute,Line,import,Is},Ts1};
attribute([{atom,L0,file}|Ts0], Line) ->
    {L1,F,Ts1} = file(Ts0, L0),
    {L1,{attribute,Line,file,F},Ts1};
attribute([{atom,L0,record}|Ts0], Line) ->
    {L,Rec,Ts1} = record(Ts0, L0),
    {L,{attribute,Line,record,Rec},Ts1};
attribute([{atom,L0,Attr}|Ts0], Line) ->	%General attribute
    {L1,Ts1} = expect(Ts0, '(', L0),
    {L2,Val,Ts2} = term(Ts1, L1),
    {L,Ts} = expect(Ts2, ')', L2),
    {L,{attribute,L0,Attr,normalise(Val)},Ts};
attribute(Ts, Line) ->
    syntax_error({missing,"attribute name"}, Ts, Line).

module_name([{atom,L,Mod}|Ts], Line) ->
    {L,Mod,Ts};
module_name(Other, Line) ->
    syntax_error({expect,module}, Other, Line).
  
export([{'(',_},{'[',L0}|Ts0], _) ->
    {L1,Es,Ts1} = export_list(Ts0, ']',L0),
    {L2,Ts2} = expect(Ts1, ')',L1),
    {L2,Es,Ts2};
export([{'(',L0}|Ts], _) ->
    export_list(Ts, ')', L0);
export(Ts, Line) ->
    syntax_error({bad,export}, Ts, Line).

export_list(Ts0, Termin, Line) ->
    {L1,E,Ts1} = farity(Ts0, export, Line),
    {L2,Es,Ts} = farity_list(Ts1, Termin, export, L1),
    {L2,[E|Es],Ts}.

import([{'(',_},{atom,_,Mod},{',',_},{'[',Line}|Ts0], _) ->
    {L1,Is,Ts} = import_list(Ts0, Line),
    {L2,Ts1} = expect(Ts, ')', L1),
    {L2,{Mod,Is},Ts1};
import(Ts, Line) ->
    syntax_error({bad,import}, Ts, Line).

import_list(Ts0, Line) ->
    {L1,I,Ts1} = farity(Ts0, import, Line),
    {L2,Is,Ts} = farity_list(Ts1, ']', import, L1),
    {L2,[I|Is],Ts}.

farity([{atom,_,Name},{'/',_},{integer,Line,Arity}|Ts], _, _) ->
    {Line,{Name,Arity},Ts};
farity(Ts, InWhat, Line) ->
    syntax_error({illegal,"function/arity"}, Ts, Line).

farity_list([{',',Line}|Ts0], Termin, InWhat, _) ->
    {L1,F,Ts1} = farity(Ts0, InWhat, Line),
    {L2,Fs,Ts} = farity_list(Ts1, Termin, InWhat, L1),
    {L2,[F|Fs],Ts};
farity_list([{Termin,L0}|Ts], Termin, _, _) ->
    {L0,[],Ts};
farity_list(Ts, Termin, InWhat, Line) ->
    syntax_error({terminator,Termin}, Ts, Line).

file([{'(',_},{string,_,Name},{',',_},{integer,_,I},{')',L}|Ts], Line) ->
    {L,{Name,I},Ts};
file(Ts, Line) ->
    syntax_error({bad,file}, Ts, Line).

record([{'(',Llp},{atom,Ln,Name},{',',Lc},{'{',Ll},{'}',Lr}|Ts0], Line) ->
    {L,Ts1} = expect(Ts0, ')', Lr),
    {L,{Name,[]},Ts1};
record([{'(',Llp},{atom,Ln,Name},{',',Lc},{'{',Ll}|Ts0], Line) ->
    {Li,I,Ts1} = record_init(Ts0, expr, Ll),
    {Lr,Is,Ts2} = record_inits(Ts1, expr, Li),
    {L,Ts3} = expect(Ts2, ')', Lr),
    {L,{Name,[I|Is]},Ts3};
record(Ts, Line) ->
    syntax_error({illegal,"record definition"}, Ts, Line).

record_init([{atom,Lf,Field},{'=',Le}|Ts0], Type, Line) ->
    {Lv,Val,Ts1} = expr(Ts0, Type, Le),
    {Lv,{{atom,Lf,Field},Val},Ts1};
record_init([{atom,Lf,Field}|Ts], Type, Line) ->
    {Lf,{{atom,Lf,Field},{atom,Lf,undefined}},Ts}; %For Joe!
record_init(Ts, Type, Line) ->
    syntax_error({bad,record}, Ts, Line).

record_inits([{'}',Lr}|Ts], Type, Line) ->
    {Lr, [], Ts};
record_inits([{',',Lc}|Ts0], Type, Line) ->
    {Lf,I,Ts1} = record_init(Ts0, Type, Lc),
    {Lr,Is,Ts} = record_inits(Ts1, Type, Lf),
    {Lr,[I|Is],Ts};
record_inits(Ts, Type, Line) ->
    syntax_error({terminator,'}'}, Ts, Line).

%% function(Tokens) ->
%%	{LastLine,Function,RestTokens}

function(Ts0, Line) ->
    {L1,C,Ts1} = function_clause(Ts0, Line),
    {L2,Cs,Ts} = function_clauses(Ts1, L1),
    {L2,build_function([C|Cs]),Ts}.

function_clauses([{';',L0}|Ts0], Line) ->
    {L1,C,Ts1} = function_clause(Ts0, L0),
    {L2,Cs,Ts2} = function_clauses(Ts1, L1),
    {L2,[C|Cs],Ts2};
function_clauses(Ts, Line) ->
    {Line,[],Ts}.

function_clause(Ts0, Line) ->
    {L1,H,Ts1} = clause_head(Ts0, Line),
    {L2,G,Ts2} = clause_guard(Ts1, L1),
    {L3,B,Ts} = clause_body(Ts2, L2),
    {L3,{clause,H,G,B},Ts}.   % build_function/1 adds Line to clause

clause_head([{atom,Line,Name},{'(',_},{')',_}|Ts], _) ->
    {Line,{call,Line,[],Name,[]},Ts};
clause_head([{atom,Line,Name},{'(',_}|Ts0], _) ->
    {L1,Arg,Ts1} = pattern(Ts0, Line),
    {L2,Rest,Ts} = pattern_args(Ts1, ')', L1),
    {L2,{call,Line,[],Name,[Arg|Rest]},Ts};
clause_head(Ts, Line) ->
    syntax_error({illegal,"function head"}, Ts, Line).

clause_guard([{'when',Line}|Ts], _) ->	%A guard follows
    guard(Ts, Line);
clause_guard(Ts, Line) ->
    {Line,[],Ts}.

%% {LastLine,BodyList,RestTokens} = clause_body(Tokens,Line)
%%  Pass the problem of detecting/handling empty bodies to the caller,
%%  may give funny error message.

clause_body([{'->',Line}|Ts], _) ->
    exprs(Ts, Line);
clause_body(Ts, Line) ->
     syntax_error({expect,body}, Ts, Line).

%% build_function(Clauses) ->
%%	{function,Name,Arity,Clauses}

build_function(Cs0) ->
    {Name,Arity,Line} = function_name(Cs0),
    {function,Line,Name,Arity,build_clauses(Cs0, Name, Arity)}.

function_name([{clause,{call,Line,_,Name,Args},_,_}|_]) ->
    {Name,length(Args),Line}.

build_clauses([{clause,{call,Line,_,Name,Args},G,B}|Cs], Name, Arity)
  when Arity == length(Args) ->
    [{clause,Line,Args,G,B}|build_clauses(Cs, Name, Arity)];
build_clauses([{clause,{call,Line,_,N1,Args},G,B}|Cs], Name, Arity) ->
    syntax_error(head, [], Line);
build_clauses([], _, _) ->
    [].

%% guard(Tokens,Line) ->
%%	{LastLine,Guard,RestTokens}
%%  A guard is either *just* the atom 'true' or a sequence of guard tests.
%%  We KNOW that '->' always comes after a guard.

guard([{atom,Line,true},{'->',L1}|Ts], _) ->
    {Line,[],[{'->',L1}|Ts]};
guard(Ts0, Line) ->
    {L1,G,Ts1} = guard_test(Ts0, Line),
    {L2,Rest,Ts} = guard_sequence(Ts1, L1),
    {L2,[G|Rest],Ts}.

guard_sequence([{',',Line}|Ts0], _) ->
    {L1,G,Ts1} = guard_test(Ts0, Line),
    {L2,Rest,Ts} = guard_sequence(Ts1, L1),
    {L2,[G|Rest],Ts};
guard_sequence(Ts, Line) ->
    {Line,[],Ts}.

%% guard_test(Tokens, Line) ->
%%	{LastLine,GuardTest,RestTokens}
%%  We do full checking of guard tests here. A guard test is either a
%%  builtin test function, or, a comparison operator where each side only
%%  consists of constants, pattern constructors, arithmetic expressions or
%%  built-in functions allowed in guards.

guard_test([{atom,Line,Name},{'(',LI1},{')',LI2}|Ts0], L0) ->
    case erlang_guard_test(Name, 0) of
	true ->
	    {Line,{test,Line,Name,[]},Ts0};
	false ->
	    {L1,C,Ts} = expr([{atom,Line,Name},{'(',LI1},{')',LI2}|Ts0],
			     guard, L0),
	    texprtl(Ts, C, L1)
    end;
guard_test([{atom,Line,Name},{'(',Li}|Ts0], L0) ->
    {L1,Arg,Ts1} = expr(Ts0, guard, Line),
    {L2,Rest,Ts} = args(Ts1, ')', guard, L1),
    case erlang_guard_test(Name, length(Rest)+1) of
	true ->
	    {L2,{test,Line,Name,[Arg|Rest]},Ts};
	false ->
	    {L3,C,Tsg} = expr([{atom,Line,Name},{'(',Li}|Ts0], guard, L0),
	    texprtl(Tsg, C, L3)
    end;
guard_test(Ts0, Line) ->
    {L1,C,Ts} = expr(Ts0, guard, Line),
    texprtl(Ts, C, L1).

texprtl([{Op,L0}|Ts0], Lhs, Line) ->
    case comp_op(Op) of
	true ->
	    {L,Rhs,Ts} = expr(Ts0, guard, L0),
	    {L,{comp,L0,Op,Lhs,Rhs},Ts};
	false ->
	    syntax_error(comp_op, [{Op,L0}|Ts0], Line)
    end;
texprtl(Ts, Lhs, Line) ->
     syntax_error(comp_op, Ts, Line).

%% Number coercing comparison operators
comp_op('==') -> true;
comp_op('/=') -> true;
comp_op('=<') -> true;
comp_op('<') -> true;
comp_op('>=') -> true;
comp_op('>') -> true;
%% Non-coercing comparison operators
comp_op('=:=') -> true;
comp_op('=/=') -> true;
comp_op(_) -> false.

exprs([T|Ts]) ->
    exprs([T|Ts], token_line(T));
exprs([]) ->
    syntax_error({missing,exprs}, [], 0).

exprs(Ts0, Line) ->
    {L1,E,Ts1} = expr(Ts0, expr, Line),
    {L2,Es,Ts} = expr_sequence(Ts1, L1),
    {L2,[E|Es],Ts}.

expr_sequence([{',',Line}|Ts0], _) ->
    {L1,C,Ts1} = expr(Ts0, expr, Line),
    {L2,Cs,Ts} = expr_sequence(Ts1, L1),
    {L2,[C|Cs],Ts};
expr_sequence(Ts, Line) ->
    {Line,[],Ts}.

%% {LastLine,Expr,RestTokens} = expr(Tokens, ExprType, Line)

expr(Ts, Type, Line) ->
    expr(Ts, 0, Type, Line).

%% {LastLine,Expr,RestTokens} = expr(Tokens, Precedence, ExprType, Line)

expr([{atom,Line,A}|Ts], Prec, Type, _) ->
    after_atom(Ts, A, Prec, Type, Line);
expr([{integer,Line,N}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {integer,Line,N}, Prec, Type, Line);
expr([{float,Line,N}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {float,Line,N}, Prec, Type, Line);
expr([{char,Line,C}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {char,Line,C}, Prec, Type, Line);
expr([{var,Line,V}|Ts], Prec, Type, _) ->
    after_var(Ts, V, Prec, Type, Line);
expr([{string,L1,S1},{string,L2,S2}|Ts], Prec, Type, Line) ->
    expr([{string,L1,append(S1, S2)}|Ts], Prec, Type, Line);
expr([{string,Line,S}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {string,Line,S}, Prec, Type, Line);
expr([{'[',Line},{']',_}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {nil, Line}, Prec, Type, Line);
expr([{'[',Line}|Ts0], Prec, Type, _) ->
    {L1,Head,Ts1} = expr(Ts0, Type, Line),
    {L2,Tail,Ts} = tail(Ts1, Type, L1),
    exprtl_max(Ts, {cons,Line,Head,Tail}, Prec, Type, L2);
expr([{'{',Line},{'}',Line1}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {tuple,Line,[]}, Prec, Type, Line1);
expr([{'{',Line}|Ts0], Prec, Type, _) ->
    {L1,Arg,Ts1} = expr(Ts0, Type, Line),
    {L2,Rest,Ts} = args(Ts1, '}', Type, L1),
    exprtl_max(Ts, {tuple,Line,[Arg|Rest]}, Prec, Type, L2);
expr([{'(',Line}|Ts0], Prec, Type, _) ->
    {L1,Expr,Ts1} = expr(Ts0, Type, Line),
    {L2,Ts} = expect(Ts1, ')', L1),
    exprtl_max(Ts, Expr, Prec, Type, L2);
%% Now come all the reserved words. All these expressions have the lowest
%% priority and must be surrounded by parentheses within expressions.
expr([{'begin',Line}|Ts0], Prec, expr, _) ->
    {L1,Block,Ts} = parse_begin(Ts0, Line),
    exprtl_max(Ts, Block, Prec, expr, L1);
expr([{'if',Line}|Ts0], Prec, expr, _) ->
    {L1,If,Ts} = parse_if(Ts0, Line),
    exprtl_max(Ts, If, Prec, expr, L1);
expr([{'case',Line}|Ts0], Prec, expr, _) ->
    {L1,Case,Ts} = parse_case(Ts0, Line),
    exprtl_max(Ts, Case, Prec, expr, L1);
expr([{'catch',Line}|Ts0], Prec, expr, _) ->
    {L1,Expr,Ts} = expr(Ts0, expr, Line),
    exprtl_max(Ts, {'catch',Line,Expr}, Prec, expr, L1);
expr([{'receive',Line}|Ts0], Prec, expr, _) ->
    {L1,Receive,Ts} = parse_receive(Ts0, Line),
    exprtl_max(Ts, Receive, Prec, expr, L1);
%% Now come all the operators. Special case +/- number.
expr([{'+',_},{integer,Line,Number}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {integer,Line,Number}, Prec, Type, Line);
expr([{'+',_},{float,Line,Number}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {float,Line,Number}, Prec, Type, Line);
expr([{'-',_},{integer,Line,Number}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {integer,Line,-Number}, Prec, Type, Line);
expr([{'-',_},{float,Line,Number}|Ts], Prec, Type, _) ->
    exprtl_max(Ts, {float,Line,-Number}, Prec, Type, Line);
expr([{'#',Lh},{atom,Ln,Name},{'{',Ll},{'}',Lr}|Ts],
     Prec, Type, Line) when Type /= guard ->
    exprtl_max(Ts, {record,Lh,Name,[]}, Prec, Type, Lr);
expr([{'#',Lh},{atom,Ln,Name},{'{',Ll}|Ts0],
     Prec, Type, Line) when Type /= guard ->
    {Lf,F,Ts1} = record_field(Ts0, Type, Ll),
    {Lr,Fs,Ts} = record_body(Ts1, Type, Lf),
    exprtl_max(Ts, {record,Lh,Name,[F|Fs]}, Prec, Type, Lr);
expr([{'#',Lh}|Ts], Prec, Type, Line) ->
    syntax_error({bad,record}, Ts, Lh);
expr([{Op,L}|Ts], Prec, Type, Line) when Op /= dot ->
    case preop_prec(Op) of
	{O,A} ->
	    after_prefix_op(Op, O, A, Ts, Prec, Type, L);
	false ->
	    syntax_error({illegal,punctuation}, [{Op,L}|Ts], Line)
    end;
expr([{dot,L}|Ts], _, _, Line) ->
    syntax_error('end', [], L).

%% after_atom(Tokens, Atom, Precedence, ExprType, AtomLine) ->
%%	{LastLine,Expression,RestTokens}
%%  This is called when an atom has been detected.

after_atom([{'(',Llp}|Ts0], Atom, Prec, Type, Line) ->
    {L1,Expr,Ts} = explicit_call(Ts0, [], Atom, Line, Type, Llp),
    exprtl_max(Ts, Expr, Prec, Type, L1);
after_atom([{':',_},{atom,Line,Name},{'(',Llp}|Ts0], Atom, Prec, expr, _) ->
    {L2,Expr,Ts} = explicit_call(Ts0, Atom, Name, Line, expr, Llp),
    exprtl_max(Ts, Expr, Prec, expr, L2);
after_atom([{':',L}|Ts], _, _, _, Line) ->
    syntax_error({illegal,"remote call"}, Ts, Line);
after_atom(Ts, Atom, Prec, Type, Line) ->
    exprtl_max(Ts, {atom,Line,Atom}, Prec, Type, Line).

%% after_var(Tokens, Var, Precedence, ExprType, VariableLine) ->
%%	{LastLine,Expression,RestTokens}
%%  This is called after a variable has been detected.

after_var([{'#',Lh},{atom,Ln,Name},{'.',Ld},{atom,Lf,Field}|Ts],
	  Var, Prec, Type, Line) ->
    exprtl_max(Ts, {record_field,Line,{var,Line,Var},Name,Field},
	       Prec, Type, Lf);
after_var([{'#',Lh},{atom,Ln,Name},{'{',Ll},{'}',Lr}|Ts],
	  Var, Prec, expr, Line) ->
    exprtl_max(Ts, {record,Line,{var,Line,Var},Name,[]},
	       Prec, expr, Lr);
after_var([{'#',Lh},{atom,Ln,Name},{'{',Ll}|Ts0], Var, Prec, expr, Line) ->
    {Lf,F,Ts1} = record_field(Ts0, expr, Ll),
    {Lr,Fs,Ts2} = record_body(Ts1, expr, Lf),
    exprtl_max(Ts2, {record,Line,{var,Line,Var},Name,[F|Fs]},
	       Prec, expr, Lr);
after_var([{'#',Lh}|Ts], Var, Prec, Type, Line) ->
    syntax_error({bad,record}, Ts, Lh);
after_var(Ts, Var, Prec, Type, Line) ->
    exprtl_max(Ts, {var,Line,Var}, Prec, Type, Line).

record_field([{atom,Lf,Field},{'=',Le}|Ts0], Type, Line) ->
    {Lv,Val,Ts1} = expr(Ts0, Type, Le),
    {Lv,{{atom,Lf,Field},Val},Ts1};
record_field(Ts, Type, Line) ->
    syntax_error({bad,record}, Ts, Line).

record_body([{'}',Lr}|Ts], Type, Line) ->
    {Lr, [], Ts};
record_body([{',',Lc}|Ts0], Type, Line) ->
    {Lf,F,Ts1} = record_field(Ts0, Type, Lc),
    {Lr,Fs,Ts} = record_body(Ts1, Type, Lf),
    {Lr,[F|Fs],Ts};
record_body(Ts, Type, Line) ->
    syntax_error({terminator,'}'}, Ts, Line).

%% explicit_call(Tokens, Module, FuncName, NameLine, ExprType, LastLine)
%% explicit_call(Tokens, Module, FuncName, NameLine, ArgList, ExprType, LastLine)
%%	{LastLine,Call,RestTokens}

explicit_call([{')',Line}|Ts], Mod, Name, Ln, Type, _) ->
    explicit_call(Ts, Mod, Name, Ln, [], Type, Line);
explicit_call(Ts0, Mod, Name, Ln, Type, Line) ->
    {L1,Arg,Ts1} = expr(Ts0, Type, Line),
    {L2,Rest,Ts} = args(Ts1, ')', Type, L1),
    explicit_call(Ts, Mod, Name, Ln, [Arg|Rest], Type, L2).

explicit_call(Ts, [], Name, Line, Args, guard, LastLine) ->
    case erlang_guard_bif(Name, length(Args)) of
	true ->
	    {LastLine,{bif,Line,Name,Args},Ts};
	false ->
	    syntax_error(guard_call, Ts, Line)
    end;
explicit_call(Ts, [], Name, Line, Args, pattern, LastLine) ->
    syntax_error({illegal,"call in pattern"}, Ts, Line);
explicit_call(Ts, [], Name, Line, Args, _, LastLine) ->
    case erlang_bif(Name, length(Args)) of
	true ->
	    {LastLine,{bif,Line,Name,Args},Ts};
	false ->
	    {LastLine,{call,Line,[],Name,Args},Ts}
    end;
explicit_call(Ts, Mod, Name, Line, Args, _, LastLine) ->
    {LastLine,{call,Line,Mod,Name,Args},Ts}.

%% args(Tokens, Terminator, Type, Line) ->
%%	{LastLine,ArgList,RestTokens}

args([{',',Line}|Ts0], Termin, Type, _) ->
    {L1,Arg,Ts1} = expr(Ts0, Type, Line),
    {L2,Rest,Ts} = args(Ts1, Termin, Type, L1),
    {L2,[Arg|Rest],Ts};
args([{Termin,Line}|Ts], Termin, _, _) ->
    {Line,[],Ts};
args(Ts, Termin, _, Line) ->
    syntax_error({terminator,Termin}, Ts, Line).

%% tail(Tokens, Type, Line) ->
%%	{LastLine,ArgList,RestTokens}

tail([{',',Line}|Ts0], Type, _) ->
    {L1,Head,Ts1} = expr(Ts0, Type, Line),
    {L2,Tail,Ts} = tail(Ts1, Type, L1),
    {L2,{cons,Line,Head,Tail},Ts};
tail([{'|',Line}|Ts0], Type, _) ->
    {L1,Tail,Ts1} = expr(Ts0, Type, Line),
    {L2,Ts} = expect(Ts1, ']', L1),
    {L2,Tail,Ts};
tail([{']',Line}|Ts], _, _) ->
    {Line,{nil,Line},Ts};
tail(Ts, _, Line) ->
    syntax_error(list, Ts, Line).

parse_begin(Ts0, Line) ->
    {L1,Es,Ts1} = exprs(Ts0, Line),
    {L2,Ts} = expect(Ts1, 'end', L1),
    {L2,{block,Line,Es},Ts}.

parse_if(Ts0, Line) ->
    {L1,C,Ts1} = if_clause(Ts0, Line),
    {L2,Cs,Ts2} = if_clauses(Ts1, L1),
    {L3,Ts} = expect(Ts2, 'end', L2),
    {L3,{'if',Line,[C|Cs]},Ts}.

if_clauses([{';',Line}|Ts0], _) ->
    {L1,C,Ts1} = if_clause(Ts0, Line),
    {L2,Cs,Ts} = if_clauses(Ts1, L1),
    {L2,[C|Cs],Ts};
if_clauses(Ts, Line) ->
    {Line,[],Ts}.

if_clause(Ts0, Line) ->
    {L1,G,Ts1} = guard(Ts0, Line),
    {L2,B,Ts} = clause_body(Ts1, L1),
    {L2,{clause,Line,G,B},Ts}.

parse_case(Ts0, Line) ->
    {L1,Expr,Ts1} = expr(Ts0, expr, Line),
    {L2,Ts2} = expect(Ts1, 'of', L1),
    {L3,C,Ts3} = cr_clause(Ts2, L2),
    {L4,Cs,Ts4} = cr_clauses(Ts3, L3),
    {L5,Ts} = expect(Ts4, 'end', L4),
    {L5,{'case',Line,Expr,[C|Cs]},Ts}.

cr_clauses([{';',Line}|Ts0], _) ->
    {L1,C,Ts1} = cr_clause(Ts0, Line),
    {L2,Cs,Ts} = cr_clauses(Ts1, L1),
    {L2,[C|Cs],Ts};
cr_clauses(Ts, Line) ->
    {Line,[],Ts}.

cr_clause(Ts0, Line) ->
    {L1,Match,Ts1} = pattern(Ts0, Line),
    {L2,G,Ts2} = clause_guard(Ts1, L1),
    {L3,B,Ts} = clause_body(Ts2, L2),
    {L3,{clause,Line,[Match],G,B},Ts}.

parse_receive([{'after',Line}|Ts0], L0) ->	%No messages
    {L1,Receive,Ts1} = receive_timeout([{'after',Line}|Ts0], [], Line, L0),
    {L2,Ts} = expect(Ts1, 'end', L1),
    {L2,Receive,Ts};
parse_receive(Ts0, Line) ->
    {L1,C,Ts1} = cr_clause(Ts0, Line),
    {L2,Cs,Ts2}= cr_clauses(Ts1, L1),
    {L3,Receive,Ts3} = receive_timeout(Ts2, [C|Cs], L2, Line),
    {L4,Ts} = expect(Ts3, 'end', L3),
    {L4,Receive,Ts}.

receive_timeout([{'after',Line}|Ts0], Cs, _, L0) -> %Have a timeout
    {L1,TimeOut,Ts1} = expr(Ts0, expr, Line),
    {L2,ToBody,Ts} = clause_body(Ts1, L1),
    {L2,{'receive',L0,Cs,TimeOut,ToBody},Ts};
receive_timeout(Ts, Cs, Last, L0) ->		%No timeout clause
    {Last,{'receive',L0,Cs},Ts}.

%%
%% The procedures which handle operators, priorities and the like.
%%

%% after_prefix_op(Op, OpPrec, OpRight, Tokens, Precedence, ExprType, Line) ->
%%	{LastLine,Expression,RestTokens}

after_prefix_op(Op, O, A, Ts0, Prec, Type, Line) when O > Prec ->
    {L1,Arg,Ts} = expr(Ts0, A, Type, Line),
    exprtl(Ts, O, {arith,Line,Op,Arg}, Prec, Type, L1);
after_prefix_op(Op, O, A, Ts, _, _, Line) ->
    syntax_error({prefix_op,Op}, Ts, Line).


%% exprtl_max(Tokens, Operand, Precedence, ExprType, Line) ->
%%	{LastLine,Expr,RestTokens}

exprtl_max([{Op,Line}|Ts], Lhs, Prec, Type, L0) ->
    case inop_prec(Op) of
	{L,O,R} ->
	    exprtl([{op,Op,L,O,R}|Ts], max_prec(), Lhs, Prec, Type, Line);
	false ->
	    {L0,Lhs,[{Op,Line}|Ts]}
    end;
exprtl_max([{Op,Line,S}|Ts], Lhs, Prec, Type, L0) ->
    case inop_prec(Op) of
	{L,O,R} ->
	    exprtl([{op,Op,L,O,R}|Ts], max_prec(), Lhs, Prec, Type, Line);
	false ->
	    {L0,Lhs,[{Op,Line,S}|Ts]}
    end;
exprtl_max([], Expr, _, _, Line) ->
    {Line,Expr,[]}.

%% exprtl(Tokens, Current, Lhs, Precedence, ExprType, Line) ->
%%	{LastLine,Expr,RestTokens}

exprtl([{op,Op,L,O,R}|Ts0], C, Lhs, Prec, Type, Line) when Prec < O, C > L ->
    {L1,Expr,Ts} = exprtl1(Ts0, Op, Lhs, R, Type, Line),
    exprtl(Ts, O, Expr, Prec, Type, L1);
exprtl([{Op,Line}|Ts], C, Lhs, Prec, Type, L0) ->
    case inop_prec(Op) of
	{L,O,R} ->
	    exprtl([{op,Op,L,O,R}|Ts], C, Lhs, Prec, Type, Line);
	false ->
	    {L0,Lhs,[{Op,Line}|Ts]}
    end;
exprtl([Op|Ts], C, Lhs, Prec, Type, L0) ->
    {L0, Lhs, [Op|Ts]};
exprtl([], _, Expr, _, _, Line) ->
    {Line,Expr,[]}.

%% exprtl1(Tokens, Operator, Lhs, RightPrec, ExprType, Line) ->
%%	{LastLine,Expr,RestTokens}
%%  Read the rhs of an infix expression, then build the correct syntatic
%%  object from the infix operator expression. Check the lhs of a match
%%  before reading rhs.

exprtl1(Ts, '=', Lhs, R, guard, Line) ->
    syntax_error(guard_call, [{'=',Line}|Ts], Line);
exprtl1(Ts0, '=', Lhs, R, Type, Line) ->
    case is_pattern(Lhs) of
	true ->
	    {L1,Rhs,Ts} = expr(Ts0, R, Type, Line),
	    {L1,{match,Line,Lhs,Rhs},Ts};
	false ->
	    syntax_error({illegal,"lhs in match"}, [{'=',Line}|Ts0], Line)
    end;
exprtl1(Ts, '!', Lhs, R, guard, Line) ->
    syntax_error(guard_call, [{'!',Line}|Ts], Line);
exprtl1(Ts0, '!', Lhs, R, Type, Line) ->
    {L1,Rhs,Ts} = expr(Ts0, R, Type, Line),
    {L1,{send,Line,Lhs,Rhs},Ts};
exprtl1(Ts0, Op, Lhs, R, Type, Line) ->
    {L1,Rhs,Ts} = expr(Ts0, R, Type, Line),
    {L1,{arith,Line,Op,Lhs,Rhs},Ts}.

inop_prec('=') -> {4,4,3};
inop_prec('!') -> {4,4,3};
inop_prec('+') -> {5,6,6};
inop_prec('-') -> {5,6,6};
inop_prec('bor') -> {5,6,6};
inop_prec('bxor') -> {5,6,6};
inop_prec('bsl') -> {5,6,6};
inop_prec('bsr') -> {5,6,6};
inop_prec('*') -> {7,8,8};
inop_prec('/') -> {7,8,8};
inop_prec('div') -> {7,8,8};
inop_prec('rem') -> {7,8,8};
inop_prec('band') -> {7,8,8};
inop_prec('^') -> {12,12,11};
inop_prec(_) ->					%Not an operator.
    false.

preop_prec('-') -> {9,10};
preop_prec('+') -> {9,10};
preop_prec('bnot') -> {9,10};
preop_prec(_) ->				%Not an operator.
    false.

max_prec() -> 20.

%% pattern(Tokens,StartLine) ->
%%	{LastLine,Pattern,RestTokens}

pattern([{atom,Line,Atom}|Toks], _) ->
    {Line,{atom,Line,Atom},Toks};
pattern([{integer,Line,Number}|Toks], _) ->
    {Line,{integer,Line,Number},Toks};
pattern([{float,Line,Number}|Toks], _) ->
    {Line,{float,Line,Number},Toks};
pattern([{char,Line,Char}|Toks], _) ->
    {Line,{char,Line,Char},Toks};
pattern([{var,Line,Var}|Toks], _) ->
    {Line, {var,Line,Var},Toks};
pattern([{string,L1,S1},{string,L2,S2}|Toks], Line) ->
    pattern([{string,L1,append(S1, S2)}|Toks], Line);
pattern([{string,Line,S}|Toks], _) ->
    {Line,{string,Line,S},Toks};
pattern([{'[',Line},{']',Line1}|Toks], _) ->
    {Line1,{nil,Line},Toks};
pattern([{'[',Line}|Toks0], _) ->
    {L1,Head,Toks1} = pattern(Toks0, Line),
    {L2,Tail,Toks} = pattern_tail(Toks1, L1),
    {L2,{cons,Line,Head,Tail},Toks};
pattern([{'{',Line},{'}',Line1}|Toks], _) ->
    {Line1,{tuple,Line,[]},Toks};
pattern([{'{',Line}|Toks0], _) ->
    {L1,Arg1,Toks1} = pattern(Toks0, Line),
    {L2,Rest,Toks} = pattern_args(Toks1, '}', L1),
    {L2,{tuple,Line,[Arg1|Rest]},Toks};
pattern([{'(',Line}|Toks0], _) ->
    {L1,Pat,Toks1} = pattern(Toks0, Line),
    {L2,Toks} = expect(Toks1, ')', L1),
    {L2,Pat,Toks};
%% Hack to allow negative (and positive) numbers as patterns.
pattern([{'+',_},{integer,Line,N}|Toks], _) ->
    {Line,{integer,Line,N},Toks};
pattern([{'+',_},{float,Line,F}|Toks], _) ->
    {Line,{float,Line,F},Toks};
pattern([{'-',_},{integer,Line,N}|Toks], _) ->
    {Line,{integer,Line,-N},Toks};
pattern([{'-',_},{float,Line,F}|Toks], _) ->
    {Line, {float,Line,-F},Toks};
pattern([{'#',Lh},{atom,Ln,Name},{'{',Ll},{'}',Lr}|Ts], _) ->
    {Lr,{record,Lh,Name,[]},Ts};
pattern([{'#',Lh},{atom,Ln,Name},{'{',Ll}|Ts0], _) ->
    {Lf,F,Ts1} = record_field(Ts0, pattern, Ll),
    {Lr,Fs,Ts} = record_body(Ts1, pattern, Lf),
    {Lr,{record,Lh,Name,[F|Fs]},Ts};
pattern(Toks, Line) ->
    syntax_error({illegal,punctuation}, Toks, Line).

pattern_args([{',',Line}|Toks0], Termin, _) ->
    {L1,Arg,Toks1} = pattern(Toks0, Line),
    {L2,Rest,Toks} = pattern_args(Toks1, Termin, L1),
    {L2,[Arg|Rest],Toks};
pattern_args([{Termin,Line}|Toks], Termin, _) ->
    {Line,[],Toks};
pattern_args(Toks, Termin, Line) ->
    syntax_error({terminator,Termin}, Toks, Line).

pattern_tail([{',',Line}|Toks0], _) ->
    {L1,Head,Toks1} = pattern(Toks0, Line),
    {L2,Tail,Toks} = pattern_tail(Toks1, L1),
    {L2,{cons,Line,Head,Tail},Toks};
pattern_tail([{'|',Line}|Toks0], _) ->
    {L1,Pat,Toks1} = pattern(Toks0, Line),
    {L2,Toks} = expect(Toks1, ']', L1),
    {L2,Pat,Toks};
pattern_tail([{']',Line}|Toks], _) ->
    {Line,{nil,Line},Toks};
pattern_tail(Toks, Line) ->
    syntax_error(list, Toks, Line).

%% is_pattern(Pat)
%% is_patternlist(Pats)
%%  Test if an abstract form is a pattern or list of patterns.

is_pattern({atom,_,_}) -> true;
is_pattern({integer,_,_}) -> true;
is_pattern({float,_,_}) -> true;
is_pattern({char,_,_}) -> true;
is_pattern({var,_,_}) -> true;
is_pattern({string,_,_}) -> true;
is_pattern({nil,_}) -> true;
is_pattern({cons,_,Head,Tail}) ->
    case is_pattern(Head) of
	true -> is_pattern(Tail);
	false -> false
    end;
is_pattern({tuple,_,Args}) ->
    is_patternlist(Args);
is_pattern({record,L,Name,Fields}) ->
    is_fieldlist(Fields);
is_pattern(Other) ->
    false.

is_patternlist([H|T]) ->
    case is_pattern(H) of
	true -> is_patternlist(T);
	false -> false
    end;
is_patternlist([]) -> true.

is_fieldlist([{_,V}|Fs]) ->
    case is_pattern(V) of
	true -> is_fieldlist(Fs);
	false -> false
    end;
is_fieldlist([]) -> true.

%% term(Tokens) ->
%% term(Tokens, StartLine) ->
%%	{LastLine,Term,RestTokens}

term([T|Ts]) ->
    term([T|Ts], token_line(T));
term([]) ->
    syntax_error({missing,term}, [], 0).

term([{atom,Line,Atom}|Toks], _) ->
    {Line,{atom,Line,Atom},Toks};
term([{integer,Line,Number}|Toks], _) ->
    {Line,{integer,Line,Number},Toks};
term([{float,Line,Number}|Toks], _) ->
    {Line,{float,Line,Number},Toks};
term([{char,Line,Char}|Toks], _) ->
    {Line,{char,Line,Char},Toks};
term([{var,Line,Var}|Toks], _) ->
    {Line, {var,Line,Var},Toks};
term([{string,L1,S1},{string,L2,S2}|Toks], Line) ->
    term([{string,L1,append(S1, S2)}|Toks], Line);
term([{string,Line,S}|Toks], _) ->
    {Line,{string,Line,S},Toks};
term([{'[',Line},{']',Line1}|Toks], _) ->
    {Line1,{nil,Line},Toks};
term([{'[',Line}|Toks0], _) ->
    {L1,Head,Toks1} = term(Toks0, Line),
    {L2,Tail,Toks} = term_tail(Toks1, L1),
    {L2,{cons,Line,Head,Tail},Toks};
term([{'{',Line},{'}',Line1}|Toks], _) ->
    {Line1,{tuple,Line,[]},Toks};
term([{'{',Line}|Toks0], _) ->
    {L1,Arg1,Toks1} = term(Toks0, Line),
    {L2,Rest,Toks} = term_args(Toks1, '}', L1),
    {L2,{tuple,Line,[Arg1|Rest]},Toks};
term([{'(',Line}|Toks0], _) ->
    {L1,Term,Toks1} = term(Toks0, Line),
    {L2,Toks} = expect(Toks1, ')', L1),
    {L2,Term,Toks};
%% Hack to allow negative (and positive) numbers as terms.
term([{'+',_},{integer,Line,N}|Toks], _) ->
    {Line,{integer,Line,N},Toks};
term([{'+',_},{float,Line,F}|Toks], _) ->
    {Line,{float,Line,F},Toks};
term([{'-',_},{integer,Line,N}|Toks], _) ->
    {Line,{integer,Line,-N},Toks};
term([{'-',_},{float,Line,F}|Toks], _) ->
    {Line, {float,Line,-F},Toks};
term(Toks, Line) ->
    syntax_error({illegal,punctuation}, Toks, Line).

term_args([{',',Line}|Toks0], Termin, _) ->
    {L1,Arg,Toks1} = term(Toks0, Line),
    {L2,Rest,Toks} = term_args(Toks1, Termin, L1),
    {L2,[Arg|Rest],Toks};
term_args([{Termin,Line}|Toks], Termin, _) ->
    {Line,[],Toks};
term_args(Toks, Termin, Line) ->
    syntax_error({terminator,Termin}, Toks, Line).

term_tail([{',',Line}|Toks0], _) ->
    {L1,Head,Toks1} = term(Toks0, Line),
    {L2,Tail,Toks} = term_tail(Toks1, L1),
    {L2,{cons,Line,Head,Tail},Toks};
term_tail([{'|',Line}|Toks0], _) ->
    {L1,Term,Toks1} = term(Toks0, Line),
    {L2,Toks} = expect(Toks1, ']', L1),
    {L2,Term,Toks};
term_tail([{']',Line}|Toks], _) ->
    {Line,{nil,Line},Toks};
term_tail(Toks, Line) ->
    syntax_error(list, Toks, Line).

%% is_term(Term)
%% is_termlist(Terms)
%%  Test if an abstract form is a term or list of terms.

is_term({atom,_,_}) -> true;
is_term({integer,_,_}) -> true;
is_term({float,_,_}) -> true;
is_term({char,_,_}) -> true;
is_term({var,_,_}) -> true;
is_term({string,_,_}) -> true;
is_term({nil,_}) -> true;
is_term({cons,_,Head,Tail}) ->
    case is_term(Head) of
	true -> is_term(Tail);
	false -> false
    end;
is_term({tuple,_,Args}) ->
    is_termlist(Args);
is_term(Other) ->
    false.

is_termlist([H|T]) ->
    case is_term(H) of
	true -> is_termlist(T);
	false -> false
    end;
is_termlist([]) -> true.

%% normalise(AbsTerm)
%% abstract(Term)
%%  Convert between the abstract form of a term and a term.

normalise({atom,_,A}) -> A;
normalise({integer,_,N}) -> N;
normalise({float,_,F}) -> F;
normalise({char,_,C}) -> C;
normalise({var,_,V}) -> {'$var',V};
normalise({string,_,S}) -> S;
normalise({nil,_}) -> [];
normalise({cons,_,Head,Tail}) ->
    [normalise(Head)|normalise(Tail)];
normalise({tuple,_,Args}) ->
    list_to_tuple(normalise_list(Args)).

normalise_list([H|T]) ->
    [normalise(H)|normalise_list(T)];
normalise_list([]) ->
    [].

abstract(T) when atom(T) ->
    {atom,0,T};
abstract(T) when integer(T) ->
    {integer,0,T};
abstract(T) when float(T) ->
    {float,0,T};
abstract([]) ->
    {nil,0};
abstract([H|T]) ->
    {cons,0,abstract(H),abstract(T)};
abstract(Tuple) when tuple(Tuple) ->
    {tuple,0,abstract_list(tuple_to_list(Tuple))}.

abstract_list([H|T]) ->
    [abstract(H)|abstract_list(T)];
abstract_list([]) ->
    [].

%% tokens(AbsTerm)
%% tokens(AbsTerm, More)
%%  Generate a list of tokens representing the abstract term.

tokens(Abs) ->
    tokens(Abs, []).

tokens({atom,L,A}, More) -> [{atom,L,A}|More];
tokens({integer,L,N}, More) -> [{integer,L,N}|More];
tokens({float,L,F}, More) -> [{float,L,F}|More];
tokens({char,L,C}, More) -> [{char,L,C}|More];
tokens({var,L,V}, More) -> [{var,L,V}|More];
tokens({string,L,S}, More) -> [{string,L,S}|More];
tokens({nil,L}, More) -> [{'[',L},{']',L}|More];
tokens({cons,L,Head,Tail}, More) ->
    [{'[',L}|tokens(Head, tokens_tail(Tail, More))];
tokens({tuple,L,[]}, More) ->
    [{'{',L},{'}',L}|More];
tokens({tuple,L,[E|Es]}, More) ->
    [{'{',L}|tokens(E, tokens_tuple(Es, More))].

tokens_tail({cons,L,Head,Tail}, More) ->
    [{',',L}|tokens(Head, tokens_tail(Tail, More))];
tokens_tail({nil,L}, More) ->
    [{']',L}|More];
tokens_tail(Other, More) ->
    [{'|',0}|tokens(Other, [{']',0}|More])].

tokens_tuple([E|Es], More) ->
    [{',',0}|tokens(E, tokens_tuple(Es, More))];
tokens_tuple([], More) ->
    [{'}',0}|More].

%%
%% Erlang Builtin functions.
%%

%% Process functions
erlang_bif(spawn, 3) -> true;
erlang_bif(spawn_link, 3) -> true;
erlang_bif(link,1) -> true;
erlang_bif(unlink, 1) -> true;
erlang_bif(exit, 1) -> true;
erlang_bif(exit, 2) -> true;
erlang_bif(process_flag, 2) -> true;

%% Global process functions
erlang_bif(register, 2) -> true;
erlang_bif(unregister, 1) -> true;
erlang_bif(whereis, 1) -> true;
erlang_bif(registered, 0) -> true;

%% Meta call functions
erlang_bif(apply, 2) -> true;
erlang_bif(apply, 3) -> true;

%% Arithmetic functions (allowed in guards)
erlang_bif(abs, 1) -> true;
erlang_bif(float, 1) -> true;
erlang_bif(trunc, 1) -> true;
erlang_bif(round, 1) -> true;

%% H/w functions
erlang_bif(open_port, 2) -> true;

%% Functions for accessing process ids and references
erlang_bif(node, 1) -> true;

%% Data access and conversion functions
erlang_bif(length, 1) -> true;
erlang_bif(hd, 1) -> true;
erlang_bif(tl, 1) -> true;
erlang_bif(size, 1) -> true;
erlang_bif(element, 2) -> true;
erlang_bif(setelement, 3) -> true;
erlang_bif(atom_to_list, 1) -> true;
erlang_bif(list_to_atom, 1) -> true;
erlang_bif(integer_to_list, 1) -> true;
erlang_bif(list_to_integer, 1) -> true;
erlang_bif(float_to_list, 1) -> true;
erlang_bif(list_to_float, 1) -> true;
erlang_bif(tuple_to_list, 1) -> true;
erlang_bif(list_to_tuple, 1) -> true;
erlang_bif(pid_to_list, 1) -> true;
erlang_bif(list_to_pid, 1) -> true;
erlang_bif(binary_to_list,1) -> true;
erlang_bif(list_to_binary,1) -> true;
erlang_bif(binary_to_list,3) -> true;
erlang_bif(split_binary,2) -> true;
erlang_bif(concat_binary,1) -> true;
erlang_bif(binary_to_term, 1) -> true;
erlang_bif(term_to_binary, 1) -> true;

%% Process dictionary functions
erlang_bif(get, 0) -> true;
erlang_bif(get, 1) -> true;
erlang_bif(put, 2) -> true;
erlang_bif(erase, 0) -> true;
erlang_bif(erase, 1) -> true;
erlang_bif(get_keys, 1) -> true;

%% Miscellaneous functions
erlang_bif(self, 0) -> true;
erlang_bif(node, 0) -> true;
erlang_bif(make_ref, 0) -> true;
erlang_bif(time, 0) -> true;
erlang_bif(date, 0) -> true;
erlang_bif(statistics, 1) -> true;
erlang_bif(garbage_collect, 0) -> true;
erlang_bif(garbage_collect, 1) -> true;
erlang_bif(processes, 0) -> true;
erlang_bif(process_info, 1) -> true;
erlang_bif(process_info, 2) -> true;
erlang_bif(throw, 1) -> true;
erlang_bif(trace, 3) -> true;
erlang_bif(halt, 0) -> true;

%% Code handling functions
%% These are all "erlang:" bifs and should not be here. /Seb
%%erlang_bif(load_module, 2) -> true;
%%erlang_bif(purge_module, 1) -> true;
%%erlang_bif(delete_module, 1) -> true;
%%erlang_bif(check_process_code, 2) -> true;
%%erlang_bif(module_loaded, 1) -> true;
%%erlang_bif(pre_loaded,0) -> true;

%% Distrubution bif's
erlang_bif(alive,2) -> true;
erlang_bif(nodes,0) ->true;
erlang_bif(is_alive,0) ->true;
erlang_bif(disconnect_node,1) -> true;

erlang_bif(internal_bif,2) -> true;
erlang_bif(group_leader,0) -> true;
erlang_bif(group_leader,2) -> true;
 
%% To compile erlang.erl theese bif's must be commented out
erlang_bif(spawn_link,4) -> true; 
erlang_bif(spawn, 4) -> true;
erlang_bif(monitor_node, 2) -> true;

erlang_bif(_, _) -> false.

%%
%% Erlang builtin functions allowed in guards.
%%

erlang_guard_bif(abs, 1) -> true;
erlang_guard_bif(float, 1) -> true;
erlang_guard_bif(trunc, 1) -> true;
erlang_guard_bif(round, 1) -> true;
erlang_guard_bif(length, 1) -> true;
erlang_guard_bif(hd, 1) -> true;
erlang_guard_bif(tl, 1) -> true;
erlang_guard_bif(size, 1) -> true;
erlang_guard_bif(element, 2) -> true;
erlang_guard_bif(self, 0) -> true;
erlang_guard_bif(node,0) -> true;
erlang_guard_bif(node,1) -> true;
erlang_guard_bif(_, _) -> false.

builtins() -> 
    [{spawn, 3},
     {spawn_link, 3},
     {link,1},
     {unlink, 1},
     {exit, 1},
     {exit, 2},
     {process_flag, 2},
     {register, 2},
     {unregister, 1},
     {whereis, 1},
     {registered, 0},
     {apply, 2},
     {apply, 3},
     {abs, 1},
     {float, 1},
     {trunc, 1},
     {round, 1},
     {open_port, 2},
     {node, 1},
     {length, 1},
     {hd, 1},
     {tl, 1},
     {size, 1},
     {element, 2},
     {setelement, 3},
     {atom_to_list, 1},
     {list_to_atom, 1},
     {integer_to_list, 1},
     {list_to_integer, 1},
     {float_to_list, 1},
     {list_to_float, 1},
     {tuple_to_list, 1},
     {list_to_tuple, 1},
     {pid_to_list, 1},
     {list_to_pid, 1},
     {binary_to_list,1},
     {list_to_binary,1},
     {binary_to_list,3},
     {split_binary,2},
     {concat_binary,1},
     {binary_to_term, 1},
     {term_to_binary, 1},
     {get, 0},
     {get, 1},
     {put, 2},
     {erase, 0},
     {erase, 1},
     {get_keys, 1},
     {self, 0},
     {node, 0},
     {make_ref, 0},
     {time, 0},
     {date, 0},
     {statistics, 1},
     {garbage_collect, 0},
     {garbage_collect, 1},
     {processes, 0},
     {process_info, 1},
     {process_info, 2},
     {throw, 1},
     {trace, 3},
     {halt, 0},
%%     {load_module, 2},
%%     {purge_module, 1},
%%     {delete_module, 1},
%%     {check_process_code, 2},
%%     {module_loaded, 1},
%%     {pre_loaded,0},
     {alive,2},
     {nodes,0},
     {is_alive,0},
     {monitor_node, 2},
     {internal_bif,2},
     {group_leader,0},
     {group_leader,2},
     {spawn_link,4},
     {spawn, 4}].

%%
%% Erlang guard tests
%%

erlang_guard_test(integer, 1) -> true;
erlang_guard_test(float, 1) -> true;
erlang_guard_test(number, 1) -> true;
erlang_guard_test(atom, 1) -> true;
erlang_guard_test(constant, 1) -> true;
erlang_guard_test(list, 1) -> true;
erlang_guard_test(tuple, 1) -> true;
erlang_guard_test(pid, 1) -> true;
erlang_guard_test(reference, 1) -> true;
erlang_guard_test(port, 1) -> true;
erlang_guard_test(binary, 1) -> true;
erlang_guard_test(record, 2) -> true;
erlang_guard_test(_, _) -> false.
