%% Copyright (C) 1990, Ellemtel Telecommunications Systems Laboratories
%% File    : pp.erl
%% Author  : Robert Virding
%% Purpose : Prettyprint Erlang code.

%% The after hook has been extended to work at each level, attribute,
%% function, clause, and expression. Lists of things, like guards, still
%% have to be lists of things. The first adding of hooks was done
%% by Joe Armstrong.

%% Modifications:
%%   930326 -- Joe Armstrong
%%      Changed {number, N} wrappers to
%%      {integer, N} or {float, N}

-module(erl_pp).

-export([form/1,form/2,attribute/1,attribute/2,function/1,function/2,
	 guard/1,guard/2,seq/1,seq/2,expr/1,expr/2,expr/4]).

-import(io_lib, [write/1,format/2,nl/0,write_char/1,write_string/1]).
-import(erl_parse,  [inop_prec/1,preop_prec/1,max_prec/0]).

%% After each "thing" has been printed the cursor is left in place. It is
%% up to what comes after, or the caller to decide how to terminate the
%% previous "thing".

form(Thing) ->
     form(Thing, none).

form({function,_,Name,Arity,Clauses}, Hook) ->
    clauses(Name, Clauses, Hook);
form({attribute,Line,Name,Arg}, Hook) ->
    attribute({attribute,Line,Name,Arg}, Hook);
form({function,Line,Name,Arity,Clauses}, Hook) ->
    clauses(Name, Clauses, Hook).
%% Now do the hook.

attribute(Thing) ->
    attribute(Thing, none).

attribute({attribute,Line,Name,Arg}, _) ->
    attribute1(Name, Arg).

attribute1(export, Falist) ->
    format("-~w(~s]).\n", [export,falist(Falist, '[')]);
attribute1(import, {From,Falist}) ->
    format("-~w(~w,~s]).\n", [import,From,falist(Falist, '[')]);
attribute1(file, {Name,Line}) ->
    format("-~w(~p, ~w).\n", [file,Name,Line]);
attribute1(record, {Name,Is}) ->
    format("-~w(~w,~s}).\n", [record,Name,record_fields(Is, '{')]);
attribute1(Name, Arg) ->
    format("-~w(~w).\n", [Name,Arg]).

falist([{Name,Arity}|Falist], P) ->
    [unquoted(P),write(Name),$/,write(Arity)|falist(Falist,',')];
falist([], _) ->
    [].

record_fields([{F,Val}|Is], P) ->
    [unquoted(P),expr(F)," = ",expr(Val)|record_fields(Is, ',')];
record_fields([], P) -> [].

function(F) ->
    function(F, none).

function({function,Name,Arity,Clauses}, Hook) ->
    clauses(Name, Clauses, Hook);
function({function,Line,Name,Arity,Clauses}, Hook) ->
    clauses(Name, Clauses, Hook).

clauses(Name, [C], Hook) ->
    [clause(Name, C, Hook),$.,nl()];
clauses(Name, [C|Cs], Hook) ->
    [clause(Name, C, Hook),$;,nl()|clauses(Name, Cs, Hook)];
clauses(_, [], _) ->				%Empty functions can not happen
    nl().

clause(Name, {clause,Head,Guard,Body}, Hook) ->
    [expr({call,[],Name,Head}, 0, 0, Hook),
     guard(Guard, 0, Hook),
     seq(Body, 4, ' ->', Hook)];
clause(Name, {clause,Line,Head,Guard,Body}, Hook) ->
    [expr({call,Line,[],Name,Head}, 0, 0, Hook),
     guard(Guard, 0, Hook),
     seq(Body, 4, ' ->', Hook)].

%% guard(GuardSequence)
%% guard(GuardSequence, Hook)
%% guard(GuardSequence, Indentation, Punctuation, Hook)

guard(Gs) ->
    guard(Gs, none).

guard(Gs, Hook) ->
    guard(Gs, 0, '', Hook).

guard([E|Es], I, Hook) ->
    guard([E|Es], I, ' when ', Hook);
guard([], _, _) ->
    [].

guard([E|Es], I, P, Hook) ->
    [unquoted(P),expr(E, I, 0, Hook)|guard(Es, I, ', ', Hook)];
guard([], _, _, _) ->
    [].

%% seq(Sequence)
%% seq(Sequence, Hook)
%% seq(Sequence, Indentation, Punctuation, Hook)
%%  Prettyprint a sequence of expressions.

seq(Es) ->
    seq(Es, none).

seq(Es, Hook) ->
    seq(Es, -10000, '', Hook).	%A hack to prohibit line breaks

seq([E|Es], I, P, Hook) ->
    [unquoted(P),nl_indent(I),expr(E, I, 0, Hook)|seq(Es, I, ',', Hook)];
seq([], _, _, _) ->
    [].

%% expr(Expression)
%% expr(Expression, Hook)
%% expr(Expression , Indentation, Precedence, Hook)
%%  Prettyprint one expr. Seeing everything is a "expr" we have to handle
%%  operator precedences for arithmetic expressions as well.

expr(E) ->
    expr(E, -10000, 0, none).

expr(E, Hook) ->
    expr(E, -10000, 0, Hook).

expr({atom,_,A}, _, _, _) ->
    write(A);
expr({integer,_,N}, _, _, _) ->
    write(N);
expr({float,_,F}, _, _, _) ->
    write(F);
expr({char,_,C}, _, _, _) ->				%When this comes
    write_char(C);
expr({var,_,V}, _, _, _) when integer(V) ->
    format('_~w', [V]);
expr({var,_,V}, _, _, _) ->
    unquoted(V);
expr({string,_,S}, _, _, _) ->
    write_string(S);
expr({nil,_}, _, _, _) ->
    "[]";
expr({cons,_,H,T}, I, _, Hook) ->
    I1 = I + 1,
    [$[,expr(H, I1, 0, Hook)|tail(T, I1, Hook)];
expr({tuple,_,Elts}, I, _, Hook) ->
    arg_list(Elts, '{', '}', I, Hook);
expr({arith,_,Op,Larg,Rarg}, I, Prec, Hook) ->
    {L,P,R} = inop_prec(Op),
    [maybe_paren(P, Prec, '('),
     expr(Larg, I, L, Hook),
     operator(Op),
     expr(Rarg, I, R, Hook),
     maybe_paren(P, Prec, ')')];
expr({arith,_,Op,Arg}, I, Prec, Hook) ->
    {P,R} = preop_prec(Op),
    [maybe_paren(P, Prec, '('),
     format('~s ', [Op]),
     expr(Arg, I, R, Hook),
     maybe_paren(P, Prec, ')')];
expr({bif,_,Name,Args}, I, _, Hook) ->
    [write(Name)|arg_list(Args, '(', ')', I, Hook)];
expr({call,_,[],Name,Args}, I, _, Hook) ->
    [write(Name)|arg_list(Args, '(', ')', I, Hook)];
expr({call,_,Mod,Name,Args}, I, _, Hook) ->
    [format('~w:~w', [Mod,Name])|arg_list(Args, '(', ')', I, Hook)];
expr({match,_,Lhs,Rhs}, I, Prec, Hook) ->
    {L,P,R} = inop_prec('='),
    [maybe_paren(P, Prec, '('),
     expr(Lhs, I, L, Hook),
     operator('='),
     expr(Rhs, I, R, Hook),
     maybe_paren(P, Prec, ')')];
expr({block,_,Es}, I, _, Hook) ->
    [seq(Es, -10000, 'begin', Hook),$ ,unquoted('end')];
expr({'if',_,Opts}, I, _, Hook) ->
    [options(Opts, I+4, 'if', Hook),
     nl_indent(I),
     unquoted('end')];
expr({'case',_,Expr,Opts}, I, _, Hook) ->
    [unquoted('case'),
     $ ,
     expr(Expr, I+5, 0, Hook),
     options(Opts, I+4, ' of', Hook),
     nl_indent(I),
     unquoted('end')];
expr({'catch',_,Expr}, I, _, Hook) ->
    [unquoted('catch'),
     $ ,
     expr(Expr, I+6, max_prec(), Hook)];
expr({send,_,To,Msg}, I, Prec, Hook) ->
    {L,P,R} = inop_prec('!'),
    [maybe_paren(P, Prec, '('),
     expr(To, I, L, Hook),
     operator('!'),
     expr(Msg, I, R, Hook),
     maybe_paren(P, Prec, ')')];
expr({'receive',_,Opts}, I, _, Hook) ->
    I1 = I + 4,
    [options(Opts, I+4, 'receive', Hook),
     nl_indent(I),
     unquoted('end')];
expr({'receive',_,Opts,To,ToOpt}, I, _, Hook) ->
    I1 = I + 4,
    [if						%Must special case no options
	 Opts /= [] ->
	     options(Opts, I1, 'receive', Hook);
	 true ->
	     unquoted('receive')
     end,
     %% Now for the timeout bit.
     nl_indent(I),
     unquoted('after'),
     nl_indent(I1),
     expr(To, I1, 0, Hook),
     seq(ToOpt, I1+4, ' ->', Hook),
     nl_indent(I),
     unquoted('end')];
%% Special case for straight values.
expr({value,_,Val}, _, _,_) ->
    write(Val);
%% Tests and comparisons in guards.
expr({test,_,Test,Args}, I, _, Hook) ->
    [write(Test)|arg_list(Args, '(', ')', I, Hook)];
expr({comp,_,Op,Lhs,Rhs}, I, _, Hook) ->
    [expr(Lhs, I, 0, Hook),
     operator(Op),
     expr(Rhs, I, 0, Hook)];
expr({record, _, Name, Fs}, I, P, Hook) ->
    format("#~w~s}", [Name,record_fields(Fs, I, P, Hook, '{')]);
expr({record, _, Var, Name, Fs}, I, P, Hook) ->
    format("~s#~w~s}", [expr(Var),Name,record_fields(Fs, I, P, Hook, '{')]);
expr({record_field, _, Var, Name, F}, I, P, Hook) ->
    format("~s#~w.~w", [expr(Var),Name,F]);
%% Now do the hook.
expr(Expr, Indentation, Precedence, {Mod,Func}) ->
    apply(Mod, Func, [Expr,Indentation,Precedence,{Mod,Func}]);
expr(Expr, Indentation, Precedence, {Mod,Func,Eas}) ->
    apply(Mod, Func, [Expr,Indentation,Precedence,{Mod,Func,Eas}|Eas]);
expr(Other, Indentation, Precedence, none) ->
    format("INVALID-FORM:~w:",[Other]).

record_fields([{F,Val}|Is], I, P, Hook, Punct) ->
    [unquoted(Punct),expr(F, I, P, Hook)," = ",expr(Val)|
     record_fields(Is, I, P, Hook, ',')];
record_fields([], I, P, Hook, Punct) -> [].

tail({cons,H,T}, I, Hook) ->
    [$,,
     expr(H, I, 0, Hook),
     tail(T, I, Hook)];
tail(nil, _, _) ->
    "]";
tail({cons,_,H,T}, I, Hook) ->
    [$,,
     expr(H, I, 0, Hook),
     tail(T, I, Hook)];
tail({nil,_}, _, _) ->
    "]";
tail(Other, I, Hook) ->
    [$|,expr(Other, I, 0, Hook),$]].

arg_list([A|As], First, Last, I, Hook) ->
    I1 = I + 1,
    [unquoted(First),expr(A, I1, 0, Hook)|arg_list(As, Last, I1, Hook)];
arg_list([], First, Last, _, _) ->
    [unquoted(First)|unquoted(Last)].

arg_list([A|As], Last, I, Hook) ->
    [$,,expr(A, I, 0, Hook)|arg_list(As, Last, I, Hook)];
arg_list([], Last, _, _) ->
    unquoted(Last).

options([{clause,G,B}|Os], I, P, Hook) ->		%If clauses
    [unquoted(P),
     nl_indent(I),
     if
	 G /= [] ->
	     guard(G, I, '', Hook);			%A small hack
	 true ->
	     write(true)
     end,
     seq(B, I+4, ' ->', Hook)|
     options(Os, I, ';', Hook)];
options([{clause,[T],G,B}|Os], I, P, Hook) ->		%Case/receive clauses
    [unquoted(P),
     nl_indent(I),
     expr(T, I, 0, Hook),
     guard(G, I, Hook),
     seq(B, I+4, ' ->', Hook)|
     options(Os, I, ';', Hook)];
options([{clause,_,G,B}|Os], I, P, Hook) ->		%If clauses
    [unquoted(P),
     nl_indent(I),
     if
	 G /= [] ->
	     guard(G, I, '', Hook);			%A small hack
	 true ->
	     write(true)
     end,
     seq(B, I+4, ' ->', Hook)|
     options(Os, I, ';', Hook)];
options([{clause,_,[T],G,B}|Os], I, P, Hook) ->		%Case/receive clauses
    [unquoted(P),
     nl_indent(I),
     expr(T, I, 0, Hook),
     guard(G, I, Hook),
     seq(B, I+4, ' ->', Hook)|
     options(Os, I, ';', Hook)];
options([], _, _, _) ->
    [].

%%
%% Utilities
%%

unquoted(A) ->
    format('~s', [A]).

operator(O) ->
    format(' ~s ', [O]).

maybe_paren(P, Prec, Par) when P =< Prec ->
    unquoted(Par);
maybe_paren(_, _, _) ->
    [].

nl_indent(I) when I >= 0 ->
    [nl()|indent(I)];
nl_indent(I) ->
    [$ ].

indent(I) when I >= 8 ->
    [$\t|indent(I-8)];
indent(I) when I > 0 ->
    [$ |indent(I-1)];
indent(_) ->
    [].
