%% Copyright (C) 1991, Ellemtel Telecommunications Systems Laboratories
%% File    : error_logger.erl
%% Author  : Claes Wikstrom
%% Purpose : Receive and do something clever with all system errors

-module(error_logger).
-export([format/2,error_msg/1,error_msg/2,logfile/1,
	 error_logger/0,start/0, tty/1]).

error_msg(Format) ->
    error_msg(Format,[]).
error_msg(Format, Args) ->
    error_logger ! {error,error,self(),group_leader(),Format,Args}, ok.

%% Just print some info, that might go into the logfile
%% The fifference between this and error_msg/2 is that the node 
%% info is not inserted here

format(Format, Args) ->
    error_logger ! {error,format,self(),group_leader(),Format,Args}, ok.

%% Log all errors to File for all eternity
logfile(File) ->
    error_logger ! {self(), {logfile, File}},
    receive
	{error_logger, R} -> R
    end.

%% Possibly turn off all tty printouts, maybe we only want the errors
%% to go to a file
%% Flag = true | false

tty(Flag) ->
    error_logger ! {self(), {tty, Flag}},
    receive
	{error_logger, ok} -> ok
    end.

%% All systems error messages go throuh here
start() ->
    Id = spawn(error_logger,error_logger,[]),
    register(error_logger,Id).

sleep(I) -> 
    receive after I -> ok  
    end.

error_logger() ->
    process_flag(trap_exit,true),
    sleep(1000),   %% This is a real good sleep call, it prevents (sometimes)
                      %% ugly looking error msgs when booting :-)
    error_logger_loop().

error_logger_loop() ->
    receive
	Anything ->
	    process_it(is_alive(), whereis(user),Anything),
	    error_logger_loop()
    end.

%% Since init is group leader of some processes started with the -s 
%% Command line option, all their messages to group_leader() will
%% go to init. .. Best thing to do .. send em' to user.
%% However, this will include all other garbage that is sent to init
%% by mistake

process_it(How, User,{init,X}) when pid(User) ->
    user ! X;  %% errors from children to init
process_it(How, undefined,{error,_,From,Gleader,Format,Args}) ->
    display2(Format,Args);
process_it(How, User,{error,error,From,Gleader,Format,Args}) when pid(User) ->
    e_format({user, node(Gleader)}, add_node(How,Format),Args);
process_it(How, User,{error,format,From,Gleader,Format,Args}) when pid(User) ->
    e_format({user, node(Gleader)},Format,Args);
process_it(How, undefined,{emulator,Gleader,Chars}) ->
    display2(add_node(How, "~s"),[Chars]);
process_it(How, User,{emulator,noproc,Chars}) when pid(User) ->
    e_format(User,"~s",[Chars]);
process_it(How, User,{emulator,Gleader,Chars}) when pid(User) ->
    e_format({user, node(Gleader)}, "~s",[Chars]);

process_it(How, _, {From, {logfile, LfReq}}) ->
    case {get(logfile), LfReq } of
	{undefined , {open, F}} -> %% Std case 
	    case file:open(F, read_write) of
		{ok,Lf} ->
		    io:format(Lf, 
			      "Starting to log errors time = ~, date = ~w~n",
			      [time(), date()]),
		    put(logfilename, F),
		    put(logfile, Lf),
		    From ! {error_logger, ok};
		{error, open} ->
		    From ! {error_logger, {error, open}}
	    end;
	{Lf, {open, F}} ->  %% Can only have one logfile
	    From ! {error_logger, {error, allready_have_logfile}};
	{undefined , close} ->
	    From ! {error_logger, {error, no_log_file}};
	{Lf, close} ->
	    io:format(Lf,"Closing log, time = ~, date = ~w~n",
		      [time(), date()]),
	    file:close(Lf),
	    erase(logfile),
	    erase(logfilename),
	    From ! {error_logger, ok};
	{undefined, filename} ->
	    From ! {error_logger, {error, no_log_file}};
	{_, filename} ->
	    From ! {error_logger, get(logfilename)};
	{_, Other} ->
	    From ! {error_logger, {error, {badrequest, Other}}}
    end;


process_it(How,  _, {'EXIT', From, _}) ->
    case get(logfile) of
	undefined -> ok;
	From -> erase(logfile)
    end;

process_it(How, _, {From, {tty, false}}) ->
    put(notty, true), From ! {error_logger, ok};

process_it(How, _, {From, {tty, _}}) ->
    erase(notty), From ! {error_logger, ok};

process_it(How, _,X) ->
    display2('',X).


%% All formatted io go through here 
e_format(U,F,A) ->
    case {get(notty), get(logfile)}  of
	{true, undefined} -> 
	    ok;                       %% Ignore completely
	{true, File} -> 
	    io:format(File, F, A);    %% To file only
	{undefined, undefined} -> 
	    io:format(U,F, A);        %% Std case
	{undefined, File} -> 
	    io:format(File, F,A), 
	    io:format(U,F,A)          %% TTy & File
    end.

add_node(true,X) -> list_to_atom(add_node2(X));
add_node(false,X) -> X.

add_node2(X) when atom(X) -> add_node2(atom_to_list(X));
add_node2(X) -> lists:concat([X,'** at node ',node(),' **~n']).

%% Can't do io_lib:format   :-(

display2(F,A) ->
    erlang:display({nice(F), nice(A)}).

nice(X) when tuple(X) ->
    nice_tuple(X);

nice([]) -> [];
nice(L) when list(L) ->
    case is_string(L) of
	true ->
	    list_to_atom(L);
	false ->
	    [H|T] = L,
	    [nice(H) | nice(T)]
    end;
nice(X) -> X.

nice_tuple(X) ->
    nice_tuple(X,1,size(X) + 1).
nice_tuple(X, Pos, Pos) -> X;
nice_tuple(X, Pos, End) ->
    nice_tuple(setelement(Pos, X, nice(element(Pos, X))), Pos+1, End).

is_string([]) -> true;
is_string([H|T]) when H > 1, H <255 -> is_string(T);
is_string(_) -> false.

