%% (location transparent) interface to the Term store BIF's
%% Author klacke@erix.ericsson.se

%% ets == Erlang Term Store

-module(ets).
-export([lookup/2,
	 insert/2,
	 delete/1,
	 delete/2,
	 first/1,
	 next/2,
	 new/2,
	 all/0,
	 match/2,
	 filter/3,
	 info/1,
	 info/2,
	 match_delete/2,
	 tab2file/2,
	 tab2list/1,
	 file2tab/1]).

lookup({T,Ref},K) when node(Ref) == node() ->         
    erlang:db_get({T,Ref},K);
lookup({T,Ref},K)  ->         
    chk(rpc:block_call(node(Ref), ets, lookup, [{T,Ref}, K])).

insert({T, Ref},V) when node(Ref) == node() ->
    erlang:db_put({T,Ref},V);
insert({T, Ref},V)  ->
    chk(rpc:block_call(node(Ref), ets, insert, [{T, Ref},V])).

delete({T, Ref},K) when node(Ref) == node() ->
    erlang:db_erase({T, Ref},K);
delete({T, Ref},K)  ->
     chk(rpc:block_call(node(Ref), ets, delete, [{T, Ref},K])).

delete({T, Ref}) when node(Ref) == node() ->      
    erlang:db_erase({T, Ref});
delete({T, Ref})  ->      
    chk(rpc:block_call(node(Ref), ets, delete, [{T, Ref}])).

first({T, Ref}) when node(Ref) == node() -> 
    erlang:db_first({T, Ref});
first({T, Ref})  -> 
    chk(rpc:block_call(node(Ref), ets, first, [{T, Ref}])).

next({T, Ref},K) when node(Ref) == node() ->
    erlang:db_next_key({T, Ref},K);
next({T, Ref},K)  ->
    chk(rpc:block_call(node(Ref), ets, next, [{T, Ref}, K])).

new(Name, Type) ->     erlang:db_create(Name, Type).
all() ->               erlang:db_all_tables().

match({T, Ref},Pattern)   when node(Ref) == node() ->
    erlang:db_match({T, Ref},Pattern);
match({T, Ref},Pattern)   ->
    chk(rpc:block_call(node(Ref), ets, match, [{T, Ref},Pattern])).

info({T, Ref}) when node(Ref) == node() ->
    Tab = {T, Ref},
    case erlang:db_info(Tab, memory) of
	undefined -> undefined;
	Mem ->
	    {{memory, Mem}, {owner, info(Tab, owner)}, {name,T},
	     {size, info(Tab, size)}, {node, node(Ref)},
	     {type, info(Tab, type)}, {protection, info(Tab, protection)}}
    end;

info({T, Ref}) ->
    chk(rpc:block_call(node(Ref), ets, info, [{T, Ref}])).

info({T, Ref}, What) when node(Ref) == node() -> 
    case What of 
	node ->
	    node(Ref);
	name ->
	    T;
	_ ->
	    erlang:db_info({T, Ref}, What)
    end;
info({T,Ref}, What) ->
    chk(rpc:block_call(node(Ref), ets, info, [{T, Ref}, What])).

match_delete({Tn, Ref}, Pattern) when node(Ref) == node() ->
    erlang:db_match_erase({Tn, Ref},Pattern);

match_delete({T, Ref}, Pattern) ->
    chk(rpc:call(node(Ref), ets , match_delete, [{T, Ref}, Pattern])).


%% Produce a list of {Key,Value} tuples from a table

tab2list({T,Ref}) when node(Ref) == node() ->
    Tab = {T,Ref},
    whole_tab(Tab, erlang:db_match(Tab, '$1'), []);
tab2list({T,Ref}) ->
    chk(rpc:call(node(Ref), ets,tab2list, [{T,Ref}])).

whole_tab(Tab, [[X]|T], Ack) -> 
    whole_tab(Tab, T, [X|Ack]);
whole_tab(Tab, [], Ack) ->
    Ack.


filter({Tn, Ref}, F, A) when node(Ref) == node() ->
    Tab = {Tn,Ref},
    do_filter(Tab,erlang:db_first(Tab),F,A, []);
filter({Tn, Ref}, F, A) -> 
    chk(rpc:call(node(Ref), ets, filter, [{Tn, Ref}, F, A])).

do_filter(Tab, '$end_of_table', _,_, Ack) -> 
    Ack;
do_filter(Tab, Key, F, A, Ack) ->
    case apply(F, [erlang:db_get(Tab, Key) | A]) of
	false ->
	    do_filter(Tab, erlang:db_next_key(Tab, Key), F,A,Ack);
	true ->
	    Ack2 = lists:append(erlang:db_get(Tab, Key), Ack),
	    do_filter(Tab, erlang:db_next_key(Tab, Key), F,A,Ack2);
	{true, Value} ->
	    do_filter(Tab, erlang:db_next_key(Tab, Key), F,A,[Value | Ack])
    end.

    
%% Dump a table to a file
tab2file(Tab, File) ->
    file:write_file(File,term_to_binary({info(Tab), tab2list(Tab)})).

%% Restore a table from a file, given that the file was written with
%% the tab2file/2 function from above

file2tab(File) ->
    case file:read_file(File) of
	{ok, Bin} ->
	    case catch binary_to_term(Bin) of
		{'EXIT',_} -> 
		    {error, badfile};
		{I,S} ->
		    Tab = mk_tab(tuple_to_list(I)),
		    insert_all(Tab, S),
		    {ok,Tab}
	    end;
	_ ->
	    {error, nofile}
    end.

mk_tab(I) ->
    {value, {name, Name}} = lists:keysearch(name, 1, I),
    {value, {type, Type}} = lists:keysearch(type, 1, I),
    {value, {protection, P}} = lists:keysearch(protection, 1, I),
    new(Name, [Type, P]).


insert_all(Tab, [Val|T]) ->
    erlang:db_put(Tab,Val),
    insert_all(Tab,T);
insert_all(Tab,[]) -> Tab.

chk({badrpc,R}) -> exit(R);
chk(X) -> X.
