%% Copyright (C) 1990, Ellemtel Telecommunications Systems Laboratories
%% File    : file.erl
%% Author  : Robert Virding
%% Purpose : A simple file server.

%% Sebastian Strollo, Made the interface to the external/driver
%%                    filer *completely* :-) synchronous

%% 930220 Klacke changed the filer to have a binary port to make
%% life nicer for the code loader and added start_slave

%% Requests (from Erlang) have the following format:
%%
%%	Op(1)   Request
%%
%% and replies
%%
%%	Op(1) Status(1) Reply

-module(file).

-export([get_cwd/0,set_cwd/1,delete/1,rename/2,make_dir/1,del_dir/1,
	 list_dir/1,file_info/1,read_file/1,write_file/2]).
-export([consult/1,path_consult/2,eval/1,path_eval/2]).
-export([path_open/3,open/2,close/1,position/2,truncate/1]).
-export([start/0,start_slave/1,server/4,file/4]).

%%
%% File management operations.
%%

get_cwd() ->
    file_request(file_server, get_cwd),
    wait_file_reply(file_server).

set_cwd(Cwd) ->
    file_request(file_server, {set_cwd,Cwd}),
    wait_file_reply(file_server).

delete(Name) ->
    file_request(file_server, {delete,Name}),
    wait_file_reply(file_server).

rename(From, To) ->
    file_request(file_server, {rename,From,To}),
    wait_file_reply(file_server).

make_dir(Name) ->
    file_request(file_server, {make_dir,Name}),
    wait_file_reply(file_server).

del_dir(Name) ->
    file_request(file_server, {del_dir,Name}),
    wait_file_reply(file_server).

list_dir(Name) ->
    file_request(file_server, {list_dir,Name}),
    wait_file_reply(file_server).

file_info(Name) ->
    file_request(file_server, {file_info,Name}),
    wait_file_reply(file_server).

read_file(Name) when atom(Name) ->
    read_file(atom_to_list(Name));
read_file(Name) when list(Name) ->
    file_request(file_server, {read_file,Name}),
    wait_file_reply(file_server).

write_file(Name, Binary) when atom(Name) ->
    write_file(atom_to_list(Name),Binary);
write_file(Name, Binary) when list(Name) ->
    file_request(file_server, {write_file,Name,Binary}),
    wait_file_reply(file_server).

%%
%% File processing operations
%%
    
consult(File) ->
    case open(File, read) of
	{ok,Fd} ->
	    R = consult_stream(Fd),
	    close(Fd),
	    R;
	{error,What} ->
	    {error,open}
    end.
    
path_consult(Path, File) ->
    case path_open(Path, File, read) of
	{ok,Fd,Full} ->
	    case consult_stream(Fd) of
		{ok,List} ->
		    close(Fd),
		    {ok,List,Full};
		{error,E} ->
		    close(Fd),
		    {error,E}
	    end;
	{error,What} ->
	    {error,open}
    end.

eval(File) ->
    case open(File, read) of
	{ok,Fd} ->
	    R = eval_stream(Fd, erl_eval:new_bindings()),
	    close(Fd),
	    R;
	{error,What} ->
	    {error,open}
    end.

path_eval(Path, File) ->
    case path_open(Path, File, read) of
	{ok,Fd,Full} ->
	    case eval_stream(Fd, erl_eval:new_bindings()) of
		ok ->
		    close(Fd),
		    {ok,Full};
		{error,E} ->
		    close(Fd),
		    {error,E}
	    end;
	{error,What} ->
	    {error,open}
    end.

consult_stream(Fd) ->
    case catch consult_stream(io:read(Fd, ''), Fd) of
	{'EXIT',Reason} ->
	    {error,read};
	error ->
	    {error,read};
	List ->
	    {ok,List}
    end.

consult_stream({ok,Term}, Fd) ->
    [Term|consult_stream(io:read(Fd, ''), Fd)];
consult_stream({error,What}, Fd) ->
    throw(error);
consult_stream(eof, Fd) ->
    [].

eval_stream(Fd, Bs) ->
    case eval_stream(Fd, 0, Bs) of
	0 -> ok;
	N when N > 0 -> {error,eval}
    end.

eval_stream(Fd, E, Bs) ->
    eval_stream(io:parse_erl_seq(Fd, ''), Fd, E, Bs).

eval_stream({ok,Form,EndLine}, Fd, E, Bs0) ->
    case catch erl_eval:seq(Form, Bs0) of
	{value,V,Bs} ->
	    eval_stream(Fd, E, Bs);
	{'EXIT',Reason} ->
	    eval_stream(Fd, E+1, Bs0)
    end;
eval_stream({error,What,EndLine}, Fd, E, Bs) ->
    eval_stream(Fd, E+1, Bs);
eval_stream({eof,EndLine}, Fd, E, Bs) ->
    E.

%%
%% File access operations.
%%

%% path_open(Path, FileName, Mode)
%%	{ok,FileDescriptor,FullName}
%%	{error,Reason}
%%  Search Path for file FileName which can be opened with Mode. Ignore
%%  the path if FileName is an absolute file name. Try to recognise both
%%  UNIX and NT type absolute names.

path_open(Ps, Name, Mode) ->
    case catch flatten(file_name(Name)) of
	{'EXIT',_} -> {error,name};
	[$/|Cs] ->
	    path_open_1([$/|Cs], Mode);
	[Dev,$:,$/|Cs] when Dev >= $a, Dev =< $z ->
	    path_open_1([Dev,$:,$/|Cs], Mode);
	[Dev,$:,$/|Cs] when Dev >= $A, Dev =< $Z ->
	    path_open_1([Dev,$:,$/|Cs], Mode);
	[Dev,$:,$\\|Cs] when Dev >= $a, Dev =< $z ->
	    path_open_1([Dev,$:,$\\|Cs], Mode);
	[Dev,$:,$\\|Cs] when Dev >= $A, Dev =< $A ->
	    path_open_1([Dev,$:,$\\|Cs], Mode);
	Fname ->
	    path_open_1(Ps, Fname, Mode)
    end.

path_open_1(Name, Mode) ->
    case open(Name, Mode) of
	{ok,Fd} -> {ok,Fd,Name};
	Error -> Error
    end.

path_open_1([P|Ps], Name, Mode) ->
    case open([P,"/",Name], Mode) of
	{ok,Fd} ->
	    {ok,Fd,flatten(file_name([P,"/",Name]))};
	{error,E} ->
	    path_open_1(Ps, Name, Mode)

    end;
path_open_1([], Name, Mode) ->
    {error,open}.

open(Name, Mode) ->
    file_request(file_server, {open,Name,Mode}),
    wait_file_reply(file_server).

close(File) ->
    file_request(File, close),
    Reply = wait_file_reply(File),
    unlink(File),
    receive
	{'EXIT', File, _} -> ok
    after 0 -> ok
    end,
    Reply.

position(File, At) ->
    file_request(File, {position,At}),
    wait_file_reply(File).

truncate(File) ->
    file_request(File, truncate),
    wait_file_reply(File).

%% io_reply(From, ReplyAs, Reply)
%% file_request(Io, Request)
%% file_reply(From, ReplyAs, Reply)
%% wait_file_reply(From)
%%  The functions for sending and waiting for i/o command acknowledgement.
%%  The messages sent have the following formats:
%%
%%	{file_request,From,ReplyAs,Request}
%%	{file_reply,ReplyAs,Reply}
%%	{io_reply,ReplyAs,Reply}

io_reply(From, ReplyAs, Reply) ->
    From ! {io_reply,ReplyAs,Reply}.

file_request(Io, Request) ->
    Io ! {file_request,self(),Io,Request}.

file_reply(From, ReplyAs, Reply) ->
    From ! {file_reply,ReplyAs,Reply}.

wait_file_reply(From) ->
    receive
	{file_reply,From,Reply} ->
	    Reply;
	{'EXIT',From,Reason} ->			%In case we are trapping exits
	    {error,terminated}
    end.

%% The basic file server and start-up.
%%
%% The file server just handles the open command/message and acts as a
%% router for messages between the port and the file processes. If a
%% file process terminates we close the associated file.


start() ->
    case init:get_flag('-master') of
	notfound ->
	    Id = spawn(file,server,[{spawn,'erlang_filer'},
				    [{packet,4},use_stdio],
				    {spawn,binary_filer},[binary]]),
	    register(file_server,Id),
	    Id;
	{'-master', Node} ->
	    start_slave([Node])
    end.

start_slave([Node]) when atom(Node) ->
    %%    If we are not alive here things will get out of hand
    Filer = rpc:call(Node,erlang,whereis,[file_server]),
    case Filer of
	{badrpc,_} ->
	    error_logger:error_msg('can not get remote filer ',[]),
	    %% Time to die
	    halt();
	XX when pid(XX) ->
	    Id = spawn(slave,relay,[Filer]),
	    register(file_server,Id)
    end.
    
server(PortName, Settings, BinPort, BinSetttings) ->
    process_flag(trap_exit, true),
    Port = open_port(PortName, Settings),
    BinP = open_port(BinPort, BinSetttings),
    server_loop(Port, BinP).

%% The file server loop must not ever never ask the code server to
%% load any code since the request to the file server that is beeing
%% handled may very well be a request to load code .... !!!!!! deadlock

server_loop(Port, BinP) ->
    receive
	{file_request,From,ReplyAs,Request} when pid(From) ->
	    server_request(Request, From, ReplyAs, Port,BinP);
	{File,{command,Op,Fd,Bytes}} ->		%File requests
	    case get(File) of
		Fd ->
		    Rep = fs_request(Port, port_command(Op, Fd, Bytes)),
		    File ! {self(),Rep};
		Other ->			%Ignore unknown/bad processes
		    true
	    end;
	{'EXIT',Port,What} ->			%Port has exited
	    file_exit(erlang_filer, What);
	{'EXIT',BinP,What} ->			%Port has exited
	    file_exit(bin_drv, What);
	{'EXIT',From,What} ->
	    case erase(From) of
		undefined ->			%From unknown, ignore
		    ignore;
		Fd ->				%From file close file
		    fs_request(Port, port_command(operation(close), Fd, [])),
		    erase(Fd)
	    end;
	Other ->				%Ignore other messages
	    true
    end,
    server_loop(Port, BinP).

file_exit(Who, Reason) ->
    unregister(file_server),
    error_logger:error_msg('Port controlling ~w terminated in file_server',
			   [Who]),
    exit(Reason).

%% server_request(Request, FromPid, ReplyAs, Port)
%%  Handle a request to file server.

server_request({open,Name,Mode}, From, ReplyAs, Port, _) ->
    %% Check basic argument formats
    case open_mode(Mode) of
	error ->
	    file_reply(From, ReplyAs, {error,mode});
	M ->
	    case catch file_name(Name) of
		{'EXIT',_} ->
		    file_reply(From, ReplyAs, {error,name});
		Chars ->
		    open_file(Chars, M, From, ReplyAs, Port)
	    end
    end;
server_request(get_cwd, From, ReplyAs, Port, _) ->
    Op = operation(get_cwd),
    case fs_request(Port, port_command(Op, [])) of
	{data,[Op,0|Dir]} ->
	    file_reply(From, ReplyAs, {ok,Dir});
	{data,[Op,1]} ->
	    file_reply(From, ReplyAs, {error,get_cwd})
    end;
server_request({read_file,Name}, From, ReplyAs, _, Binp) ->
    Req = {command,catch [$f|file_name(Name)]},
    case fs_request(Binp,Req) of
	{data,B}  ->
	    {Head,Tail} = split_binary(B,1),
	    case binary_to_list(Head) of
		[$y] ->
		    file_reply(From, ReplyAs, {ok,Tail}),
		    erlang:garbage_collect(self());		
						%This might help to cleanup
		[$n] ->
		    file_reply(From, ReplyAs, {error,read_file})
	    end;
	badsig ->
	    file_reply(From, ReplyAs, {error,read_file});
	error ->
	    file_reply(From, ReplyAs, {error,read_file})
    end;
server_request({write_file,Name,Binary}, From, ReplyAs, Port, Binp) ->
    Req = {command, [$w, catch file_name(Name), 0, Binary]},
    case fs_request(Binp,Req) of
	{data,B}  ->
	    {Head,Tail} = split_binary(B,1),
	    case binary_to_list(Head) of
		[$y] ->
		    file_reply(From, ReplyAs, ok);
		[$n] ->
		    file_reply(From, ReplyAs, {error,write_file})
	    end;
	badsig ->
	    file_reply(From, ReplyAs, {error,write_file});
	error ->
	    file_reply(From, ReplyAs, {error,write_file})
    end;
server_request({set_cwd,Name}, From, ReplyAs, Port, Binp)  ->
    case catch file_name(Name)  of
	{'EXIT',_} ->
	    file_reply(From, ReplyAs, {error,set_cwd});
	X ->
	    Req = {command,[$c|X]},
	    case fs_request(Binp,Req) of
		{data,D} ->
		    case binary_to_list(D) of
			[$y] ->
			    fm_op(catch file_name(Name), set_cwd,
				  From, ReplyAs,Port);
			_ ->
			    file_reply(From, ReplyAs, {error,set_cwd})
		    end;
		_ ->
		    file_reply(From, ReplyAs, {error,set_cwd})
	    end
    end;
server_request({delete,Name}, From, ReplyAs, Port, _) ->
    fm_op(catch file_name(Name), delete, From, ReplyAs, Port);
server_request({rename,Fr,To}, From, ReplyAs, Port, _) ->
    fm_op(catch file_name([Fr,$\000,To]), rename, From, ReplyAs, Port);
server_request({make_dir,Name}, From, ReplyAs, Port, _) ->
    fm_op(catch file_name(Name), make_dir, From, ReplyAs, Port);
server_request({del_dir,Name}, From, ReplyAs, Port, _) ->
    fm_op(catch file_name(Name), del_dir, From, ReplyAs, Port);
server_request({list_dir,Name}, From, ReplyAs, Port, _) ->
    list_dir_op(catch file_name(Name), From, ReplyAs, Port);
server_request({file_info,Name}, From, ReplyAs, Port, _) ->
    file_info_op(catch file_name(Name), From, ReplyAs, Port);
server_request(_, From, _, _, _) ->
    exit(From, badio).

bin_op({'EXIT',Reason}, _, _) -> ignore;
bin_op(Chars, Command, Port) ->
    Port ! {self(),{command,[Command|Chars]}}.

%% file_name(FileName)
%%  Generate a file name from a deep list of atoms and characters (integers).

file_name([C|T]) when integer(C), C >= 0, C =< 255 ->
    [C|file_name(T)];
file_name([H|T]) ->
    [file_name(H)|file_name(T)];
file_name([]) -> [];
file_name(N) when atom(N) -> atom_to_list(N).

flatten(List) ->
    flatten(List, []).

flatten([H|T], Cont) when list(H) ->
    flatten(H, [T|Cont]);
flatten([H|T], Cont) ->
    [H|flatten(T, Cont)];
flatten([], [H|Cont]) ->
    flatten(H, Cont);
flatten([], []) -> [].
		  
%% open_file(Name, Mode, From, ReplyAs, Port)
%%  Try to open a file. Start the file process if file is opened.

open_file(Name, Mode, From, ReplyAs, Port) ->
    Op = operation(open),
    case fs_request(Port, port_command(Op, Mode, Name)) of
	{data,[Op,0|Bs]} ->
	    Fd = get_int32(Bs),
	    Id = spawn_link(file, file, [Fd,From,ReplyAs,self()]),
	    put(Fd, Id),
	    put(Id, Fd);
	{data,[Op,1]} ->			%Open failed
	    file_reply(From, ReplyAs, {error,open});
	badsig ->
	    file_reply(From, ReplyAs, {error,open})
    end.

%% fm_op(Req, OpName, FromPid, ReplyAs, Port)
%%  Do a file management operation that only has a status value as return.

fm_op({'EXIT',_}, OpName, From, ReplyAs, Port) ->
    file_reply(From, ReplyAs, {error,name});
fm_op(Chars, OpName, From, ReplyAs, Port) ->
    Op = operation(OpName),
    case fs_request(Port, port_command(Op, Chars)) of
	{data,[Op,0]} ->
	    file_reply(From, ReplyAs, ok);
	{data,[Op,1]} ->
	    file_reply(From, ReplyAs, {error,OpName})
    end.

%% list_dir_op(Req, FromPid, ReplyAs, Port)
%%  List all the files in a directory, each file comes in its own message.

list_dir_op({'EXIT',_}, From, ReplyAs, Port) ->
    file_reply(From, ReplyAs, {error,name});
list_dir_op(Chars, From, ReplyAs, Port) ->
    Op = operation(list_dir),
    case fs_request(Port, port_command(Op, Chars)) of
	{data,[Op,0]} ->
	    file_reply(From, ReplyAs, {ok,list_dir_op(Port)});
	{data,[Op,1]} ->
	    file_reply(From, ReplyAs, {error,list_dir})
    end.

list_dir_op(Port) ->
    Op = operation(get_next),
    case fs_request(Port, port_command(Op, "")) of
	{data,[Op,0]} ->
	    [];
	{data,[Op,0|Name]} ->
	    [Name|list_dir_op(Port)]
    end.

%% file_info_op(Req, FromPid, ReplyAs, Port)
%%  List all the files in a directory, each file comes in its own message.

file_info_op({'EXIT',_}, From, ReplyAs, Port) ->
    file_reply(From, ReplyAs, {error,name});
file_info_op(Chars, From, ReplyAs, Port) ->
    Op = operation(file_info),
    case fs_request(Port, port_command(Op, Chars)) of
	{data,[Op,0,S1,S2,S3,S4,T,A]} ->
	    S = (S1 bsl 24) bor (S2 bsl 16) bor (S3 bsl 8) bor S4,
	    {Atime,Mtime} = file_info_op(Port),
	    file_reply(From, ReplyAs,
		     {ok,{S,file_type(T),file_access(A),Atime,Mtime,
			  unused,unused}});
	{data,[Op,1]} ->
	    file_reply(From, ReplyAs, {error,file_info})
    end.

file_info_op(Port) ->
    Op = operation(get_next),
    {data,[_,_,Ay,Am,Ad,Ah,Ami,As]} = fs_request(Port, port_command(Op, "")),
    {data,[_,_,My,Mm,Md,Mh,Mmi,Ms]} = fs_request(Port, port_command(Op, "")),
    {{Ay+1900,Am,Ad,Ah,Ami,As},{My+1900,Mm,Md,Mh,Mmi,Ms}}.

file_type(1) -> device;
file_type(2) -> directory;
file_type(3) -> regular;
file_type(4) -> symlink;
file_type(T) -> other.

file_access(3) -> read_write;
file_access(2) -> read;
file_access(1) -> write;
file_access(0) -> none.    

%% fs_request(Port, Request)
%% fs_reply(Port)

fs_request(_,{'EXIT',_}) -> error;
fs_request(Port, Req) ->
    Port ! {self(),Req},                        %Send the request
    fs_reply(Port).


fs_reply(Port) ->
    receive					% and wait for the reply
	{Port,Rep} ->				%Valid reply
	    Rep;
	{'EXIT',Port,badsig} ->			%Only from the port
	    badsig;
	{'EXIT',Port,Reason} ->			%"Port" down
	    file_exit(erlang_filer, Reason)			% just exit
    end.

port_command(Op, List) ->
    {command,[Op|List]}.

port_command(Op, Fd, List) ->
    {command,[Op,Fd| List]}.

%% The actual file process with which the user communicates.
%%
%% All the commands/messages are processed here except 'close'. This
%% just results in the process terminating, we let the file server
%% close the actual file when it detects our terminatin.
%%
%% The file process is linked to the opener, if this (or any other linked)
%% process dies the file is closed.

%% file(Descriptor, Opener, ReplyAs, Server)

file(Fd, Pid, ReplyAs, Server) ->
    process_flag(trap_exit, true),		%Trap all exits
    link(Pid),
    file_reply(Pid, ReplyAs, {ok,self()}),	%Reply that file is open
    file_loop(Fd, Server, []).

%% file_loop(Descriptor, Server, ReadBuffer)

file_loop(Fd, Server, Buf0) ->
    receive
	{file_request,From,ReplyAs,Request} when pid(From) ->
	    Buf = file_request(Request, From, ReplyAs, Fd, Server, Buf0),
	    file_loop(Fd, Server, Buf);
	{io_request,From,ReplyAs,Request} when pid(From) ->
	    Buf = io_request(Request, From, ReplyAs, Fd, Server, Buf0),
	    file_loop(Fd, Server, Buf);
	%% An exit signal from anyone causes us to exit.
	{'EXIT',_,Reason} ->
	    exit(Reason);
	Other ->				%Ignore other messages
	    file_loop(Fd, Server, Buf0)
    end.

%% file_request(Request, Sender, ReplyAs, Descriptor, Server, Buffer)
%%  Returns the new buffer.

file_request(Req, From, ReplyAs, Fd, Server, Buf0) ->
    case file_request(Req, Fd, Server, Buf0) of
	{Status,Reply,Buf} ->
	    file_reply(From, ReplyAs, Reply),
	    Buf;
	{exit,normal} ->
	    file_reply(From, ReplyAs, ok),
	    exit(normal);
	{exit,R} ->
	    file_reply(From, ReplyAs, {error,terminated}),
	    exit(normal)
    end.
file_request({position,At}, Fd, Server, Buf) ->
    position_file(At, Fd, Server, Buf);
file_request(truncate, Fd, Server, Buf) ->
    case f_request(Server, command(truncate, Fd, [])) of
	{data,[_,0]} ->
	    {ok,ok,[]};
	{exit,R} ->
	    {exit,R};
	Other ->				%File/data error
	    {error,{error,truncate},Buf}
    end;
file_request(close, Fd, Server, Buf) ->
    {exit,normal};
file_request(R, Fd, Server, Buf) ->		%Unknown request
    {ok,{error,{request,R}},Buf}.		%Ignore but give error (?)

%% io_request(Request, Sender, ReplyAs, Descriptor, Server, Buffer)
%%  Returns the new buffer.

io_request(Req, From, ReplyAs, Fd, Server, Buf0) ->
    case io_request(Req, Fd, Server, Buf0) of
	{Status,Reply,Buf} ->
	    io_reply(From, ReplyAs, Reply),
	    Buf;
	{exit,normal} ->
	    io_reply(From, ReplyAs, ok),
	    exit(normal);
	{exit,R} ->
	    io_reply(From, ReplyAs, {error,terminated}),
	    exit(normal)
    end.

io_request({put_chars,Chars}, Fd, Server, Buf) ->
    case f_request(Server, command(put, Fd, Chars)) of
	{data,[_,0]} ->				%Success
	    {ok,ok,Buf};
	{exit,R} ->
	    {exit,R};
	Other ->				%Write/data error
	    {error,{error,put_chars},Buf}
    end;
io_request({put_chars,Mod,Func,Args}, Fd, Server, Buf) ->
    case catch apply(Mod,Func,Args) of
	Chars when list(Chars) ->
	    io_request({put_chars,Chars}, Fd, Server, Buf);
	Other ->
	    {error,{error,Func}, Buf}
    end;
io_request({get_until,Prompt,M,F,As}, Fd, Server, Buf) ->
    get_until(M, F, As, Fd, Server, Buf);
io_request({requests,Reqs}, Fd, Server, Buf) ->
    io_requests(Reqs, {ok,ok,Buf}, Fd, Server);
io_request(R, Fd, Server, Buf) ->		%Unknown request
    {ok,{error,{request,R}},Buf}.		%Ignore but give error (?)

%% Status = io_requests(RequestList, PrevStat, Descriptor, Server)
%%  Process a list of output requests as long as the previous status is 'ok'.

io_requests([R|Rs], {ok,Res,Buf}, Fd, Server) ->
    io_requests(Rs, io_request(R, Fd, Server, Buf), Fd, Server);
io_requests([_|_], Error, _, _) ->
    Error;
io_requests([], Stat, _, _) ->
    Stat.

%% f_request(To, Request)

f_request(To, Req) ->
    To ! {self(),Req},				%Send the request
    receive					% and wait for the reply
	{To,Rep} ->				%Valid reply (even badsig)
	    Rep;
	{'EXIT',_,Reason} ->			%So thing exited
	    {exit,Reason}
    end.

%% Command = command(OpName, Descriptor, Arguments)

command(Op, Fd, Args) ->
    {command,operation(Op),Fd,Args}.

%%
%% The codes for operations and operands.
%%

%% file_position(At, Offset, VirtualOffset)

file_position({From,Offset}, _, Voff) ->
    file_position(From, Offset, Voff);
file_position(bof, Offset, Voff) when integer(Offset) ->
    [0|put_int32(Offset)];
file_position(cur, Offset, Voff) when integer(Offset) ->
    [1|put_int32(Offset-Voff)];
file_position(eof, Offset, Voff) when integer(Offset) ->
    [2|put_int32(Offset)];
file_position(Pos, _, Voff) when integer(Pos) ->
    file_position(bof, Pos, Voff);
file_position(_, _, _) ->
    error.

operation(open) -> 0;
operation(close) -> 1;
operation(position) -> 2;
operation(truncate) -> 3;
operation(get) -> 4;
operation(put) -> 5;
operation(get_cwd) -> 6;
operation(set_cwd) -> 7;
operation(delete) -> 8;
operation(rename) -> 9;
operation(make_dir) -> 10;
operation(del_dir) -> 11;
operation(list_dir) -> 12;
operation(file_info) -> 13;
operation(get_next) -> 14;
operation(read_file) -> 15;
operation(write_file) -> 16.


open_mode(read) -> 0;
open_mode(write) -> 1;
open_mode(read_write) -> 2;
open_mode(_) -> error.

put_int32(I) ->
    [(I bsr 24) band 16#ff,
     (I bsr 16) band 16#ff,
     (I bsr 8) band 16#ff,
     I band 16#ff].

get_int32([B1,B2,B3,B4|_]) ->
    (B1 bsl 24) bor (B2 bsl 16) bor (B3 bsl 8) bor B4.

put_int16(I) ->
    [(I bsr 8) band 16#ff,I band 16#ff].

get_int16(B1, B2) ->
    (B1 bsl 8) bor B2.

%% position_file(At, Descriptor, Server, ReadBuffer)
%%  Position the file at At. Returns:
%%
%%	{ok,NewPosition,NewReadBuffer}
%%	{error,{error,position},NewReadBuffer}
%%
%%  NewReadBuffer is always [] here. This is less efficient but makes
%%  life easier.

position_file(At, Fd, Server, Buf) ->
    case file_position(At, 0, length(Buf)) of
	error ->				%Ignore errors
	    {error,{error,position},[]};
	Pos ->
	    case f_request(Server, command(position, Fd, Pos)) of
		{data,[_,0|Bs]} ->
		    {ok,{ok,get_int32(Bs)},[]};
		{exit,R} ->
		    {exit,R};
		Other ->
		    {error,{error,position},[]}
	    end
    end.

%% get_until(Module, Function, Arguments, Descriptor, Server, Buffer)
%%  Gets characters from the input file as long as the applied function
%%  returns {more,Continuation}. Returns:
%%
%%	{Status,Result,NewSaveBuffer}
%%	{exit,Reason}

get_until(M, F, As, Fd, Server, Buf) ->
    get_until1(catch apply(M, F, [[],Buf|As]), M, F, As, Fd, Server).

get_until1({more,Cont}, M, F, As, Fd, Server) ->
    case f_request(Server, command(get, Fd, put_int16(128))) of
	{data,[_,0]} ->				%[] means eof here
	    get_until1(catch apply(M, F, [Cont,eof|As]), M, F, As, Fd, Server);
	{data,[_,0|Bs]} ->
	    get_until1(catch apply(M, F, [Cont,Bs|As]), M, F, As, Fd, Server);
	{exit,R} ->
	    {exit,R};
	Other ->
	    {error,{error,F},[]}
    end;
get_until1({done,Result,Buf}, M, F, As, Fd, Server) ->
    {ok,Result,Buf};
get_until1(Other, M, F, As, Fd, Server) ->
    {error,{error,F},[]}.
