%% This is a generic server, this code can be used to 
%% start,stop and request any server, The advantage of
%% using the generic server is that we get a consitent interface
%% to all servers, we get additional logging and tracing functionality
%% on all servers, and we get very nice error reports

%% Author: Claes Wikstrom, klacke@erix.ericsson.se

-module(generic).

-export([start/4,
	 sync_start/2,
	 stop/1,
	 request/2,
	 request/3,
	 reply/3,
	 tty_trace/2,
	 print_log/1,
	 log/2,
	 debug/2,
	 log_to_file/2,
	 no_log_to_file/1,
	 info/1,
	 statistics/2,
	 timeout_loop/4,
	 timeout_loop/5,
	 loop/3,
	 loop/4]).

%% Internale export
-export([start_it/6]).

-import(error_logger ,[error_msg/2, format/2]).


%% Start a local (registered) server
start(Name, Mod, Fun, Args) when atom(Name) ->
    case whereis(Name) of
	undefined ->
	    Rfunc = {erlang, register},
	    P = spawn(generic, start_it, [self(),Rfunc,Name, Mod, Fun, Args]),
	    receive 
		{Name, error} ->
		    catch global:unregister(Name),
		    error;
		{Name, Status} -> 
		    {P, Status}
	    after 12000 ->
		    format("** Generic server ~w probably forgot ", [Name]),
		    format("to call generic:sync_start/2~n",[]),
		    catch global:unregister(Name),
		    error
	    end;
	Pid -> 
	    {Pid, already_started}
    end;

%% Use these clauses when we want to start a global server
start({global, Name}, Mod, Fun, Args) ->
    start({global, Name, node()}, Mod, Fun, Args);
start({global, Name, Node}, Mod, Fun, Args) ->
    case global:whereis_name(Name) of
	undefined ->
	    Rfunc = {global, register_name},
	    P = spawn(Node,generic, start_it, 
		      [self(),Rfunc,Name, Mod, Fun, Args]),
	    receive 
		{Name, error} ->
		    global:unregister_name(Name),
		    error;
		{Name, Status} -> 
		    {P, Status}
	    after 12000 ->
		    format("** Generic global server ~w probably forgot ", 
			      [Name]),
		    format("to call generic:started/1~n",[]),
		    global:unregister_name(Name),
		    error	
	    end;
	Pid -> 
	    {Pid, allready_started}
    end.



start_it(Top,Rfunc,Name, Mod, Fun, Args) ->
    apply(Rfunc, [Name, self()]),   %% register the process
    put(start_process, Top),
    apply(Mod, Fun, Args).


sync_start(Name,Status) ->
    case get(start_process) of
	undefined ->
	    error_msg("Illegal call to generic:sync_start(~w, ~w)\n",
		      [Name, Status]), 
	    exit(error);
	Pid when pid(Pid) ->
	    erase(start_process),
	    Pid ! {Name, Status}
    end.
		   
stop(Name) when atom(Name) ->
    case whereis(Name) of
	undefined -> stopped;
	Pid -> request(Name, stop)
    end;
stop({global, Name}) ->
    case global:whereis_name(Name) of
	undefined -> stopped;
	Pid -> request({global, Name}, stop)
    end;

stop(Pid) ->
    request(Pid, stop).


%% The request function, It calls exit/1 on any type
%% of error that can occur. or returns the value rerturned by the
%% server

request(Name, Req) -> request(Name, Req, infinity).

request(Name, Request, T) when atom(Name) ->  %% For locally registered procs
    Name ! {self(), Request},
    rec(Name, T);

request({global, Name}, Request, T) ->      %% For globally registered servers
    Pid = global:send(Name, {self(),Request}),
    Node = node(Pid),
    monitor_node(Node, true),
    rrec(Name, Node, T);

request({Name, Node}, Request, T) ->
    monitor_node(Node, true),
    {Name, Node} ! {self(), Request},
    rrec(Name, Node, T);

request(Name, Request, T) ->      %% Maybe use with pids as well
    Name ! {self(), Request},
    rec(Name, T).

rec(Name, T) ->
    receive
	{Name, Reply} -> Reply;
	{Name, Node, Reply} -> Reply
    after T ->
	    exit(generic_timeout)
    end.

rrec(Name, Node, T) ->
    R  = receive
	     {Name, Reply} -> Reply;
	     {Name, Node, Reply} -> Reply;
	     {nodedown, Node} -> exit({nodedown, Node})
	 after T ->
		 exit(generic_timeout)
	 end,
    monitor_node(Node, false),
    R.



reply(Name, To, Reply) -> To ! {Name, node(), Reply}.

tty_trace(Name, Flag)  ->  irequest(Name, {trace, Flag}).
print_log(Name) ->         irequest(Name, print_events).
log(Name, Flag) ->         irequest(Name, {log, Flag}).
log_to_file(Name, File) -> irequest(Name, {logfile, File}).
no_log_to_file(Name) ->    irequest(Name, no_log_to_file).
info(Name) ->              irequest(Name, info).
statistics(Name, Flag) ->  irequest(Name, {statistics, Flag}).

debug(Name, Flag) ->
    log(Name, Flag), statistics(Name, Flag).
    
%% internal requests
%% We use the format of {ping, Request} here 
%% so that we can receive joe's ping messages without any
%% extra tests.

irequest({global, Name}, Req) ->
    global:send(Name, {ping, {self(), Req}}),
    irec(Name);
irequest(Name, Req) when atom(Name) -> 
    Name ! {ping, {self(), Req}},
    irec(Name);
irequest({Name, Node}, Req) ->
    monitor_node(Node, true),
    {Name, Node} ! {ping, {self(), Req}},
    receive
	{Name, Node, Reply} -> 
	    monitor_node(Node, false),
	    Reply;
	{Name, Reply} ->
	    monitor_node(Node, false),
	    Reply;
	{nodedown, Node} ->
	    format("** Server ~w is unreachable \n", [{Name, Node}]),
	    error
    after 4000 ->
	    format("** Server ~w might be deadlocked ** \n", [Name]),
	    error
    end;
irequest(Name, Req)  ->  %% Possibly use with pids as well
    Name ! {ping, {self(), Req}},
    irec(Name).

irec(Name) ->
    receive
	{Name, Reply} -> Reply
    after 4000 ->
	    format("** Server ~w might be deadlocked ** \n", [Name]),
	    error
    end.

loop(Name, State, F) ->
    case init:get_flag('-generic_debug') of
	notfound ->
	    loop(Name, State, F, []);
	{'-generic_debug', true} ->
	    loop(Name, State, F, debug);
	{'-generic_debug', full} ->
	    case file:open(lists:concat([Name,".log"]), write) of
		{error, _} ->
		    error_msg("** Server ~w can't open logfile \n", [Name]),
		    loop(Name, State, F, debug);
		{ok, File} ->
		    S = {statistics, {{date(), time()}, reds(),0,0}},
		    loop(Name, State, F, [{logfile, File}, S,{log, []}])
	    end;
	_ ->
	    loop(Name, State, F, debug)
    end.

loop(Name, State, F, debug) ->
    S = {statistics, {{date(), time()}, reds(),0,0}},
    loop(Name, State, F, [S,{log, []}]);  %% Log and statistics on 
loop(Name,State,F, Trace) ->  %% Let the user provide the args
    loop1(Name,State,F, Trace).

loop1(Name,State,F, Trace) ->
    receive
	{ping, Req} ->
	    Trace2 = do_internal(Name,State,F, Trace, Req),
	    loop1(Name, State, F, Trace2);
	Msg when Trace == [] ->
	    handle(Msg, catch apply(F, [Msg, State]), Name, State, F, []);
	Msg ->
	    Trace2 = input(Name,Msg, Trace),
	    handle(Msg, catch apply(F, [Msg, State]), Name, State, F, Trace2)
    end.


%% Standard case first for effeciency

handle(Msg, {reply, To, Reply, Nstate}, Name, State, F, []) ->
    case catch To ! {Name, node(),Reply} of
	{'EXIT', _} -> 
	    error(Name, F,State, Msg, badreply, []);
	_ -> 
	    loop1(Name, Nstate, F, [])
    end;

handle(_, {noreply, Nstate}, Name, State, F, Trace) ->
    loop1(Name, Nstate, F, Trace);

%% Now standard case for timeout replies

handle(Msg, {reply, To, Reply, Nstate, Ntimeout}, Name, State, F, []) ->
    case catch To ! {Name, node(),Reply} of
	{'EXIT', _} -> 
	    error(Name, F,State, Msg, badreply, []);
	_ ->
	    timeout_loop1(Name, Nstate, F, Ntimeout, [])
    end;

handle(_, {noreply, Nstate, NTimeout}, Name, State, F, Trace) ->
    timeout_loop1(Name, Nstate, F, NTimeout, Trace);

handle(_, {stopped, From}, Name, State, F, _) ->
    From ! {Name, stopped},
    exit(normal);

handle(Msg, {'EXIT', Reason}, Name, State, F, Trace) ->
    error(Name, F,State, Msg, Reason, Trace);

%% Now handle the case where Trace != []

handle(Msg, {reply, To,Reply, Nstate}, Name, State, F, Trace) ->
    Trace3 = output(Name,To,Reply,Trace),
    case catch To ! {Name, node(),Reply} of
	{'EXIT', _} -> 
	    error(Name, F,State, Msg, badreply, Trace3);
	_ ->
	    loop1(Name, Nstate, F, Trace3)
    end;

%% Catch all cases

handle(Msg, Other, Name, State, F, Trace) ->
    Reason = {badreply, Other},
    error(Name, F,State, Msg, Reason, Trace).


timeout_loop(Name, State, F, Timeout) ->
    loop(Name, State, F, []).
timeout_loop(Name, State, F, Timeout,debug) ->
    S = {statistics, {{date(), time()}, reds(),0,0}},
    timeout_loop(Name, State, F, [S,{log, []}]);  %% Log and statistics on 
timeout_loop(Name,State,F, Timeout,Trace) ->  %% Let the user provide the args
    timeout_loop1(Name,State,F, Timeout,Trace).


timeout_loop1(Name,State,F, Timeout,Trace) ->
    receive
	{ping, Req} ->
	    Trace2 = do_internal(Name,State,F, Trace, Req),
	    timeout_loop1(Name, State, F, Timeout, Trace2);
	Msg when Trace == [] ->
	    handle(Msg, catch apply(F, [Msg, Timeout,State]), 
		   Name, State, F, []);
	Msg ->
	    Trace2 = input(Name,Msg, Trace),
            handle(Msg, catch apply(F, [Msg, Timeout,State]), 
		   Name, State, F,Trace2)
    after Timeout ->
	    case Trace of
		[] ->
		    handle(timeout, catch apply(F, [timeout, Timeout, State]),
			   Name, State, F, []);
		_ ->
		    Trace2 = input(Name,timeout, Trace),
         	    handle(timeout, catch apply(F, [timeout, Timeout, State]),
		           Name, State, F,Trace2)
             end
    end.


input(Name, Msg, [trace | Tail]) ->
    io:format("generic: Server ~w got ~w~n", [Name, Msg]),
    [trace | input(Name, Msg, Tail)];
input(Name, Msg, [{log, L} | Tail]) ->
    L2 = [{input, Msg} | trim(L)],
    [{log, L2} | input(Name, Msg, Tail)];
input(Name, Msg, [{logfile, F} |Tail]) ->
    io:format(F, "Server ~w got ~w~n", [Name, Msg]),
    [{logfile, F} | input(Name, Msg, Tail)];
input(Name, Msg, [{statistics, {Time, Reds,In, Out}} | Tail]) ->
    [{statistics, {Time, Reds,In+1, Out}} | input(Name, Msg, Tail)];
input(Name, Msg, []) -> [].


output(Name,To, Reply, [trace | Tail]) ->
    io:format("generic: Server ~w Replying ~w to ~w~n", [Name,Reply,To]),
    [trace | output(Name,To, Reply, Tail)];
output(Name, To, Reply, [{log, L} | Tail]) ->
    L2 = [{output, To, Reply} | trim(L)],
    [{log, L2} |output(Name,To, Reply, Tail)];
output(Name, To, Reply, [{logfile, F} | Tail]) ->
    io:format(F,"Server ~w Replying ~w to ~w~n", [Name,Reply,To]),
    [{logfile, F}  |output(Name,To, Reply, Tail)];
output(Name, To, Reply, [{statistics, {Time, Reds,In, Out}} | Tail]) ->
    [{statistics, {Time, Reds,In, Out+1}} | output(Name,To, Reply, Tail)];
output(Name, To, Reply, []) -> [].

trim([X1,X2,X3,X4,X5,X6,X7,X8,X9,X10 | Tail]) ->
    [X1,X2,X3,X4,X5,X6,X7,X8,X9,X10];
trim(X) -> X.
	
reds() -> 
    process_info(self(),reductions).
reds(R) -> 
    {reductions, R2} = process_info(self(),reductions),
    {reductions, R2 - R}.

do_internal(Name,State, F,Trace,Req) ->
    case Req of
	From when pid(From) -> %% Joe's message
	    From ! {pong, self()},
	    Trace;
	{From, no_log_to_file} ->
	    case keyget(logfile,Trace) of
		{logfile, File} ->
		    file:close(File);
		_ -> 
		    ok
	    end,
	    From ! {Name, ok},
	    keydel(logfile, Trace);
	{From, {trace,true}} ->
	    From ! {Name, ok},
	    [trace | Trace];
	{From, {trace, false}} ->
	    From ! {Name, ok},
	    lists:delete(trace, Trace);
	{From,{log, true}} ->
	    From ! {Name, ok},
	    [{log, []} | Trace];  
	{From,{log, false}} ->
	    From ! {Name, ok},
	    keydel(log, Trace);
	{From,{statistics, true}} ->
	    From ! {Name, ok},	
	    [{statistics, {{date(), time()}, reds(),0, 0}} | Trace];
	{From,{statistics, false}} ->
	    From ! {Name, ok},	
	    keydel(statistics, Trace);
	{From,info} ->
	    From ! {Name, [{state, State}, {'fun', F}, {trace, Trace}]},
	    Trace;
	{From,print_events} ->
	    case keyget(log,Trace) of
		{log, L} ->
		    print_events(Name, lists:reverse(L));
		_ ->
		    error_msg("** Server ~w not logging events \n", [Name])
	    end,
	    From ! {Name, ok},	
	    Trace;
	{From,{logfile ,FileName}} ->
	    case file:open(FileName, write) of
		{error, _} -> 
		    error_msg("** Server ~w can't open logfile \n", [Name]),
		    From ! {Name, error},
		    Trace;
		{ok,File} ->
		    From ! {Name, ok},
		    [{logfile, File} | Trace]
	    end;
	{From, {statistics, get}} ->
	    case keyget(statistics, Trace) of
		{statistics, {Time,{reductions, Reds},In, Out}} ->
		    From ! {Name, [{from,Time}, reds(Reds), 
				   {messages_in, In}, {messages_out,Out}]};
		no ->
		    error_msg("** Server ~w not tracing statistics\n",[Name]),
		    From ! {Name, error}
	    end,
	    Trace;
	Other ->
	    error_msg("** Internal error in generic, got ~w **\n", [Other]),
	    exit(error)
    end.

print_events(Name, E) ->
    format("The ~w last events for the ~w server were ~n", 
	      [length(E), Name]),
    print_events(E).

print_events([{input, I} |Tail]) ->
    format("<< ~w~n",[I]),
    print_events(Tail);
print_events([{output, To, M} |Tail]) ->
    format(">> ~w ! ~w~n", [To, M]),
    print_events(Tail);
print_events([]) -> done.


keyget(Key, [{Key, V} | _]) -> {Key, V};
keyget(Key, [_|T]) -> keyget(Key, T);
keyget(Key, []) -> no.

keydel(Key, [{Key, V} | T]) -> T;
keydel(Key, [H|T]) -> [H|keydel(Key, T)];
keydel(Key, []) -> [].

%% Use error_logger:format here, If we want all generic
%% server crashes to be logged to a file, we gotta tell the
%% error logger to log to a file instead

%% Provide two means for the process to call exit, one normal
%% and one abnormal.

error(_,_,_,_, stopgeneric, _) ->
    exit(stopgeneric);
error(_,_,_,_, normal, _) ->
    exit(normal);
error(Name, F,State, Request, Reason, Trace) ->
    format("** Generic server ~w terminating \n", [Name]),
    format("** Last message in was ~w~n", [Request]),
    format("** When Server state == ~w~n", [State]),
    format("** Reason for termination = ~n** ~w~n", [Reason]),
    case keyget(log, Trace) of
	{log, L} ->
	    print_events(Name, lists:reverse(L));
	_ -> ok
    end,
    exit(Reason).


	
    
