%% Copyright (C) 1993,1994 Ellemtel Telecommunications Systems Laboratories
%% File     : global.erl
%% Author   : klacke@erix.ericsson.se
%% Purpose  : A Global register that allows the global registration
%%            of pid's and name's
%%


-module(global).
-export([start/0, 
	 stop/0, 
	 whereis_name/1, 
	 register_name/2, 
	 re_register_name/2,
	 unregister_name/1,
         registered_names/0, 
	 send/2]).

%% Local exports
-export([init/0, 
	 doit/2,
	 local_unregister/1]).


-import(error_logger, [error_msg/2]).
-import(generic, [request/2]).

%% A global name manager, that dynamically keeps up to date
%% with the entire network
%% Global is started from the net kernel.


start() ->        generic:start(global_name_server, global, init, []).
stop()  ->        generic:stop(global_name_server).
send(To, Mess) -> 
    case request(global_name_server, {send, To, Mess}) of
	{ok, Pid} -> 
	    Pid;
	undefined ->
	    error_handler:undefined_global_name({global, To}, Mess)
    end.

whereis_name(Name) -> 
    request(global_name_server, {global_pid, Name}).

register_name(Name,Pid)   ->
    {L,_} = 
	rpc:multi_server_call(global_name_server, {pre_register, Name, Pid}),
    reg_loop(L, Name, Pid).

reg_loop([], Name, Pid) ->
    rpc:multi_server_call(global_name_server, {do_register, Name, Pid}),
    yes;

reg_loop([{no, Node} |Tail], Name, Pid) when node() < Node ->
    reg_loop(Tail, Name, Pid);

reg_loop([{no, Node} |T], Name, Pid) ->
    rpc:multi_server_call(global_name_server, {pre_unregister, Name, Pid}),
    no;

reg_loop([occupied|T], Name, Pid) ->
    rpc:multi_server_call(global_name_server, {pre_unregister,Name, Pid}),
    error_msg("** Global name ~w already registered ** \n", [Name]),
    no;
reg_loop([ok|T], Name, Pid) ->
    reg_loop(T, Name, Pid).


unregister_name(Name) ->
    rpc:multi_server_call(global_name_server, {do_unregister, Name}),
    ok.

%%
%% Assuming a fully connected network
%%

re_register_name(Name, Pid) ->
    rpc:multicall(global, local_unregister, [Name]),
    N = node(),
    case lists:min([node()|nodes()]) of
	N -> %% We're the one
	    register_name(Name, Pid);
	Other ->
	    {othernode_chosen, Other}
    end.

local_unregister(Name) ->
    request(global_name_server, {local_unregister, Name}).

%% Returns all globally registered names 
registered_names() ->
    request(global_name_server, globs).

init() ->
    case init:get_flag('-connect_all') of
	{'-connect_all', false} ->
	    generic:sync_start(global_name_server, error);
	_ ->
	    process_flag(trap_exit, true),
	    Tab = ets:new(global, [set, protected]),
	    generic:sync_start(global_name_server, started),
	    generic:loop(global_name_server, Tab, {global, doit})
    end.

doit({From, {send, Name, Message}}, Tab) ->
    case ets:lookup(Tab,Name) of
	[] -> 
	    {reply,From,undefined, Tab};
	[{Name,Pid}] -> 
	    Pid ! Message,
	    {reply, From, {ok,Pid}, Tab}
    end;

doit({From, {global_pid, Name}}, Tab) ->
    case ets:lookup(Tab, Name) of
	[] ->
	    {reply, From, undefined, Tab};
	[{_, Pid}] ->
	    {reply, From, Pid, Tab}
    end;


doit({From, {pre_register, Name, Pid}}, Tab) ->
    dolink(link, Pid),
    case ets:lookup(Tab, {pre_alloc, Name}) of
	[{{pre_alloc, Name}, Pid2, Node}] ->
	    {reply, From, {no, Node}, Tab};
	[] ->
	    case ets:lookup(Tab, Name) of
		[] ->
		    ets:insert(Tab, {{pre_alloc, Name}, Pid, node(Pid)}),
		    {reply, From, ok, Tab};
		_ ->
		    {reply, From, occupied, Tab}
	    end
    end;

doit({From, {do_register, Name, Pid}}, Tab) ->
    ets:delete(Tab, {pre_alloc, Name}),
    ets:insert(Tab, {Name, Pid}),
    {reply, From, yes, Tab};

doit({From, {pre_unregister, Name, Pid}}, Tab) ->
    dolink(unlink, Pid),
    case ets:lookup(Tab, {pre_alloc, Name}) of
	[{{pre_alloc, Name}, Pid, Node}] ->
	    ets:delete(Tab, {pre_alloc, Name});
	_ ->
	    ok
    end,
    {reply, From, yes, Tab};

doit({From, {do_unregister, Name}}, Tab) ->
    dolink(unlink, ets:lookup(Tab, Name)),
    ets:delete(Tab, Name),
    {reply, From, ok, Tab};


doit({From, globs}, Tab) ->
    {reply, From, extract(ets:tab2list(Tab)), Tab};

doit({'EXIT', From, Reason}, Tab) when pid(From) ->
    check_exit(Tab, From),
    {noreply, Tab};

doit({nodedown, Node}, Tab) ->
    do_node_down(Tab, Node),
    {noreply, Tab};

doit({nodeup, Node}, Tab) ->    
    %% Make sure we're all connected 
    L = rpc:call(Node, erlang, nodes, []),
    catch lists:map({net_adm, ping}, [], fixl(node(), L)),
    rpc:abcast(lists:delete(Node, nodes()), global_name_server,
	       {newnode, Node}),

    %% Now we asyncronously send all nodes a copy of our names
    case ets:tab2list(Tab) of
	[] -> 
	    ignore;
	List ->
	    rpc:abcast(nodes(), global_name_server, {exchange, List})
    end,
    {noreply, Tab};

doit({newnode, Node}, Tab) ->
    lists:map({net_adm, ping}, [], fixl(node(), [Node])),
    {noreply, Tab};


doit({exchange, List}, Tab) ->
    exchange(List, Tab),
    {noreply, Tab};

%% We get this message from the other globa_name_server;s when either
%% a node (with a registered pid) goes down or
%% when a globally registered (or pre_alloced) process exits

doit({async_del, Name, Pid}, Tab) ->
    ets:delete(Tab, Name),
    case ets:lookup(Tab, {pre_alloc, Name}) of  %% !!!!
	[{{pre_alloc, Name}, Pid, Node}] ->
	    ets:delete(Tab, {pre_alloc, Name});
	_ ->
	    ok
    end,
    {noreply, Tab};

doit({From, {local_unregister, Name}}, Tab) ->
    ets:delete(Tab, Name),
    {reply, From, ok, Tab};

doit({From, stop}, Tab) ->
    {stopped, From}.

%% Here is marvelous trick to ensure that the pinging only occurs
%% in one direction .... :-)

fixl(Self, []) -> [];
fixl(Self, [Node|Nodes]) when Node < Self -> fixl(Self, Nodes);
fixl(Self, [Node|Nodes]) -> [Node|fixl(Self, Nodes)].


exchange([{Name, Pid} |Tail], Tab) ->
    case ets:lookup(Tab, Name) of
	[{Name, Pid}] -> 
	    exchange(Tail, Tab);
	[{Name, Pid2}] ->
	    Pid = min(Pid, Pid2),
	    error_msg("*** global: Name conflict terminating ~w~n", 
		      [{Name, Pid2}]),
	    exit(Pid2, kill),
	    ets:insert(Tab, {Name, Pid}),
	    exchange(Tail, Tab);
	_ ->
	    ets:insert(Tab, {Name, Pid}),
	    exchange(Tail, Tab)
    end;
exchange([], _) -> ok.
	    
min(P1,P2) ->
    if node(P1) < node(P2) -> P1; true -> P2 end.

%% Only link to pids on our own node
dolink(Op, Pid) when node(Pid) == node() ->
    apply(erlang, Op, [Pid]);
dolink(Op, [{_, Pid}]) when node(Pid) == node() ->
    apply(erlang, Op, [Pid]);
dolink(_,_) -> ok.

%% check_exit() removes the Pid from affected tables
%% This function needs to abcast the thingie since only the
%% local server is linked to the registerd process
%% all the other servers rely on the nodedown mechanism

check_exit(Tab,Pid) ->
    del_names(Tab, ets:tab2list(Tab), Pid).

del_names(_,[],Pid) -> done;

del_names(Tab, [{Name, Pid} |Tail], Pid) when atom(Name) ->
    ets:delete(Tab, Name),
    rpc:abcast(nodes(), global_name_server, {async_del, Name, Pid}),
    del_names(Tab, Tail, Pid);

del_names(Tab, [{{pre_alloc, Name}, Pid, Node} |Tail], Pid) ->
    ets:delete(Tab, {pre_alloc, Name}),
    rpc:abcast(nodes(), global_name_server, {async_del, Name, Pid}),
    del_names(Tab, Tail, Pid);

del_names(Tab, [_|T], Pid) ->
    del_names(Tab, T, Pid).
	       
%% Unregister all Name/Pid pairs such that node(Pid) == Node
do_node_down(Tab, Node) ->
    do_node_down(Tab, Node, ets:tab2list(Tab)).

do_node_down(Tab, Node, [{Name, Pid}|T]) when node(Pid) == Node ->
    ets:delete(Tab, Name),
    do_node_down(Tab, Node, T);

do_node_down(Tab, Node, [{{pre_alloc,Name}, _, Node}|T]) ->
    ets:delete(Tab, {pre_alloc, Name}),
    do_node_down(Tab, Node, T);

do_node_down(Tab,Node, [_|T]) ->
    do_node_down(Tab, Node, T);

do_node_down(_,_,[]) -> ok.

extract([{Name, Pid}|T])  when atom(Name) -> [Name |extract(T)];
extract([_|T]) -> extract(T);
extract([]) -> [].


