%% Copyright (C) 1990, Ellemtel Telecommunications Systems Laboratories
%% File    : io_lib.erl
%% Author  : Robert Virding
%% Purpose : Library of useful i/o functions.

%% Modifications:
%%   930326 -- Joe Armstrong
%%      Changed {number, N} wrappers to
%%      {integer, N} or {float, N}

%% This module is a library of useful i/o functions. It is hoped that the
%% functions defined in it are basic enough to be used without modification
%% as components of more complex utilities.

%% It is completely self-contained and uses no other modules. Its own
%% utilities are exported.

%% Most of the code here is derived from the original prolog versions and
%% from similar code written by Joe Armstrong and myself.

-module(io_lib).

-export([fwrite/2,fread/2,fread/3,format/2]).
-export([write/1,write/3,nl/0,write_string/1,write_string/2,write_char/1]).
-export([char_list/1,deep_char_list/1,printable_list/1]).
-export([scan/1,scan/2,scan/3,reserved_word/1]).
-export([collect_chars/3,collect_line/2]).

%% Backward compatibility functions.

scan(Cont, Chars, Pos) -> erl_scan:tokens(Cont, Chars, Pos).

scan(Chars) -> erl_scan:string(Chars).

scan(Chars, StartPos) -> erl_scan:string(Chars, StartPos).

reserved_word(Atom) -> erl_scan:reserved_word(Atom).

%% Interface calls to sub-modules.

fwrite(Format, Args) ->
    io_lib_format:fwrite(Format, Args).

fread(Chars, Format) ->
    io_lib_format:fread(Chars, Format).

fread(Cont, Chars, Format) ->
    io_lib_format:fread(Cont, Chars, Format).

format(Format, Args) ->
    io_lib_format:fwrite(Format, Args).

%% write(Term)
%% write(Term, Depth, Pretty)
%%  Return a (non-flattened) list of characters giving a printed
%%  representation of the term.

write(Term) -> write(Term, -1, false).

write([], D, P) ->
    "[]";
write(Term, D, P) when integer(Term) ->
    integer_to_list(Term);
write(Term, D, P) when float(Term) ->
    io_lib_format:fwrite_g(Term, none, none, none); %We KNOW
write(Atom, D, P) when atom(Atom) ->
    List = atom_to_list(Atom),
    case quote_atom(Atom, List) of
	true ->
	    write_string(List, $');
	false ->
	    List
    end;
write(Term, D, P) when port(Term) ->
    "#Port";
write(Term, D, P) when pid(Term) ->
    pid_to_list(Term);
write(Term, D, P) when reference(Term) ->
    "#Ref";
write(Term, D, P) when binary(Term) ->
    "#Bin";
write({'$var',Name}, D, P) ->			%Symmetric with read (in parse)
    atom_to_list(Name);
write({}, D, P) ->
    "{}";
write(Term, 0, P) ->
    "...";
write([H|T], D, true) ->
    case printable_list([H|T]) of
	true -> write_string(lists:flatten([H|T]), $");
	false -> [$[,write(H, D-1, true)|write_tail(T, D-1, true)]
    end;
write([H|T], D, false) ->
    [$[,write(H, D-1, false)|write_tail(T, D-1, false)];
write(T, D, P) when tuple(T) ->
    [${,write(element(1, T), D-1, P)|write_tuple(T, 2, size(T), D-1, P)].

%% write_tail(List, Depth, Pretty)
%%  Test the terminating case first as this looks better with depth.

write_tail([], D, P) ->
    "]";
write_tail(List, 0, P) ->
    "|...]";
write_tail([H|T], D, P) ->
    [$,,write(H, D-1, P)|write_tail(T, D-1, P)];
write_tail(Other, D, P) ->
    [$|,write(Other, D-1, P),$]].

%% write_tuple(Tuple, Index, Size, Depth, Pretty)
%%  Test the terminating case first as this looks better with depth.

write_tuple(T, I, Size, D, P) when I > Size ->
    "}";
write_tuple(T, I, Size, 0, P) ->
    "|...}";
write_tuple(T, I, Size, D, P) ->
    [$,,write(element(I, T), D-1, P)|write_tuple(T, I+1, Size, D-1, P)].

%% quote_atom(Atom, CharList)
%%  Return 'true' if atom with chars in CharList needs to be quoted, else
%%  return 'false'.

quote_atom(Atom, Cs0) ->
    case reserved_word(Atom) of
	true -> true;
	false ->
	    case Cs0 of
		[C|Cs] when C >= $a, C =< $z ->
		    quote_atom(Cs);
		_ -> true
	    end
    end.

quote_atom([C|Cs]) when C >= $a, C =< $z ->
    quote_atom(Cs);
quote_atom([C|Cs]) when C >= $A, C =< $Z ->
    quote_atom(Cs);
quote_atom([C|Cs]) when C >= $0, C =< $9 ->
    quote_atom(Cs);
quote_atom([$_|Cs]) ->
    quote_atom(Cs);
quote_atom([$@|Cs]) ->
    quote_atom(Cs);
quote_atom([_|_]) ->
    true;
quote_atom([]) ->
    false.

%% PrintList = write_string(CharList)
%%  Generate the list of characters needed to print a string.

write_string(S) ->
    write_string(S, $").

write_string(S, Q) ->
    [Q|write_string1(S, Q)].

write_string1([], Q) ->
    [Q];
write_string1([C|Cs], Q) ->
    write_char(C, Q, write_string1(Cs, Q)).

%% PrintList = write_char(Char)
%%  Generate the list of characters needed to print a character constant.

write_char(C) when C >= 0, C =< 255 ->
    [$$|write_char(C, -1, [])].

write_char(Q, Q, Tail) ->			%Must check this first
    [$\\,Q|Tail];
write_char($\\, _, Tail) ->			%In printable character range
    [$\\,$\\|Tail];
write_char(C, _, Tail) when C >= $ , C =< $~ ->
    [C|Tail];
write_char($\n, Q, Tail) ->			%\n = LF
    [$\\,$n|Tail];
write_char($\r, _, Tail) ->			%\r = CR
    [$\\,$r|Tail];
write_char($\t, _, Tail) ->			%\t = TAB
    [$\\,$t|Tail];
write_char($\v, _, Tail) ->			%\v = VT
    [$\\,$v|Tail];
write_char($\b, _, Tail) ->			%\b = BS
    [$\\,$b|Tail];
write_char($\f, _, Tail) ->			%\f = FF
    [$\\,$f|Tail];
write_char($\e, _, Tail) ->			%\e = ESC
    [$\\,$e|Tail];
write_char($\d, _, Tail) ->			%\d = DEL
    [$\\,$d|Tail];
write_char(C, _, Tail) when C < $  ->
    C1 = (C bsr 3) + $0,
    C2 = (C band 7) + $0,
    [$\\,$0,C1,C2|Tail];
write_char(C, _, Tail) when C > $~ ->
    C1 = (C bsr 6) + $0,
    C2 = ((C bsr 3) band 7) + $0,
    C3 = (C band 7) + $0,
    [$\\,C1,C2,C3|Tail].

%% char_list(CharList)
%% deep_char_list(CharList)
%%  Return true if CharList is a (possibly deep) list of characters, else
%%  false.

char_list([C|Cs]) when integer(C), C >= 0, C =< 255 ->
    char_list(Cs);
char_list([]) -> true;
char_list(Other) -> false.			%Everything else is false

deep_char_list(Cs) ->
    deep_char_list(Cs, []).

deep_char_list([C|Cs], More) when list(C) ->
    deep_char_list(C, [Cs|More]);
deep_char_list([C|Cs], More) when integer(C), C >= 0, C =< 255 ->
    deep_char_list(Cs, More);
deep_char_list([], [Cs|More]) ->
    deep_char_list(Cs, More);
deep_char_list([], []) -> true;
deep_char_list(Other, More) ->			%Everything else is false
    false.

%% printable_list(CharList)
%%  Return true if CharList is a list of printable characters, else
%%  false.

printable_list([C|Cs]) when integer(C), C >= $ , C =< 255 ->
    printable_list(Cs);
printable_list([$\n|Cs]) ->
    printable_list(Cs);
printable_list([$\r|Cs]) ->
    printable_list(Cs);
printable_list([$\t|Cs]) ->
    printable_list(Cs);
printable_list([$\v|Cs]) ->
    printable_list(Cs);
printable_list([$\b|Cs]) ->
    printable_list(Cs);
printable_list([$\f|Cs]) ->
    printable_list(Cs);
printable_list([$\e|Cs]) ->
    printable_list(Cs);
printable_list([]) -> true;
printable_list(Other) -> false.			%Everything else is false

%% List = nl()
%%  Return a list of characters to generate a newline.

nl() ->
    "\n".

%%
%% Utilities for collecting characters in input files
%%

%% collect_chars(Continuation, MoreChars, Count)
%%  Returns:
%%	{done,Result,RestChars}
%%	{more,Continuation}

collect_chars([], Chars, N) ->
    collect_chars1(N, Chars, []);
collect_chars({Left,Sofar}, Chars, N) ->
    collect_chars1(Left, Chars, Sofar).

collect_chars1(N, Chars, Stack) when N =< 0 ->
    {done,lists:reverse(Stack, []),Chars};
collect_chars1(N, [C|Rest], Stack) ->
    collect_chars1(N-1, Rest, [C|Stack]);
collect_chars1(N, eof, []) ->
    {done,eof,[]};
collect_chars1(N, eof, Stack) ->
    {done,lists:reverse(Stack, []),[]};
collect_chars1(N, [], Stack) ->
    {more,{N,Stack}}.

%% collect_line(Continutation, MoreChars)
%%  Returns:
%%	{done,Result,RestChars}
%%	{more,Continuation}

collect_line([], Chars) ->
    collect_line1(Chars, []);
collect_line({SoFar}, More) ->
    collect_line1(More, SoFar).

collect_line1([$\n|Rest], Stack) ->
    {done,lists:reverse([$\n|Stack], []),Rest};
collect_line1([C|Rest], Stack) ->
    collect_line1(Rest, [C|Stack]);
collect_line1(eof, []) ->
    {done,eof,[]};
collect_line1(eof, Stack) ->
    {done,lists:reverse(Stack, []),[]};
collect_line1([], Stack) ->
    {more,{Stack}}.
