%% File    : io_lib_format.erl
%% Author  : Robert Virding
%% Purpose : Formatting functions of io library.

-module(io_lib_format).

-export([fwrite/2,fwrite_g/4,fread/2,fread/3]).

-import(string, [chars/2,chars/3]).

%% fwrite(Format, ArgList)
%%  Format the arguments in ArgList after string Format. Just generate
%%  an error if there is an error in the arguments.

fwrite(Format, Args) when atom(Format) ->
    fwrite(atom_to_list(Format), Args);
fwrite(Format, Args) ->
    fwrite(Format, Args, []).

fwrite([$~|Format0], Args0, Tail) ->
    {List,Format,Args} = fwrite_format(Format0, Args0),
    [List|fwrite(Format, Args, Tail)];
fwrite([C|Format], Args, Tail) ->
    [C|fwrite(Format, Args, Tail)];
fwrite([], [], Tail) ->
    Tail.

fwrite_format(Fmt0, Args0) ->
    {F,A,Fmt1,Args1} = fwrite_width(Fmt0, Args0),
    {P,Fmt2,Args2} = fwrite_precision(Fmt1, Args1),
    {Pc,Fmt3,Args3} = fwrite_pad(Fmt2, Args2),
    fwrite_control(Fmt3, Args3, F, P, Pc, A).

fwrite_width([$-|Fmt0], Args0) ->
    {F,Fmt,Args} = fwrite_field(Fmt0, Args0),
    fwrite_width(-F, Fmt, Args);
fwrite_width(Fmt0, Args0) ->
    {F,Fmt,Args} = fwrite_field(Fmt0, Args0),
    fwrite_width(F, Fmt, Args).

fwrite_width(F, Fmt, Args) when F < 0 ->
    {-F,left,Fmt,Args};
fwrite_width(F, Fmt, Args) when F >= 0 ->
    {F,right,Fmt,Args}.

fwrite_precision([$.|Fmt], Args) ->
    fwrite_field(Fmt, Args);
fwrite_precision(Fmt, Args) ->
    {none,Fmt,Args}.

fwrite_field([$*|Fmt], [A|Args]) when integer(A) ->
    {A,Fmt,Args};
fwrite_field([C|Fmt], Args) when C >= $0, C =< $9 ->
    fwrite_field([C|Fmt], Args, 0);
fwrite_field(Fmt, Args) ->
    {none,Fmt,Args}.

fwrite_field([C|Fmt], Args, F) when C >= $0, C =< $9 ->
    fwrite_field(Fmt, Args, 10*F + (C - $0));
fwrite_field(Fmt, Args, F) ->		%Default case
    {F,Fmt,Args}.

fwrite_pad([$.,$*|Fmt], [A|Args]) ->
    {A,Fmt,Args};
fwrite_pad([$.,Pc|Fmt], Args) ->
    {Pc,Fmt,Args};
fwrite_pad(Fmt, Args) ->
    {$ ,Fmt,Args}.

fwrite_control([C|Fmt], Args0, F, P, Pc, A) ->
    {List,Args} = fwrite1(C, Args0, F, P, Pc, A),
    {List,Fmt,Args}.

%% fwrite1(FormatChar, Arguments, FieldWidth, Precision, PadChar, Adjust) ->
%%	{CharList,RestArgs}
%%  This is the main dispatch function for the various formatting commands.
%%  Field widths and precisions have already been calculated.

fwrite1($w, [A|Args], F, P, Pc, Adj) ->
    {fwrite_term(io_lib:write(A, -1, false), F, P, Pc, Adj),Args};
fwrite1($p, [A|Args], F, P, Pc, Adj) ->
    {fwrite_term(io_lib:write(A, -1, true), F, P, Pc, Adj),Args};
fwrite1($W, [A,Depth|Args], F, P, Pc, Adj) when integer(Depth) ->
    {fwrite_term(io_lib:write(A, Depth, false), F, P, Pc, Adj),Args};
fwrite1($P, [A,Depth|Args], F, P, Pc, Adj) when integer(Depth) ->
    {fwrite_term(io_lib:write(A, Depth, true), F, P, Pc, Adj),Args};
fwrite1($s, [A|Args], F, P, Pc, Adj) when atom(A) ->
    {fwrite_string(atom_to_list(A), F, P, Pc, Adj),Args};
fwrite1($s, [L|Args], F, P, Pc, Adj) when list(L) ->
    true = io_lib:deep_char_list(L),		%Check if L a character list
    {fwrite_string(L, F, P, Pc, Adj),Args};
fwrite1($e, [A|Args], F, P, Pc, Adj) when float(A) ->
    {fwrite_e(A, F, P, Pc, Adj),Args};
fwrite1($f, [A|Args], F, P, Pc, Adj) when float(A) ->
    {fwrite_f(A, F, P, Pc, Adj),Args};
fwrite1($g, [A|Args], F, P, Pc, Adj) when float(A) ->
    {fwrite_g(A, F, P, Pc, Adj),Args};
fwrite1($c, [A|Args], F, P, Pc, Adj) when integer(A) ->
    {fwrite_char(A band 255, F, P, Pc, Adj),Args};
fwrite1($n, Args, F, P, Pc, Adj) ->
    {fwrite_char($\n, F, P, Pc, Adj),Args};
fwrite1($~, Args, F, P, Pc, Adj) ->
    {fwrite_char($~, F, P, Pc, Adj),Args};
fwrite1($i, [_|Args], F, P, Pc, Adj) ->
    {[],Args}.

%% fwrite_term(TermList, Field, Precision, PadChar, Adjust)

fwrite_term(T, none, none, Pc, A) ->
    T;
fwrite_term(T, none, P, Pc, A) ->
    fwrite_term(T, P, P, Pc, A);
fwrite_term(T, F, none, Pc, A) ->
    fwrite_term(T, F, min(flat_length(T), F), Pc, A);
fwrite_term(T, F, P, Pc, A) when F >= P ->
    L = flat_length(T),
    if
	L > P ->
	    adjust(chars($*, P), chars(Pc, F - P), A);
	true ->
	    adjust(T, chars(Pc, F - L), A)
    end.

%% fwrite_e(Float, Field, Precision, PadChar, Adjust)

fwrite_e(Fl, F, P, Pc, A) ->
    fwrite_e(Fl, F, P, Pc, A, true).

fwrite_e(Fl, none, none, Pc, A, Pad) ->		%Default values
    fwrite_e(Fl, none, 6, Pc, A, Pad);
fwrite_e(Fl, none, P, A, Pc, Pad) when P >= 2 ->
    float_e(Fl, float_data(Fl), P);
fwrite_e(Fl, F, none, Pc, A, Pad) ->
    fwrite_e(Fl, F, 6, Pc, A, Pad);
fwrite_e(Fl, F, P, Pc, A, Pad) when P >= 2 ->
    L = length(Fs = float_e(Fl, float_data(Fl), P)),
    if
	L > F ->
	    chars($*, F);
	true ->
	    adjust(Fs, chars(Pc, F - L), A)
    end.

float_e(Fl, Fd, P) when Fl < 0.0 ->		%Negative numbers
    [$-|float_e(-Fl, Fd, P)];
float_e(Fl, {Ds,Dc,E}, P) when P < Dc ->
    case float_e(Ds, P, true, E) of
	{[$0|Fs],true} ->			%Carry all the way up
	    [$1,$.|Fs];
	{[F|Fs],false} ->
	    [F,$.|Fs]
    end;
float_e(Fl, {[D|Ds],Dc,E}, P) ->
    [D,$.|append(Ds, chars($0, P - Dc, float_exp(E)))].

%% float_e(Digits, Precision, RollOver, Exponent)
%%  Step down the digit list checking if we need to roll over exponent when
%%  rounding upwards, calculate "new" exponent string and go back up the
%%  digit string building number.

float_e([D|_], 0, true, E) when D >= $5 ->
    {float_exp(E + 1),true};
float_e([D|_], 0, R, E) when D >= $5 ->
    {float_exp(E),true};
float_e([D|_], 0, R, E) ->
    {float_exp(E),false};
float_e([$9|Ds], Dc, R, E) ->
    case float_e(Ds, Dc - 1, R, E) of
	{Fs,true} ->
	    {[$0|Fs],true};
	{Fs,false} ->
	    {[$9|Fs],false}
    end;
float_e([D|Ds], Dc, R, E) ->
    case float_e(Ds, Dc - 1, false, E) of
	{Fs,true} ->
	    {[D + 1|Fs],false};
	{Fs,false} ->
	    {[D|Fs],false}
    end.

float_exp(E) when E >= 0 ->
    [$e,$+|integer_to_list(E)];
float_exp(E) ->
    [$e|integer_to_list(E)].

%% fwrite_f(FloatData, Field, Precision, PadChar, Adjust)

fwrite_f(Fl, none, none, Pc, A) ->		%Default values
    fwrite_f(Fl, none, 6, Pc, A);
fwrite_f(Fl, none, P, Pc, A) when P >= 1 ->
    float_f(Fl, float_data(Fl), P);
fwrite_f(Fl, F, none, Pc, A) ->
    fwrite_f(Fl, F, 6, Pc, A);
fwrite_f(Fl, F, P, Pc, A) when P >= 1 ->
    L = length(Fs = float_f(Fl, float_data(Fl), P)),
    if
	L > F ->
	    chars($*, F);
	true ->
	    adjust(Fs, chars(Pc, F - L), A)
    end.

float_f(Fl, Fd, P) when Fl < 0.0 ->
    [$-|float_f(-Fl, Fd, P)];
float_f(Fl, {Ds,Dc,E}, P) when E < 0 ->
    case float_f_p(chars($0, -E - 1, Ds), P) of
	{Fs,true} ->
	    [$1,$.|Fs];
	{Fs,false} ->
	    [$0,$.|Fs]
    end;
float_f(Fl, {Ds,Dc,E}, P) ->
    case float_f_d(Ds, E + 1, P) of
	{Fs,true} ->
	    [$1|Fs];
	{Fs,false} ->
	    Fs
    end.

float_f_d(Ds, 0, P) ->
    {Fs,Carry} = float_f_p(Ds, P),
    {[$.|Fs],Carry};
float_f_d([$9|Ds], E, P) when E > 0 ->
    case float_f_d(Ds, E - 1, P) of
	{Fs,true} ->
	    {[$0|Fs],true};
	{Fs,false} ->
	    {[$9|Fs],false}
    end;
float_f_d([D|Ds], E, P) when E > 0 ->
    case float_f_d(Ds, E - 1, P) of
	{Fs,true} ->
	    {[D + 1|Fs],false};
	{Fs,false} ->
	    {[D|Fs],false}
    end;
float_f_d([], E, P) ->
    case float_f_d([], E - 1, P) of
	{Fs,true} ->
	    {[$1|Fs],false};
	{Fs,false} ->
	    {[$0|Fs],false}
    end.

float_f_p([D|Ds], 0) when D >= $5 ->
    {[],true};
float_f_p([D|Ds], 0) ->
    {[],false};
float_f_p([], P) ->				%Pad with 0's
    {chars($0, P),false};
float_f_p([$9|Ds], P) ->
    case float_f_p(Ds, P - 1) of
	{Fs,true} ->
	    {[$0|Fs],true};
	{Fs,false} ->
	    {[$9|Fs],false}
    end;
float_f_p([D|Ds], P) ->
    case float_f_p(Ds, P - 1) of
	{Fs,true} ->
	    {[D + 1|Fs],false};
	{Fs,false} ->
	    {[D|Fs],false}
    end.

float_data(Fl) ->
    float_data(float_to_list(Fl), [], 0).

float_data([$e|E], Ds, Dc) ->
    {reverse(Ds),Dc,list_to_integer(E)};
float_data([D|Cs], Ds, Dc) when D >= $0, D =< $9 ->
    float_data(Cs, [D|Ds], Dc + 1);
float_data([D|Cs], Ds, Dc) ->
    float_data(Cs, Ds, Dc).

%% fwrite_g(Float, Field, Precision, Adjust)
%% fwrite_g(Float, Field, Precision, PadChar, Adjust)
%%  Use the f form if Float is > 0.1 and < 10^4, else the e form.
%%  Precision always means the # of significant digits.

fwrite_g(Fl, F, P, A) ->
    fwrite_g(Fl, F, P, $ , A).

fwrite_g(Fl, F, none, Pc, A) ->
    fwrite_g(Fl, F, 6, Pc, A);
fwrite_g(Fl, F, P, Pc, A) when abs(Fl) < 0.1 ->
    fwrite_e(Fl, F, P, Pc, A);
fwrite_g(Fl, F, P, Pc, A) when abs(Fl) < 1.0 ->
    fwrite_f(Fl, F, P, Pc, A);
fwrite_g(Fl, F, P, Pc, A) when abs(Fl) < 10.0 ->
    fwrite_f(Fl, F, P-1, Pc, A);
fwrite_g(Fl, F, P, Pc, A) when abs(Fl) < 100.0 ->
    fwrite_f(Fl, F, P-2, Pc, A);
fwrite_g(Fl, F, P, Pc, A) when abs(Fl) < 1000.0 ->
    fwrite_f(Fl, F, P-3, Pc, A);
fwrite_g(Fl, F, P, Pc, A) when abs(Fl) < 10000.0 ->
    fwrite_f(Fl, F, P-4, Pc, A);
fwrite_g(Fl, F, P, Pc, A) ->
    fwrite_e(Fl, F, P, Pc, A).

%% fwrite_string(String, Field, Precision, PadChar, Adjust)

fwrite_string(S, none, none, Pc, A) ->
    S;
fwrite_string(S, F, none, Pc, A) ->
    fwrite_string(S, F, min(flat_length(S), F), Pc, A);
fwrite_string(S, none, P, Pc, A) ->
    string:left(flatten(S), P, Pc);
fwrite_string(S, F, P, Pc, A) when F >= P ->
    adjust(string:left(flatten(S), P, Pc), chars(Pc, F - P), A).

%% fwrite_char(Char, Field, Precision, PadChar, Adjust)

fwrite_char(C, none, none, Pc, A) ->
    C;
fwrite_char(C, F, none, Pc, A) ->
    chars(C, F);
fwrite_char(C, none, P, Pc, A) ->
    chars(C, P);
fwrite_char(C, F, P, Pc, A) when F >= P ->
    adjust(chars(C, P), chars(Pc, F - P), A).

adjust(Data, Pad, left) ->
    [Data,Pad];
adjust(Data, Pad, right) ->
    [Pad,Data].

%% fread(Continuation, CharList, FormatString)
%%  This is the main function into the re-entrant formatted reader. It
%%  repeatedly collects lines and calls fread/2 to format the input until
%%  all the format string has been used. 

fread([], Chars, Format) ->
    fread({[],Format,0,[]}, Chars, Format);
fread({Rest,RestFormat,N,Inputs}, MoreChars, Format) ->
    %%io:format("FREAD: ~w `~s'~n", [{Rest,RestFormat,N,Inputs},MoreChars]),
    fread_collect(MoreChars, [], Rest, RestFormat, N, Inputs).

fread_collect([$\r|More], Stack, Rest, RestFormat, N, Inputs) ->
    fread(RestFormat, append(Rest, reverse(Stack)), N, Inputs, More);
fread_collect([$\n|More], Stack, Rest, RestFormat, N, Inputs) ->
    fread(RestFormat, append(Rest, reverse(Stack)), N, Inputs, More);
fread_collect([C|More], Stack, Rest, RestFormat, N, Inputs) ->
    fread_collect(More, [C|Stack], Rest, RestFormat, N, Inputs);
fread_collect([], Stack, Rest, RestFormat, N, Inputs) ->
    {more,{reverse(Stack, Rest),RestFormat,N,Inputs}};
fread_collect(eof, Stack, Rest, RestFormat, N, Inputs) ->
    fread(RestFormat, append(Rest, append(reverse(Stack), eof)), N, Inputs, []).

fread(Format, Line, N0, Inputs0, More) ->
    %%io:format("FREAD1: `~s' `~s'~n", [Format,Line]),
    case fread(Format, Line, N0, Inputs0) of
	{ok,Input,Rest} ->
	    {done,{ok,Input},append(Rest, More)};
	{more,RestFormat,N,Inputs} ->
	    %% Don't forget the newline.
	    {more,{More,RestFormat,N+1,Inputs}};
	Other ->				%An error has occurred
	    {done,Other,More}
    end.

%% Conventions
%%   ~s 	String White terminated
%%   ~d         Integer terminated by ~[0-9]
%%   ~f         Float
%%   ~a         as ~s but converted to an atom
%%   ~c		characters without any stripping
%%   ~n		number of characters scanned
%%   WHITE      Skip white space
%%   X          Literal X

fread(Format, Line) ->
    fread(Format, Line, 0, []).

fread([$~|Format0], Line, N, Results) ->
    {Format,F,Sup} = fread_field(Format0),
    fread1(Format, F, Sup, Line, N, Results, Format0);
fread([$ |Format], Line, N, Results) ->
    fread_skip_white(Format, Line, N, Results);
fread([C|Format], [C|Line], N, Results) ->
    fread(Format, Line, N+1, Results);
fread([F|Format], [C|Line], N, Results) ->
    fread_error(input);
fread([], Line, N, Results) ->
    {ok,reverse(Results),Line}.

fread_skip_white(Format, [$ |Line], N, Results) ->
    fread_skip_white(Format, Line, N+1, Results);
fread_skip_white(Format, [$\t|Line], N, Results) ->
    fread_skip_white(Format, Line, N+1, Results);
fread_skip_white(Format, Line, N, Results) ->
    fread(Format, Line, N, Results).

%% fread_field(Format) 
%%  Reads the field specification paramters. Returns:
%%
%%	{RestFormat,FieldWidth,Suppress}

fread_field([$*|Format]) -> fread_field(Format, true);
fread_field(Format) -> fread_field(Format, false).

fread_field([C|Format], Sup) when C >= $0, C =< $9 ->
    fread_field(Format, C - $0, Sup);
fread_field(Format, Sup) -> {Format,none,Sup}.

fread_field([C|Format], F, Sup) when C >= $0, C =< $9 ->
    fread_field(Format, 10*F + C - $0, Sup);
fread_field(Format, F, Sup) ->
    {Format,F,Sup}.

%% fread1(Format, FieldWidth, Suppress, Line, N, Results, AllFormat)
%% fread1(Format, FieldWidth, Suppress, Line, N, Results)
%%  The main dispatch function for the formatting commands. Done in two
%%  stages so format commands that need no input can always be processed.

fread1([$l|Format], F, Sup, Line, N, Res, AllFormat) ->
    fread(Format, Line, N, fread_result(Sup, N, Res));
fread1(Format, F, Sup, [], N, Res, AllFormat) ->
    %% Need more input here.
    {more,[$~|AllFormat],N,Res};
fread1(Format, F, Sup, eof, N, [], AllFormat) ->
    %% This is at start of format string so no error.
    eof;
fread1(Format, F, Sup, eof, N, Res, AllFormat) ->
    %% This is an error as there is no more input.
    fread_error(input);
fread1(Format, F, Sup, Line, N, Res, AllFormat) ->
    fread1(Format, F, Sup, Line, N, Res).

fread1([$f|Format], none, Sup, Line0, N0, Res) ->
    {Line,N,Cs} = fread_float_cs(Line0, N0),
    fread_float(Cs, Sup, Format, Line, N, Res);
fread1([$f|Format], F, Sup, Line0, N, Res) ->
    {Line,Cs} = fread_chars(Line0, F),
    fread_float(Cs, Sup, Format, Line, N+F, Res);
fread1([$d|Format], none, Sup, Line0, N0, Res) ->
    {Line,N,Cs} = fread_int_cs(Line0, N0),
    fread_integer(Cs, Sup, Format, Line, N, Res);
fread1([$d|Format], F, Sup, Line0, N, Res) ->
    {Line,Cs} = fread_chars(Line0, F),
    fread_integer(Cs, Sup, Format, Line, N+F, Res);
fread1([$s|Format], none, Sup, Line0, N0, Res) ->
    {Line,N,Cs} = fread_string_cs(Line0, N0),
    fread_string(Cs, Sup, Format, Line, N, Res);
fread1([$s|Format], F, Sup, Line0, N, Res) ->
    {Line,Cs} = fread_chars(Line0, F),
    fread_string(Cs, Sup, Format, Line, N+F, Res);
fread1([$a|Format], none, Sup, Line0, N0, Res) ->
    {Line,N,Cs} = fread_string_cs(Line0, N0),
    fread_atom(Cs, Sup, Format, Line, N, Res);
fread1([$a|Format], F, Sup, Line0, N, Res) ->
    {Line,Cs} = fread_chars(Line0, F),
    fread_atom(Cs, Sup, Format, Line, N+F, Res);
fread1([$c|Format], none, Sup, Line0, N, Res) ->
    {Line,Cs} = fread_chars(Line0, 1),
    fread_chars(Cs, Sup, Format, Line, N+1, Res);
fread1([$c|Format], F, Sup, Line0, N, Res) ->
    {Line,Cs} = fread_chars(Line0, F),
    fread_chars(Cs, Sup, Format, Line, N+F, Res);
fread1([$~|Format], F, Sup, [$~|Line], N, Res) ->
    fread(Format, Line, N+1, Res);
fread1(Format, F, Sup, Line, N, Res) ->
    fread_error(format).

%% fread_float(FloatChars, Suppress, Format, Line, N, Results)

fread_float(Cs, Sup, Format, Line, N, Res) ->
    case catch list_to_float(fread_skip_white(reverse(Cs))) of
	{'EXIT',R} ->
	    fread_error(float);
	Float ->
	    fread(Format, Line, N, fread_result(Sup, Float, Res))
    end.

%% fread_integer(IntegerChars, Suppress, Format, Line, N, Results)

fread_integer(Cs, Sup, Format, Line, N, Res) ->
    case catch list_to_integer(fread_skip_white(reverse(Cs))) of
	{'EXIT',R} ->
	    fread_error(integer);
	Integer ->
	    fread(Format, Line, N, fread_result(Sup, Integer, Res))
    end.

%% fread_string(StringChars, Suppress, Format, Line, N, Results)

fread_string(error, Sup, Format, Line, N, Res) ->
    fread_error(string);
fread_string(Cs0, Sup, Format, Line, N, Res) ->
    Cs = fread_skip_white(reverse(fread_skip_white(Cs0))),
    fread(Format, Line, N, fread_result(Sup, Cs, Res)).

%% fread_atom(AtomChars, Suppress, Format, Line, N, Results)

fread_atom(error, Sup, Format, Line, N, Res) ->
    fread_error(atom);
fread_atom(Cs0, Sup, Format, Line, N, Res) ->
    Cs = fread_skip_white(reverse(fread_skip_white(Cs0))),
    fread(Format, Line, N, fread_result(Sup, list_to_atom(Cs), Res)).

%% fread_chars(Characters, Suppress, Format, Line, N, Results)

fread_chars(error, Sup, Format, Line, N, Res) ->
    fread_error(character);
fread_chars(Cs, Sup, Format, Line, N, Res) ->
    fread(Format, Line, N, fread_result(Sup, reverse(Cs), Res)).

%% fread_chars(Line, Count)

fread_chars(Line, C) ->
    fread_chars(C, Line, []).

fread_chars(0, Line, Cs) -> {Line,Cs};
fread_chars(N, [$\n|Line], Cs) -> {[$\n|Line],error};
fread_chars(N, [C|Line], Cs) ->
    fread_chars(N-1, Line, [C|Cs]);
fread_chars(N, [], Cs) -> {[],error}.

%% fread_int_cs(Line, N)

fread_int_cs(Line0, N0) ->
    {Line1,N1} = fread_skip_white(Line0, N0),
    {Line,N,Cs} = fread_sign(Line1, N1, []),
    fread_digits(Line, N, Cs).

%% fread_float_cs(Line, N)
%%  A float is "[+|-][0-9]+.[0-9]+[[E|e][+|-][09-]+]

fread_float_cs(Line0, N0) ->
    {Line1,N1} = fread_skip_white(Line0, N0),
    {Line2,N2,Cs2} = fread_sign(Line1, N1, []),
    {Line,N,Cs} = fread_digits(Line2, N2, Cs2),
    fread_float_cs_1(Line, N, Cs).

fread_float_cs_1([$.|Line0], N0, Cs0) ->
    {Line,N,Cs} = fread_digits(Line0, N0+1, [$.|Cs0]),
    fread_float_cs_2(Line, N, Cs);
fread_float_cs_1(Line, N, Cs) ->
    {Line,N,Cs}.

fread_float_cs_2([$e|Line0], N0, Cs0) ->
    {Line,N,Cs} = fread_sign(Line0, N0+1, [$e|Cs0]),
    fread_digits(Line, N, Cs);
fread_float_cs_2([$E|Line0], N0, Cs0) ->
    {Line,N,Cs} = fread_sign(Line0, N0+1, [$E|Cs0]),
    fread_digits(Line, N, Cs);
fread_float_cs_2(Line, N, Cs) ->
    {Line,N,Cs}.

%% fread_string_cs(Line, N)

fread_string_cs(Line0, N0) ->
    {Line,N} = fread_skip_white(Line0, N0),
    fread_skip_nonwhite(Line, N, []).

%% fread_skip_white(Line)
%% fread_skip_white(Line, N)
%% fread_skip_nonwhite(Line, N, Characters)
%% fread_sign(Line, N, Characters)
%% fread_digits(Line, N, Characters)
%%  Read segments of things, return "thing" characters in reverse order.

fread_skip_white([$ |Line]) -> fread_skip_white(Line);
fread_skip_white([$\t|Line]) -> fread_skip_white(Line);
fread_skip_white([$\n|Line]) -> fread_skip_white(Line);
fread_skip_white(Line) -> Line.

fread_skip_white([$ |Line], N) ->
    fread_skip_white(Line, N+1);
fread_skip_white([$\t|Line], N) ->
    fread_skip_white(Line, N+1);
fread_skip_white([$\n|Line], N) ->
    fread_skip_white(Line, N+1);
fread_skip_white(Line, N) -> {Line,N}.

fread_skip_nonwhite([$ |Line], N, Cs) -> {[$ |Line],N,Cs};
fread_skip_nonwhite([$\t|Line], N, Cs) -> {[$\t|Line],N,Cs};
fread_skip_nonwhite([$\n|Line], N, Cs) -> {[$\n|Line],N,Cs};
fread_skip_nonwhite([C|Line], N, Cs) ->
    fread_skip_nonwhite(Line, N+1, [C|Cs]);
fread_skip_nonwhite([], N, Cs) -> {[],N,Cs}.

fread_sign([$+|Line], N, Cs) -> {Line,N+1,[$+|Cs]};
fread_sign([$-|Line], N, Cs) -> {Line,N+1,[$-|Cs]};
fread_sign(Line, N, Cs) -> {Line,N,Cs}.

fread_digits([C|Line], N, Cs) when C >= $0, C =< $9 ->
    fread_digits(Line, N+1, [C|Cs]);
fread_digits(Line, N, Cs) -> {Line,N,Cs}.

%% fread_result(Suppress, Value, Results)

fread_result(true, V, Res) -> Res;
fread_result(false, V, Res) -> [V|Res].

fread_error(In) ->
    {error,{fread,In}}.

%%
%% Utilities
%%

min(L, R) when L < R -> L;
min(L, R) -> R.

append([H|T], Tail) ->
    [H|append(T, Tail)];
append([], Tail) ->
    Tail.

reverse(List) ->
    reverse(List, []).

reverse([], Stack) ->
    Stack;
reverse([H|T], Stack) ->
    reverse(T, [H|Stack]).

%% flatten(List)
%%  Flatten a list.

flatten(List) ->
    flatten(List, [], []).

flatten([H|T], Cont, Tail) when list(H) ->
    flatten(H, [T|Cont], Tail);
flatten([H|T], Cont, Tail) ->
    [H|flatten(T, Cont, Tail)];
flatten([], [H|Cont], Tail) ->
    flatten(H, Cont, Tail);
flatten([], [], Tail) ->
    Tail.

%% flat_length(List)
%%  Calculate the length of a list of lists.

flat_length(List) ->
    flat_length(List, 0).

flat_length([H|T], L) when list(H) ->
    flat_length(H, flat_length(T, L));
flat_length([H|T], L) ->
    flat_length(T, L + 1);
flat_length([], L) ->
    L.
