%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%% File     : net_adm.erl
%% Author   : Claes Wikstrom klacke@erix.ericsson.se
%% Date     : 9300915
%% Purpose  : Various network admin functions

-module(net_adm).
-export([host_file/0,
	 localhost/0,
	 ping_list/1,
	 world/0,world/1,
	 dns_hostname/1,
	 epmd/1,
	 ping/1]).

%% Kept here for compatibility purposes
-export([start_node/2, start_node/3, 
	 epmd/2,start_node/4]). 

%% Try to read .hosts.erlang file in 
%% 1. cwd , 2. $HOME 3. init:root_dir() 

host_file() ->
    Home0 = lib:nonl(unix:cmd("echo $HOME")),
    Home = case Home0 of
	       [] -> 
		   case init:get_flag('-HOME') of
		       {'-HOME', H} ->
			   H;
		       notfound ->
			   exit(nohome)
		   end;
	       Other ->
		   Home0
	   end,
    case file:path_consult([".", Home, code:root_dir()], ".hosts.erlang") of
	{ok, Hosts,_} -> Hosts;
	_ -> {error, "Cant find or parse file .hosts.erlang"}
    end.

	
%%% Check whether a node is up or down
%%% side effect: set up a connection to Node if there 
%%  not yet is one.

ping(Node)  when atom(Node) ->
    case auth:is_auth(Node) of
	yes -> pong;
	Other -> erlang:disconnect_node(Node), pang
    end.

localhost() -> epmd("-localhost").

dns_hostname(H) ->
    case epmd(lists:concat(["-ghbn ", H])) of
	[] -> {error, H};
	X -> {ok, X}
    end.


%% These funcs are kept here for compatibility reasons
start_node(Name, Host) ->
    slave:start(Host, Name).
start_node(Name, Host, local) ->
    D = lists:concat([localhost(),":0"]),
    slave:start(Host, Name, [{"DISPLAY", D}]);
start_node(Name, Host, remote) ->
    slave:start(Host, Name).
start_node(Name, Host,D0, Pseudoservers) ->
    D = case D0 of
	local -> [{"DISPLAY", lists:concat([localhost(),":0"])}];
	remote -> []
    end,
    slave:start(Host, Name,D, Pseudoservers).


%% a common situation in "life" is to have a configuration file with a list
%% of nodes, and then at startup, all nodes in the list are ping'ed
%% this can lead to no end of troubles if two disconnected nodes
%% simultaneously ping each other.
%% Use this function in order to do it safe,
%% It assumes a working global.erl which ensures a fully 
%% connected network.
%% Had the erlang runtime system been able to fully cope with 
%% the possibility of two simultaneous (unix) connects, this function would
%% merley  be lists:map({net_adm, ping}, [], Nodelist).
%% It is also assumed, that the same (identical) Nodelist is given to all
%% nodes which are to perform this call. (possibly simultaneously)
%% Even this code has a flaw, and that is the case where two
%% nodes simultaneously and without *any* other already 
%% running nodes execute this code. :-(


ping_list(Nodelist) ->
    net_kernel:monitor_nodes(true),
    Sofar = ping_first(Nodelist, nodes()),
    collect_new(Sofar, Nodelist).

ping_first([], S) ->
	[];

ping_first([Node|Nodes], S) ->
    case lists:member(Node, S) of
	true -> [Node | ping_first(Nodes, S)];
	false ->
	    case ping(Node) of
		pong -> [Node];
		pang -> ping_first(Nodes, S)
	    end
    end.
    
collect_new(Sofar, Nodelist) ->
    receive
	{nodeup, Node} ->
	    case lists:member(Node, Nodelist) of
		true ->
		    collect_new(Sofar, Nodelist);
		false ->
		    collect_new([Node | Sofar], Nodelist)
	    end
    after 3000 ->
	    net_kernel:monitor_nodes(false),
	    Sofar
    end.

%% This function polls a set of hosts according to a file called
%% .hosts.erlang that need to reside either in the current directory
%% or at your home directory ( The current dir is tried first )
%% world() returns a list of all nodes on the network that can be 
%% found (including ourselves) obs. $HOME variable is inspected
    
world() ->
    world(silent).

world(Verbose) ->
    case net_adm:host_file() of
        {error,R} -> exit({error, R});
        Hosts -> lists:flatten(collect_nodes(Hosts,Verbose))
    end.

collect_nodes([],_) -> [];
collect_nodes([Host|Tail],Verbose) ->
    case collect_host_nodes(Host,Verbose) of
        nil ->
            collect_nodes(Tail,Verbose);
        L ->
            [L|collect_nodes(Tail,Verbose)]
    end.
collect_host_nodes(Host,Verbose) ->
    Chars = epmd(lists:concat(["-names ",Host])),
    case Chars of
        [$e,$p,$m,$d|_] -> 
            ok_chars(Chars,Host,Verbose);
        _ ->
            nil
    end.

ok_chars(Chars,Host,Verbose) ->
    {Line,More} = getline(Chars),
    do_ping(getline(More),atom_to_list(Host),Verbose).

do_ping([],Host,_) -> [];
do_ping({[],Cont},Host,Verbose) -> 
    do_ping(getline(Cont),Host,Verbose);
do_ping(done,Host,_) -> [];
do_ping({Line,Cont},Host,Verbose) ->
    Node = list_to_atom(lists:concat([first_word(Line),'@',longshort(Host)])),
    case Verbose of
        verbose -> io:format("Pinging ~w  ",[Node]);
        _ -> ok
    end,
    case net:ping(Node) of
        pong ->
            case Verbose of
                verbose ->
                    io:format("pong from  ~w ~n",[Node]);
                _ ->
                    ok
            end,
            [Node | do_ping(getline(Cont),Host,Verbose)];
        pang -> 
           do_ping(getline(Cont),Host,Verbose)
    end.
         
longshort(Host) ->
    case net_kernel:longnames() of
	false ->
	    uptodot(Host);
	true ->
	    Host
    end.

uptodot([$.|T]) -> [];
uptodot([])-> [];
uptodot([H|T]) -> [H|uptodot(T)].
	
first_word([]) -> [];
first_word([H|T]) ->
    case H of
        $  -> [];
        _ -> [H| first_word(T)]
    end.

getline(Chars) ->
    getline2(Chars,[]).

getline2([10|Tail],Line) ->
    {lists:reverse(Line),Tail};
getline2([],Line) ->
    done;
getline2([X|Tail],Line) ->
    getline2(Tail,[X|Line]).
    
%% Call 'epmd Chars'

epmd(Chars) ->
    P = spawn(net_adm, epmd, [self(), Chars]),
    receive
	{P, Reply} -> Reply
    end.

epmd(Top, Chars) ->
    process_flag(trap_exit, true),
    P = open_port({spawn, lists:append("epmd ", Chars)}, [stream, eof]),
    Top ! {self(), lists:flatten(epmd_collect(P))}.

epmd_collect(P) ->
    receive 
	{P, {data, Bytes}} ->
	    [Bytes | epmd_collect(P)];
	{P, eof} ->
	    []
    end.
