%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%% File     : net_kernel.erl
%% Author   : Claes Wikstrom klacke@erix.ericsson.se
%% Date     : 9300915 (Modified 941101)
%% Purpose  : net_kernel 

-module(net_kernel).

%% User Interface Exports
-export([start/1, stop/0,
	 kernel_apply/3,
	 monitor_nodes/1,
	 longnames/0,
	 allow/1]).


%% Internal Exports 
-export([do_spawn_link/5, 
	 ticker/2,
	 do_nodeup/3,
	 kernel0/4]).

	 
-import(error_logger,[error_msg/2]).

%% This is the main startup routine for net_kernel
%% The defaults are longnames and a ticktime of 10 secs to the tcp_drv.

start([Name]) ->
    start([Name, longnames]);

start([Name, LongOrShortNames]) ->
    start([Name, LongOrShortNames, 15000]);

start([Name, LongOrShortNames, Ticktime]) ->
    start_kernel(Name, LongOrShortNames, Ticktime),
    ddd:start(),        %% May or may not be allready started 
    global:start(),     %% May or may not be allready started 
    auth:start(),
    rpc:start(),
    started.


%% The central spawning facility, All requests to spawn a process 
%% Or to send to a registered process goes
%% through this server.
%% This server need to be up for distributed erlang to work.
%%

start_kernel(Name, LongOrShortNames, Ticktime) ->
    register(net_kernel, 
	     K = spawn(net_kernel,kernel0,[self(),Name, 
					   LongOrShortNames, Ticktime])),
    receive
	{net_kernel, continue} -> ok;
	{net_kernel, error} -> exit(nodistribution)
    end.

kernel0(Top,Name, LongOrShortNames, Ticktime) ->
    put(long_short, LongOrShortNames),
    process_flag(trap_exit,true),
    P = open_port({spawn,tcp},[{packet, 4}]),
    erlang:alive(Name, P),  % a == ALIVE_REQUEST
    Cmd = lists:concat(["a", Name, " ", LongOrShortNames,[0]]),
    P ! {self(), {command, Cmd}},
    receive
	{P, {data, "alive"}} ->
	    spawn_link(net_kernel, ticker, [self(), Ticktime]),
	    Top ! {net_kernel, continue},
	    process_flag(priority, max),
	    kernel(P, std_monitors());
	{P, {data, "notalive"}} ->
	    Top ! {net_kernel, error};
	{'EXIT', P, _} ->
	    Top ! {net_kernel, error}
    end.


std_monitors() -> [ddd_server, global_name_server].

kernel(P, Monitor) ->
    receive
	{From,registered_send,To,Mess} when pid(From) ->
	    send(From,To,Mess),
	    kernel(P, Monitor);

	{From,spawn,M,F,A,Gleader} when pid(From) ->      %% spawning
	    From ! {spawn_reply,Id = catch spawn(M,F,A)},
	    group_leader(Gleader,Id),
	    kernel(P, Monitor);
	{From,spawn_link,M,F,A,Gleader} ->
	    catch spawn(net_kernel,do_spawn_link,[From,M,F,A,Gleader]),
	    kernel(P, Monitor);

	{nodeup, Node} ->
	    spawn(net_kernel, do_nodeup, [get(allowed), Monitor, Node]),
	    kernel(P, Monitor);

	{nodedown, Node} ->
	    case get(allowed) of
		undefined ->
		    send_list(Monitor, {nodedown, Node});
		_ -> 
		    ignore
	    end,
	    kernel(P, Monitor);

	{From, {is_auth, Node}} ->                     %% authentication
	    From ! {auth_reply,node(),yes},
	    kernel(P, Monitor);
	{From, badcookie, net_kernel, {is_auth, Node}} ->
	    From ! {auth_reply,node(),no},
	    kernel(P, Monitor);

	%% Pass on all messages to the auth server allthough it's a badcookie
	{From, badcookie, auth, Mess} ->
	    safesend(auth, Mess),
	    kernel(P, Monitor);

	{From, badcookie, To ,Mess} ->
	    safesend(auth, {From, badcookie, To ,Mess}),
	    kernel(P, Monitor);
	{From,{apply,Mod,Fun,Args}} when pid(From),node(From) == node() ->
	    From ! {net_kernel, ok},
	    catch apply(Mod,Fun,[P|Args]),
	    kernel(P, Monitor);
	{'EXIT', P, _}  when port(p) ->
	    error_msg("** Netkernel terminating ... **\n", []),
	    exit(no_network);

        %% For restricted systems ......

	{From, {allow, Nodes}} ->
	    case all_atoms(Nodes) of
		true ->
		    case get(allowed) of
			undefined -> 
			    put(allowed, Nodes);
			List ->
			    put(allowed, lists:append(List, Nodes))
		    end,
		    From ! {net_kernel, ok};
		false ->
		    From ! {net_kernel, error}
	    end,
	    kernel(P, Monitor);

	%% Used to monitor all changes in the nodes list

	{From, {monitor_nodes, true}} ->
	    link(From),
	    From ! {net_kernel, ok},
	    kernel(P, [From | Monitor]);
	{From, {monitor_nodes, false}} ->
	    unlink(From),
	    From ! {net_kernel, ok},
	    kernel(P, lists:delete(From, Monitor));
	{From, {set_monitors, L}} ->
	    From ! {net_kernel, done},
	    kernel(P, L);
	{'EXIT', From, _} when pid(From) ->
	    kernel(P, delete_all(From, Monitor));

	{From, stop} ->
	    P ! {self(), close},
	    receive
		{P, closed} ->
		    global:stop(),
		    ddd:stop(),
		    rpc:stop(),  
		    auth:stop(),
		    stop_dist(nodes(), Monitor),
		    From ! {net_kernel, stopped},
		    exit(normal)
	    end;
	tick ->
	    P ! {self(), {command, "T"}},
	    kernel(P, Monitor);
	{From, longnames} ->
	    case get(long_short) of
		longnames -> 
		    From ! {net_kernel, true};
		_ ->
		    From ! {net_kernel, false}
	    end,
	    kernel(P, Monitor);
	X -> 
	    error_msg("Net kernel got ~w~n",[X]),
	    kernel(P, Monitor)
    end.

kernel_apply(M,F,A) ->         request({apply,M,F,A}).
allow(Nodes) ->                request({allow, Nodes}).
monitor_nodes(How)   ->        request({monitor_nodes, How}).
stop() ->                      request(stop).
longnames() ->                 request(longnames).

delete_all(From, [From |Tail]) -> delete_all(From, Tail);
delete_all(From, [H|Tail]) ->  [H|delete_all(From, Tail)];
delete_all(_, []) -> [].

all_atoms([]) -> true;
all_atoms([N|Tail]) when atom(N) ->
    all_atoms(Tail);
all_atoms(_) -> false.

do_nodeup(Allowed , Monitor, Node) ->
    receive
	after 1000 -> ok   %% sleep a sec, 
    end,	 
    case lists:member(Node, nodes()) of
	false -> exit(normal);
	true -> ok
    end,
    case auth:is_auth(Node) of
	yes -> ok;
	Other -> exit(normal)
    end,
    case Allowed of
	undefined ->
	    send_list(Monitor, {nodeup, Node});
	L ->
	    case lists:member(Node, L) of
		true ->
		    send_list(Monitor, {nodeup, Node});
		false ->
		    error_msg("** Connection attempt with/from ", []),
		    error_msg("disallowed node ~w ** ~n", [Node]),
		    erlang:disconnect_node(Node)
	    end
    end.


%% If the net_kernel isn't running we ignore all requests to the 
%% kernel, thus basically accepting them :-)
request(Req) ->
    case whereis(net_kernel) of
	P when pid(P) ->
	    P ! {self(), Req},
	    receive
		{net_kernel, Reply} -> Reply
	    end;
	Other -> ignored
    end.

stop_dist([], _) -> ok;
stop_dist([Node|Nodes], Monitor) ->
    send_list(Monitor, {nodedown, Node}),
    stop_dist(Nodes, Monitor).

ticker(Kernel, Tick) ->
    process_flag(priority, max),
    ticker1(Kernel, to_integer(Tick)).

to_integer(T) when integer(T) -> T;
to_integer(T) when atom(T) -> 
    list_to_integer(atom_to_list(T)).

ticker1(Kernel, Tick) ->
    receive
	after Tick -> 
		Kernel ! tick,
		ticker1(Kernel, Tick)
    end.

send(From,To,Mess) ->
    case whereis(To) of
	undefined ->
	    auth:print(node(From),
		'** Undefined registered name {~w,~w} when sending ~w **~n',
		 [To,node(), Mess]);
	P when pid(P) ->
	    P ! Mess
    end.

safesend(Name,Mess) when atom(Name) ->
    case whereis(Name) of 
	undefined ->
	    error_msg(
	       '** Undefined registered name ~w when sending ~w **~n',
	       [Name,Mess]);
	P when pid(P) ->
	    P ! Mess
    end;
safesend(Pid, Mess) -> Pid ! Mess.

send_list([P|T], M) -> safesend(P, M), send_list(T, M);
send_list([], _) -> ok.

%% This code is really intricate. The link will go first and then comes
%% the pid, This means that the client need not do a network link.
%% If the link message would not arrive, the runtime system  shall
%% generate a nodedown message

do_spawn_link(From,M,F,A,Gleader) ->
    link(From),
    From ! {spawn_link_reply,self()},  %% ahhh
    group_leader(Gleader,self()),
    apply(M,F,A).

