%% Copyright (C) 1993, Ellemtel Telecommunications Systems Laboratories
%% Author: Claes Wikstrom  klacke@erix.ericsson.se
%% Purpose:  Supply a computational pool of processors 
%% The chief user interface function here is get_node()
%% Which returns the name of the nodes in the pool
%% with the least load !!!!
%% This function is callable from any node including the master
%% That is part of the pool
%% nodes are scheduled on a per usgae basis and per load basis,
%% Whenever we use a node, we put at the end of the queue, and whenever
%% a node report a change in load, we insert it accordingly

%% Last Changed: March 1995


-module(pool).

% User interface Exports ...
-export([start/1, 
	 start/2, 
	 stop/0,
	 get_nodes/0, 
	 get_nodes_and_load/0, 
	 get_node/0,
	 pspawn/3, 
	 attach/1,
	 pspawn_link/3, 
	 get_node/0]).

%% Internal Exports 
-export([statistic_collector/0,
	 start_it/0,
	 master/2,
	 do_spawn/4]).
	 
-import(generic, [request/2]).

%% User interface 

%% Start up using the .hosts.erlang file
start(Name) ->    
    start(Name,[]).
start(Name, local) -> %% For compatibility with 4.2
    start(Name, [{"DISPLAY", lists:concat([net_adm:localhost(), ":0"])}]);
start(Name, remote) ->  %% For compatibily with 4.2
    start(Name, []);
start(Name,Env) when atom(Name) ->
    generic:start({global, pool_master}, pool, start_it, []),
    Hosts = net_adm:host_file(),
    Nodes = start_nodes(Hosts, Name, Env),
    lists:map({pool,attach}, [], Nodes),
    Nodes.

%%
%% Interface functions ...
%%
get_nodes() ->          get_elements(2,get_nodes_and_load()).
attach(Node) ->         request({global, pool_master}, {attach, Node}).
get_nodes_and_load() -> request({global, pool_master},get_nodes).
get_node() ->           request({global, pool_master},get_node).
pspawn(M,F,A) ->        request({global, pool_master}, 
				{spawn, group_leader(), M,F,A}).
pspawn_link(M,F,A) ->   P = pspawn(M,F,A),link(P), P.

start_nodes([], _, _) -> [];
start_nodes([Host|Tail], Name, Env) -> 
    case slave:start(Host, Name, Env) of 
	{error, R} ->
	    io:format("Can't start node on host ~w due to ~w~n",[Host, R]),
	    start_nodes(Tail, Name, Env);
	{ok, Node} -> 
	    [Node | start_nodes(Tail, Name, Env)]
    end.

stop() -> generic:stop({global, pool_master}).

get_elements(Pos,[]) -> [];
get_elements(Pos,[E|T]) -> [element(Pos,E) | get_elements(Pos,T)].

stop_em([]) -> stopped;
stop_em([N|Tail]) ->
    rpc:cast(N,erlang,halt,[]),
    stop_em(Tail).

start_it() ->
    process_flag(trap_exit, true),
    spawn_link(pool, statistic_collector, []),
    generic:sync_start(pool_master, started),
    generic:loop(pool_master, [{0, node()}], {pool, master}).

master({Node,load,Load}, Nodes) ->
    Nodes2 = insert_node({Load,Node},Nodes),
    {noreply, Nodes2};

master({nodedown, Node}, Nodes) ->
    {noreply, lists:keydelete(Node,2,Nodes)};

master({From,get_nodes} , Nodes)->
    {reply, From, Nodes, Nodes};

master({From,get_node}, [{Load,N}|Tail]) ->
    {reply, From, N,lists:append(Tail, [{Load+1, N}])};

master({From, {attach, Node}}, Nodes) ->
    case lists:keysearch(Node,2,Nodes) of
	{value,_} ->
	    {reply, From, allready_attached, Nodes};
	false ->
	    erlang:monitor_node(Node, true),
	    spawn_link(Node, pool, statistic_collector, []),
	    {reply, From, attached, lists:append(Nodes,[{999999,Node}])}
    end;

master({From, {spawn,Gl, M, F, A}}, Nodes) ->
    [{Load,N}|Tail] = Nodes,
    Pid = spawn(N, pool, do_spawn, [Gl, M, F, A]),
    {reply, From, Pid, lists:append(Tail, [{Load+1, N}]) };

master({From, {spawn_link, Gl, M, F, A}}, Nodes) ->
    [{Load,N}|Tail] = Nodes,
    Pid = spawn(N, pool, do_spawn, [From,Gl, M, F, A]),
    {reply, From, Pid, lists:append(Tail, [{Load+1, N}])};

master({From, stop}, Nodes) ->
    N = lists:delete(node(), get_elements(2, Nodes)),
    stop_em(N),
    {stopped, From};

master(_, Nodes) ->  %% The EXIT signals etc.etc
    {noreply, Nodes}.

do_spawn(Gl, M, F, A) ->
    group_leader(Gl,self()),
    apply(M,F,A).

insert_node({Load,Node},[{L,Node}|Tail]) when Load > L ->
    %% We have a raised load here
    pure_insert({Load,Node},Tail);
insert_node({Load,Node},[{L,N}|Tail]) when Load =< L ->
    %% Move forward in the list
    T = lists:keydelete(Node,2,[{L,N}|Tail]),
    [{Load,Node} | T];
insert_node(Ln,[H|T]) ->
    [H | insert_node(Ln,T)];
insert_node(X,[]) ->          % Can't happen
    error_logger:error_msg('Pool_master: Bad node list X=~w\n',[X]),
    exit(crash).


pure_insert({Load,Node},[]) ->
    [{Load,Node}];
pure_insert({Load,Node},[{L,N}|Tail]) when Load < L ->
    [{Load,Node} , {L,N} | Tail];
pure_insert(L,[H|T]) -> [H|pure_insert(L,T)].

%% Really should not meassure the contributions from
%% the back ground processes here .... which we do :-(
%% We don';t have to monitor the master, since we're slaves anyway

statistic_collector() ->
    statistic_collector(5).

statistic_collector(0) -> exit(normal);
statistic_collector(I) ->
    sleep(300),  
    case global:whereis_name(pool_master) of
	undefined ->
	    statistic_collector(I-1);
	M ->
	    stat_loop(M,999999)
    end.

%% Do not tell the master about our load if it has not  changed

stat_loop(M,Old) ->
    sleep(2000),
    case statistics(run_queue) of
	Old ->
	    stat_loop(M,Old);
	NewLoad ->
	    M ! {node(),load,NewLoad}, %% async 
	    stat_loop(M,NewLoad)
    end.

sleep(I) -> receive after I -> ok end.

    
    



