%% Copyright (C) 1990, Ellemtel Telecommunications Systems Laboratories
%% File	   : shell.erl
%% Author  : Robert Virding
%% Purpose : Simple Erlang shell.

%% This is a simple history mechanism which uses the process dictionary to
%% save the last 20 commands.

%% added tell_debugger() -- Joe 970728
%% And klacke removed it 1/12 1994
%% Klacke added the .bindings thing 25/11 1994

-module(shell).

-export([start/0,server/0,evaluator/3,local_func/4]).

start() ->
    code:ensure_loaded(user_default),
    spawn(shell, server, []).

server() ->
    %% Our spawner has fixed the process groups.
    Bs = erl_eval:new_bindings(),
    process_flag(trap_exit, true),
    io:fwrite("Eshell V~s  (abort with ^G)~n", [erlang:info(version)]),
    server_loop(0, start_eval(Bs, []), Bs, []).

server_loop(N0, Eval0, Bs0, Ds0) ->
    N = N0 + 1,
    case io:parse_erl_seq(prompt(N)) of
	{ok,Es0,EndLine} ->			%Commands
	    case expand_hist(Es0, N) of
		{ok,Es} ->
		    {V,Eval,Bs,Ds} = shell_cmd(Es, Eval0, Bs0, Ds0, N),
		    del_cmd(N - 20),
		    add_cmd(N, Es, V),
		    server_loop(N, Eval, Bs, Ds);
		{error,E} ->
		    io:fwrite("** ~s **\n", [E]),
		    server_loop(N0, Eval0, Bs0, Ds0)
	    end;
	{error,{Line,Mod,What},EndLine} ->
	    io:fwrite("** ~w: ~s **\n", [Line,apply(Mod,format_error,[What])]),
	    server_loop(N0, Eval0, Bs0, Ds0);
	{error,terminated} ->			%Io process terminated
	    exit(Eval0, kill),
	    terminated;
	{error,interrupted} ->			%Io process interrupted us
	    exit(Eval0, kill),
	    {_,Eval,_,_} = shell_rep(Eval0, Bs0, Ds0),
	    server_loop(N0, Eval, Bs0, Ds0);
	{eof,EndLine} ->
	    io:fwrite("** Terminating erlang **\n"),
	    halt();
	eof ->
	    io:fwrite("** Terminating erlang **\n"),
	    halt()
    end.

prompt(N) ->
    case is_alive() of
	true -> {format,"(~s)~w> ",[node(),N]};
	false -> {format,"~w> ",[N]}
    end.

%% expand_hist(Expressions, CommandNumber)
%%  Preprocess the expression list replacing all history list commands
%%  with their expansions.

expand_hist(Es, C) ->
    catch {ok,expand_seq(Es, C)}.

expand_seq([E|Es], C) ->
    [expand_expr(E, C)|expand_seq(Es, C)];
expand_seq([], C) ->
    [].

expand_expr({cons,L,H,T}, C) ->
    {cons,L,expand_expr(H, C),expand_expr(T, C)};
expand_expr({tuple,L,Elts}, C) ->
    {tuple,L,expand_seq(Elts, C)};
expand_expr({bif,L,B,Args}, C) ->
    {bif,L,B,expand_seq(Args, C)};
expand_expr({arith,L,Op,Larg,Rarg}, C) ->
    {arith,L,Op,expand_expr(Larg, C),expand_expr(Rarg, C)};
expand_expr({arith,L,Op,Arg}, C) ->
    {arith,L,Op,expand_expr(Arg, C)};
expand_expr({call,L,[],e,[N]}, C) ->
    case get_cmd(N, C) of
	{[Ce],V} ->
	    Ce;
	{Ces,V} ->
	    {block,L,Ces};
	undefined ->
	    no_command(N)
    end;
expand_expr({call,L,[],v,[N]}, C) ->
    case get_cmd(N, C) of
	{Ces,V} ->
	    {value,L,V};
	undefined ->
	    no_command(N)
    end;
expand_expr({match,L,Lhs,Rhs}, C) ->
    {match,L,Lhs,expand_expr(Rhs, C)};
expand_expr({call,L,M,F,Args}, C) ->
    {call,L,M,F,expand_seq(Args, C)};
expand_expr({block,L,Es}, C) ->
    {block,L,expand_seq(Es, C)};
expand_expr({'if',L,Cs}, C) ->
    {'if',L,expand_cs(Cs, C)};
expand_expr({'case',L,E,Cs}, C) ->
    {'case',L,expand_expr(E, C),expand_cs(Cs, C)};
expand_expr({'catch',L,E}, C) ->
    {'catch',L,expand_expr(E, C)};
expand_expr({send,L,T,M}, C) ->
    {send,L,expand_expr(T, C),expand_expr(M, C)};
expand_expr(E, C) ->				%All constants
    E.

expand_cs([{clause,L,G,B}|Cs], C) ->
    [{clause,L,G,expand_seq(B, C)}|expand_cs(Cs, C)];
expand_cs([{clause,L,P,G,B}|Cs], C) ->
    [{clause,L,P,G,expand_seq(B, C)}|expand_cs(Cs, C)];
expand_cs([], C) ->
    [].

no_command(N) ->
    throw({error,io_lib:fwrite("~s: command not found", [erl_pp:expr(N)])}).

%% add_cmd(Number, Expressions, Value)
%% get_cmd(Number)
%% del_cmd(Number)

add_cmd(N, Es, V) ->
    put({command,N}, {Es, V}).

get_cmd({integer,_,N}, C) ->
    get_cmd(N, C);
get_cmd(N, C) when N < 0 ->
    get({command,C+N});
get_cmd(N, C) ->
    get({command,N}).

del_cmd(N) ->
    erase({command,N}).

%% shell_cmd(Sequence, Evaluator, Bindings, Dictionary, CommandNumber)
%% shell_rep(Evaluator) ->
%%	{Value,Evaluator,Bindings,Dictionary}
%%  Send a command to the evaluator and wait for the reply. Start a new
%%  evaluator if necessary.

shell_cmd(Es, Eval, Bs, Ds, N) ->
    Eval ! {shell_cmd,self(),{eval,Es}},
    shell_rep(Eval, Bs, Ds).

shell_rep(Ev, Bs0, Ds0) ->
    receive
	{shell_rep,Ev,{value,V,Bs,Ds}} ->
	    io:fwrite("~p~n", [V]),
	    {V,Ev,Bs,Ds};
	{shell_req,Ev,get_cmd} ->
	    Ev ! {shell_rep,self(),get()},
	    shell_rep(Ev, Bs0, Ds0);
	{shell_req,Ev,exit} ->
	    Ev ! {shell_rep,self(),exit},
	    io:fwrite("** Terminating shell **\n", []),
	    exit(normal);
	{'EXIT',Ev,Reason} ->			%It has exited unnaturally
	    io:fwrite("** exited: ~p **\n", [Reason]),
	    {{'EXIT',Reason},start_eval(Bs0, Ds0), Bs0, Ds0};
	{'EXIT',Id,interrupt} ->		%Someone interrupted us
	    exit(Ev, kill),
	    shell_rep(Ev, Bs0, Ds0);
	{'EXIT',Id,R} ->
	    exit(Ev, R),
	    exit(R);
	Other ->				%Ignore everything else
	    shell_rep(Ev, Bs0, Ds0)
    end.

start_eval(Bs, Ds) ->
    spawn_link(shell, evaluator, [self(),Bs,Ds]).

%% evaluator(Shell, Bindings, ProcessDictionary)
%%  Evaluate expressions from the shell. Use the "old" variable bindings
%%  and ductionary.

evaluator(Shell, Bs, Ds) ->
    init_dict(Ds),
    eval_loop(Shell, Bs).

eval_loop(Shell, Bs0) ->
    receive
	{shell_cmd,Shell,{eval,Es}} ->
	    {value,V,Bs} = erl_eval:seq(Es, Bs0, {shell,local_func,[Shell]}),
	    Shell ! {shell_rep,self(),{value,V,Bs,get()}},
	    eval_loop(Shell, Bs)
    end.

init_dict([{K,V}|Ds]) ->
    put(K, V),
    init_dict(Ds);
init_dict([]) -> true.

%% local_func(Function, Args, Bindings, Shell) ->
%%	{value,Val,Bs}
%%  Evaluate local functions, including shell commands.

local_func(h, [], Bs, Shell) ->
    Cs = shell_req(Shell, get_cmd),
    {value,list_commands(lists:keysort(1, Cs)),Bs};
local_func(b, [], Bs, Shell) ->
    {value,list_bindings(erl_eval:bindings(Bs)),Bs};
local_func(f, [], Bs, Shell) ->
    {value,ok,[]};
local_func(f, [{var,_,Name}], Bs, Shell) ->
    {value,ok,erl_eval:del_binding(Name, Bs)};
local_func(f, [Other], Bs, Shell) ->
    exit({function_clause,{shell,f,1}});
local_func(exit, [], Bs, Shell) ->
    shell_req(Shell, exit),			%This terminates us
    exit(normal);
local_func(F, As0, Bs0, Shell) ->
    {As,Bs} = erl_eval:arg_list(As0, Bs0, {shell,local_func,[Shell]}),
    case erlang:function_exported(user_default, F, length(As)) of
	true ->
	    {value,apply(user_default, F, As),Bs};
	false ->
	    {value,apply(shell_default, F, As),Bs}
    end.

shell_req(Shell, Req) ->
    Shell ! {shell_req,self(),Req},
    receive
	{shell_rep,Shell,Rep} -> Rep
    end.

list_commands([{{command,N},{Es,V}}|Ds]) ->
    io:requests([{format,"~w: ~s~n",[N,erl_pp:seq(Es)]},
		 {format,"-> ~p~n",[V]}]),
    list_commands(Ds);
list_commands([D|Ds]) ->
    list_commands(Ds);
list_commands([]) -> ok.

list_bindings([{Name,Val}|Bs]) ->
    io:fwrite("~s = ~p~n", [Name,Val]),
    list_bindings(Bs);
list_bindings([]) ->
    ok.
