%% Copyright (C) 1994, Ellemtel Telecommunications Systems Laboratories
%% Author: Claes Wikstrom  klacke@erix.ericsson.se
%% Purpose:   Utility functions for slave stuff,

-module(slave).
-export([send_rsh_originator/1 ,
	 monitor_master/1,
	 pseudo/1,
	 slave_start/1,
	 start/1,
	 start/2,
	 start/3,
	 start/4,
	 start2/4,
	 stop/1,
	 relay/1,
	 take_over/1, take_over/2,
	 do_take_over/2]).

%% Internal Exports 
-export([ send_rsh_originator2/2,
	 mon_master/1]).
-import(error_logger,[error_msg/2]).


%% This is used when starting up slave nodes.

pseudo([Master | ServerList]) ->
    pseudo(Master , ServerList);
pseudo(_) ->
    error_msg("No master node give to slave:pseudo/1~n",[]).

pseudo(_, []) -> ok;
pseudo(Master, [S|Tail]) ->
    start_pseudo(S, whereis(S), Master),
    pseudo(Master, Tail).

start_pseudo(Name, undefined, Master) ->
    X = rpc:call(Master,erlang, whereis,[Name]),
    register(Name, spawn(slave, relay, [X]));

start_pseudo(_,_,_) -> ok.  %% It's allready there


%% This relay can be used to relay all messages directed to a process
%% For instance a node with a remote filer will relay all it's
%% messages directed to the local filer to the remote filer , or tty
%% type of io as well
%% Since the code loading now goes through the filer the code server
%% load requests will be relayed as well !!!!

relay({badrpc,Reason}) ->
    error_msg(' ** exiting relay server ~w :~w  **~n',
			   [self(),Reason]),
    exit(Reason);
relay(undefined) ->
    error_msg(' ** exiting relay server ~w  **~n',
			   [self()]),
    exit(undefined);
relay(Pid) when pid(Pid) ->
    relay1(Pid).

relay1(Pid) ->
    receive
        X ->
            Pid ! X
    end,
    relay1(Pid).

%% Run by all slaves... If master dies take down this node.

monitor_master([Master]) ->
    spawn(slave,mon_master,[Master]).

mon_master(Master) when atom(Master) ->
    erlang:monitor_node(Master, true),
    receive
	{nodedown, Master} ->
	    halt()
    end.

%% This function is used when starting up slave nodes via rsh,
%% to be able to know when the node is up and running.
%% The function needs to sit in a loop waiting for the auth server
%% to be registered, since the code for the auth server 
%% may come over the net from the original master,

send_rsh_originator([Node]) ->
    spawn(slave, send_rsh_originator2, [Node, whereis(auth)]).

send_rsh_originator2(Node, undefined) ->
    receive after 200 -> ok end,
    send_rsh_originator2(Node, whereis(auth));

send_rsh_originator2(Node, Auth) ->
    case auth:is_auth(Node) of
	yes ->
	    {rsh_starter, Node} ! {node(), hello};
	no ->
	    io:format("Not authorized to talk to originator ~n",[])
    end.


start(Host) ->
    L = atom_to_list(node()),
    Name = upto($@, L),
    start(Host, Name).

start(Host, Name) ->
    start(Host, Name,[]).

start(Host, Name,Env) ->
    start(Host, Name,Env, []).

start(Host, Name,Env, PseudoServers) ->
    start2(Host, Name,Env, PseudoServers).

start2(Host0 , Name, Env, PseudoServers) ->
    Host = case net_kernel:longnames() of
		    true -> dns(Host0);
		    false -> strip_host_name(to_list(Host0))
		end,
    check_pseudo_servers(PseudoServers),
    E2 = fix_env(Env),
    Node = list_to_atom(lists:concat([Name, "@", Host])),
    case auth:exists(Node) of
	yes -> 
	    case auth:is_auth(Node) of
		yes ->                        %% WE act as master
		    take_over(Node), 
		    case rpc:call(Node,slave, pseudo, 
				  [[node() | PseudoServers]]) of
			{badrpc, _} ->
			    {error, nodedown};
			_ ->
			    {ok, Node}
		    end;
		no ->
		    {error, {noauthority, Node}}
	    end;
	no ->
	    case start_it(Host, Name, Node, E2) of
		{ok, N} ->                    %% WE act as master
		    rpc:call(N,slave, pseudo, [[node() | PseudoServers]]),
		    rpc:call(N,code, set_path,[code:get_path()]),
		    {ok, N};
		Other -> Other
	    end
    end.

dns(H) -> {ok, Host} = net_adm:dns_hostname(H), Host.

strip_host_name([]) -> [];
strip_host_name([$.|_]) -> [];
strip_host_name([H|T]) -> [H|strip_host_name(T)].

%% We allways set display if not allready set by the user
fix_env(E) ->
    case lists:keysearch("DISPLAY", 1, E) of
	false ->
	    fix_env2([{"DISPLAY", ":0"} |E], []);
	_ ->
	    fix_env2(E, [])
    end.
fix_env2([], Ack) -> lists:flatten(Ack);
fix_env2([{Var, Val} |Tail], Ack) ->
    fix_env2(Tail, [" -env ",to_list(Var), " ",to_list(Val) | Ack]);
fix_env2([H|_],_) -> 
    io:format("Bad environment to slave:start/N ~w~n",[H]),
    exit(badenv).

stop(Node) -> rpc:cast(Node, erlang, halt, []).

check_pseudo_servers([]) -> ok;
check_pseudo_servers([S|Tail]) ->
    case whereis(S) of
	undefined ->
	    error_logger:error_msg("No server ~w registered~n",[S]),
	    exit(bad_pseudo_server);
	_ ->
	    check_pseudo_servers(Tail)
    end.

to_list(X) when list(X) -> X;
to_list(X) when atom(X) -> atom_to_list(X).

upto(Char,[]) -> [];
upto(Char, [Char|T]) -> [];
upto(Char, [H|T]) -> [H|upto(Char, T)].

start_it(Host, Name, Node, Env) ->
    do_register(rsh_starter , self()),
    Cmd = mk_cmd(Host,Name, Env),
    P = open_port({spawn,Cmd},[stream]),
    Result = rec_loop(P, Node),
    unregister(rsh_starter),
    Result.

rec_loop(P, Node) ->
    receive
	{'EXIT', Port, _} ->
	    rec_loop(P, Node);
	{Port, {data, D}} ->
	    io:format("~s",[D]),
	    rec_loop(P, Node);
	{Node, hello} -> 
	    {ok, Node}
    after 20000 ->
	    %% something's gone wrong !!!
	    timeout_action(P, Node)
    end.

timeout_action(P, Node) ->
    io:format("Timeout .... maybe trouble \n"),
    case auth:is_auth(Node) of
	yes -> {ok, Node};
	no ->  {error, Node}
    end.


%% A bit of unix stuff here .....
%% Assumes that we are allowed to do an rsh to Host without
%% beeing prompted for pasword

mk_cmd(Host, Name, Env) ->
    Prog = lib:progname(),  %% assuming the path is right 
    %%Prog = "/home/super/klacke/erl",
    Master = lists:concat([" -master ",node()," "]),
    Cmd = lists:concat([rsh(), " ",Host," \"exec ",Prog," -noshell ",
			Env," -master ", node(),
			long_or_short(), Name,
			 " -s slave slave_start ",node(),
			 " >&! $HOME/.erlang.slave.out.",cut(to_list(Host)),
			 "  < /dev/null &\""]),
    %%io:format("Cmd is : ~s~n",[Cmd]),
    Cmd.

%% Give the user an opportunity to run another program
%% than the (default) which is "rsh", On i.e. HP-UX it's called remsh
%% So HP user must start erlang as erl -rsh remsh .........
%% and on stuff like WinNT there are no equivalents at all.

rsh() ->
    case init:get_flag('-rsh') of
	{'-rsh', Prog} -> Prog;
	_ -> "rsh"
    end.

long_or_short() -> 
    case net_kernel:longnames() of
	true -> " -name ";
	false -> " -sname "
    end.

slave_start([Master]) ->
    spawn(slave, mon_master, [Master]),
    {rsh_starter, Master} ! {node(), hello}.  %% We're in the air 
    

%% Just the short hostname, not the qualified. for conveniance.
cut([]) -> [];
cut([$.|_]) -> [];
cut([X|Tail]) -> [X|cut(Tail)].


do_register(Name, Pid) ->
    case erlang:process_info(Pid, registered_name) of
	{registered_name, N} ->
	    error_logger:error_msg("Can't slave:start() from a process",[]),
	    error_logger:error_msg(
                " which already is registered under a name ~n",[]),
	    exit(already_registered);
	_ ->
	    register(Name, Pid)
    end.


%% Call this function to take over all code loading filing and tty io
%% from Node, Once we terminate the calling node all will be reset,
%% The function returns {ok, Pid} and if we send 'reset' to the Pid all 
%% servers will be reset to the original server as well

take_over(Node) -> 
    take_over(Node, [file_server,user]).
take_over(Node, Servers) ->
    P = spawn(Node, slave, do_take_over, [self(), Servers]),
    receive
	{P, error, Reason} ->
	    error_logger:format("** Can't take over node ~w it is ~w~n",
				[Node, Reason]),
	    Reason;
	{P, Pid, Status} ->
	    {ok, Pid}
    after 8000 ->    %% long time if we're running on a WAN
	    timeout
    end.

do_take_over(From, Servers) ->  
    case whereis(take_over_monitor) of
	Pid when pid(Pid) ->
	    Pid ! {self(), master},
	    receive
		{Pid, Mnode} when node(From) == Mnode ->
		    From ! {self(), Pid, already_runing};
		{Pid, Mnode} ->
		    From ! {self(), error, occupied}
	    end;
	undefined ->
	    Master = node(From),
	    erlang:monitor_node(Master, true),
	    Oldcode = whereis(code_server),
	    Oldpath = code:get_path(),
	    Newpath = rpc:call(Master, code, get_path, []),
	    S2 = get_server_pids(Master,Servers),
	    lists:map({erlang, unregister}, [], Servers),
	    create_pseudo(S2),
	    code:set_path(Newpath),
	    register(take_over_monitor, self()),
	    From ! {self(), self(), running},
	    monitor_loop(S2, Master, Oldpath)
    end.

monitor_loop(S2, Master, Oldpath) ->
    receive
	{nodedown, Master} ->
	    reset(S2, Oldpath);
	reset ->
	    reset(S2, Oldpath);
	{From, master} ->
	    From ! {self(), Master},
	    monitor_loop(S2, Master, Oldpath)
    end.

reset([], Oldpath) -> 
    code:set_path(Oldpath);  %% terminate and unregister silently
reset([{Name, _, Old} | T], Opath) ->
    catch exit(whereis(Name), kill),
    catch unregister(Name),
    catch register(Name, Old),
    reset(T, Opath).

create_pseudo([]) -> done;
create_pseudo([{Name, New, Old}|T]) ->
    register(Name, spawn(slave, relay, [New])),
    create_pseudo(T).

get_server_pids(_,[]) -> [];
get_server_pids(Master, [Name |Tail]) ->
    case rpc:call(Master, erlang, whereis, [Name]) of
	{badrpc, _} -> exit(normal);
	Rp -> [{Name, Rp, whereis(Name)} | get_server_pids(Master, Tail)]
    end.

