%% Copyright (C) 1994, Ellemtel Telecommunications Systems Laboratories
%% File     :   timer.erl
%% Author   :   Sebastian Strollo (seb@erix.ericsson.se)
%% Purpose  :   Timer functions

-module(timer).

-export([apply_after/4,
	 send_after/3, send_after/2,
	 exit_after/3, exit_after/2, kill_after/2, kill_after/1,
	 apply_interval/4, send_interval/3, send_interval/2,
	 cancel/1, sleep/1, tc/3,
	 seconds/1, minutes/1, hours/1, hms/3]).

-export([start/0, timer_init/0, do_apply_at/2, send/2]).

%% ------------------------------------------------------ %%

start() ->
    start(whereis(timer_server)).

apply_after(Time, M, F, A) ->
    req(apply_after, {Time, {M, F, A}}).

send_after(Time, Pid, Message) ->
    req(apply_after, {Time, {timer, send, [Pid, Message]}}).

send_after(Time, Message) ->
    send_after(Time, self(), Message).

exit_after(Time, Pid, Reason) ->
    req(apply_after, {Time, {erlang, exit, [Pid, Reason]}}).

exit_after(Time, Reason) ->
    exit_after(Time, self(), Reason).

kill_after(Time, Pid) ->
    exit_after(Time, Pid, kill).

kill_after(Time) ->
    exit_after(Time, self(), kill).

apply_interval(Time, M, F, A) ->
    req(apply_interval, {Time, {M, F, A}}).

send_interval(Time, Pid, Message) ->
    req(apply_interval, {Time, {timer, send, [Pid, Message]}}).

send_interval(Time, Message) ->
    send_interval(Time, self(), Message).

cancel(Ref) ->
    req(cancel, Ref).

sleep(T) ->
    send_after(T, self(), R = make_ref()),
    receive
	R ->
	    ok
    end.

tc(M, F, A) ->
    Before = erlang:now(),
    Val = catch apply(M, F, A),
    After = erlang:now(),
    Elapsed =
	(element(1,After)*1000000000000
	 +element(2,After)*1000000+element(3,After)) -
	(element(1,Before)*1000000000000
	 +element(2,Before)*1000000+element(3,Before)),
    {Elapsed, Val}.
    

seconds(Seconds) ->
    1000 * Seconds.
minutes(Minutes) ->
    1000 * 60 * Minutes.
hours(Hours) ->
    1000 * 60 * 60 * Hours.
hms(H, M, S) ->
    hours(H) + minutes(M) + seconds(S).

%% ------------------------------------------------------ %%

req(Req, Arg) ->
    req(whereis(timer_server), Req, Arg).
req(Pid, Req, Arg) when pid(Pid) ->
    Pid ! {self(), Req, Arg},
    receive
	{Pid, Reply} ->
	    Reply
    end;
req(undefined, Req, Arg) ->
    timer:start(),
    req(Req, Arg).

reply_ok(Pid, Result) ->
    Pid ! {self(), {ok, Result}}.
reply_error(Pid, Why) ->
    Pid ! {self(), {error, Why}}.

%% ------------------------------------------------------ %%

system_time() ->
    %%erlang:now().
    %%element(1, statistics(wall_clock)).
    {M,S,U} = erlang:now(),
    1000000000 * M + 1000 * S + (U div 1000).
max_timeout() ->
    16#0800000.

start(Pid) when pid(Pid) ->
    ok;
start(_) ->
    Pid = spawn(timer, timer_init, []),
    register(timer_server, Pid),
    ok.

timer_init() ->
    process_flag(trap_exit, true),
    timer_server([]).

timer_server(TL) ->
    Start = system_time(),
    TimeOut = case TL of
		  [] -> infinity;
		  _ ->  element(1, hd(TL))
	      end,
    timer_receive(TL, Start, TimeOut).

timer_receive(TL, Start, TimeOut) when TimeOut =< 0 ->
    %% Handle timeouts before any new requests!
    timer_timeout(TL);
timer_receive(TL, Start, TimeOut) ->
    %%io:format("timer_server: ~w~n~w~n~n", [TimeOut,TL]),
    receive
	{Pid, apply_after, {Time, MFA}} when integer(Time), Time>0 ->
	    When = Time + system_time(),
	    reply_ok(Pid, Ref = make_ref()),
	    NewTL = do_insert(Time, Ref, timeout, {When,MFA},
			      elapsed(system_time()-Start, TL)),
	    timer_server(NewTL);
	{Pid, apply_interval, {Time, MFA}} when integer(Time), Time>0 ->
	    When = Time + system_time(),
	    reply_ok(Pid, Ref = make_ref()),
	    NewTL = do_insert(Time, Ref, {repeat, Time}, {When,MFA},
			      elapsed(system_time() - Start, TL)),
	    timer_server(NewTL);
	{Pid, cancel, Ref} ->
	    reply_ok(Pid, cancel),
	    timer_server(elapsed(system_time() - Start, delete(Ref, TL)));
	{Pid, apply_after, _} ->
	    reply_error(Pid, badarg),
	    timer_server(elapsed(system_time() - Start, TL));
	{Pid, apply_interval, _} ->
	    reply_error(Pid, badarg),
	    timer_server(elapsed(system_time() - Start, TL));
	O ->
	    timer_server(elapsed(system_time() - Start, TL))
    after TimeOut ->
	    timer_timeout(TL)
    end.

timer_timeout([{_, _, timeout, {When, MFA}}|Rest]) ->
    spawn(timer, do_apply_at, [When - system_time(), MFA]),
    timer_server(Rest);
timer_timeout([{TimeOut, Ref, {repeat, Interv}, {When,MFA}}|Rest]) ->
    spawn(timer, do_apply_at, [When - system_time(), MFA]),
    timer_server(do_insert(Interv, Ref, {repeat, Interv},
			   {When+Interv, MFA}, Rest));
timer_timeout([{_, Ref, {keep_until, Op}, {When, MFA}}|Rest]) ->
    timer_server(do_insert(When - system_time(), Ref, Op, {When, MFA}, Rest)).


do_apply_at(TimeOut, {M, F, A}) when TimeOut > 0 ->
    receive
    after TimeOut ->
	    apply(M, F, A)
    end;
do_apply_at(TimeOut, {M, F, A}) ->
    apply(M, F, A).

send(Pid, Msg) ->
    Pid ! Msg.

do_insert(TimeOut, Ref, Op, Item, TL) ->
    case max_timeout() of
%%	_ when TimeOut > 10000 ->
%%	    insert(10000, Ref, {keep_until, Op}, Item, TL);
	Max when TimeOut > Max ->
	    insert(Max, Ref, {keep_until, Op}, Item, TL);
	_ ->
	    insert(TimeOut-100, Ref, Op, Item, TL)
    end.

%% Delta list functions

elapsed(Elapsed, [{Time, Ref, Op, Item}|T]) when Elapsed =< Time ->
    [{Time - Elapsed, Ref, Op, Item}|T];
elapsed(Elapsed, [{Time, Ref, Op, Item}|T]) ->
    [{0, Ref, Op, Item}|elapsed(Elapsed-Time, T)];
elapsed(Elapsed, []) ->
    [].

insert(Time, Ref, Op, Item, [{Time1, Ref1, Op1, Item1}|T]) when Time < Time1 ->
    [{Time, Ref, Op, Item}, {Time1 - Time, Ref1, Op1, Item1}|T];
insert(Time, Ref, Op, Item, [{Time1, Ref1, Op1, Item1}|T]) ->
    [{Time1, Ref1, Op1, Item1}|insert(Time - Time1, Ref, Op, Item, T)];
insert(Time, Ref, Op, Item, []) ->
    [{Time, Ref, Op, Item}].

delete(Ref, [{Time, Ref, Op, Item}, {Time1, Ref1, Op1, Item1}|T]) ->
    [{Time + Time1, Ref1, Op1, Item1}|T];
delete(Ref, [{Time, Ref, Op, Item}]) ->
    [];
delete(Ref, [H|T]) ->
    [H|delete(Ref, T)];
delete(Ref, []) ->
    [].

%% ------------------------------------------------------ %%
