
/*
** $Id: erlnode.c,v 1.24 2004/09/13 14:08:01 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/erlnode.c,v $
** $Revision: 1.24 $
** $Date: 2004/09/13 14:08:01 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1998
** geoff@serc.rmit.edu.au
** 
** Erlang Node Server: 
**      handle multiple incoming messages and buffer them
**      notify clients of their arrival and deliver when
**      the client is ready to receive them
**
** Note: This should really be built in Erlang 
**      But I maintain lots of evil state..
**      Also the connect structure (C) should be a list
**
** Must be linked to liberl.so
**
** FIX: not robust to weird string lengths at the moment
**      or random signals (SIGPIPEs etc)
**
** Also spawning should be handled here so Unix signals (for failure)
** can be caught and passed onto the appropriate processes..
*/
#include <stdio.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>
#include <netdb.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>

#include "erlib.h"
#include "message.h"
#include "ehash.h"
#include "meshash.h"
#include "erlnode.h"

#define DEBUG 1

#define IP_PROT 0

#define MAX_CONNECTIONS 20000
#define MAX_CONNECTED 128
#define MAX_MESSAGE 2048
#define MESS_QUEUE_PRIME 617

int Connection[MAX_CONNECTIONS+1];
int connected = 0;
int high_connect = 0;
int empty = 0;

int debug = 0;
    // Debug flag.

Hash_table mess_queue;
    /* Message queues */

int Bound_socket;

/* private functions */
int handle_data(unsigned int Cfd, char * buf, int bufsize);
int process_message(unsigned int Cfd, int length, const char * buf);
void serve();


void queue(unsigned int pid, unsigned int fd, int length, const char * buf)
{
    Message_bucket tmp, newb, mb;

    newb = (Message_bucket) malloc(sizeof(struct MESSAGE_BUCKET));
    if (newb == NULL)
    {
        printf("Nasty: Erlang node can't queue a message.\n");
        return;
    }

    newb->pid = pid;
    newb->fd = fd;
    newb->b = (unsigned char *) malloc(sizeof(unsigned char) * length);
    if (newb->b == NULL)
    {
        printf("Nasty: Erlang node can't queue a message.\n");
        return;
    }
    newb->length = length;
    newb->next = NULL;
    memcpy(newb->b, buf, length);

    mb = find_hash(mess_queue, (void *)pid);
    if (mb != NULL)
    {
        /* tail insertion - we do head removal */
        /* should we use a heap instead?       */
        if (debug) printf("queue l=%d: tail insertion for (%d), length (%d)\n", mess_queue->entries, newb->pid, length);
        while (mb != NULL)
        {
            tmp = mb;
            mb = mb->next;
        }
        tmp->next = newb; 
    }
    else
    {
        if (debug) printf("queue l=%d: for (%d), length (%d)\n", mess_queue->entries, newb->pid, length);
        add_hash(mess_queue, newb);
    }
}

int main(int argc, char * argv[])
{
    int x;
    struct sockaddr_in sin;

    if (argc < 2)
    {
        fprintf(stderr, "server [-d] <port>\n");
        exit(1);
    }

    for (x = 1; x < argc; x++)
    {
        if (argv[x][0] == '-')
        {
            switch (argv[x][1])
            {
                case 'd':
                    debug = 1;
                    break;
                default:
                    fprintf(stderr, "server [-d] <port>\n");
                    exit(1);
            }
        }
        else break;
    }

    /* initialise structures */
    mess_queue = make_hash_table(MESS_QUEUE_PRIME, mes_hash, mes_key, mes_match);

    /* create, bind and listen to a port */
    Bound_socket = socket(AF_INET, SOCK_STREAM, IP_PROT);
    memset((char *) &sin, '\0', sizeof sin);
    sin.sin_family = AF_INET;
    sin.sin_port = htons(atoi(argv[x]));
    sin.sin_addr.s_addr = INADDR_ANY;
    if (bind(Bound_socket, (struct sockaddr *)&sin, sizeof sin) == -1)
    {
        perror("server bind");
        exit(1);
    }
    listen(Bound_socket,128);

    printf("Bound to port\n");

    /* non blocking on accepts */
    if (fcntl(Bound_socket, F_SETFL, O_NDELAY) == -1) 
    {
        perror("fnctl(O_NDELAY)");
        exit(1);
    }

    serve();
    return 0;
}


void serve()
{
    struct sockaddr_in addr;
    int length = sizeof addr;

    fd_set readfds, totalfds, exceptfds;
    int maxfd = 0;
    int sleepcount = 0;
    int T;

    FD_ZERO(&totalfds);

    /* main listen loop */
    while (1)
    {
        struct timeval timeout;
        int res, i;
        int flag;
        char buf[MAX_MESSAGE];
        
        flag = 0;
        T = 0;
        if (connected < MAX_CONNECTED)
        {
            T = accept(Bound_socket,(struct sockaddr *)&addr,&length);
        }
        else
        {
            if (debug) printf("At MAXIMUM connections connected = %d, high_connect = %d, empty = %d.\n", connected, high_connect, empty);
        }

        if (T == -1 && errno != EAGAIN)
        {
            perror("accept");
            // continue;
        }

        if (T > 0)
        {
            fcntl(T, F_SETFL, O_NDELAY);

            Connection[empty++] = T;
            connected++;
            if (empty > high_connect) high_connect = empty;

            FD_SET(T, &totalfds);
            if (T >= maxfd) maxfd = T+1;
#ifdef DEBUG
            if (debug) printf("Accepted connection (%d), connected = %d, high_connect = %d, empty = %d.\n", T, connected, high_connect, empty);
            flag = 1;
#endif
        }

        readfds = totalfds;

        timeout.tv_sec = 0;
        timeout.tv_usec = 50000;
        FD_ZERO(&exceptfds);
        res = select(maxfd, &readfds, 0, &exceptfds, &timeout);
        if (res == -1)
        {
            flag = 1;
            perror("select");
        } 
        if (res > 0)
        {
#ifdef DEBUG
            // if (debug) printf("select result = %d\n", res);
#endif
            for (i = 0; i < high_connect; i++)
            {
                if (!Connection[i]) continue;
                if (FD_ISSET(Connection[i], &exceptfds))
                {
                    FD_CLR(Connection[i], &totalfds);
                    //shutdown(Connection[i],2);
                    close(Connection[i]);
                    Connection[i] = 0;
                    connected--;
                    if (i < empty) 
                    {
                        empty = i;
                        flag = 0;
                    }
                }
                if (FD_ISSET(Connection[i], &readfds))
                {
                    if (handle_data(Connection[i], buf, MAX_MESSAGE) == 0)
                    {
                        FD_CLR(Connection[i], &totalfds);
                        //shutdown(Connection[i],2);
                        close(Connection[i]);
                        Connection[i] = 0;
                        connected--;
                        if (i < empty) 
                        {
                            empty = i;
                            flag = 0;
                        }
                    }
                }

            }
        }

        if (flag)
        {
            // find next empty slot
            int count;
            count = 0;
            if (T > 0 && connected < MAX_CONNECTIONS)
                while (count++ < MAX_CONNECTIONS && Connection[empty] != 0) empty = (empty + 1) % MAX_CONNECTIONS;
        }

        // try and reach them again?
        // (hack because SIGUSR goes missing so often)
        if (!((sleepcount+5) % 20) && T < 1 && res < 1)
        {
            Message_bucket buck;
            for (i = 0; i < MESS_QUEUE_PRIME; i++)
            {
                if (mess_queue->table[i])
                {
                    buck = mess_queue->table[i]->element;
                    while (buck)
                    {
                        kill(buck->pid, SIGUSR1);
                        buck = buck->next;
                    }

                }
            }
            
        }

        // Don't busy wait
        if (!(sleepcount++ % 10))
            usleep(5);
    }
}

int handle_data(unsigned int Cfd, char * buf, int bufsize)
{
    int res;

    res = recv(Cfd, buf, bufsize, 0);
    if (res > 0)
    {
        buf[res] = 0;
        if (debug) printf("processing message(s) on fd %d (len=%d)\n", Cfd, res);
        return process_message(Cfd, res, buf);
    }
    else if (res == 0) 
    {
        return 0;
    }
    else
    {
        perror("recvfrom");
        return 0;
    }

    return 1;
}


#define LEN_BYTES 4

int process_message(unsigned int Cfd, int length, const char * buf)
{
    int topid;
    int messlen = 0;
    int offset = 0;

    do 
    {
    messlen = buf[offset+0] | (buf[offset+1] << 8) 
            | (buf[offset+2] << 16) | (buf[offset+3] << 24);
    offset = offset + LEN_BYTES;

#ifdef DEBUG
    if (debug) printf("processing message: %d of %d[%d] in length (from %d)\n",
                messlen, length, buf[offset], Cfd);
#endif

    switch (buf[offset])
    {
        case C_link:
            /* FIX: set up a watch on the process? */
            topid = extract_4(&(buf[1+offset]));
            break;

        case C_spawn:
            assert(1 == 0);
            break;

        case C_send:
            /* buffer the message to be processed by the receiving process */
            topid = extract_4(&(buf[1+offset]));
#ifdef DEBUG
            if (debug) printf( "C_send (%s): to %d, length = %d\n", &buf[7+offset], topid, length);
#endif
            queue(topid, Cfd, messlen, &(buf[5+offset]));
            kill(topid, SIGUSR1);
            break;

        case C_exit:
            /* send a kill to the process in question */
            topid = extract_4(&(buf[1+offset]));
            /* FIX: should be queueing a tuple */
            queue(topid, Cfd, messlen, &(buf[5+offset]));
            break;

        case C_unlink:
            /* FIX: remove a watch on a process */
            topid = extract_4(&(buf[1+offset]));
            break;

        case C_node_link:
            /* link two node - we don't do this */
            break;

        case C_reg_send:
            /* FIX: a registered send */
            break;

        case C_group_leader:
            {
                unsigned int frompid;
                /* FIX: set a processes group leader */
                frompid = extract_4(&(buf[1+offset]));
                topid = extract_4(&(buf[1+offset]));
                break;
            }
        case C_exit2:
            /* exit 2 a process */
            topid = extract_4(&(buf[1+offset]));
            /* FIX: should be queueing a tuple */
            queue(topid, Cfd, messlen, &(buf[5+offset]));
            break;

        case C_magic:
            /* magic number notification */
            topid = extract_4(&(buf[1+offset]));
#ifdef DEBUG
            if (debug) printf("C_magic: to %d, length = %d\n", topid, length);
#endif
            queue(topid, Cfd, messlen, &(buf[5+offset]));
            break;

        case C_goahead:
        case C_goahead_magic:
            /* request for a message - send it to requestor */
            /* or fetch a message with a magic pid and send it */
            {
                Message_bucket l, m;
                int x;

#ifdef DEBUG
            //if (debug) printf( "C_goahead: length = %d\n", length);
#endif
                topid = extract_4(&(buf[9+offset]));
                m = find_hash(mess_queue, (void *)topid);
                if (m == NULL)
                {
                    if (debug) printf( "C_goahead: Unable to locate messages for (%d)\n", topid);
                    /* FIX: should we send back a throw message? probably */
                    /* Possibly: send back a NAK */
                    return 0;
                }
                /* should this be a non-blocking send? probably */
                /* x = send(m->fd, m->b, m->length, 0); */
                x = send(Cfd, m->b, m->length, 0);

                if (x == -1 && errno == EAGAIN) break;
                if (x == -1) 
                {
                    perror("send (C_go_ahead)");
                    remove_hash(mess_queue, (void *)topid);
                    return 0;
                }

                if (remove_hash(mess_queue, (void *)topid) == 0)
                    perror("Unable to deque sent message\n");

                l = m;
                m = m->next;
                if (m != NULL)
                {
                    add_hash(mess_queue, m);
                }

                free(l);

#ifdef DEBUG
            if (debug) printf("C_goahead: to %d -> sent %d bytes on connection %d\n", topid, x, Cfd);
#endif
                break;
            }
        default:
            printf("Illegal message received (%d: %s) and discarded\n", length, buf);
            break;
    }

    offset = offset + messlen;
    } while (length - offset > 0);
    return 0;
}


