

/*
** $Id: message.c,v 1.59 2004/09/13 14:08:01 geoff Exp $
**
** $Source: /home/cvs/gerl/lib/message.c,v $
** $Revision: 1.59 $
** $Date: 2004/09/13 14:08:01 $
** $State: Exp $
**
** Author: Geoff Wong
** Copyright(C) 1998
** geoff@serc.rmit.edu.au
*/

/*
** Messaging Scheme 1
**
** Single process per node which handles all incoming messages
** Processes are signalled (using SIGUSR1) if a message arrives
** Processes typically ignore SIGUSR1 unless they're in a receive
** wait state (either starting or waiting).
** The process then retrieves a message from the node which does
** all the message buffering (how can we monitor this?)
** sendto/recvfrom pair. Then receive can then be matched.
** Initially send a message to the Node process to attempt to retrieve
** any data that is already buffered.
** with SIGUSR1 if any messages are waitingreceive attempts to fetch data 
** (if available)
**
** Scheme 2
**
** Still with a centralised message reception and buffering.
** In the event of a message arriving SIGUSR1 isn't ignored,
** indeed it fetches the message immediately and buffers for itself.
** Receiving simply checks the buffer.
*/



#include <sys/types.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/wait.h>
#include <string.h>
#include <fcntl.h>
#include <assert.h>
#ifdef SunOS
#include <netinet/in.h>
#endif
#ifdef Linux
#include <linux/in.h>
#endif
#include "erlib.h"
#include "message.h"
#include "esocket.h"
#include "code.h"

#ifndef _ALLOCS_H
extern EBinary binaryalloc(int);
#endif

static int Esendpidbin(Val pid, EBinary str);
static int Esendpidbinrecv(Val pid, EBinary str);
static LenVal lvextract(const unsigned char * loc);
static void message_handler(int signo);

static unsigned int messages_waiting = 0;

Val zero = { LT_number, { 0 } };

Socket _Eserver = NULL;
    /* Socket connecting us to our local node */

Socket _Eexternal = NULL;
    /* a persistent connection to the last place we send a message to */

Val  _Eserverpid;
    /* Address info */

int  _Emonitor = 0;
    // Monitor available

int  _ENomonitor = 0;
    // Don't ever monitor
    
Val  _Emonitorpid;
    /* Address info */

Val  _Egroupleader;
    /* Address info of group leader */


static fd_set port_ifds;
static fd_set port_ofds;
static fd_set port_efds;
static int fd_max = -1;
    // File descriptor sets for handling ports

/*
** Initiliase an Erlang node at a specific port 
** Fork (& exec?) if we don't exist already and create an Erlang node handler
*/
void Einitserver(unsigned int addr, unsigned int port)
{
#if 0
fprintf(stderr,"Einitserver: %d, %d\n", addr, port);
#endif
    if (_Eserver != 0) Eclose(_Eserver);
    Einitcall();

    FD_ZERO(&port_ifds);
    FD_ZERO(&port_ofds);
    FD_ZERO(&port_efds);

    if (addr == 0)
    {
        /* addr = LOCAL_HOST; */
        addr = INADDR_ANY;
    }
    if (port == 0)
    {
        port = DEFAULT_NODE_PORT;
    }

    /* FIX: getpid() gets our own pid! Not the servers... */
    /* Do we need to have the servers PID at all? No */
#ifdef _ALLOCS_H
    _Eserverpid = Emake_tuple(tuplemalloc(3), EValint(addr), 
                                 EValint(port), 
                                 EValint(getpid()));
#else
    _Eserverpid = Emake_tuple(3, EValint(addr), 
                                 EValint(port), 
                                 EValint(getpid()));
#endif
    _Eserver = Ecreate_socket(AF_INET, SOCK_STREAM, 0);

    /* Create a socket (_Eserver) point at the server */
    if (Econnect(_Eserver, _Eserverpid) == 0)
    {
        /* throw an error? - or start something up*/
        printf("Warning: now running in non-message passing, uni-process mode.\n");
        return;
    }

#ifdef NONBLOCKING
    /* should they be non-blocking?   */
    if (fcntl(_Eserver->socket_no, F_SETFL, O_NDELAY) == -1)
            perror("fnctl(_Eserver: O_NDELAY)");
#endif

    /* should these even be seperate? */
    // _Eexternal = _Eserver;
    signal(SIGUSR1, message_handler);
}

void Ecloseserver()
{
#ifdef _ALLOCS_H
    Esendmonitor(EValatom(_tx0, "stop"));
#else
    Esendmonitor(EValatom("stop"));
#endif

    // Eshutdown(_Eserver, 2);
    Eclose(_Eserver);

    Eclosecall();
    // should close down all dlopen stuff too ...
    
    while (waitpid(0, 0, WNOHANG) > 0);
}

#if 0
{
    _Eserver = Ecreate_socket(AF_INET, SOCK_STREAM, 0);
    if (Ebind(_Eserver, port) == 0)
    {
        fprintf(stderr, "Erlang node unable to bind to port (%d), no messaging!\n", port);
        return;
    }
}
#endif

void Einitmonitor(unsigned int addr, unsigned int port, unsigned int pid)
{
}

/*
** The following functions relate to actually packaging
** messages to send and receive
**
** Note: FIX: upper message size so that it calculates the size dynamically
** Returns: the amount of the buffer we used up (in bytes)
**
*/
#define MAX_FLAT 65536


/* copying of values into char * */
static void copy_in_4(unsigned char * loc, int val)
{
    loc[0] = (val & 0xff000000) >> 24;
    loc[1] = (val & 0x00ff0000) >> 16;
    loc[2] = (val & 0x0000ff00) >> 8;
    loc[3] = (val & 0x000000ff);
}

static void copy_in_2(unsigned char * loc, int val)
{
    loc[0] = (val & 0x0000ff00) >> 8;
    loc[1] = (val & 0x000000ff);
}

int extract_4(const unsigned char * loc)
{
    return (loc[0] << 24) | (loc[1] << 16) | (loc[2] << 8) | (loc[3]);
}

int extract_2(const unsigned char * loc)
{
    return (loc[0]) << 8 | (loc[1]);
}

static int set_value(char * loc, Val message)
{
    /* FIX: need to check in various places we aren't too long */
    switch (message.tag & LT_single)
    {
        case LT_bool:
            loc[0] = Ext_bool;
            loc[1] = message.u.number; 
            return 2;
        case LT_number:
            if (message.u.number > 0 && message.u.number < 255)
            {
                loc[0] = Ext_small_integer;
                loc[1] = message.u.number;
                return 2;
            }
            else
            {
                loc[0] = Ext_integer;
                copy_in_4(&(loc[1]), message.u.number);
                return 5;
            }
        case LT_real:
            loc[0] = Ext_float;
            sprintf(&loc[1], "%.20e", message.u.real);
            return 32;
        case LT_atom:
            {
                int len;
                loc[0] = Ext_atom;
                len = strlen(message.u.atom);  /* BAD: atoms can contain '\0' */
                copy_in_2(&(loc[1]), len);
#if 0
fprintf(stderr, "packing LT_atom (%d, %d): %s\n", loc[1], loc[2], message->u.atom);
#endif
                memcpy(&(loc[3]), message.u.atom, len);
                return 3 + len;
            }
        case LT_tuple:
            {
                int result, offset, i;
                if (message.u.tuple->size < 256)
                {
                    loc[0] = Ext_small_tuple;
                    loc[1] = message.u.tuple->size;
                    result = 2;
                }
                else
                {
                    loc[0] = Ext_large_tuple;
                    copy_in_4(&(loc[1]), message.u.tuple->size);
                    result = 5;
                }
                for (i = 0; i < message.u.tuple->size; i++)
                {
                    offset = set_value(&(loc[result]), message.u.tuple->e[i]);
                    result = result + offset;
                }
                return result;
            }
        case LT_list:
#ifdef _CONS_H
            {
                int result, i, offset, size = 0;
                _Lelement * le;

                loc[0] = Ext_list;
                result = 5;
                le = message.u.list;
                while (le != NULL)
                {
                    offset = set_value(&(loc[result]), le->e);
                    result = result + offset;
                    le = le->next;
                    size++;
                }
                copy_in_4(&(loc[1]), size);
                return result;
            }
#else
            {
                int result, i, offset;

                loc[0] = Ext_list;
                copy_in_4(&(loc[1]), message.u.list->length);
                result = 5;
                for (i = 0; i < message.u.list->length; i++)
                {
                    offset = set_value(&(loc[result]), Eindex_list(message,i));
                    result = result + offset;
                }
                return result;
            }
#endif
        case LT_string:
            {
                int len;

                loc[0] = Ext_string;
                len = strlen(message.u.string);
                copy_in_2(&(loc[1]), len);
                strncpy(&loc[3], message.u.string, len);
                return 3 + len;
            }
        case LT_pid:
            {
                int i;

                assert (message.u.tuple->size == 3);
            
                loc[0] = Ext_pid;
                for (i = 0; i < 3; i++)
                {
                    copy_in_4(&(loc[1]), message.u.tuple->e[0].u.number);
                    copy_in_4(&(loc[5]), message.u.tuple->e[1].u.number);
                    copy_in_4(&(loc[9]), message.u.tuple->e[2].u.number);
                }
                return 13;
            }
        case LT_binary:
            {
                int len;

                loc[0] = Ext_binary;
                len = message.u.binary->size;
                copy_in_4(&(loc[1]), len);
                memcpy(&loc[3], message.u.binary->b, len);
                return 5 + len;
            }
#if 0
        case LT_control:
            loc[0] = Ext_control;
            loc[1] = message.u.number;
            return 2;
#endif
        default:
            Ethrowv("Illegal basic value (type %d)\n", message.tag);
            break;
    }
    return 0;
}

static int set_binary(char * loc, Val message)
{
    /* FIX: need to check in various places we aren't too long */
    switch (message.tag & LT_single)
    {
        case LT_bool:
            loc[0] = message.u.number; 
            return 1;
        case LT_number:
            if (message.u.number > 0 && message.u.number < 255)
            {
                loc[0] = message.u.number;
                return 1;
            }
            else
            {
                copy_in_4(&(loc[0]), message.u.number);
                return 4;
            }
        case LT_real:
            sprintf(&loc[0], "%.20e", message.u.real);
            return 31;
        case LT_atom:
            {
                int len;
                len = strlen(message.u.atom);  /* BAD: atoms can contain '\0' */
                memcpy(&(loc[0]), message.u.atom, len);
                loc[len] = '\0';
                return 1 + len;
            }
        case LT_tuple:
            {
                int result, offset, i;
                loc[0] = '{';
                result = 1;
                for (i = 0; i < message.u.tuple->size; i++)
                {
                    offset = set_value(&(loc[result]), message.u.tuple->e[i]);
                    result = result + offset;
                    loc[result++] = ',';
                }
                if (message.u.tuple->size == 0)
                    loc[result] = ']';
                else
                    loc[result-1] = ']';
                return result;
            }
        case LT_list:
#ifdef _CONS_H
            {
                int result, i, offset, size = 0;
                _Lelement * le;

                loc[0] = '['
                result = 1;
                le = message.u.list;
                while (le != NULL)
                {
                    offset = set_value(&(loc[result]), le->e);
                    result = result + offset;
                    loc[result++] = ',';
                    le = le->next;
                    size++;
                }
                loc[result-1] = ']';
                return result;
            }
#else
            {
                int result, i, offset;

                loc[0] = '[';
                result = 1;
                for (i = 0; i < message.u.list->length; i++)
                {
                    offset = set_value(&(loc[result]), Eindex_list(message,i));
                    result = result + offset;
                    loc[result++] = ',';
                }
                if (message.u.list->length == 0)
                    loc[result] = ']';
                else
                    loc[result-1] = ']';
                return result;
            }
#endif
        case LT_string:
            {
                int len;

                len = strlen(message.u.string);
                strncpy(&loc[0], message.u.string, len);
                loc[len] = '\0';
                return 1+len;
            }
        case LT_pid:
            {
                int i, result, offset;

                result = 0;
                for (i = 0; i < message.u.tuple->size; i++)
                {
                    offset = set_value(&(loc[result]), message.u.tuple->e[i]);
                    result = result + offset;
                    loc[result++] = ',';
                }
                loc[result-1] = '\0';
                return result;
            }
        case LT_binary:
            {
                int len;

                len = message.u.binary->size;
                memcpy(&loc[0], message.u.binary->b, len);
                return len;
            }
#if 0
        case LT_control:
            loc[0] = Ext_control;
            loc[1] = message.u.number;
            return 2;
#endif
        default:
            Ethrowv("Illegal basic value (type %d)\n", message.tag);
            break;
    }
    return 0;
}
static Val extract_value(const unsigned char * loc)
{
    LenVal x;

    x = lvextract(loc);

    return x.v;
}

static LenVal lvextract(const unsigned char * loc)
{
    LenVal ret;

#if 0
fprintf(stderr, "lvextract: type(%d)\n", loc[0]);
#endif
    switch (loc[0])
    {
        case Ext_bool:
            ret.len = 2;
            ret.v = EValbool(loc[1]);
            return ret;
        case Ext_small_integer:
            ret.len = 2;
            ret.v = EValint((unsigned int)loc[1]); 
            return ret;
        case Ext_integer:
            ret.len = 5;
            ret.v = EValint(extract_4(&(loc[1])));
            return ret;
        case Ext_float:
            {
                Val v = EValreal(0);

                ret.len = 5;
                sscanf(&loc[1], "%lf", &(v.u.real));
                ret.v = v;
                return ret;
            }
        case Ext_atom:
            {
                int len;
                len = extract_2(&(loc[1]));
#if 0
fprintf(stderr, "Ext_atom: len(%d)\n", len);
#endif
                /* FIX: we don't use the length at the moment */
                ret.len = 3 + len;
                ret.v = EValmatom(&(loc[3]), len);
                return ret;
            }
        case Ext_pid:
            {
                Val v;
                
                v = tuplemalloc(3);
                v.u.tuple->e[0].u.number = extract_4(&loc[1]);
                v.u.tuple->e[1].u.number = extract_4(&loc[5]);
                v.u.tuple->e[2].u.number = extract_4(&loc[9]);
                v.tag = LT_pid;
                
                ret.len = 13;
                ret.v = v;
                return ret;
            }
        case Ext_small_tuple:
            {
                int len, i, offset;
                Val v;

                len = loc[1];
                v = tuplemalloc(len);                  
                offset = 2;
                for (i = 0; i < len; i++)
                {
                    LenVal x;
                    x = lvextract(&(loc[offset]));
                    v.u.tuple->e[i] = x.v;
                    offset = offset + x.len;
                }
                ret.v = v;
                ret.len = offset;
                return ret;
            }
        case Ext_large_tuple:
            {
                int len, i, offset;
                Val v;

                len = extract_4(&(loc[1]));
                v = tuplemalloc(len);                  
                offset = 5;
                for (i = 0; i < len; i++)
                {
                    LenVal x;
                    x = lvextract(&(loc[offset]));
                    v.u.tuple->e[i] = x.v;
                    offset = offset + x.len;
                }
                ret.v = v;
                ret.len = offset;
                return ret;
            } 
        case Ext_list:
#ifdef _CONS_H
            {
                int len, i, offset, start = 0;
                _Lelement * lblock, * trav;
                Val v;

                len = extract_4(&(loc[1]));
                v.tag = LT_list;
                lblock = make_block_lelement(len);
                v.u.list = lblock;
                trav = lblock;
                i = 0;
                offset = 5;
                while (trav)
                {
                    LenVal x;
                    x = lvextract(&(loc[offset]));
                    trav->e = EValcopy(x.v);
                    if (start == 0)
                    {
                        trav->e.tag |= LT_on_heap;
                        start = 1;
                    }
                    else
                    {
                        trav->e.tag &= ~LT_on_heap;
                    }
                    offset = offset = x.len;
                    trav = trav->next;
                }
                ret.v = v;
                ret.len = offset;
                return ret;
            } 
#else
            {
                int len, i, offset;
                Val v;

                len = extract_4(&(loc[1]));
                v = listmalloc(len);                  
                offset = 5;
                for (i = 0; i < len; i++)
                {
                    LenVal x;
                    x = lvextract(&(loc[offset]));
                    Eindex_list(v,i) = x.v;
                    offset = offset = x.len;
                }
                ret.v = v;
                ret.len = offset;
                return ret;
            } 
#endif
        case Ext_string:
            {
                int len;
                len = extract_2(&(loc[1]));
                /* FIX: we don't use the length at the moment */
                ret.len = 1 + len;
                ret.v = EValmstring(&(loc[3])); 
                return ret;
            }
        case Ext_binary:
            {
                int len;
                Val v;

                len = extract_4(&(loc[1]));
                v = EValmbinary(&(loc[5]), len);
                // memcpy(v.u.binary->b, &(loc[5]), len);
                ret.v = v;
                ret.len = 5 + len;
                return ret;
            }
#if 0
        case Ext_control:
            {
                return EValcontrol(loc[1]);
            }
#endif
        default:
            Ethrowv("Illegal external value %d.\n", loc[0]);
            break;
    }
    ret.len = 0;
    return ret;
}

/*
** Link to a process for error reporting
*/
void EClink(Val topid)
{
    EBinary chunk;

    chunk = binaryalloc(5);
    chunk->b[0] = C_link;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(topid, Pid_pid).u.number);
    Esendpidbin(topid, chunk);
    binaryfree(chunk);
}

/*
** Send a message to another pid
*/
int ECsend(Val topid, Val term)
{
    EBinary chunk;
    int size;

    chunk = binaryalloc(MAX_FLAT);
    chunk->b[0] = C_send;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(topid, Pid_pid).u.number);
    size = set_value(&(chunk->b[5]), term) + 5;
    chunk->size = size;
#if 0
    printf("ECsend: to %d (%d bytes)\n", 
        Eindex_tuple(topid, Pid_pid).u.number, chunk->size);
#endif
    if (Esendpidbin(topid, chunk) < chunk->size)
    {
        perror("ECsend");
        binaryfree(chunk);
        return 0;
    }

    binaryfree(chunk);
    return 1;
}

/*
** Send a message to an open port
*/
void Psend(Val port, Val term)
{
    EBinary chunk;
    int size;
    int binary = 1;

    // FIX: should determine if port is a binary port or not
    //      determines style of setting
    
    printf("port send to fd %d\n", port.u.tuple->e[1].u.number);

    if (binary)
    {
        chunk = binaryalloc(MAX_FLAT);
        size = set_binary(&(chunk->b[0]), term);
        chunk->size = size;
        if (write(port.u.tuple->e[1].u.number, &chunk->b[0], size) == -1)
        {
            perror("Psend (binary)");
        }
        binaryfree(chunk);
    }
    else
    {
        chunk = binaryalloc(MAX_FLAT);
        size = set_value(&(chunk->b[0]), term);
        chunk->size = size;
        // printf("port send to fd %d\n", port.u.tuple->e[1].u.number);
        if (write(port.u.tuple->e[1].u.number, &chunk->b[0], size) == -1)
        {
            perror("Psend");
        }
        binaryfree(chunk);
    }
}

void Pread(Val port, Val term)
{
    EBinary chunk;
    int size;

    chunk = binaryalloc(MAX_FLAT);
    size = set_value(&(chunk->b[0]), term);
    chunk->size = size;
    if (read(port.u.tuple->e[2].u.number, &chunk->b[0], size) == -1)
    {
        perror("Pread");
    }
    binaryfree(chunk);
}

/*
** Force a process to exit
*/
void ECexit(Val topid)
{
    EBinary chunk;

    chunk = binaryalloc(5);
    chunk->b[0] = C_exit;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(topid, Pid_pid).u.number);
    Esendpidbin(topid, chunk);
    binaryfree(chunk);
}

/*
** Unlink a process from error reporting to another
*/
void ECunlink(Val topid)
{
    EBinary chunk;

    chunk = binaryalloc(5);
    chunk->b[0] = C_unlink;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(topid, Pid_pid).u.number);
    Esendpidbin(topid, chunk);
    binaryfree(chunk);
}

/*
** Send a message to a registered name
*/
void ECreg_send(Val topid, Val toname, Val term)
{
    EBinary chunk;
    int size;

    chunk = binaryalloc(MAX_FLAT);
    chunk->b[0] = C_send;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(topid, Pid_pid).u.number);
    size = set_value(&(chunk->b[5]), term) + 5;
    chunk->size = size;
    Esendpidbin(topid, chunk);
    binaryfree(chunk);
}

void ECgroup_leader(Val topid)
{
    EBinary chunk;
    int frompid = getpid();

    chunk = binaryalloc(17);
    chunk->b[0] = C_group_leader;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(_Eserverpid, Pid_ipaddr).u.number);
    copy_in_4(&(chunk->b[5]), Eindex_tuple(_Eserverpid, Pid_port).u.number);
    copy_in_4(&(chunk->b[9]), frompid);
    copy_in_4(&(chunk->b[13]), Eindex_tuple(topid, Pid_pid).u.number);
    Esendpidbin(topid, chunk);
    binaryfree(chunk);
}

/*
** Send an exit message with a reason
*/
void ECexit2(Val topid, Val reason)
{
    EBinary chunk;

    chunk = binaryalloc(MAX_FLAT);
    chunk->b[0] = C_exit2;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(topid, Pid_pid).u.number);
    chunk->size = set_value(&(chunk->b[5]), reason) + 5;
    Esendpidbin(topid, chunk);
    binaryfree(chunk);
}

void ECmagic(int magic_no, Val call)
{
    EBinary chunk;
    int size;
    Val topid = tuplecopy(_Eserverpid);
    
#if 0
    fprintf(stderr, "ECmagic: %d\n", magic_no);
#endif

    topid.u.tuple->e[Pid_pid].u.number = magic_no;
    chunk = binaryalloc(MAX_FLAT);
    chunk->b[0] = C_magic;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(topid, Pid_pid).u.number);
    size = set_value(&(chunk->b[5]), call) + 5;
    chunk->size = size;
    Esendpidbin(topid, chunk);
    binaryfree(chunk);
}

/*
** Send a "go-ahead" message to _Eserver
*/
void ECgoahead(int magic)
{
    EBinary chunk;
    int frompid = getpid();

    if (magic != 0)
        frompid = magic;

    chunk = binaryalloc(13);
    chunk->b[0] = C_goahead;
    copy_in_4(&(chunk->b[1]), Eindex_tuple(_Eserverpid, Pid_ipaddr).u.number);
    copy_in_4(&(chunk->b[5]), Eindex_tuple(_Eserverpid, Pid_port).u.number);
    copy_in_4(&(chunk->b[9]), frompid);
    Esendpidbinrecv(_Eserverpid, chunk);
    binaryfree(chunk);
}

/*
** Send a `magic-pid' message to _Eserver 
** Used in starting up processes so functions & args can be passed
**
** Note: magicnum should be unique && > max process #.
*/
void ECgoahead_magic(int magicnum, Val mod, Val fun, Val args)
{
    EBinary chunk;
    int loc;

    chunk = binaryalloc(MAX_FLAT);
    chunk->b[0] = C_goahead_magic;
    copy_in_4(&(chunk->b[1]), magicnum);
    loc = 5;
    loc = loc + set_value(&(chunk->b[loc]), mod);
    loc = loc + set_value(&(chunk->b[loc]), fun);
    loc = loc + set_value(&(chunk->b[loc]), args);
    chunk->size = loc;
    Esendpidbinrecv(_Eserverpid, chunk);
    binaryfree(chunk);
}

/*
** Below are the calls related to language support for 
** facilities send/receive 
*/

static void message_handler(int signo)
{
    signal(SIGUSR1, message_handler);
    messages_waiting++;
    /* alternatively we could actually fetch the message here? */
}

/*
** Ereceive: performs the internal receive mechanism (in milliseconds)
** Apparently 0 should be instantaneous release
** and infinity should be wait.
**
*/
#define M_SLEEP_TIME 1

Val Ereceive(Val waittime)
{
    Val ret = valalloc(LT_none);
    EBinary retbuf;
    unsigned int sleeptime = 0;

    TC(waittime, LT_number, 1, "receive"); 
    /* note: we should check for atom `infinity' too strictly speaking */
    
    while ((messages_waiting == 0) 
            && ((sleeptime<=(unsigned int)waittime.u.number))) 
    {
        sleeptime += M_SLEEP_TIME;
        usleep(M_SLEEP_TIME * 1000);
    }

    /* check to see if we timed out */
    if (messages_waiting == 0 && random()%100) return ret;

    /* otherwise get the message    */
    ECgoahead(0);
    retbuf = Erecv(_Eexternal, MAX_MESSAGE);
    Eclose(_Eexternal);
    _Eexternal = 0;
#if 1
    if (retbuf == NULL) return ret;
#else
    if (retbuf == NULL)
    {
        messages_waiting = 0;
    }
#endif
    ret = extract_value(retbuf->b);
    /* free retbuf? */

    return ret;
}

/* 
** Message(Port, Msg) when integer(Port) ->
** Pids < ip#, port, unixpid >
**
** Should be build support for persistent connections?
** (ie. some sort of delayed shutdown of the socket)
**
** Returns: the message we sent (it doesn't block and returns immediately)
** 
*/
Val Esend(Val pid, Val message)
{
    /* TC(pid, LT_pid, 1, "message send"); */
    TC(pid, LT_tuple, 1, "message send"); 
    // Port send or Process send?
    if (pid.u.tuple->size == 4) Psend(pid, message);
    else ECsend(pid, message);
    return message;
}

void Esetmonitor(int pid)
{
    if (!_ENomonitor)
    {
        _Emonitor = pid;
        // printf("Esetmonitor %d\n", pid);
        // FIX: address/port should match server ..
#ifdef _ALLOCS_H
        _Emonitorpid = Emake_tuple(tuplemalloc(3), EValint(INADDR_ANY),
                    EValint(DEFAULT_NODE_PORT), EValint(pid));
#else
        _Emonitorpid = Emake_tuple(3, EValint(INADDR_ANY),
                    EValint(DEFAULT_NODE_PORT), EValint(pid));
#endif
    }
}

void ENomonitor()
{
    _ENomonitor = 1;
}

int MonitoringOn()
{
    return !_ENomonitor;
}

Val Esendmonitor(Val message)
{
    if (_Emonitor != 0)
    {
        // printf("Esendmonitor %d: %s\n", _Emonitor, message.u.atom);
        if (!ECsend(_Emonitorpid, message))
        {
            // kill the monitor anyway?
            kill(Eindex_tuple(_Emonitorpid, Pid_pid).u.number, SIGUSR1);
        }
    }
    return message;
}

#if 0
static int Esendport(Val pid, EBinary str)
{
    return 0;
}
#endif

static int Esendpidbin(Val pid, EBinary str)
{
    int retval = 0;

    if (_Eexternal)
    {
        if (!((_Eexternal->int_address == Eindex_tuple(pid, 
                                                Pid_ipaddr).u.number) &&
             (_Eexternal->port == Eindex_tuple(pid, Pid_port).u.number)))
        {
            fprintf(stderr, "Esendpidbin: shutting down _Eexternal (to server)\n");
            fprintf(stderr, "Esendpidbin: addr: %d = %d, port: %d = %d\n",
                        _Eexternal->int_address, Eindex_tuple(pid, Pid_ipaddr).u.number, _Eexternal->port, Eindex_tuple(pid, Pid_port).u.number);
            Eclose(_Eexternal);
            _Eexternal = 0;
        }
    }

    /* should we make this non-blocking? */
    if (_Eexternal == 0)
    {
        // fprintf(stderr, "Esendpidbin: creating new socket\n");
        _Eexternal = Ecreate_socket(AF_INET, SOCK_STREAM, 0);
        if (Econnect(_Eexternal, pid) == 0) return 0;
    }

#if 0
    fprintf(stderr, "Esendpidbin: send binary on (%d)\n",
            _Eexternal->socket_no);
#endif

    retval = send(_Eexternal->socket_no, str, str->size+sizeof(str->size), 0);
    Eclose(_Eexternal);
    _Eexternal = 0;

    return retval;
}

static int Esendpidbinrecv(Val pid, EBinary str)
{
    if (_Eexternal)
    {
        if (!((_Eexternal->int_address == Eindex_tuple(pid, 
                                                Pid_ipaddr).u.number) &&
             (_Eexternal->port == Eindex_tuple(pid, Pid_port).u.number)))
        {
            fprintf(stderr, "Esendpidbin: shutting down _Eexternal (to server)\n");
            fprintf(stderr, "Esendpidbin: addr: %d = %d, port: %d = %d\n",
                        _Eexternal->int_address, Eindex_tuple(pid, Pid_ipaddr).u.number, _Eexternal->port, Eindex_tuple(pid, Pid_port).u.number);
            Eclose(_Eexternal);
            _Eexternal = 0;
        }
    }

    /* should we make this non-blocking? */
    if (_Eexternal == 0)
    {
        // fprintf(stderr, "Esendpidbin: creating new socket\n");
        _Eexternal = Ecreate_socket(AF_INET, SOCK_STREAM, 0);
        if (Econnect(_Eexternal, pid) == 0) return 0;
    }

#if 0
    fprintf(stderr, "Esendpidbin: send binary on (%d)\n",
            _Eexternal->socket_no);
#endif
    return send(_Eexternal->socket_no, str, str->size+sizeof(str->size), 0);
    // return Esendbinary(_Eexternal, str);
}


/*
** Returns: a 3-tuple which is {module,function,{list of args}}
*/
Val Ereceivemagic(int magic)
{
    Val ret;
    EBinary retbuf;

    ECgoahead(magic);
#ifdef NONBLOCKING
    while ((retbuf = Erecv(_Eexternal, MAX_MESSAGE)) != -1 && 
            errno == EWOULDBLOCK)
    {
        sleep(1);
        ECgoahead(magic);
    }
#else
    retbuf = Erecv(_Eexternal, MAX_MESSAGE);
    Eclose(_Eexternal);
    _Eexternal = 0;
#endif
    if (retbuf == NULL)
    {
        perror("Ereceivemagic failed\n");
        Ethrowv("Unable to receive startup instructions from server.\n");
    }
    ret = extract_value(retbuf->b);
    /* FIX: free retbuf? */
    messages_waiting--; 


    return ret;
}


//
// FIX: incomplete
// Some port handling stuff.
//
void Eportregister(Val port)
{
    FD_SET(port.u.tuple->e[1].u.number,&port_ifds);
    FD_SET(port.u.tuple->e[2].u.number,&port_ofds);
    FD_SET(port.u.tuple->e[3].u.number,&port_efds);

    if (port.u.tuple->e[1].u.number > fd_max) fd_max = port.u.tuple->e[1].u.number;
    // etc
}

void Eportremove(Val port)
{
    FD_CLR(port.u.tuple->e[1].u.number, &port_ifds);
    FD_CLR(port.u.tuple->e[2].u.number, &port_ofds);
    FD_CLR(port.u.tuple->e[3].u.number, &port_efds);
}

Val Eportreceive()
{
    struct timeval tm;
    int i;
    fd_set ofds, ifds, efds;
    Val dummy;
 
    tm.tv_sec = 0;
    tm.tv_usec = 10;
 
    ifds = port_ifds;
    //ofds = port_ofds;
    //efds = port_efds;
    select(fd_max, &ifds, 0, 0, &tm);
 
    for (i = 0; i < fd_max; i++)
    {
        if (FD_ISSET(i, &ifds))
        {
            // read and return 
        }
    }

    return dummy;
}


