%{


//
// $Id: erlang.y,v 1.128 2004/09/05 02:42:56 geoff Exp $
//
// $Source: /home/cvs/gerl/src/erlang.y,v $
// $Revision: 1.128 $
// $Date: 2004/09/05 02:42:56 $
// $State: Exp $
//
// Author: Geoff Wong
//
// Geoff Wong - Copyright 1997 - 2002.
//
//
// @@DOC
// <h2> The Parser </h2>
// 
// The grammar for the Erlang parser was taken from the book
// "Concurrent Programming in ERLANG", Armstrong et al, 1993.
// It implements the Erlang 4.2 (or 3?) specifiction.
// The yacc grammar itself constructs an abstract syntax tree
// which is further processed depending on the type of compilation
// being performed.
// @@ENDDOC
//


#include <stdarg.h>
#include <stdio.h>
#include <list>
#include <iostream>

#include "nodes.h"
#include "guard.h"
#include "binary.h"
#include "pattern.h"
#include "match.h"
#include "program.h"
#include "expr.h"
#include "case.h"
#include "farity.h"
#include "pragma.h"
#include "define.h"
#include "yy.h"
#include "str.h"
#include "tuple.h"
#include "function.h"
#include "catch.h"
#include "namer.h"
#include "record.h"
#include "unary.h"
#include "list_comp.h"

extern int current_line;
extern char * current_file;

extern int debug;

Program 			* yy_program;
static int			clause_id = 0;
static int			cr_id = 0;
static int			if_id = 0;
static int          tmp_val_count = 0;
// static int          tmp_alloc_count = 0;
static Symbol       * retsymbol = 0;
// static SymbolTable  * function_table;
static FunctionList * active_function;

int yyerrorf(const char *fmt, ...);

%}
%union {
	int number;
    float real;
	char * cstring;
	char * var;
	string * stringp;
	int truth;
	Node * node;
	NodeList * nodelist;
	CaseList * caselist;
	ExprList * exprlist;;
	FunctionList * funclist;
	FormList * formlist;
	Attribute * attribute;
	DefClause * defclause;
}


%token T_bor  T_bxor T_bsl T_bsr T_equal T_notequal T_le T_lt
%token T_ge T_gt T_exactequal T_div T_rem T_band T_bnot T_atom
%token T_number T_string T_var T_bool T_begin T_catch T_concat T_listsub
%token T_case T_of T_if T_receive T_after T_end T_when T_arrow T_dot
%token T_define T_defbody T_record T_fun T_such T_generator T_rechash

%token T_form T_plus T_minus T_multiply T_divide T_idiv T_exprtail
%token T_patternlist T_patterntail T_pattern T_tuple T_bracket T_exprs
%token T_list T_funcall T_remotefuncall T_ifclause T_match T_many
%token T_guardtail T_guardcall T_biftest T_guardtest T_guardtests
%token T_funcclause T_clausehead T_clausebody T_farity T_attribute
%token T_patterntuple T_function T_send T_pragma T_endpragma T_text T_real

%nonassoc CATCH
%right '='
%right SEND_OP '!' T_send
%left COMP_OP T_lt T_le T_gt T_ge T_notequal T_equal T_exactequal
%left LIST_OP T_concat T_listsub
%left ADD_OP '+' '-' T_bor T_bxor T_bsl T_bsr
%left MULT_OP '*' '/' T_div T_band T_rem
%nonassoc UNARY UPLUS UMINUS T_bnot

%type <defclause> def_clause
%type <stringp> T_atom T_real T_defbody
%type <number> T_number T_bool 
%type <stringp> mult_op add_op prefix_op comp_op 
%type <stringp> T_string T_var T_text T_define
%type <node> form pragma 
%type <node> pattern_tuple pattern cr_head
%type <node> expr tuple basic_type record_expr
%type <node> function_call cr_clause 
%type <node> case_expr if_expr receive_expr
%type <node> match_expr send_expr guard_expr 
%type <node> guard_expr_tuple 
%type <node> guard_call bif_test guard_test
%type <node> guard function_clause clause_head 
%type <node> clause_guard clause_body receive_body
%type <node> farity if_clause if_head
%type <node> record_field basic_field anon_function
%type <nodelist> expr_tail if_clauses 
%type <nodelist> guard_tests list farity_list
%type <nodelist> guard_expr_tail guard_expr_list list_comp_elements
%type <caselist> cr_clauses
%type <exprlist> exprs parameter_list guard_parameter_list guard_exprs 
%type <exprlist> pragma_text
%type <funclist> function
%type <formlist> form_list 
%type <nodelist> formal_parameter_list patterns pattern_tail pattern_list 
%type <nodelist> def_params record_fields basic_fields possible_basic_fields
%type <attribute> attribute
%%
program : form_list
	{ 	
		yy_program->ast = $1; 
	}

form_list : 	form T_dot
	{	
		$$ = new FormList($1);
		$$->set_program(yy_program);
        $$->type(LT_number);
	}
	|	form T_dot form_list
	{ 	
		$3->prepend($1);
		$$ = $3;	
	}

form :	'-' T_atom '(' attribute ')'
	{ 	
        if (*($2) == "module")
        {
//          printf("module value:%s\n", ((Form *)(*i))->value->emit_code().chars());
            yy_program->set_module($4->print());
        }
		else if (*($2) == "import")
		{
            yy_program->add_imports($4);
		}
		else if (*($2) == "export")
		{
            yy_program->add_exports($4);
		}
		$$ = new Form($2, $4); 
        $$->type(LT_number);
	}
    |  T_record '(' T_atom ',' '{' record_fields '}' ')'
    {
        if (debug)
            fprintf(stderr,"adding -record(%s, exprs)\n", (*$3).c_str());
        yy_program->add_record(*($3), $6);
        $$ = NULL;
    }
    |  T_define '(' def_clause T_defbody ')'
    {
        if (debug)
            fprintf(stderr,"adding -define(X, %s)\n", (*$4).c_str());
        $3->body(*$4);
        add_define($3);
        $$ = NULL;
    }
	|	
    {
        active_function = new FunctionList(); 
    }
    function
	{	 
		$$ = $2;
		$<funclist>$->set_program(yy_program);
	}
    |   pragma
    {
        $$ = $1;
        $$->type(LT_number);
    }

record_fields:
    record_field
    {   
        $$ = new NodeList($1); 
        $$->type($1->type());
    }
	|	record_field ',' record_fields
    {   
		$3->prepend($1); 
        $3->type($1->type() | $3->type());
		$$ = $3;
    }

record_field: 
    T_atom
    {   
        $$ = new Field(*($1), NULL); 
        $$->type(LT_all);
    }
    | T_atom '=' expr 
    {   
        $$ = new Field(*($1), $3); 
        $$->type($3->type());
    }

add_op :	'+'	
	{	$$ = new string("Eplus"); }
	|	'-'
	{	$$ = new string("Eminus"); }
	|	T_bor
	{	$$ = new string("Ebor"); }
	|	T_bxor
	{	$$ = new string("Ebxor"); }
	|	T_bsl
	{	$$ = new string("Ebsl"); }
	|	T_bsr
	{	$$ = new string("Ebsr"); }

mult_op	:	'*'
	{	$$ = new string("Emultiply"); }
	|	'/'
	{	$$ = new string("Edivide"); }
	|	T_div
	{	$$ = new string("Ediv"); }
	|	T_rem
	{	$$ = new string("Erem"); }
	|	T_band
	{	$$ = new string("Eband"); }

comp_op :	T_equal
	{	$$ = new string("Eequal"); }
	|	T_notequal
	{	$$ = new string("Enotequal"); }
	|	T_le
	{	$$ = new string("Elessequal"); }
	|	T_lt
	{	$$ = new string("Elessthan"); }
	|	T_ge
	{	$$ = new string("Egreaterequal"); }
	|	T_gt
	{	$$ = new string("Egreaterthan"); }
	|	T_exactequal
	{	$$ = new string("Eexactequal"); }

prefix_op :	'+'
	{	$$ = new string("Epreplus"); }
	|	'-'
	{	$$ = new string("Epreminus"); }
	|	T_bnot
	{	$$ = new string("Ebnot"); }

basic_type :	T_atom
	{ 
        string s;

        if (CurrentTable != NULL)
        {
#ifdef _ALLOCS_H
		    s = CurrentTable->add_temp();
#endif
            $$ = new Atom(s, $1); 
        }
        else
        {
            s = "__vvv";
            $$ = new Atom(s, $1); 
        }
        $$->type(LT_atom | LT_constant | LT_on_heap);
    }
	|	T_number
	{ 
        $$ = new Number($1); 
        $$->type(LT_number | LT_constant);
    }
	|	T_real
	{ 
        $$ = new Real(*($1)); 
        $$->type(LT_real | LT_constant);
    }
	|	T_string
	{ 
        string s;

#ifdef _ALLOCS_H
        if (CurrentTable != NULL)
        {
		    s = CurrentTable->add_temp();
        }
#endif

        $$ = new Estring(s, $1); 
        $$->type(LT_list | LT_on_heap);
    }
	|	T_bool
	{ 
        $$ = new Boolean($1); 
        $$->type(LT_bool | LT_constant);
    }
    | T_define
    {
        $$ = new Define(*($1));
    }
    | T_define '(' parameter_list ')'
    {
        $$ = new Define(*($1));
    }
    | record_expr
    {
        $$ = $1;
    }

record_expr:
    T_var T_rechash T_atom T_dot T_atom
    {
        $$ = new Record(new Variable($1), *($3), *($5)); 
        //$$ = new Record($1, *($3), *($5)); 
        //printf("blah#rec.x executed\n");
    }
    | '(' record_expr ')' T_rechash T_atom T_dot T_atom
    {
        $$ = new Record($2, *($5), *($7)); 
    }
	|	T_var
	{ 
        Symbol * s = NULL;
        
        if (CurrentTable != NULL)
        { 
            s = CurrentTable->find(*($1), S_Variable);
        }

        if (s != NULL) 
        {
            $$ = s->associated();
            // $$->type(s->type());
        }
        else
        {
            $$ = new Variable($1); 
            $$->type(LT_normal);
        }
    }
    | T_rechash T_atom '{' possible_basic_fields '}'
    {
        $$ = new NewRecord(*($2), $4); 
    }


possible_basic_fields: /* empty */
    { $$ = NULL; }
    | basic_fields
    { $$ = $1; }

basic_fields:
    basic_field
    {   
        $$ = new NodeList($1); 
        $$->type($1->type());
    }
	|	basic_field ',' basic_fields
    {   
		$3->prepend($1); 
        $3->type($1->type() | $3->type());
		$$ = $3;
    }

basic_field: 
    T_atom '=' expr 
    {   
        $$ = new Field(*($1), $3); 
        $$->type($3->type());
    }

pattern :	basic_type
	{ $$ = $1; }
	|	pattern_list
	{ $$ = $1; }
	|	pattern_tuple
	{ 
        $$ = $1; 
        $$->type($1->type() & ~LT_on_heap);
    }

pattern_list :	'[' ']'
	{ $$ = new PatternTail; }
	|	'[' pattern pattern_tail ']'
	{ 
		$3->prepend($2);
		$$ = $3;
	}

pattern_tail :	'|' pattern
	{ 	
        $$ = new PatternTail($2); 
        //$2->type($2->type() | LT_on_heap);
        $2->type($2->type() & ~LT_on_heap);
    }
	|	',' pattern pattern_tail
	{ 	
		$3->prepend($2); 
		$$ = $3;
	}
	|	/* empty */
	{	$$ = new PatternTail; }

pattern_tuple :	'{' '}'
	{ 
        string s;

        if (CurrentTable != NULL)
        {
		    s = CurrentTable->add_temp();
        }
        $$ = new PatternTuple(s); 
    }
	| 	'{' patterns '}'
	{ 
        string s;

        if (CurrentTable != NULL)
        {
		    s = CurrentTable->add_temp();
        }
        $$ = new PatternTuple(s, $2); 
        $$->type(LT_normal);
    }

patterns :	pattern
	{ $$ = new PatternList($1); }
	| 	pattern ',' patterns
	{ 
		$3->prepend($1); 
		$$ = $3;
	}

expr :	
    basic_type
    {   $$ = $1; }
    | T_var T_rechash T_atom '{' possible_basic_fields '}'
    {
        $$ = new NewRecord(*($1), *($3), $5); 
    }
	|	list
	{	
        $$ = $1; 
        $$->type(LT_list | LT_on_heap);
    }
	|	tuple
	{	
        $$ = $1; 
        $$->type(LT_tuple | LT_on_heap);
    }
    |   anon_function
    {
        $$ = $1;
        $$->type(LT_anyvar);
    }
	|	function_call
	{	
        $$ = $1; 
        $$->type($1->type() | LT_on_heap);
    }
	|	case_expr
	{ $$ = $1;  }
	|	if_expr
	{ $$ = $1; }
	|	receive_expr
	{ $$ = $1; }
	|	match_expr
	{ $$ = $1; }
	|	send_expr %prec SEND_OP
	{ $$ = $1; }
	|	expr '+' expr 	    %prec ADD_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new PlusExpr($1, $3); 
        if ( (($1->type() & LT_single) == LT_real) 
           ||(($3->type() & LT_single) == LT_real) )
            $$->type(LT_real);
        else if ( (($1->type() & LT_single) == LT_number) 
           && (($3->type() & LT_single) == LT_number) )
            $$->type(LT_number);
        else
            $$->type(LT_number | LT_real);
    }
	|	expr '-' expr 	    %prec ADD_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new MinusExpr($1, $3); 
        if ( (($1->type() & LT_single) == LT_real) 
           || (($3->type() & LT_single) == LT_real) )
            $$->type(LT_real);
        else if ( (($1->type() & LT_single) == LT_number) 
           && (($3->type() & LT_single) == LT_number) )
            $$->type(LT_number);
        else
            $$->type(LT_number | LT_real);
    }
	|	expr T_bor expr 	    %prec ADD_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new BorExpr($1, $3); 
        $$->type(LT_number);
    }
	|	expr T_bxor expr 	    %prec ADD_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new BxorExpr($1, $3); 
        $$->type(LT_number);
    }
	|	expr T_bsl expr 	    %prec ADD_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new BslExpr($1, $3); 
        $$->type(LT_number);
    }
	|	expr T_bsr expr 	    %prec ADD_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new BsrExpr($1, $3); 
        $$->type(LT_number);
    }
	|	expr '*' expr 	        %prec MULT_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new MultiplyExpr($1, $3); 
        if ( (($1->type() & LT_single) == LT_real) 
           ||(($3->type() & LT_single) == LT_real) )
            $$->type(LT_real);
        else if ( (($1->type() & LT_single) == LT_number) 
           && (($3->type() & LT_single) == LT_number) )
            $$->type(LT_number);
        else 
            $$->type(LT_number | LT_real);
        
    }
	|	expr '/' expr 	        %prec MULT_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new DivideExpr($1, $3); 
        if ( (($1->type() & LT_single) == LT_real) 
           ||(($3->type() & LT_single) == LT_real) )
            $$->type(LT_real);
        else if ( (($1->type() & LT_single) == LT_number) 
           && (($3->type() & LT_single) == LT_number) )
            $$->type(LT_number);
        else 
            $$->type(LT_number | LT_real);
    }
	|	expr T_rem expr 	    %prec MULT_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new RemExpr($1, $3); 
        $$->type(LT_number);
    }
	|	expr T_div expr 	    %prec MULT_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new DivExpr($1, $3); 
        $$->type(LT_number);
    }
	|	expr T_band expr 	    %prec MULT_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new BandExpr($1, $3); 
        $$->type(LT_number);
    }
    |   expr T_concat expr      %prec LIST_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new ConcatExpr($1, $3); 
        $$->type(LT_list);
    }
    |   expr T_listsub expr      %prec LIST_OP
    {
        $1->set_depth(1);
        $3->set_depth(1);
		$$ = new ListSubtract($1, $3); 
        $$->type(LT_list);
    }
    |   '-' expr            %prec UNARY
    {
        $2->set_depth(1);
		$$ = new UnaryMinus($2); 
        $$->type($2->type());
    }
    |   '+' expr            %prec UNARY
    {
        $2->set_depth(1);
		$$ = new UnaryPlus($2); 
        $$->type($2->type());
    }
    |   T_bnot expr            %prec UNARY
    {
        $2->set_depth(1);
		$$ = new UnaryBnot($2); 
        $$->type($2->type());
    }
	|	'(' expr ')'
	{ 
        $$ = new Bracket($2); 
        $$->type($2->type());
    }
	|	T_begin exprs T_end
	{ 
        // $2->delimiter(", ");
        $2->set_depth(1);
        $2->assign(retsymbol);
        $$ = $2; 
        $$->type($2->type());
    }
	|	T_catch expr 		%prec CATCH
	{ 
        $2->set_depth(1);
        $$ = new Catch($2); 
        // $$->assign(retsymbol);
        $$->type($2->type());
    }
    |   '[' expr T_such list_comp_elements ']'
    {
        $$ = new List_Comp($2, $4);
    }

pragma : T_pragma pragma_text T_endpragma
    { $$ = new Pragma($2); }

pragma_text : T_text
    { 
        string s;

        $$ = new ExprList(new Estring(s,$1)); 
    }
    | T_text pragma_text
    { 
        string s;
        Estring * p = new Estring(s,$1);

        $2->prepend(p);
        $$ = $2;
    }
 
list :		'[' ']'
	{	
        string s;

#ifdef _ALLOCS_H
        if (CurrentTable != NULL)
        {
		    s = CurrentTable->add_temp();
        }
#endif
        $$ = new EList(s); 
        // $$->type($$->type() | LT_on_heap);
    }
	| 	'[' expr expr_tail ']'
	{	
        $2->set_depth(1);
		$3->prepend($2);
        $3->set_depth(1);
		$$ = $3;
        // $$->type($$->type() | LT_on_heap);
	}

list_comp_elements :
    list_comp_element
    | list_comp_element ',' list_comp_element

list_comp_element :
    T_var T_generator expr 
//    | conditional_expr

expr_tail :	'|' expr
	{ 	
        string s, s1, s2, s3;

        if (CurrentTable != NULL)
        {
#ifdef _ALLOCS_H
		    s = CurrentTable->add_temp();
#endif
		    s1 = CurrentTable->add_temp();
		    s2 = CurrentTable->add_temp();
		    s3 = CurrentTable->add_temp();
        }
        $2->set_depth(1);
        $$ = new EList(s,s1,s2,s3,$2); 
    }
	| 	',' expr expr_tail
	{ 	
        $2->set_depth(1);
		$3->prepend($2);
		$$ = $3;
	}
	|	/* empty */
	{	
        string s;

        if (CurrentTable != NULL)
        {
#ifdef _ALLOCS_H
		    s = CurrentTable->add_temp();
#endif
        }
        $$ = new EList(s); 
    }

tuple :		'{' '}'
	{	
        string s;

#ifdef _ALLOCS_H
        if (CurrentTable != NULL)
        {
		    s = CurrentTable->add_temp();
        }
#endif
        $$ = new Tuple(s); 
    }
	|	'{' exprs '}'
	{	
        string s;

#ifdef _ALLOCS_H
        if (CurrentTable != NULL)
        {
		    s = CurrentTable->add_temp();
        }
#endif
        $2->set_depth(1);
        $$ = new Tuple(s, $2); 
    }

function_call : T_atom '(' parameter_list ')'
	{	
		string module;
        string fname = Namer::semi_mangle(*($1),$3->size());
        int x;
        string ts;
        Symbol * other, * freevar = NULL;
        // SymbolTable * function_table;
        FunctionCall * fcall;

		// Need to do some clever stuff in here with imported modules	
        if ( (other = yy_program->Table.find(fname, S_Function)) )
        {
            x = other->used();
            module = other->module();
            other->used(x+1);
        }

        if (module.length() == 0)
        {
		    module = yy_program->Module();
        }
#if 0
        fprintf(stderr, "CT->name() = %s_%d (%s:%s_%d)\n", 
            CurrentTable->name().c_str(), CurrentTable->params(),
            module.c_str(), (*($1)).c_str(), $3->size());
#endif
        if (CurrentTable == NULL)
        {
            $$ = NULL;
            break;
        }

        freevar = CurrentTable->add_freetemp();
		fcall = new FunctionCall(&module, $1, $3, freevar); 
        fcall->scope(CurrentTable);

        // need to find the Function symbol table 
        if ((module == yy_program->Module())
                && (CurrentTable->params() == $3->size())
                && (CurrentTable->name() == (*($1))))
        {
#if 0
            fprintf(stderr, "%s_%d setting LCO_localfun\n", (*($1)).c_str(),
                    $3->size());
#endif
            active_function->lco(true);
            fcall->LCO(LCO_localfun);
            fcall->stmt_type(ST_goto);
        }
        else if (module == yy_program->Module() || module == "bif" 
                 || module == "erlang")
        {
#if 0
            fprintf(stderr, "%s_%d setting LCO_local\n", (*($1)).c_str(),
                    $3->size());
#endif
            fcall->LCO(LCO_local);
        }
        $$ = fcall;
        $$->line(current_line);
	}
	|  	T_atom ':' T_atom '(' parameter_list ')'
	{	
        Symbol * freevar;

        FunctionCall * fcall;
        freevar = CurrentTable->add_freetemp();
        fcall = new FunctionCall($1, $3, $5, freevar); 
        fcall->scope(CurrentTable);
        fcall->LCO(LCO_external);
        $$ = fcall;
        $$->line(current_line);

    }
	|  	T_var ':' T_atom '(' parameter_list ')'
	{	
        Symbol * freevar;

        FunctionCall * fcall;
        freevar = CurrentTable->add_freetemp();
        fcall = new FunctionCall($1, $3, $5, freevar); 
        fcall->scope(CurrentTable);
        fcall->LCO(LCO_external);
        $$ = fcall;
        $$->line(current_line);

    }

parameter_list : exprs
	{	
        int i;
        string ts;
        Symbol * s;
        list<Node *>::iterator nit;

        $1->set_depth(1);
        $1->allocate_locals(tmp_val_count);
        for (i = 0, nit = $1->nodes.begin(); i < $1->size(); i++,nit++)
        {
 		    s = new Symbol("_vx" + itoa(tmp_val_count++), S_Variable, ts);
            // FIX: should be ok - but it propogates
            // s->type((*nit)->type());
            // (*nit)->freesym(s));
		    CurrentTable->add(s);
        }
        $$ = $1; 	
    }
	| 	/* empty */
	{	$$ = new ExprList(); 	}

case_expr :	T_case expr T_of cr_clauses T_end
	{	
		string ts = "";
		string label;
		Symbol * s;
		Case * cs;
		cs = new Case($2, $4);
        cs->assign(retsymbol);

		label = cs->label();
        if (debug) fprintf(stderr,"case(%s) at line %d\n", label.c_str(), current_line);
        assert(CurrentTable != NULL);

 		s = new Symbol(label, S_Variable, ts);
		CurrentTable->add(s);
 		s = new Symbol(label + "L", S_Boolean, ts);
		CurrentTable->add(s);

        $2->parent(cs);
        $4->set_depth(1);
        $4->set_parent(cs);

		$$ = cs;
        $$->type($4->type());
	}
	
cr_clause : cr_head clause_guard clause_body
	{	
		$$ = new CrClause($1, $2, $3);
        ((ClauseBody *)$3)->scope(CurrentTable);
        TableStack.pop_front();
        CurrentTable = TableStack.front();
        $1->add_symbol(CurrentTable);
        $$->type($3->type());
        // $3->parent($1);
	}


cr_head : 
    pattern
    {
		SymbolTable * s;
		string chid;
		string ts = "";

		chid = "cH" + itoa(cr_id++);
 		s = new SymbolTable(chid, S_Clause, ts, 0);
        // s->chain(CurrentTable);
        TableStack.push_front(s);
		CurrentTable = s;
		$$ = new CrHead(CurrentTable->symbol(),CurrentTable,$1); 
        $1->add_symbol(CurrentTable);
    }

cr_clauses : 	cr_clause
	{	
        $$ = new CaseList($1); 
        $$->type($1->type());
    }
	| 	cr_clause ';' cr_clauses
	{	
		$3->prepend($1); 
		$$ = $3;
        $$->type($1->type() | $3->type());
	}

if_expr : 	T_if if_clauses T_end
	{	
        $$ = new If($2); 
        $$->type($2->type());

        string ts;
 		Symbol * ifsymbol = new Symbol("_ifval" + itoa(if_id++), S_Variable, ts);
        $$->assign(ifsymbol);
        CurrentTable->add(ifsymbol);
    }

if_clause : 	if_head clause_body
	{	
        $$ = new IfClause($1, $2); 
        $2->set_depth(1);
        $2->set_flat(true);
        $$->type($2->type());
        // $2->parent($$);
        ((ClauseBody *)$2)->scope(CurrentTable);
        TableStack.pop_front();
        CurrentTable = TableStack.front();
    }

if_head :
    {
    }
    guard 
    {
		string ifid;
		string ts = "";
        SymbolTable * s;

		ifid = "iH" + itoa(cr_id++);
 		s = new SymbolTable(ifid, S_Clause, ts, 0);

        // s->chain(CurrentTable);

        TableStack.push_front(s);
		CurrentTable = s;
        $$ = $2;
    }

if_clauses :	if_clause
	{ 	
        $$ = new NodeList($1); 
        $$->type($1->type());
    }
	| 	if_clause ';' if_clauses
	{	
		$3->prepend($1); 
        $3->type($1->type() | $3->type());
		$$ = $3;
	}

receive_expr : 	T_receive T_after expr receive_body T_end
	{	
        Receive * r;
        string label, empty;
        Symbol * s;

        r = new Receive(NULL, $3, $4); 
        r->assign(retsymbol);
        label = r->label();
        if (CurrentTable)
        {
            s = new Symbol(label, S_Variable, empty);
            CurrentTable->add(s);
            s = new Symbol(label + "L", S_Boolean, empty);
            CurrentTable->add(s);
        }                          
        $4->set_depth(1);
        $$ = r;
        $$->type($4->type());
    }
	| 	T_receive cr_clauses T_end
	{
        Receive * r;
        string label, empty;
        Symbol * s;

        r = new Receive($2, NULL, NULL); 
        r->assign(retsymbol);
        label = r->label();
        if (CurrentTable)
        {
            s = new Symbol(label, S_Variable, empty);
            CurrentTable->add(s);
            s = new Symbol(label + "L", S_Boolean, empty);
            CurrentTable->add(s);
        }                          
        $2->set_depth(1);
        $2->set_parent(r);
        $$ = r;
        $$->type($2->type());
    }
	| 	T_receive cr_clauses T_after expr receive_body T_end
	{	
        Receive * r;
        string label, empty;
        Symbol * s;

        r = new Receive($2, $4, $5); 
        r->assign(retsymbol);
        label = r->label();
        $4->add_symbol(CurrentTable);
        if (CurrentTable)
        {
            s = new Symbol(label, S_Variable, empty);
            CurrentTable->add(s);
            s = new Symbol(label + "L", S_Boolean, empty);
            CurrentTable->add(s);
        }                          
        $2->set_depth(1);
        $2->set_parent(r);
        $4->set_depth(1);
        $5->set_depth(1);
        $$ = r;
        $$->type($2->type() | $5->type());
    }

receive_body :
    {
		string ifid;
		string ts = "";
        SymbolTable * s;

		ifid = "rH" + itoa(cr_id++);
 		s = new SymbolTable(ifid, S_Clause, ts, 0);

        TableStack.push_front(s);
		CurrentTable = s;
    }
    clause_body
    {
        ((ClauseBody *)$2)->scope(CurrentTable);
        TableStack.pop_front();
        CurrentTable = TableStack.front();
        $$ = $2;
    }


match_expr : 	expr '=' expr
	{ 
        string ts = "";
        string label;
        Match * mtch;
        Symbol * s;

        $1->set_depth(1);
        $3->set_depth(1);

        // Create a new "match" variable which is the RHS
        mtch = new Match($1, $3); 
		label = mtch->label();
 		s = new Symbol(label, S_Variable, ts);
		if (CurrentTable)
		{
			CurrentTable->add(s);
		}
        $$ = mtch;

        mtch->Table.chain(CurrentTable);

        // Now add the symbols into the Matches symbol table.
        $1->add_symbol(&(mtch->Table), $3->type() & LT_single);
        // Now add all variables in LHS into the symbol table
        // $1->add_symbol(CurrentTable, $3->type() & LT_single);

        $$->type($3->type());
    }

send_expr : 	expr T_send expr
	{ 
        $1->set_depth(1);
        $3->set_depth(1);
        $$ = new Send($1, $3); 
        $$->type($3->type());
     }

exprs :		expr
	{	
        $1->line(current_line);
        $$ = new ExprList($1); 
        $$->type($1->type());
        $1->parent($$);
    }
	|	expr ',' exprs
	{	
        $1->line(current_line);
        $1->parent($3);
		$3->prepend($1);
		$$ = $3; 
	}
    | expr error
    {
        yyerrorf("Syntax error: expecting a `,' between expressions");
        return 0;
    }

guard_expr :	basic_type
	{	$$ = $1; }
	|	guard_expr_list
	{	$$ = $1; }
	|	guard_expr_tuple
	{	$$ = $1; }
	|	guard_call
	{	$$ = $1; }
	|	'(' guard_expr ')'
	{	$$ = $2; }
	|	guard_expr add_op guard_expr %prec ADD_OP
	{
		$$ = new Binary($2, $1, $3); 
		delete $2;
	}
	| 	guard_expr mult_op guard_expr %prec MULT_OP
	{
		$$ = new Binary($2, $1, $3); 
		delete $2;
	}
	|	prefix_op guard_expr %prec UNARY
	{ 
		$$ = new Unary($1, $2); 
		delete $1;
	}

guard_expr_list : '[' ']'
	{	
        string s;

        if (CurrentTable != NULL)
        {
#ifdef _ALLOCS_H
		    s = CurrentTable->add_temp();
#endif
        }
        $$ = new EList(s); 
    }
	| '[' guard_expr guard_expr_tail ']'
	{	
		$3->prepend($2);
		$$ = $3;
	}

guard_expr_tail : '|' guard_expr
	{	$$ = new GuardTail($2); }
	|	',' guard_expr guard_expr_tail
	{	
		$3->prepend($2);
		$$ = $3;
	}
	| 	/* empty */
	{	$$ = new GuardTail; }
	;

guard_expr_tuple : '{' '}'
	{	$$ = new GuardTuple; }
	| '{' guard_exprs '}'
	{	$$ = new GuardTuple($2); }

guard_exprs :	guard_expr
	{	$$ = new ExprList($1); }
	|	guard_expr ',' guard_exprs
	{	
		$3->prepend($1); 
		$$ = $3;
	}
	
guard_call :	T_atom '(' guard_parameter_list ')'
	{	
        string fname;
        int x, size;
        Symbol * other;

        if ($3) size = $3->size();
        else size = 0;

        fname = Namer::semi_mangle(*($1),size);

        $$ = new GuardCall($1, $3); 
		// Need to do some clever stuff in here with imported modules	
        if ( (other = yy_program->Table.find(fname, S_Function)) )
        {
            x = other->used();
            other->used(x+1);
        }
    }

guard_parameter_list : guard_exprs
	{	$$ = $1; }
	|	/* empty */
	{	$$ = NULL; }
	;

bif_test :	T_atom '(' guard_parameter_list ')'
	{	
        string fname = Namer::semi_mangle(*($1),$3->size());
        Symbol * other;
        BifTest * bt;
        int x, tbt;

        bt = new BifTest($1, $3); 

        if ((tbt = bt->is_internal()))
        {
            $3->type_all(tbt);
        }

		// Need to do some clever stuff in here with imported modules	
        if ( (other = yy_program->Table.find(fname, S_Function)) )
        {
            x = other->used();
            other->used(x+1);
        }
        $$ = bt;
    }

guard_test : 	bif_test
	{	$$ = $1; }
	| 	guard_expr comp_op guard_expr
	{	
		$$ = new GuardTest(*($2), $1, $3); 
		delete $2;
	}

guard_tests : 	guard_test
	{	$$ = new GuardTestList($1); }
	|	guard_test ',' guard_tests
	{	
		$3->prepend($1);
		$$ = $3;
	}

guard :		T_bool
	{	$$ = new GuardBool($1); }
	|	guard_tests
	{ 	$$ = $1; }

anon_function : T_fun '(' 
    {
        SymbolTable * s;
        string anonname = Namer::anon_function();
		string ts = "";

 		s = new SymbolTable(anonname, S_Function, ts, 0);
 		retsymbol = new Symbol("_ret", S_Variable, ts);
        s->add(retsymbol);
        s->name(anonname);
        TableStack.push_front(s);
		CurrentTable = s;
    }
    formal_parameter_list ')' clause_guard clause_body T_end
    {
        ClauseHead * ch;
        FunctionClause * fc;
        AnonFunctionList * anonfun;
        string anonname;

        // add param symbols to local scope ..
        $4->add_symbol(CurrentTable, LT_param);

        anonname = CurrentTable->name();
		ch = new ClauseHead(anonname, anonname, $4); 
        fc = new FunctionClause(ch, $6, $7);
        $7->parent(fc);
        fc->type($7->type() | LT_on_heap);
        ((ClauseBody *) $7)->scope(CurrentTable);
        ((ClauseBody *) $7)->add_final(CurrentTable);

        fc->parent(yy_program->local);

        anonfun = new AnonFunctionList();
        anonfun->parent(yy_program->local);

        anonfun->append(fc);

        yy_program->local->append(anonfun);

        $$ = new AnonFunction(fc);
        $$->type(fc->type());

        // FIX: find "undeclared" variables in anon function
        //      add them as parameter
        
        TableStack.pop_front();
    }

function_clause : clause_head clause_guard clause_body
	{	
        $$ = new FunctionClause($1, $2, $3); 
        $3->parent($$);
        $$->type($3->type() | LT_on_heap);
        ((ClauseBody *) $3)->scope(CurrentTable);
        ((ClauseBody *) $3)->add_final(CurrentTable);
        TableStack.pop_front();
     }

clause_head :	
    T_atom 
    {
		SymbolTable * s;
		string ts = "";
		string chid;

        tmp_val_count = 0;
		chid = *($1) + itoa(clause_id++);

 		s = new SymbolTable(chid, S_Function, ts, 0);
 		retsymbol = new Symbol("_ret", S_Variable, ts);
        s->add(retsymbol);

        TableStack.push_front(s);
		CurrentTable = s;
    }
    '(' formal_parameter_list ')'
	{	
		Symbol * sf;
		string ts = "";
        string func, fname;
		int params = 0;
		string chid = *($1) + itoa(clause_id-1);

		if ($<nodelist>4 != NULL)
		{
			params = $<nodelist>4->size();
		}

        CurrentTable->params(params);
        fname = *($1);
		func = Namer::semi_mangle(fname, params);
        // fprintf(stderr, "CT->set_name(%s)\n", func.c_str());

        CurrentTable->name(fname);

		// fprintf(stderr, "yy_program add: %s\n", CurrentTable->full_symbol().c_str());
		yy_program->Table.add(CurrentTable);

		if (yy_program->Table.find(func, S_Function) == 0)
		{
			// fprintf(stderr, "Adding (%s)\n", func.c_str());
 			sf = new Symbol(func, S_Function, ts, params);
			yy_program->Table.add(sf);
		}
        else
        {
			//fprintf(stderr, "FOUND (%s)\n", func.c_str());
        }

        // This adds variables into the local scope
        // and has them set as derived from param
        $4->add_symbol(CurrentTable, LT_param);

		$$ = new ClauseHead(chid,fname,$4); 
        $$->line(current_line);
	}

formal_parameter_list : /* empty */
	{	$$ = new PatternList(); }
	|	patterns
	{	$$ = $1; }

clause_guard : 	/* empty */	
	{	$$ = NULL; }
	|	T_when guard
	{	$$ = new When($2); }
	
clause_body : T_arrow exprs
	{	
        $$ = new ClauseBody($2); 
        $$->type($2->type());
    }
    | T_arrow pragma
    {   
        $$ = $2; 
        $$->type(LT_anyvar);
    }

function :	function_clause
	{	
        $$ = active_function;
        $$->prepend($1);
        $$->type($1->type());
        $1->parent($$);
    }
	| 	function_clause ';' function
	{	
        // FIX: check $3 and $1 have the same # of parameters..
		$3->prepend($1); 
        $1->parent($3);
        $3->type($1->type() | $3->type());
		$$ = $3;
	}

def_params:	T_atom
	{ 
        $$ = new NodeList(new Define(*($1))); 
    }
	| 	T_atom ',' def_params
    {   
        $3->prepend(new Define(*($1)));
        $$ = $3;
    }

def_clause:
    T_atom
    {
        if (debug)
            fprintf(stderr,"def_clause T_atom\n");
        $$ = new DefClause(*($1), NULL);
    }
    | T_atom '(' def_params ')'
    {
        if (debug)
            fprintf(stderr,"def_clause T_atom ( def_params )\n");
        $$ = new DefClause(*($1), $3);
    }

attribute :	pattern
	{	$$ = new Attribute($1); }
	| 	'[' farity_list ']'
	{	$$ = new Attribute(NULL, $2); }
	|	T_atom ',' '[' farity_list ']'
	{	$$ = new Attribute($1, $4); }
    |   T_atom ',' '{' exprs '}'
    {   $$ = new Attribute($1, $4); }

farity_list : 	farity
	{	$$ = new FarityList($1); }
	|	farity ',' farity_list
	{	
		$3->prepend($1);
		$$ = $3;
	}

farity :	T_atom '/' T_number
	{	$$ = new Farity($1, $3); }
%%

void yyerror(const char *s)
{
	yyerrorf(s);
}

int yyerrorf(const char *fmt, ...)
{
    char buff[1024];
    va_list args;

    va_start(args, fmt);
    vsprintf(buff, fmt, args);
    va_end(args);
    fputs(buff, stderr);
    fprintf(stderr, " in %s at line %d\n", current_file, current_line);
	fprintf(stderr, "yytext(%s)\n", yytext);
	yy_program->errors++;
	return 0;
}

