//
// $Id: node.h,v 1.40 2004/09/05 02:42:56 geoff Exp $
//
// $Source: /home/cvs/gerl/src/node.h,v $
// $Revision: 1.40 $
// $Date: 2004/09/05 02:42:56 $
// $State: Exp $
//
// Geoff Wong (C) Copyright 1996,1997,1998.
// See the file "Copying" distributed with this code.
//

// @@DOC
// <h2> Node Class </h2>
// The Node class is the super class for all classes in the AST.
// It contains a number of key abstract methods for implementing the
// Interpreter Pattern (see Gamma et al).
// Nodes for the AST for Erlang - derived from the abstract base node
// Types of nodes available:
//         - unary
//        - binary
//        - n-ary (using the list class)
// @@ENDDOC

#ifndef _NODE_H
#define _NODE_H

#include <list>
#include <assert.h>
#include "String.h"
#include "etype.h"
#include "stmt.h"

class Symbol;
class SymbolTable;

// The abstract class at the top of the node hierarchy
class Node
{
protected:
    static int tablevel;
        // PURPOSE: shared variable indicating current tab depth
        //   for pretty printing (in print())
    string tabs();
        // PURPOSE: a string with the appropriate # of tabs

    int depth;
        // PURPOSE: what depth are we in the tree
        //  useful when emitting stuff

    int _line;
        // PURPOSE: record the line number relevant to this node

    int _type;
        // PURPOSE: what type am I?

    int _stmt;
        // PURPOSE: what type of code do I emit?

    string _extra;
        // PURPOSE: `remember' or provide extra information
        //      sort of a general purpose variable.

    Symbol * _assign;
        // should this node be assigned to a temporary variable?

    Symbol * _freesym;
        // should this node be freed (have an allocated symbol to free)?

    Node * _parent;
        // PURPOSE: if i'm a sub-node this is a *possible*
        //      pointer to my parent (if it bothered to set it)

    list<string> _pre_code;
        // List of code that must be pre-emitted (as statements in C)
        
    bool _flat;

public:
    Node();
        // constructor - to initialise stuff

    virtual ~Node();
        // virtual destructor

    virtual list<string> emit_code() = 0;
        // PURPOSE: Emit a list of code fragments generated by this node
        
    virtual list<string> pre_code();
        // PURPOSE: return emit_code() list if a we have an assignable
        //      symbol.
        
    virtual void add_pre_code(list<string> l);
        // Purpose: add pre_code to our pre_code list
        
    virtual string expr_code();
        // PURPOSE: Return the symbol we're assigned to (if we have one)
        // otherwise returns emit_code() built as a string

    virtual string print() = 0;
        // PURPOSE: Pretty print it - use for debugging

    virtual string emit_pattern(string x);
        // PURPOSE: Emit as pattern matching code.
        // (often as a parameter)

    virtual string emit_guard(string x);
        // PURPOSE: Emit as a guard

    virtual void extra(const string&);
        // set _extra statement(s) for output

    virtual string extra();
        // return _extra

    virtual void add_symbol(SymbolTable * st, int default_type = LT_none);
        // PURPOSE: Adds any relevant symbols from this AST (fragment)
        //  to the given symbol table
        //  Walks down the given tree to leaf Variables and adds them
        //  to the current symbol table (with a type)

    virtual void set_depth(int);
        // PURPOSE: set the node depth (at the moment only 0 or 1 is used)

    virtual void line(int);
        // PURPOSE: set the line number

    virtual int line();
        // PURPOSE: return the line number

    virtual void type(int);
        // PURPOSE: set my type 

    virtual int type();
        // PURPOSE: return my type

    virtual int stmt_type();
        // PURPOSE: general description of the type of node 
    
    virtual void stmt_type(int);
        // set the statement type

    virtual void assign(Symbol *);
        // Assign this code to a temporary var.

    virtual Symbol * assign();
        // Return the temporary var the code assigns to.
        
    virtual void freesym(Symbol *);
        // Assign this code to a temporary var to be freed.

    virtual Symbol * freesym();
        // Return the temporary var we may want to free
        // or are associated with ...

    virtual bool typecheck(Etype check, Etype restrict = LT_all);
        // PURPOSE: resolve my type with given type?

    virtual string emit_typecheck(string str, int restrict, 
                                  int param, string where);
        // PURPOSE: generate appropriate typechecking code

    virtual void parent(Node *);
        // PURPOSE: tell me who my parent is

    virtual Node * parent();
        // PURPOSE: return the parent

    virtual bool flat_code() { return _flat; }
        // Purpose: if true we can be emitted as direct C code ";" separated.
        
    virtual void set_flat(bool f) { _flat = f; }
        // Purpose: if true we can be emitted as direct C code ";" separated.
};


// @@DOC
// <h2> NodeList Class </h2>
// The class is a basic wrapper to maintain a list of Node (or derived)
// classes
// @@ENDDOC

class NodeList : public Node
{
protected:
    string _delimiter;

public:
    list<Node *> nodes;

    NodeList();
    NodeList(Node * n);
    NodeList(Node * n, Node *);
        // constructors

    list<string> emit_code();
        // code emission
    string emit_pattern(string);
        // emit the pattern matching code
    string emit_guard(string x);
        // Emit as a guard
    string print();
        // Pretty print
    void add_symbol(SymbolTable * st, int default_type = LT_none);
        // Add all the symbols in the list in..
    void set_depth(int x);
        // Set the depth of everything in the list 
        
    Symbol * assign();
        // Return the temporary var the code assigns to.
        
    void assign(Symbol *);
        // Make all elements assigned?

    // just wrappers for push_front and push_back
    void prepend(Node * n);
    void append(Node * n);
    int size();

    virtual void type_all(int t);
        // set the type of all elements in the list..

    virtual int type_all();
        // return the or'ed set of all elements in the list..

    virtual void delimiter(const string& delim); 
        // set up a delimiter between individual nodes being outputted
};

#endif
